#! /usr/bin/python
import sys

from edge import eventlog
from edge.exceptions import *
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.api_terminal import TerminalApi
from edge.webserver.sysconf.dgmod import dg


class _ArgSet:
    ACTION_CREATE = 'create'
    ACTION_CREATE_WAGA = 'create-waga'
    ACTION_REMOVE = 'remove'
    ACTION_REGENERATE = 'regenerate'

    _CREATE_ACTION_VENDOR = '--vendor'
    _CREATE_ACTION_MODEL = '--model'
    _CREATE_ACTION_SERIAL = '--serial'
    _CREATE_ACTION_UUID = '--uuid'

    _REMOVE_ACTION_PID = '--pid'

    _ARG_VALUE_SEPARATOR = '='

    def __init__(self, action, create_vendor=None, create_model=None, create_serial=None, remove_pid=None,
                 create_uuid=None):
        """
        :type action: str
        :type create_vendor: str | None
        :type create_model: str | None
        :type create_serial: str | None
        :type remove_pid: str | None
        :type create_uuid: str | None
        """
        self.action = action
        self.create_vendor = create_vendor
        self.create_model = create_model
        self.create_serial = create_serial
        self.create_uuid = create_uuid
        self.remove_pid = remove_pid

    @staticmethod
    def show_usage():
        print 'pipeline-ctl create-waga [%s%s<uuid>]' % (_ArgSet._CREATE_ACTION_UUID, _ArgSet._ARG_VALUE_SEPARATOR)
        print 'pipeline-ctl create %s%s<vendor> %s%s<model> %s%s<serial> [%s%s<uuid>]' \
              % (_ArgSet._CREATE_ACTION_VENDOR, _ArgSet._ARG_VALUE_SEPARATOR,
                 _ArgSet._CREATE_ACTION_MODEL, _ArgSet._ARG_VALUE_SEPARATOR,
                 _ArgSet._CREATE_ACTION_SERIAL, _ArgSet._ARG_VALUE_SEPARATOR,
                 _ArgSet._CREATE_ACTION_UUID, _ArgSet._ARG_VALUE_SEPARATOR)

        print 'pipeline-ctl remove %s%s<pipeline-id>' \
              % (_ArgSet._REMOVE_ACTION_PID, _ArgSet._ARG_VALUE_SEPARATOR)

        print 'pipeline-ctl regenerate'

    @staticmethod
    def _deserialize_create(raw_arg_set):
        """
        :type raw_arg_set: list of str
        :rtype _ArgSet
        """
        if len(raw_arg_set) < 3:
            raise ValueError('invalid number of arguments')

        vendor = None
        model = None
        serial = None
        new_uuid = None
        for raw_arg in raw_arg_set:
            raw_name_value = raw_arg.split(_ArgSet._ARG_VALUE_SEPARATOR, 1)
            if len(raw_name_value) != 2:
                raise ValueError('invalid argument %s' % raw_arg)

            raw_name, raw_value = raw_name_value
            raw_name = raw_name.strip()
            raw_value = raw_value.strip()

            if raw_name == _ArgSet._CREATE_ACTION_VENDOR:
                vendor = raw_value
                continue

            if raw_name == _ArgSet._CREATE_ACTION_SERIAL:
                serial = raw_value
                continue

            if raw_name == _ArgSet._CREATE_ACTION_MODEL:
                model = raw_value
                continue

            if raw_name == _ArgSet._CREATE_ACTION_UUID:
                import uuid
                new_uuid = raw_value
                uuid.UUID(new_uuid)
                continue

            raise ValueError('invalid argument %s' % raw_arg)

        if not reduce(lambda x, y: x and y, [vendor, model, serial]):
            raise ValueError('invalid number of arguments')

        return _ArgSet(_ArgSet.ACTION_CREATE, create_vendor=vendor, create_model=model, create_serial=serial, create_uuid=new_uuid)

    @staticmethod
    def _deserialize_remove(raw_arg_set):
        """
        :type raw_arg_set: list of str
        :rtype _ArgSet
        """
        if len(raw_arg_set) != 1:
            raise ValueError('invalid number of arguments')

        raw_name_value = str(raw_arg_set[0]).split(_ArgSet._ARG_VALUE_SEPARATOR, 1)
        if len(raw_name_value) != 2:
            raise ValueError('invalid argument')

        raw_name, raw_value = raw_name_value
        raw_name = raw_name.strip()
        raw_value = raw_value.strip()

        if raw_name != _ArgSet._REMOVE_ACTION_PID:
            raise ValueError('invalid argument')

        return _ArgSet(_ArgSet.ACTION_REMOVE, remove_pid=raw_value)

    @staticmethod
    def _deserialize_regenerate(raw_arg_set):
        """
        :type raw_arg_set: list of str
        :rtype _ArgSet
        """
        if len(raw_arg_set) != 0:
            raise ValueError('invalid number of arguments')

        return _ArgSet(_ArgSet.ACTION_REGENERATE)

    @staticmethod
    def _deserialize_create_waga(raw_arg_set):
        """
        :type raw_arg_set: list of str
        :rtype _ArgSet
        """
        if len(raw_arg_set) not in [0, 1]:
            raise ValueError('invalid number of arguments')

        new_uuid = None
        for raw_arg in raw_arg_set:
            raw_name_value = raw_arg.split(_ArgSet._ARG_VALUE_SEPARATOR, 1)
            if len(raw_name_value) != 2:
                raise ValueError('invalid argument %s' % raw_arg)

            raw_name, raw_value = raw_name_value
            raw_name = raw_name.strip()
            raw_value = raw_value.strip()

            if raw_name == _ArgSet._CREATE_ACTION_UUID:
                import uuid
                new_uuid = raw_value
                uuid.UUID(new_uuid)
                continue

            raise ValueError('invalid argument %s' % raw_arg)

        return _ArgSet(_ArgSet.ACTION_CREATE_WAGA, create_uuid=new_uuid)

    @staticmethod
    def deserialize(raw_arg_set):
        """
        :type raw_arg_set: list of str
        :rtype _ArgSet
        """
        raw_arg_count = len(raw_arg_set)
        if raw_arg_count < 2:
            raise ValueError('invalid number of arguments')

        raw_action = str(raw_arg_set[1])

        if raw_action == _ArgSet.ACTION_CREATE:
            return _ArgSet._deserialize_create(raw_arg_set[2:])

        elif raw_action == _ArgSet.ACTION_REMOVE:
            return _ArgSet._deserialize_remove(raw_arg_set[2:])

        elif raw_action == _ArgSet.ACTION_REGENERATE:
            return _ArgSet._deserialize_regenerate(raw_arg_set[2:])

        elif raw_action == _ArgSet.ACTION_CREATE_WAGA:
            return _ArgSet._deserialize_create_waga(raw_arg_set[2:])

        raise ValueError('invalid command-line')


def _handle_regenerate():
    """
    :rtype: None
    """
    from edge.webserver.sysconf.edge_creator import generate

    root_info = ph.RootInfoUiconf.get_info()
    terms_info = ph.InfoUiconf.map_units_r(root_info, unit_types=[ph.UI_TYPE_TERMINAL])

    for term_info in terms_info:
        term_uimod = ph.UiMod.create(term_info['ref'])
        eventlog.log_d('exec: {}'.format(term_uimod.ref))
        # noinspection PyBroadException
        try:
            # create edge-jnode-set
            factory_uimod = ph.UiMod(term_uimod.pid, ph.UI_TYPE_FACTORY, '0')
            uirefs = ph.InfoUiconf.find_unit_refs_r(term_info)
            generate(factory_uimod.create_backup_dirpath(), uirefs, forced=True)

            eventlog.log_d('done: {}'.format(term_uimod.ref))

        except BusyError:
            eventlog.log_d('fail: {} (lock error)'.format(term_uimod.ref))

        except Exception:
            eventlog.log_d('fail: {} (traceback)'.format(term_uimod.ref))
            eventlog.log_x()

    eventlog.log_d('done')


def _main(raw_arg_set):
    arg_set = None

    try:
        arg_set = _ArgSet.deserialize(raw_arg_set)
    except ValueError:
        eventlog.log_x()
        _ArgSet.show_usage()
        exit(1)

    if arg_set.action == _ArgSet.ACTION_CREATE:
        root_info = ph.RootInfoUiconf.get_info()
        ctrl_info = ph.RootInfoUiconf.find_unit_r(root_info, unit_type=ph.UI_TYPE_CONTROLLER)
        TerminalApi.create_radar_term(ctrl_info['ref'], {
            'vendor': arg_set.create_vendor,
            'model': arg_set.create_model,
            'serial': arg_set.create_serial,
            'internal': True,
            'uuid': arg_set.create_uuid
        })
    elif arg_set.action == _ArgSet.ACTION_CREATE_WAGA:
        root_info = ph.RootInfoUiconf.get_info()
        ctrl_info = ph.RootInfoUiconf.find_unit_r(root_info, unit_type=ph.UI_TYPE_CONTROLLER)
        TerminalApi.create_waga_rtsp_term(ctrl_info['ref'], {
            'internal': True,
            'uuid': arg_set.create_uuid,
            'with-intrinsic': True
        })

    elif arg_set.action == _ArgSet.ACTION_REMOVE:
        TerminalApi.drop_terminal2(arg_set.remove_pid)
        dg.LifeCycle.sighup_ss_dgmod()
        dg.LifeCycle.restart_ws_dgmod()

    elif arg_set.action == _ArgSet.ACTION_REGENERATE:
        _handle_regenerate()

    exit(0)


if __name__ == '__main__':
    eventlog.open_log('pipeline-ctl', 'debug')
    # noinspection PyBroadException
    try:
        _main(sys.argv)

    except Exception:
        eventlog.log_x()
        exit(1)

    finally:
        eventlog.close_log()
