#!/usr/bin/python2

import os
import datetime
import iso8601
import sys

import ujson
import psycopg2
import psycopg2.extras

from edge import eventlog
from edge.exceptions import *
from edge.webserver import cgi
from edge.webserver.sysconf.uiconf.time import TimeUiconf

HTTP_GET = 'GET'

DBNAME = 'autocode'
DBUSER = 'autocode'
DBPASS = 'autocode'


class ReportError(RuntimeError):
    pass


def _normalize_iso_datetime(iso_datetime):
    """
    :type iso_datetime: str
    :rtype: (str, int)
    :raises: ReportError
    """
    # define tz-offset
    aware_dt = iso8601.parse_date(iso_datetime, default_timezone=None)
    if aware_dt.tzinfo is None:
        # in case of 'YYYY-MM-DDThh:mm:ss' use edge-tz-offset
        dt_tzinfo = TimeUiconf.get_tz()
        if dt_tzinfo is None:
            eventlog.log_e('fail: define tz-info')
            raise ReportError

    else:
        # in case of 'YYYY-MM-DDThh:mm:ss<tz-offset>' use supplied tz-offset
        dt_tzinfo = aware_dt.tzinfo

    tz_offset = dt_tzinfo.utcoffset(aware_dt)
    if tz_offset is None:
        eventlog.log_d('fail: define tz-offset')
        raise ReportError

    # remove tz-offset from iso_datetime if present
    naive_iso_datetime = aware_dt.replace(tzinfo=None).isoformat()

    return naive_iso_datetime, int(tz_offset.total_seconds())


class _ReportArgs:
    def __init__(self, origin, min_datetime, max_datetime, min_ts_gap, min_total_count):
        """
        :type origin: str
        :type min_datetime: datetime.date
        :type max_datetime: datetime.date
        :type min_ts_gap: int
        :type min_total_count: int
        """
        self.origin = origin
        self.min_datetime = min_datetime
        self.max_datetime = max_datetime
        self.min_ts_gap = min_ts_gap
        self.min_total_count = min_total_count

    @staticmethod
    def deserialize(query_string):
        """
        :type query_string: str
        :rtype: _ReportArgs
        """
        min_time = None
        max_time = None
        min_ts_gap = None
        min_total_count = None
        origin = None

        if len(query_string) > 1:
            entry_set = query_string[1:].split('&')
            for entry in entry_set:
                key_value = entry.split('=', 1)
                if len(key_value) != 2:
                    eventlog.log_d("fail: invalid query-arg")
                    return None

                key, value = key_value
                if key == "min-datetime":
                    iso_datetime_min = value
                    try:
                        iso8601.parse_date(iso_datetime_min)
                    except iso8601.ParseError as x:
                        eventlog.log_d("fail: invalid min-ts [%s]" % x.message)
                        return None

                    min_time = iso_datetime_min

                elif key == "max-datetime":
                    iso_datetime_max = value
                    try:
                        iso8601.parse_date(iso_datetime_max)
                    except iso8601.ParseError as x:
                        eventlog.log_d("fail: invalid max-ts [%s]" % x.message)
                        return None

                    max_time = iso_datetime_max

                elif key == "min-ts-gap":
                    try:
                        raw_min_ts_gap = int(value)
                        if 0 <= raw_min_ts_gap <= 65535:
                            eventlog.log_d("raw_min_ts_gap: %s" % raw_min_ts_gap)
                            min_ts_gap = raw_min_ts_gap
                        else:
                            eventlog.log_d("fail: invalid min-ts-gap")
                            return None
                    except ValueError:
                        eventlog.log_d("fail: invalid min-ts-gap")
                        return None

                elif key == "min-total-count":
                    try:
                        raw_min_total_count = int(value)
                        if 1 <= raw_min_total_count <= 65535:
                            min_total_count = raw_min_total_count
                        else:
                            eventlog.log_d("fail: invalid min-total-count")
                            return None

                    except ValueError:
                        eventlog.log_d("fail: invalid min-total-count")
                        return None

                elif key == "origin":
                    try:
                        raw_origin = value
                        if len(raw_origin) == 36:
                            origin = raw_origin
                        else:
                            eventlog.log_d("fail: invalid origin")
                            return None

                    except ValueError:
                        eventlog.log_d("fail: invalid origin")
                        return None

                else:
                    eventlog.log_d("fail: invalid query-arg")
                    return None

        # define iso-datetime-max
        if max_time is None:
            max_time = datetime.datetime.now().isoformat()

        naive_iso_dt_max, api_query_tz_offset = _normalize_iso_datetime(max_time)

        # define iso-datetime-min
        if min_time is None:
            datetime_max = iso8601.parse_date(naive_iso_dt_max)
            datetime_min = datetime_max - datetime.timedelta(days=7)
            min_time = datetime_min.isoformat()

        naive_iso_dt_min, api_query_tz_offset = _normalize_iso_datetime(min_time)

        # validate date period
        datetime_max = iso8601.parse_date(naive_iso_dt_max)
        datetime_min = iso8601.parse_date(naive_iso_dt_min)
        if datetime_max - datetime_min > datetime.timedelta(days=7):
            eventlog.log_d('fail: report period is too long')
            return None

        # assign default values
        if min_ts_gap is None:
            min_ts_gap = 10
        if min_total_count is None:
            min_total_count = 2

        # eventlog.log_d("min_ts %s" % min_ts)
        # eventlog.log_d("max_ts %s" % max_ts)
        # eventlog.log_d("min_ts_gap %s" % min_ts_gap)
        # eventlog.log_d("min_total_count %s" % min_total_count)

        # make report
        if datetime_min and datetime_max and min_ts_gap and min_total_count and origin:
            return _ReportArgs(origin, datetime_min, datetime_max, min_ts_gap, min_total_count)

        else:
            eventlog.log_d("fail: invalid number of query-arg")
            return None


def _get_report(report_args):
    """
    :type report_args: _ReportArgs
    :rtype: list of dict | None
    """
    dbapi_jnode = {
        'name': 'plate.scurry',
        'param': {
            'origin': report_args.origin,
            'fromtime': report_args.min_datetime.isoformat(),
            'totime': report_args.max_datetime.isoformat(),
            'mingap': report_args.min_ts_gap,
            'mindet': report_args.min_total_count
        }
    }
    psycopg2.extras.register_default_jsonb(globally=True)
    try:
        with psycopg2.connect(database=DBNAME, user=DBUSER, password=DBPASS, port=5433) as conn:
            cursor = conn.cursor()
            cursor.execute(
                "select errcode, errmsg, result from api_report_get(%s::jsonb)",
                (psycopg2.extras.Json(dbapi_jnode, dumps=ujson.dumps),))
            errcode, errmsg, raw_report = cursor.fetchone()

        if not raw_report or errcode != 0:
            eventlog.log_d('fail: api_report_get (%s)' % errmsg)
            return None

        # eventlog.log_d('raw_report: %s' % ujson.dumps(raw_report))

        result_jnode = raw_report
        if len(result_jnode) > 0:
            expected_set = frozenset(['plate-text', 'total-count', 'last-guid', 'last-datetime'])
            actual_set = set(result_jnode[0].keys())
            if len(actual_set.symmetric_difference(expected_set)) > 0:
                eventlog.log_d("fail: invalid report from db")
                return None

        return result_jnode

    except psycopg2.Error as e:
        eventlog.log_d('fail: db exception (%s)' % e.message)

    return None


def _run(http_method, query_string):
    """
    :type http_method: str
    :type query_string: str
    :rtype: edge.webserver.cgi.CgiReply
    """
    # Dispatch http-method.
    if http_method != HTTP_GET:
        eventlog.log_d("fail: unsupported http-method")
        return cgi.CgiReply(cgi.R405_METHOD_NOT_ALLOWED)

    report_args = _ReportArgs.deserialize(query_string)
    if not report_args:
        eventlog.log_e('fail: deserialize report-args')
        return cgi.CgiReply(cgi.R400_BAD_REQUEST)

    report = _get_report(report_args)
    if report is None:
        eventlog.log_e('fail: get-report')
        return cgi.CgiReply(cgi.R500_INTERNAL_SERVER_ERROR)

    return cgi.CgiReply(cgi.R200_SUCCESS, report)


# noinspection PyBroadException
def _main():
    reply = cgi.CgiError500().reply
    try:
        http_mode = os.environ['REQUEST_METHOD']
        path_info = os.environ['PATH_INFO']
        query_string = os.environ['QUERY_STRING']
        eventlog.log_d("%s|%s|%s" % (http_mode, path_info, query_string))

        if path_info:
            eventlog.log_e("fail: invalid path-info")
            reply = cgi.CgiReply(cgi.R404_NOT_FOUND)
        else:
            reply = _run(http_mode, query_string)

    except ParamsError:
        reply = cgi.CgiError400().reply

    except NotFoundError:
        reply = cgi.CgiError404().reply

    except BusyError:
        reply = cgi.CgiError409().reply

    except CamServiceError:
        reply = cgi.CgiCamError().reply

    except LicServiceError:
        reply = cgi.CgiLicError().reply

    except ServiceError:
        reply = cgi.CgiError503().reply

    except cgi.CgiError as x:
        reply = x.reply

    except EdgeError:
        pass

    except Exception:
        eventlog.log_x()

    finally:
        reply.send()

    sys.exit(0)


if __name__ == '__main__':
    eventlog.open_log('eventstore-fact', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
