#!/usr/bin/python2

from edge import eventlog
from edge.webserver import cgi
from edge.webserver.sysconf.api_license import LicenseApi


class _Module(cgi.CgiModule):
    """
    GET      /api/v2/sys/conf/license   /<uiref>/keys
    POST     /api/v2/sys/conf/license   /<uiref>/keys/<keyref>/attached
    DELETE   /api/v2/sys/conf/license   /<uiref>/keys/<keyref>/attached
    DELETE   /api/v2/sys/conf/license   /<uiref>/keys/<keyref>
    GET      /api/v2/sys/conf/license   /<uiref>/fingerprint?key=<keyref>
    POST     /api/v2/sys/conf/license   /<uiref>/keys
    GET      /api/v2/sys/conf/license   /<uiref>?key=<keyref>?type=lpr&type=radar

    """
    def __init__(self, handler):
        """
        :type handler: () -> cgi.CgiReply
        """
        cgi.CgiModule.__init__(self, handler)

    @classmethod
    def factory(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        method_to_handlers_map = {
            'GET': [
                cls._license__get_uiref,
                cls._license__get_uiref_keys,
                cls._license__get_uiref_fingerprint
            ],
            'POST': [
                cls._license__post_uiref_keys,
                cls._license__post_uiref_keys_keyref_attached
            ],
            'DELETE': [
                cls._license__delete_uiref_keys_keyref_attached,
                cls._license__delete_uiref_keys_keyref
            ]
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request)

            except cgi.CgiError404:
                pass

        eventlog.log_d('fail: http404')
        raise cgi.CgiError404

    @classmethod
    def _license__get_uiref_keys(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        # GET   /api/v2/sys/conf/license/   :unitRef/keys
        if len(request.path_info) != 2:
            raise cgi.CgiError404

        elif request.path_info[0] != 'license.0':
            raise cgi.CgiError404

        elif request.path_info[1] != 'keys':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            keys_info = LicenseApi.get_keys_info()
            return cgi.CgiReply(body=keys_info)

        return cls(handler)

    @classmethod
    def _license__get_uiref(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        # GET   /api/v2/sys/conf/license/   :unitRef?key=keyRef
        if len(request.path_info) != 1:
            raise cgi.CgiError404

        elif request.path_info[0] != 'license.0':
            raise cgi.CgiError404

        # request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            if not request.query_params:
                ui_lic_info = LicenseApi.get_licinfo_key_all()
                return cgi.CgiReply(body=ui_lic_info)

            key_refs = []
            for name_value in request.query_params:
                name, value = name_value.split('=', 1)
                if name == 'key':
                    key_refs.append(value)

            if len(key_refs) == 1:
                ui_lic_info = LicenseApi.get_licinfo_key_keyref(key_refs[0])
                return cgi.CgiReply(body=ui_lic_info)
            else:
                raise cgi.CgiError400

        return cls(handler)

    @classmethod
    def _license__get_uiref_fingerprint(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        # GET   /api/v2/sys/conf/license/   :unitRef/fingerprint?key=keyRef
        if len(request.path_info) != 2:
            raise cgi.CgiError404

        elif request.path_info[0] != 'license.0':
            raise cgi.CgiError404

        elif request.path_info[1] != 'fingerprint':
            raise cgi.CgiError404

        # request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            if not request.query_params:
                host_c2v = LicenseApi.get_host_c2v_base64()
                return cgi.CgiReply.create200({
                    'c2v': host_c2v
                })

            key_refs = []
            for name_value in request.query_params:
                name, value = name_value.split('=', 1)
                if name == 'key':
                    key_refs.append(value)

            if len(key_refs) == 1:
                c2v_base64 = LicenseApi.get_key_c2v_base64(key_refs[0])
                return cgi.CgiReply.create200({
                    'c2v': c2v_base64
                })
            else:
                raise cgi.CgiError400

        return cls(handler)

    @classmethod
    def _license__post_uiref_keys(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        # POST   /api/v2/sys/conf/license/   :unitRef/keys
        if len(request.path_info) != 2:
            raise cgi.CgiError404

        elif request.path_info[0] != 'license.0':
            raise cgi.CgiError404

        elif request.path_info[1] != 'keys':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_present()

        v2c_base64 = request.body_jnode.get('v2c')
        if not v2c_base64:
            eventlog.log_d('fail: http400 (invalid body#/v2c)')
            raise cgi.CgiError400

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            LicenseApi.install_v2c_base64(v2c_base64)
            return cgi.CgiReply.create200()

        return cls(handler)

    @classmethod
    def _license__post_uiref_keys_keyref_attached(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        # POST   /api/v2/sys/conf/license/   :unitRef/keys/:keyRef/attached
        if len(request.path_info) != 4:
            raise cgi.CgiError404

        elif request.path_info[0] != 'license.0':
            raise cgi.CgiError404

        elif request.path_info[1] != 'keys':
            raise cgi.CgiError404

        elif request.path_info[3] != 'attached':
            raise cgi.CgiError404

        request.assert_query_absent()
        # request.assert_body_absent()

        if request.body:
            # TODO: fix {} in request-body from frontend
            if len(request.body_jnode) > 0:
                request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            LicenseApi.attach_key(request.path_info[2])
            return cgi.CgiReply.create200()

        return cls(handler)

    @classmethod
    def _license__delete_uiref_keys_keyref_attached(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        # DELETE   /api/v2/sys/conf/license/   :unitRef/keys/:keyRef/attached
        if len(request.path_info) != 4:
            raise cgi.CgiError404

        elif request.path_info[0] != 'license.0':
            raise cgi.CgiError404

        elif request.path_info[1] != 'keys':
            raise cgi.CgiError404

        elif request.path_info[3] != 'attached':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            LicenseApi.detach_key(request.path_info[2])
            return cgi.CgiReply.create200()

        return cls(handler)

    @classmethod
    def _license__delete_uiref_keys_keyref(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        # DELETE   /api/v2/sys/conf/license/   :unitRef/keys/:keyRef
        if len(request.path_info) != 3:
            raise cgi.CgiError404

        elif request.path_info[0] != 'license.0':
            raise cgi.CgiError404

        elif request.path_info[1] != 'keys':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            LicenseApi.remove_key(request.path_info[2])
            return cgi.CgiReply.create200()

        return cls(handler)


if __name__ == '__main__':
    eventlog.open_log('cgi-license', 'debug')
    try:
        cgi.run_cgi(_Module.factory)
    finally:
        eventlog.close_log()
