import os

from edge import eventlog
from edge.webserver import cgi
from edge.webserver.sysconf.api_openvpn import OpenVpnApi


class _Module(cgi.CgiModule):
    """
    GET      /api/v2/sys/conf/openvpn   /<uiref>
    POST     /api/v2/sys/conf/openvpn   /<uiref>/profiles
    DELETE   /api/v2/sys/conf/openvpn   /<uiref>/profiles/<ref>

    POST     /api/v2/sys/conf/openvpn   /<uiref>/profiles/<ref>/active
    DELETE   /api/v2/sys/conf/openvpn   /<uiref>/profiles/<ref>/active
    """
    def __init__(self, handler):
        """
        :type handler: () -> cgi.CgiReply
        """
        cgi.CgiModule.__init__(self, handler)

    @classmethod
    def factory(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        method_to_handlers_map = {
            'GET': [
                cls._openvpn__get_uiref
            ],
            'POST': [
                cls._openvpn__post_uiref_profiles,
                cls._openvpn__post_uiref_profiles_ref_active
            ],
            'DELETE': [
                cls._openvpn__delete_uiref_profiles_ref,
                cls._openvpn__delete_uiref_profiles_ref_active
            ]
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request)

            except cgi.CgiError404:
                pass

        eventlog.log_d('fail: http404')
        raise cgi.CgiError404

    @classmethod
    def _openvpn__get_uiref(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        if len(request.path_info) != 1:
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            uiconf_pub = OpenVpnApi.load_unit(request.path_info[0])
            return cgi.CgiReply(body=uiconf_pub)

        return cls(handler)

    @classmethod
    def _openvpn__post_uiref_profiles(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        if len(request.path_info) != 2:
            raise cgi.CgiError404

        elif request.path_info[1] != 'profiles':
            raise cgi.CgiError404

        request.assert_query_absent()

        prof_path = os.getenv('PROF_PATH')
        if prof_path is None:
            eventlog.log_e('fail: http500 (env.PROF_PATH not found)')
            raise cgi.CgiError500

        prof_name = os.getenv('PROF_NAME')
        if prof_name is None:
            eventlog.log_e('fail: http500 (env.PROF_NAME not found)')
            raise cgi.CgiError500

        prof_user = os.getenv('PROF_USER')
        prof_pass = os.getenv('PROF_PASS')

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            uiconf_pub = OpenVpnApi.make_profile(request.path_info[0], prof_name, prof_path, prof_user, prof_pass)
            return cgi.CgiReply(body=uiconf_pub)

        return cls(handler)

    @classmethod
    def _openvpn__delete_uiref_profiles_ref(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        if len(request.path_info) != 3:
            raise cgi.CgiError404

        elif request.path_info[1] != 'profiles':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            uiconf_pub = OpenVpnApi.drop_profile(
                unit_uiref=request.path_info[0],
                profile_ref=request.path_info[2])

            return cgi.CgiReply(body=uiconf_pub)

        return cls(handler)

    @classmethod
    def _openvpn__post_uiref_profiles_ref_active(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        if len(request.path_info) != 4:
            raise cgi.CgiError404

        elif request.path_info[1] != 'profiles':
            raise cgi.CgiError404

        elif request.path_info[3] != 'active':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            uiconf_pub = OpenVpnApi.make_profile_active(
                unit_uiref=request.path_info[0],
                profile_ref=request.path_info[2],
                active=True)

            return cgi.CgiReply(body=uiconf_pub)

        return cls(handler)

    @classmethod
    def _openvpn__delete_uiref_profiles_ref_active(cls, request):
        """
        :type request: cgi.CgiRequest
        :rtype: cgi.CgiModule
        """
        if len(request.path_info) != 4:
            raise cgi.CgiError404

        elif request.path_info[1] != 'profiles':
            raise cgi.CgiError404

        elif request.path_info[3] != 'active':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            uiconf_pub = OpenVpnApi.make_profile_active(
                unit_uiref=request.path_info[0],
                profile_ref=request.path_info[2],
                active=False)

            return cgi.CgiReply(body=uiconf_pub)

        return cls(handler)


if __name__ == '__main__':
    eventlog.open_log('cgi-openvpn', 'debug')
    try:
        cgi.run_cgi(_Module.factory)
    finally:
        eventlog.close_log()
