import os

from edge import eventlog
from edge.com import fs
from edge.webserver import cgi
from edge.webserver.sysdb import wanted

_LOAD_ACTION_NONE = 'none'
_LOAD_ACTION_RM = 'rm'
_LOAD_ACTION_CP = 'cp'

_API_TO_DB_ACTION_MAP = {
    _LOAD_ACTION_CP: 'set',
    _LOAD_ACTION_RM: 'unset',
    _LOAD_ACTION_NONE: 'none'
}


def _get_time(qs, key_to_find):
    """
    :type qs: str
    :type key_to_find: str
    :rtype: str
    """
    import urlparse
    import iso8601
    url_query = urlparse.parse_qs(qs[1:], strict_parsing=True)

    iso_datetime = None
    for key, values in url_query.iteritems():
        # deserialize datetime-max
        if key == key_to_find:
            if len(values) != 1:
                eventlog.log_d('fail: http400')
                raise cgi.CgiError404

            # url parse replases '+' to ' '
            raw_iso_datetime = values.pop()
            iso_datetime = raw_iso_datetime.replace(' ', '+')
            try:
                iso8601.parse_date(iso_datetime)
            except iso8601.ParseError:
                eventlog.log_d('fail: http400')
                raise cgi.CgiError404

            continue

    return iso_datetime


class _Module(cgi.CgiModule):
    """
    GET     /api/v2/sys/db/wanted
    PUT     /api/v2/sys/db/wanted  /list/:owner/:name
    DELETE  /api/v2/sys/db/wanted  /lists/:owner/:name

    GET     /api/v2/sys/db/wanted  /profiles/:name
    PUT     /api/v2/sys/db/wanted  /profiles/:name
    DELETE  /api/v2/sys/db/wanted  /profiles/:name
    DELETE  /api/v2/sys/db/wanted  /profiles/:name/log
    """
    def __init__(self, handler):
        """
        :type handler: () -> cgi.CgiReply
        """
        cgi.CgiModule.__init__(self, handler)

    @classmethod
    def factory(cls, request, wantedpack):
        """
        :type request: cgi.CgiRequest
        :type wantedpack: WantedPack
        :rtype: cgi.CgiModule
        """
        method_to_handlers_map = {
            'GET': [
                cls._wanted__get,
                cls._wanted__get_profiles_ref
            ],
            'PUT': [
                cls._wanted__put_list,
                cls._wanted__put_profiles_ref
            ],
            'DELETE': [
                cls._wanted__delete_list_owner_name,
                cls._wanted__delete_list_owner_name_log__cleanup,
                cls._wanted__delete_list_owner_name_log__rollback,
                cls._wanted__delete_profiles_ref
            ]
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request, wantedpack)

            except cgi.CgiError404:
                pass

        eventlog.log_d('fail: http404')
        raise cgi.CgiError404

    @classmethod
    def _wanted__get(cls, request, wantedpack):
        """
        :type request: cgi.CgiRequest
        :type wantedpack: WantedPack
        :rtype: cgi.CgiModule
        """
        if len(request.path_info) != 0:
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            reply_jnode = wantedpack.get()
            return cgi.CgiReply(body=reply_jnode)

        return cls(handler)

    @classmethod
    def _wanted__delete_list_owner_name(cls, request, wantedpack):
        """
        :type request: cgi.CgiRequest
        :type wantedpack: WantedPack
        :rtype: cgi.CgiModule
        """
        if len(request.path_info) != 3:
            raise cgi.CgiError404

        elif request.path_info[0] != 'list':
            raise cgi.CgiError404

        elif '*' in request.path_info[1]:
            eventlog.log_d('fail: http400')
            raise cgi.CgiError400

        elif '*' in request.path_info[2]:
            eventlog.log_d('fail: http400')
            raise cgi.CgiError400

        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            try:
                reply_jnode = wantedpack.delete_watch_list(request.path_info[1], request.path_info[2])
            except wanted.WantedError:
                raise cgi.CgiError404

            return cgi.CgiReply(body=reply_jnode)

        return cls(handler)

    @classmethod
    def _wanted__put_list(cls, request, wantedpack):
        """
        :type request: cgi.CgiRequest
        :type wantedpack: WantedPack
        :rtype: cgi.CgiModule
        """

        if len(request.path_info) != 1:
            raise cgi.CgiError404

        elif request.path_info[0] != 'list':
            raise cgi.CgiError404

        request.assert_query_absent()

        list_owner = os.getenv('LIST_OWNER')
        list_name = os.getenv('LIST_NAME')
        list_path = os.getenv('LIST_PATH')
        prof_path = os.getenv('PROF_PATH')
        prof_name = os.getenv('PROF_NAME', 'default')
        comment = os.getenv('LIST_COMMENT')

        prof_same = os.getenv('PROF_SAME', _LOAD_ACTION_NONE)
        prof_new = os.getenv('PROF_NEW', _LOAD_ACTION_NONE)
        prof_old = os.getenv('PROF_OLD', _LOAD_ACTION_NONE)

        if prof_same not in [_LOAD_ACTION_CP, _LOAD_ACTION_RM, _LOAD_ACTION_NONE]:
            eventlog.log_d('fail: http400')
            raise cgi.CgiError400

        if prof_new not in [_LOAD_ACTION_CP, _LOAD_ACTION_NONE]:
            eventlog.log_d('fail: http400')
            raise cgi.CgiError400

        if prof_old not in [_LOAD_ACTION_RM, _LOAD_ACTION_NONE]:
            eventlog.log_d('fail: http400')
            raise cgi.CgiError400

        prof_tpl = None
        if prof_path:
            if os.path.exists(prof_path):
                prof_tpl = fs.load_json3(prof_path)

        prof_same = _API_TO_DB_ACTION_MAP[prof_same]
        prof_new = _API_TO_DB_ACTION_MAP[prof_new]
        prof_old = _API_TO_DB_ACTION_MAP[prof_old]

        if None in [list_owner, list_name, list_path]:
            eventlog.log_i('fail: http400')
            raise cgi.CgiError400

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            reply_jnode = {}
            try:
                wantedpack.load_watch_list(list_owner, list_name, list_path,
                                           prof_tpl, prof_name, comment,
                                           prof_same, prof_new, prof_old)
            except wanted.WantedError:
                raise cgi.CgiError400

            return cgi.CgiReply(body=reply_jnode)

        return cls(handler)

    @classmethod
    def _wanted__delete_profiles_ref(cls, request, wantedpack):
        """
        :type request: cgi.CgiRequest
        :type wantedpack: WantedPack
        :rtype: cgi.CgiModule
        """
        if len(request.path_info) != 2:
            raise cgi.CgiError404

        elif request.path_info[0] != 'profiles':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            try:
                reply_jnode = wantedpack.drop_profile(request.path_info[1])
            except wanted.WantedError:
                raise cgi.CgiError404

            return cgi.CgiReply(body=reply_jnode)

        return cls(handler)

    @classmethod
    def _wanted__delete_list_owner_name_log__cleanup(cls, request, wantedpack):
        """
        :type request: cgi.CgiRequest
        :type wantedpack: WantedPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo
            :raise CgiError:
            """
            if len(path_info) != 4:
                raise cgi.CgiError404

            if path_info[0] != 'list':
                raise cgi.CgiError404

            if path_info[3] != 'log':
                raise cgi.CgiError404

        def check_args():
            if request.body:
                eventlog.log_d('fail: http400')
                raise cgi.CgiError400

            if request.query_string is None:
                eventlog.log_d('fail: http400')
                raise cgi.CgiError400

            if not request.query_string.startswith('?'):
                eventlog.log_d('fail: http400')
                raise cgi.CgiError400

            if '*' in request.path_info[1] or '*' in request.path_info[2]:
                eventlog.log_d('fail: hppt400')
                raise cgi.CgiError400

        check_path_info(request.path_info)
        check_args()

        max_time = _get_time(request.query_string, 'max-time')
        if max_time is None:
            raise cgi.CgiError404

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            rc = cgi.R200_SUCCESS
            try:
                wantedpack.cleanup_list(request.path_info[1], request.path_info[2], max_time)
            except wanted.WantedError:
                rc = cgi.R404_NOT_FOUND

            jnode = wantedpack.get()
            return cgi.CgiReply(status=rc, body=jnode)
        # handler()

        return cls(handler)

    @classmethod
    def _wanted__delete_list_owner_name_log__rollback(cls, request, wantedpack):
        """
        :type request: cgi.CgiRequest
        :type wantedpack: WantedPack
        :rtype: cgi.CgiModule
        """
        if len(request.path_info) != 4:
            raise cgi.CgiError404

        if request.path_info[0] != 'list':
            raise cgi.CgiError404

        if request.path_info[3] != 'log':
            raise cgi.CgiError404

        def check_args():
            if request.body:
                eventlog.log_d('fail: http400')
                raise cgi.CgiError400

            if request.query_string is None:
                eventlog.log_d('fail: http400')
                raise cgi.CgiError400

            if not request.query_string.startswith('?'):
                eventlog.log_d('fail: http400')
                raise cgi.CgiError400

            if '*' in request.path_info[1] or '*' in request.path_info[2]:
                eventlog.log_d('fail: hppt400')
                raise cgi.CgiError400

        check_args()

        min_time = _get_time(request.query_string, 'min-time')
        if min_time is None:
            raise cgi.CgiError404

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            rc = cgi.R200_SUCCESS
            try:
                # do rollback
                wantedpack.revert_list(request.path_info[1], request.path_info[2], min_time)
            except wanted.WantedError:
                rc = cgi.R404_NOT_FOUND

            jnode = wantedpack.get()
            return cgi.CgiReply(status=rc, body=jnode)

        return cls(handler)

    @classmethod
    def _wanted__get_profiles_ref(cls, request, wantedpack):
        """
        :type request: cgi.CgiRequest
        :type wantedpack: WantedPack
        :rtype: cgi.CgiModule
        """
        if len(request.path_info) != 2:
            raise cgi.CgiError404

        elif request.path_info[0] != 'profiles':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            try:
                reply_jnode = wantedpack.load_profile(request.path_info[1])
            except wanted.WantedError:
                raise cgi.CgiError404

            return cgi.CgiReply(body=reply_jnode)

        return cls(handler)

    @classmethod
    def _wanted__put_profiles_ref(cls, request, wantedpack):
        """
        :type request: cgi.CgiRequest
        :type wantedpack: WantedPack
        :rtype: cgi.CgiModule
        """

        if len(request.path_info) != 2:
            raise cgi.CgiError404

        elif request.path_info[0] != 'profiles':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_present()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            try:
                reply_jnode = wantedpack.save_profile(
                    request.path_info[1], request.body_jnode['args']['template'])
            except wanted.WantedError:
                raise cgi.CgiError404

            return cgi.CgiReply(body=reply_jnode)

        return cls(handler)


def _main():
    wanted_pack = wanted.init()
    try:
        cgi.run_cgi(lambda arg: _Module.factory(arg, wanted_pack))
    finally:
        wanted.fini(wanted_pack)


if __name__ == '__main__':
    eventlog.open_log('cgi-sys-db-wanted', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
