import os
import iso8601

from edge import eventlog
from edge.webserver import cgi
from edge.webserver.systest.api import (init as init_testpack, fini as fini_testpack)
from edge.webserver.systest.test import TestError, TestArgsError, TestLimitsError


class _PathInfo(cgi.CgiPathInfo):
    def __init__(self, raw_path_info):
        """
        :type raw_path_info: str
        """
        cgi.CgiPathInfo.__init__(self, raw_path_info)

    @property
    def session_ref(self):
        return self[1]

    @property
    def test_type(self):
        return self[3]

    @property
    def test_run(self):
        return self[4]

    @property
    def test_ref(self):
        return self[3], self[4]

    @property
    def test_step_ref(self):
        return self[6]


class _Module(cgi.CgiModule):
    """
    GET     /api/v2/sys/test
    PUT     /api/v2/sys/test  /results
    POST    /api/v2/sys/test  /results/report
    DELETE  /api/v2/sys/test  /results/report/:ref
    POST    /api/v2/sys/test  /sessions
    PUT     /api/v2/sys/test  /sessions/:session
    DELETE  /api/v2/sys/test  /sessions/:session

    POST    /api/v2/sys/test  /sessions/:session/tests
    GET     /api/v2/sys/test  /sessions/:session/tests/:test/:run
    PUT     /api/v2/sys/test  /sessions/:session/tests/:test/:run
    DELETE  /api/v2/sys/test  /sessions/:session/tests/:test/:run

    POST    /api/v2/sys/test  /sessions/:session/tests/:test/:run/log
    PUT     /api/v2/sys/test  /sessions/:session/tests/:test/:run/log/all
    PUT     /api/v2/sys/test  /sessions/:session/tests/:test/:run/log/:item
    DELETE  /api/v2/sys/test  /sessions/:session/tests/:test/:run/log
    DELETE  /api/v2/sys/test  /sessions/:session/tests/:test/:run/log/:item
    """
    def __init__(self, handler):
        """
        :type handler: () -> cgi.CgiReply
        """
        cgi.CgiModule.__init__(self, handler)

    @classmethod
    def unpack(cls, request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule

        :raise cgi.CgiError:
        """
        method_to_handlers_map = {
            'GET': [
                cls._get,
                cls._get_sessions_session_tests_test_run
            ],
            'POST': [
                cls._post_sessions,
                cls._post_report,
                cls._post_sessions_session_tests,
                cls._post_sessions_session_tests_test_run_log
            ],
            'PUT': [
                cls._put_sessions_session,
                cls._put_test_results,
                cls._put_sessions_session_tests_test_run,
                cls._put_sessions_session_tests_test_run_log_all,
                cls._put_sessions_session_tests_test_run_log_item
            ],
            'DELETE': [
                cls._delete_report,
                cls._delete_sessions_session,
                cls._delete_sessions_session_tests_test_run,
                cls._delete_sessions_session_tests_test_run_log_item,
                cls._delete_sessions_session_tests_test_run_log
            ],
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request, testpack)

            except cgi.CgiError400:
                raise

            except cgi.CgiError404:
                pass

        eventlog.log_i('fail: http404')
        raise cgi.CgiError404

    @staticmethod
    def _get(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo
            :raise cgi.CgiError:
            """
            if len(path_info) != 0:
                raise cgi.CgiError404

        check_path_info(request.path_info)
        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            reply_jnode = testpack.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _post_report(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 2:
                raise cgi.CgiError404

            if path_info[0] != 'results':
                raise cgi.CgiError404

            if path_info[1] != 'report':
                raise cgi.CgiError404

        def check_args():
            """
            :rtype: str
            """
            request.assert_query_absent()

            try:
                tmp_path = os.environ['BLOB_PATH']
            except KeyError:
                eventlog.log_e('fail: http500 (env.BLOB_PATH not found)')
                raise cgi.CgiError500

            return tmp_path

        check_path_info(request.path_info)
        blob_filepath = check_args()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            testpack.add_report(blob_filepath)

            reply_jnode = testpack.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _delete_report(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 3:
                raise cgi.CgiError404

            if path_info[0] != 'results':
                raise cgi.CgiError404

            if path_info[1] != 'report':
                raise cgi.CgiError404

        check_path_info(request.path_info)
        request.assert_query_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            try:
                testpack.delete_report(request.path_info[2])

                reply_jnode = testpack.pack()
                return cgi.CgiReply(body=reply_jnode)

            except TestError:
                eventlog.log_e('fail: delete report')
                raise cgi.CgiError500

        return _Module(handler)

    @staticmethod
    def _post_sessions(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 1:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'sessions':
                raise cgi.CgiError404

        check_path_info(request.path_info)
        request.assert_query_absent()

        auto_tests = False
        if request.body:
            auto_tests = request.body_jnode.get('tests', False)
            if type(auto_tests) != bool:
                eventlog.log_i('fail: http400')
                raise cgi.CgiError400

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            try:
                session = testpack.add_session()
                if auto_tests:
                    session.add_tests()

                reply_jnode = testpack.pack()
                return cgi.CgiReply(body=reply_jnode)

            except TestLimitsError:
                eventlog.log_i('fail: http409')
                raise cgi.CgiError409

            except TestError:
                eventlog.log_e('fail: add session')
                raise cgi.CgiError500

        return _Module(handler)

    @staticmethod
    def _put_sessions_session(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 2:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'sessions':
                raise cgi.CgiError404

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise cgi.CgiError404

        check_path_info(request.path_info)
        request.assert_query_absent()
        request.assert_body_present()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                eventlog.log_i('fail: http404')
                raise cgi.CgiError404

            session.set_summary(request.body_jnode)

            reply_jnode = testpack.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _put_test_results(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 1:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'results':
                raise cgi.CgiError404

        def check_args():
            request.assert_query_absent()

            if request.body is not None:
                certificate_jnode = request.body_jnode.get('certificate')
                if certificate_jnode:
                    ref_number = certificate_jnode.get('ref-number')
                    max_time = certificate_jnode.get('max-time')
                    try:
                        iso8601.parse_date(max_time)

                    except iso8601.ParseError as x:
                        eventlog.log_i('fail: <{}>'.format(x.message))
                        raise cgi.CgiError400

                    if None in [ref_number, max_time]:
                        raise cgi.CgiError400

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            testpack.add_results(request.body_jnode if request.body else None)
            reply_jnode = testpack.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _delete_sessions_session(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 2:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'sessions':
                raise cgi.CgiError404

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise cgi.CgiError404

        check_path_info(request.path_info)
        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            path_info = request.path_info  # type: _PathInfo
            try:
                if not testpack.del_session(path_info.session_ref):
                    eventlog.log_i("fail: http404")
                    raise cgi.CgiError404

            except TestError:
                eventlog.log_e('fail: del session')
                raise cgi.CgiError500

            reply_jnode = testpack.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _post_sessions_session_tests(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 3:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'sessions':
                raise cgi.CgiError404

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise cgi.CgiError404

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise cgi.CgiError404

        check_path_info(request.path_info)
        request.assert_query_absent()
        request.assert_body_present()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                eventlog.log_i('fail: http404')
                raise cgi.CgiError404

            session.add_test(request.body_jnode)

            reply_jnode = testpack.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _get_sessions_session_tests_test_run(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo
            :raise cgi.CgiError:
            """
            if len(path_info) != 5:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'sessions':
                raise cgi.CgiError404

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise cgi.CgiError404

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise cgi.CgiError404

        check_path_info(request.path_info)
        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)

            test = session.get_test(path_info.test_ref)
            if test is None:
                eventlog.log_i('fail: http404')
                raise cgi.CgiError404

            reply_jnode = test.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _put_sessions_session_tests_test_run(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 5:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'sessions':
                raise cgi.CgiError404

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise cgi.CgiError404

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise cgi.CgiError404

        def check_args():
            request.assert_query_absent()

            body_jnode = request.body_jnode if request.body else None
            if body_jnode:
                summary_jnode = body_jnode.get('summary') or dict()
                if summary_jnode.get('status') not in ['failed', 'passed']:
                    eventlog.log_i('fail: http400')
                    raise cgi.CgiError400

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)

            test = session.get_test(path_info.test_ref)
            if test is None:
                eventlog.log_i('fail: http404')
                raise cgi.CgiError404

            test.set_summary(request.body_jnode if request.body else None)
            reply_jnode = test.pack()

            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _delete_sessions_session_tests_test_run(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 5:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'sessions':
                raise cgi.CgiError404

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise cgi.CgiError404

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise cgi.CgiError404

        check_path_info(request.path_info)
        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            session.del_test(path_info.test_ref)

            reply_jnode = testpack.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _post_sessions_session_tests_test_run_log(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo
            :raise cgi.CgiError:
            """
            if len(path_info) != 6:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'sessions':
                raise cgi.CgiError404

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise cgi.CgiError404

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run/log
            if path_info[5] != 'log':
                raise cgi.CgiError404

        check_path_info(request.path_info)
        request.assert_query_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            test = session.get_test(path_info.test_ref)

            if test is None:
                eventlog.log_i('fail: http404')
                raise cgi.CgiError404

            try:
                test.add_step({
                    'args': request.body_jnode if request.body else None
                })

            except TestArgsError:
                eventlog.log_i('fail: http400')
                raise cgi.CgiError400

            except TestLimitsError:
                eventlog.log_i('fail: http409')
                raise cgi.CgiError409

            except TestError:
                eventlog.log_i('fail: http500')
                raise cgi.CgiError500

            reply_jnode = test.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _put_sessions_session_tests_test_run_log_item(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 7:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'sessions':
                raise cgi.CgiError404

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise cgi.CgiError404

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run/log
            if path_info[5] != 'log':
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run/log/:item
            test_step = test.get_step(path_info.test_step_ref)
            if test_step is None:
                raise cgi.CgiError404

        check_path_info(request.path_info)
        request.assert_query_absent()
        request.assert_body_present()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise cgi.CgiError404

            step = test.get_step(path_info.test_step_ref)
            if step is None:
                raise cgi.CgiError404

            step.edit(request.body_jnode)

            reply_jnode = test.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _put_sessions_session_tests_test_run_log_all(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 7:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'sessions':
                raise cgi.CgiError404

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise cgi.CgiError404

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run/log
            if path_info[5] != 'log':
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run/log/all
            if path_info.test_step_ref != 'all':
                raise cgi.CgiError404

        report = False

        def check_args():
            request.assert_query_absent()

            body_jnode = request.body_jnode if request.body else dict()
            add_report = body_jnode.get('report', False)
            return add_report

        check_path_info(request.path_info)
        report = check_args()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise cgi.CgiError404

            if not test.set_steps_report(report):
                raise cgi.CgiError404

            reply_jnode = test.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _delete_sessions_session_tests_test_run_log_item(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 7:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'sessions':
                raise cgi.CgiError404

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise cgi.CgiError404

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run/log
            if path_info[5] != 'log':
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run/log/:item
            test_step = test.get_step(path_info.test_step_ref)
            if test_step is None:
                raise cgi.CgiError404

        check_path_info(request.path_info)
        request.assert_query_absent()
        request.assert_body_absent()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)

            test = session.get_test(path_info.test_ref)
            if test is None:
                raise cgi.CgiError404

            if not test.del_step(path_info.test_step_ref):
                raise cgi.CgiError404

            reply_jnode = test.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _delete_sessions_session_tests_test_run_log(request, testpack):
        """
        :type request: cgi.CgiRequest
        :type testpack: TestPack
        :rtype: cgi.CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise cgi.CgiError:
            """
            if len(path_info) != 6:
                raise cgi.CgiError404

            # /sessions
            if path_info[0] != 'sessions':
                raise cgi.CgiError404

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise cgi.CgiError404

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise cgi.CgiError404

            # /sessions/:session/tests/:test/:run/log
            if path_info[5] != 'log':
                raise cgi.CgiError404

        def unpack_args():
            """
            :rtype: list[str]
            """
            if request.body:
                eventlog.log_i('fail: http400')
                raise cgi.CgiError400

            if not request.query_string:
                return []

            if not request.query_string.startswith('?'):
                eventlog.log_i('fail: http400')
                raise cgi.CgiError400

            from collections import defaultdict
            name_to_value = defaultdict(list)
            for name_value in request.query_string[1:].split('&'):
                name, value = name_value.split('=', 1)
                name_to_value[name] += [value]

            if len(name_to_value) != 1:
                eventlog.log_i('fail: http400')
                raise cgi.CgiError400

            items = name_to_value.get('item', [])
            if not items:
                eventlog.log_i('fail: http400')
                raise cgi.CgiError400

            return items

        check_path_info(request.path_info)
        step_refs = unpack_args()

        def handler():
            """
            :rtype: cgi.CgiReply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise cgi.CgiError404

            if not test.del_steps(step_refs):
                raise cgi.CgiError404

            reply_jnode = test.pack()
            return cgi.CgiReply(body=reply_jnode)

        return _Module(handler)


def _main():
    testpack = init_testpack()
    try:
        cgi.run_cgi(lambda arg: _Module.unpack(arg, testpack), _PathInfo)
    finally:
        fini_testpack(testpack)


if __name__ == '__main__':
    eventlog.open_log('cgi-sys-test', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
