import os
import datetime
import iso8601
import urlparse
import sys

from edge import eventlog
from edge.com import db as edge_db
from edge.exceptions import *
from edge.webserver import cgi
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.uiconf.time import TimeUiconf
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf


_MAX_REPORT_DAYS = 14


class ReportError(RuntimeError):
    pass


def _fail_param(param_key):
    """
    :type param_key: str
    :rtype: str
    """
    return 'fail: param <%s> has invalid value' % param_key


def _fail_param2(param_key, reason):
    """
    :type param_key: str
    :type reason: str
    :rtype: str
    """
    return 'fail: param <%s> has invalid value (%s)' % (param_key, reason)


def _normalize_iso_datetime(iso_datetime):
    """
    :type iso_datetime: str
    :rtype: (str, int)
    :raises: ReportError
    """
    # define tz-offset
    aware_dt = iso8601.parse_date(iso_datetime, default_timezone=None)
    if aware_dt.tzinfo is None:
        # in case of 'YYYY-MM-DDThh:mm:ss' use edge-tz-offset
        dt_tzinfo = TimeUiconf.get_tz()
        if dt_tzinfo is None:
            eventlog.log_e('fail: define tz-info')
            raise ReportError
    else:
        # in case of 'YYYY-MM-DDThh:mm:ss<tz-offset>' use supplied tz-offset
        dt_tzinfo = aware_dt.tzinfo

    tz_offset = dt_tzinfo.utcoffset(aware_dt)
    if tz_offset is None:
        eventlog.log_d('fail: define tz-offset')
        raise ReportError

    # remove tz-offset from iso_datetime if present
    naive_iso_datetime = aware_dt.replace(tzinfo=None).isoformat()

    return naive_iso_datetime, int(tz_offset.total_seconds())


def _deserialize_traffic_flow_args(query_string):
    """
    :type query_string: str
    :rtype: dict, dict, int
    :raises: ReportError
    """
    url_query = urlparse.parse_qs(query_string[1:], strict_parsing=True)

    limits_jnode = dict()
    api_query_jnode = {
        'ordinate': 'none',
        'datetime-step': 'day'
    }

    iso_datetime_min = None
    iso_datetime_max = None
    for key, values in url_query.iteritems():
        # deserialize origin
        if key == 'origin':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            origin = values.pop()
            if len(origin) != 36:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            limits_jnode['origin'] = origin
            api_query_jnode['origin'] = origin
            continue

        # deserialize datetime-max
        if key == 'datetime-max':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            # url parse replases '+' to ' '
            raw_iso_datetime_max = values.pop()
            iso_datetime_max = raw_iso_datetime_max.replace(' ', '+')
            try:
                iso8601.parse_date(iso_datetime_max)
            except iso8601.ParseError as x:
                eventlog.log_d(_fail_param2(key, str(x)))
                raise ReportError

            limits_jnode['datetime-max'] = iso_datetime_max
            continue

        # deserialize datetime-min
        if key == 'datetime-min':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            # url parse replases '+' to ' '
            raw_iso_datetime_min = values.pop()
            iso_datetime_min = raw_iso_datetime_min.replace(' ', '+')
            try:
                iso8601.parse_date(iso_datetime_min)
            except iso8601.ParseError as x:
                eventlog.log_d(_fail_param2(key, str(x)))
                raise ReportError

            limits_jnode['datetime-min'] = iso_datetime_min
            continue

        # deserialize datetime-step
        if key == 'datetime-step':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            datetime_step = values.pop()
            if datetime_step not in ['hour', 'day']:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            limits_jnode['datetime-step'] = datetime_step
            api_query_jnode['datetime-step'] = datetime_step
            continue

        # deserialize plate-stat
        if key == 'plate-stat':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            plate_stat = values.pop()
            if plate_stat not in ['undefined', 'defined', 'any']:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            limits_jnode['plate-stat'] = plate_stat
            api_query_jnode['plate-stat'] = plate_stat
            continue

        # deserialize speed-stat
        if key == 'speed-stat':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            speed_stat = values.pop()
            if speed_stat not in ['undefined', 'defined', 'any']:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            limits_jnode['speed-stat'] = speed_stat
            api_query_jnode['speed-stat'] = speed_stat
            continue

        # deserialize speed-min
        if key == 'speed-min':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            try:
                speed_min = int(values.pop())
            except ValueError:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            if not 0 <= speed_min <= 255:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            limits_jnode['speed-min'] = speed_min
            api_query_jnode['speed-min'] = speed_min
            continue

        # deserialize speed-max
        if key == 'speed-max':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            try:
                speed_max = int(values.pop())
            except ValueError:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            if not 0 <= speed_max <= 255:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            limits_jnode['speed-max'] = speed_max
            api_query_jnode['speed-max'] = speed_max
            continue

        # deserialize lane-stat
        if key == 'lane-stat':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            lane_stat = values.pop()
            if lane_stat not in ['undefined', 'defined', 'any']:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            limits_jnode['lane-stat'] = lane_stat
            api_query_jnode['lane-stat'] = lane_stat
            continue

        # deserialize lane
        if key == 'lane':
            lane_jnode = []
            for raw_lane in values:
                if len(raw_lane) != 1:
                    eventlog.log_d(_fail_param(key))
                    raise ReportError

                lane_jnode.append(raw_lane)

            limits_jnode['lane'] = lane_jnode
            api_query_jnode['lane'] = lane_jnode
            continue

        eventlog.log_d('fail: param <%s> is not supported' % key)
        raise ReportError

    # define iso-datetime-max
    if iso_datetime_max is None:
        iso_datetime_max = datetime.datetime.now().isoformat()

    naive_iso_dt_max, api_query_tz_offset = _normalize_iso_datetime(iso_datetime_max)
    api_query_jnode['datetime-max'] = naive_iso_dt_max

    # define iso-datetime-min
    if iso_datetime_min is None:
        datetime_max = iso8601.parse_date(api_query_jnode['datetime-max'])
        datetime_min = datetime_max - datetime.timedelta(days=7)
        iso_datetime_min = datetime_min.isoformat()

    naive_iso_dt_min, api_query_tz_offset = _normalize_iso_datetime(iso_datetime_min)
    api_query_jnode['datetime-min'] = naive_iso_dt_min

    # validate date period
    datetime_max = iso8601.parse_date(api_query_jnode['datetime-max'])
    datetime_min = iso8601.parse_date(api_query_jnode['datetime-min'])
    if datetime_max - datetime_min > datetime.timedelta(days=_MAX_REPORT_DAYS):
        eventlog.log_d('fail: report period is too long')
        raise ReportError

    return limits_jnode, api_query_jnode, api_query_tz_offset


def _deserialize_event_flow_args(query_string):
    """
    :type query_string: str
    :rtype: dict, dict, int
    :raises ReportError
    """
    url_query = urlparse.parse_qs(query_string[1:], strict_parsing=True)

    api_query_jnode = {
        'ordinate': 'type',
        'datetime-step': 'day'
    }

    limits_jnode = dict()

    iso_datetime_min = None
    iso_datetime_max = None

    for key, values in url_query.iteritems():
        # deserialize origin
        if key == 'origin':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            origin = values.pop()
            if len(origin) != 36:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            limits_jnode['origin'] = origin
            api_query_jnode['origin'] = origin
            continue

        # deserialize datetime-max
        if key == 'datetime-max':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            iso_datetime_max = values.pop()
            try:
                iso8601.parse_date(iso_datetime_max)
            except iso8601.ParseError as x:
                eventlog.log_d(_fail_param2(key, str(x)))
                raise ReportError

            limits_jnode['datetime-max'] = iso_datetime_max
            continue

        # deserialize datetime-min
        if key == 'datetime-min':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            iso_datetime_min = values.pop()
            try:
                iso8601.parse_date(iso_datetime_min)
            except iso8601.ParseError as x:
                eventlog.log_d(_fail_param2(key, str(x)))
                raise ReportError

            limits_jnode['datetime-min'] = iso_datetime_min
            continue

        # deserialize datetime-step
        if key == 'datetime-step':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            datetime_step = values.pop()
            if datetime_step not in ['hour', 'day']:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            limits_jnode['datetime-step'] = datetime_step
            api_query_jnode['datetime-step'] = datetime_step
            continue

        # deserialize lane-stat
        if key == 'lane-stat':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            lane_stat = values.pop()
            if lane_stat not in ['undefined', 'defined', 'any']:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            limits_jnode['lane-stat'] = lane_stat
            api_query_jnode['lane-stat'] = lane_stat
            continue

        # deserialize lane
        if key == 'lane':
            lane_jnode = []
            for raw_lane in values:
                if len(raw_lane) != 1:
                    eventlog.log_d(_fail_param(key))
                    raise ReportError

                lane_jnode.append(raw_lane)

            limits_jnode['lane'] = lane_jnode
            api_query_jnode['lane'] = lane_jnode
            continue

        # deserialize event types
        if key == 'type':
            types_jnode = []
            for event_type in values:
                if event_type not in ['transit', 'parking', 'spot-speed', 'wrong-way',
                                      'red-light', 'wanted', 'section-speed', 'bus-lane']:
                    eventlog.log_d(_fail_param(key))
                    raise ReportError

                types_jnode.append(event_type)

            limits_jnode['type'] = types_jnode
            api_query_jnode['type'] = types_jnode
            continue

        eventlog.log_d('fail: param <%s> is not supported' % key)
        raise ReportError

    # define iso-datetime-max
    if iso_datetime_max is None:
        iso_datetime_max = datetime.datetime.now().isoformat()

    naive_iso_dt_max, api_query_tz_offset = _normalize_iso_datetime(iso_datetime_max)
    api_query_jnode['datetime-max'] = naive_iso_dt_max

    # define iso-datetime-min
    if iso_datetime_min is None:
        datetime_max = iso8601.parse_date(api_query_jnode['datetime-max'])
        datetime_min = datetime_max - datetime.timedelta(days=7)
        iso_datetime_min = datetime_min.isoformat()

    naive_iso_dt_min, api_query_tz_offset = _normalize_iso_datetime(iso_datetime_min)
    api_query_jnode['datetime-min'] = naive_iso_dt_min

    # validate date period
    datetime_max = iso8601.parse_date(api_query_jnode['datetime-max'])
    datetime_min = iso8601.parse_date(api_query_jnode['datetime-min'])
    if datetime_max - datetime_min > datetime.timedelta(days=_MAX_REPORT_DAYS):
        eventlog.log_d('fail: report period is too long')
        raise ReportError

    return limits_jnode, api_query_jnode, api_query_tz_offset


def _get_report(api_query_jnode, api_query_tz_offset):
    """
    :type api_query_jnode: dict
    :type api_query_tz_offset: int
    :rtype: dict
    :raises: ReportError
    """
    dbapi_query_jnode = {
        'name': 'plate.traffic.combined',
        'param': api_query_jnode
    }

    db_args = edge_db.DbArgs.deserialize2()

    def db_handler(db_link):
        """
        :rtype: dict
        :raises: edge_db.DbError
        :raises: ReportError
        """
        with db_link, db_link.cursor() as db_cursor:
            db_cursor.execute(
                "SET LOCAL TIME ZONE INTERVAL '%sS';select errcode, errmsg, result from api_report_get(%s)",
                [api_query_tz_offset, edge_db.DbJsonArg(dbapi_query_jnode)])

            errcode, errmsg, api_reply_jnode = db_cursor.fetchone()
            if not api_reply_jnode or errcode != 0:
                eventlog.log_d('fail: api_report_get (%s)' % errmsg)
                raise ReportError

            report_jnode = api_reply_jnode
            return report_jnode

    try:
        return edge_db.with_db(db_args, db_handler)

    except edge_db.DbError:
        eventlog.log_d('fail: db')
        raise ReportError


def _get_limits(require_types):
    """
    :type require_types: bool
    :rtype: dict 
    """
    origins_jnode = []
    limits_jnode = {
        'origins': origins_jnode
    }

    root_info = ph.RootInfoUiconf.get_info()
    term_infos = ph.InfoUiconf.map_units_r(root_info, unit_types=[ph.UI_TYPE_TERMINAL])
    for term_info in term_infos:
        term_uimod = ph.UiMod.create(term_info['ref'])
        term_uiconf = TerminalUiconf(term_uimod.conf_path)
        term_uuid = term_uiconf.get_uuid()
        origin_jnode = {
            'uuid': term_uuid,
            'name': term_uiconf.get_name(default=term_uimod.pid),
            'lanes': _get_lanes(term_uimod.pid)  # TODO: fix this function
        }

        ra_info = ph.InfoUiconf.find_unit_r(term_info, unit_type=ph.UI_TYPE_RADAR)
        # event-feed limits
        if require_types:
            origin_jnode['types'] = ph.RootInfoUiconf.get_term_event_types(term_info)

        # transit-feed limits
        elif ra_info is not None:
            origin_jnode['speed-max'] = 250

        origins_jnode.append(origin_jnode)

    return limits_jnode


def _get_lanes(pid):
    """
    :param pid: str 
    :rtype: list[int]
    """
    ui_roadway = ph.UiMod(pid, ph.UI_TYPE_ROADWAY, '0')
    ui_jnode = ui_roadway.load_conf()
    lanes_jnode = ui_jnode['rw']['args']['lanes']
    lane_list = map(lambda lane: lane['name'], lanes_jnode)
    lane_list.sort()

    return lane_list


def _run(http_method, query_string, path_info):
    """
    :type http_method: str
    :type query_string: str
    :type path_info: list of str
    :rtype: cgi.CgiReply
    """

    # validate url-path
    valid_url_path_list = [
        ['transit-flow'],
        ['transit-flow', 'limits'],
        ['event-flow'],
        ['event-flow', 'limits']
    ]

    if path_info not in valid_url_path_list:
        eventlog.log_d("fail: not-found")
        return cgi.CgiReply(cgi.R404_NOT_FOUND)

    # validate http-method
    if http_method != 'GET':
        eventlog.log_d('fail: method-not-allowed (%s)' % http_method)
        return cgi.CgiReply(cgi.R405_METHOD_NOT_ALLOWED)

    # reply report-limits
    if len(path_info) == 2:
        try:
            if path_info[0] == 'transit-flow':
                reply_jnode = _get_limits(require_types=False)

            elif path_info[0] == 'event-flow':
                reply_jnode = _get_limits(require_types=True)

            else:
                eventlog.log_d('fail: report is not supported')
                raise ReportError

            return cgi.CgiReply(cgi.R200_SUCCESS, reply_jnode)

        except ReportError:
            eventlog.log_d('fail: get report-limits')
            raise

    # define report-args
    try:
        if path_info[0] == 'transit-flow':
            limits_jnode, api_query_jnode, api_query_tz_offset \
                = _deserialize_traffic_flow_args(query_string)

        elif path_info[0] == 'event-flow':
            limits_jnode, api_query_jnode, api_query_tz_offset = \
                _deserialize_event_flow_args(query_string)

        else:
            eventlog.log_d('fail: report is not supported')
            raise ReportError

    except ReportError:
        eventlog.log_d('fail: bad-request (deserialize query-string)')
        return cgi.CgiReply(cgi.R400_BAD_REQUEST)

    # reply report
    try:
        report_jnode = _get_report(api_query_jnode, api_query_tz_offset)
        report_jnode['limits'] = limits_jnode
        return cgi.CgiReply(cgi.R200_SUCCESS, report_jnode)

    except ReportError:
        eventlog.log_d('fail: get-report')
        raise


# noinspection PyBroadException
def _main():
    reply = cgi.CgiError500().reply
    try:
        http_method = os.environ['REQUEST_METHOD']
        raw_path_info = os.environ['PATH_INFO']
        query_string = os.environ['QUERY_STRING']
        eventlog.log_d("%s|%s|%s" % (http_method, raw_path_info, query_string))

        path_info = raw_path_info.split('/')[1:]
        reply = _run(http_method, query_string, path_info)

    except ReportError:
        pass

    except ParamsError:
        reply = cgi.CgiError400().reply

    except NotFoundError:
        reply = cgi.CgiError404().reply

    except BusyError:
        reply = cgi.CgiError409().reply

    except CamServiceError:
        reply = cgi.CgiCamError().reply

    except LicServiceError:
        reply = cgi.CgiLicError().reply

    except ServiceError:
        reply = cgi.CgiError503().reply

    except cgi.CgiError as x:
        reply = x.reply

    except EdgeError:
        pass

    except Exception:
        eventlog.log_x()

    finally:
        reply.send()

    sys.exit(0)


if __name__ == '__main__':
    eventlog.open_log('eventstore-traffic', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
