import os
import datetime
import iso8601
import urlparse
import sys

from edge import eventlog
from edge.com import db as edge_db
from edge.exceptions import *
from edge.webserver import cgi
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.uiconf.time import TimeUiconf
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf


_MAX_REPORT_DAYS = 14


class ReportError(RuntimeError):
    pass


def _fail_param(param_key):
    """
    :type param_key: str
    :rtype: str
    """
    return 'fail: param <%s> has invalid value' % param_key


def _fail_param2(param_key, reason):
    """
    :type param_key: str
    :type reason: str
    :rtype: str
    """
    return 'fail: param <%s> has invalid value (%s)' % (param_key, reason)


def _normalize_iso_datetime(iso_datetime):
    """
    :type iso_datetime: str
    :rtype: (str, int)
    :raises: ReportError
    """
    # define tz-offset
    aware_dt = iso8601.parse_date(iso_datetime, default_timezone=None)
    if aware_dt.tzinfo is None:
        # in case of 'YYYY-MM-DDThh:mm:ss' use edge-tz-offset
        dt_tzinfo = TimeUiconf.get_tz()
        if dt_tzinfo is None:
            eventlog.log_e('fail: define tz-info')
            raise ReportError
    else:
        # in case of 'YYYY-MM-DDThh:mm:ss<tz-offset>' use supplied tz-offset
        dt_tzinfo = aware_dt.tzinfo

    tz_offset = dt_tzinfo.utcoffset(aware_dt)
    if tz_offset is None:
        eventlog.log_d('fail: define tz-offset')
        raise ReportError

    # remove tz-offset from iso_datetime if present
    naive_iso_datetime = aware_dt.replace(tzinfo=None).isoformat()

    return naive_iso_datetime, int(tz_offset.total_seconds())


def _deserialize_transit_rate2_args(query_string):
    """
    :type query_string: str
    :rtype: dict, dict, int
    :raises: ReportError
    """
    url_query = urlparse.parse_qs(query_string[1:], strict_parsing=True)

    limits_jnode = dict()
    api_query_jnode = {
        'ordinate': 'lane.category.speed'
    }

    iso_datetime_min = None
    iso_datetime_max = None
    for key, values in url_query.iteritems():
        # deserialize origin
        if key == 'origin':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            origin = values.pop()
            if len(origin) != 36:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            limits_jnode['origin'] = origin
            api_query_jnode['origin'] = origin
            continue

        # deserialize datetime-max
        if key == 'datetime-max':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            # url parse replases '+' to ' '
            raw_iso_datetime_max = values.pop()
            iso_datetime_max = raw_iso_datetime_max.replace(' ', '+')
            try:
                iso8601.parse_date(iso_datetime_max)
            except iso8601.ParseError as x:
                eventlog.log_d(_fail_param2(key, str(x)))
                raise ReportError

            limits_jnode['datetime-max'] = iso_datetime_max
            continue

        # deserialize datetime-min
        if key == 'datetime-min':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            # url parse replases '+' to ' '
            raw_iso_datetime_min = values.pop()
            iso_datetime_min = raw_iso_datetime_min.replace(' ', '+')
            try:
                iso8601.parse_date(iso_datetime_min)
            except iso8601.ParseError as x:
                eventlog.log_d(_fail_param2(key, str(x)))
                raise ReportError

            limits_jnode['datetime-min'] = iso_datetime_min
            continue

        # deserialize lane-stat
        if key == 'lane-stat':
            if len(values) != 1:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            lane_stat = values.pop()
            if lane_stat not in ['undefined', 'defined', 'any']:
                eventlog.log_d(_fail_param(key))
                raise ReportError

            limits_jnode['lane-stat'] = lane_stat
            api_query_jnode['lane-stat'] = lane_stat
            continue

        # deserialize lane
        if key == 'lane':
            lane_jnode = []
            for raw_lane in values:
                if len(raw_lane) != 1:
                    eventlog.log_d(_fail_param(key))
                    raise ReportError

                lane_jnode.append(raw_lane)

            limits_jnode['lane'] = lane_jnode
            api_query_jnode['lane'] = lane_jnode
            continue

        eventlog.log_d('fail: param <%s> is not supported' % key)
        raise ReportError

    if limits_jnode.get('lane-stat', 'any') == 'undefined':
        limits_jnode.pop('lane', None)

    # define iso-datetime-max
    if iso_datetime_max is None:
        iso_datetime_max = datetime.datetime.now().isoformat()

    naive_iso_dt_max, api_query_tz_offset = _normalize_iso_datetime(iso_datetime_max)
    api_query_jnode['datetime-max'] = naive_iso_dt_max

    # define iso-datetime-min
    if iso_datetime_min is None:
        datetime_max = iso8601.parse_date(api_query_jnode['datetime-max'])
        datetime_min = datetime_max - datetime.timedelta(days=7)
        iso_datetime_min = datetime_min.isoformat()

    naive_iso_dt_min, api_query_tz_offset = _normalize_iso_datetime(iso_datetime_min)
    api_query_jnode['datetime-min'] = naive_iso_dt_min

    # validate date period
    datetime_max = iso8601.parse_date(api_query_jnode['datetime-max'])
    datetime_min = iso8601.parse_date(api_query_jnode['datetime-min'])
    if datetime_max - datetime_min > datetime.timedelta(days=_MAX_REPORT_DAYS):
        eventlog.log_d('fail: report period is too long')
        raise ReportError

    return limits_jnode, api_query_jnode, api_query_tz_offset


def _get_report(api_query_jnode, api_query_tz_offset):
    """
    :type api_query_jnode: dict
    :type api_query_tz_offset: int
    :rtype: dict
    :raises: ReportError
    """
    # return {
    #     "limits": {
    #         "origin": "69e86fa7-e1ad-4e68-96b7-b910f40bdb49",
    #         "datetime-min": "2016-06-14T10:52:00.000000+00:00",
    #         "datetime-max": "2016-06-14T12:52:00.000000+00:00",
    #         "lanes": []
    #     },
    #     "items":
    #         [
    #             {
    #                 "lane": "1",
    #                 "count": 153,
    #                 "items":
    #                     [
    #                         {
    #                             "class": "CAR",
    #                             "count": 115,
    #                             "speed": 45
    #                         },
    #                         {
    #                             "class": "BUS",
    #                             "count": 38,
    #                             "speed": 21
    #                         }
    #                     ]
    #             }
    #         ]
    # }

    dbapi_query_jnode = {
        'name': 'plate.traffic.combined',
        'param': api_query_jnode
    }

    db_args = edge_db.DbArgs.deserialize2()

    def db_handler(db_link):
        """
        :rtype: dict
        :raises: edge_db.DbError
        :raises: ReportError
        """
        with db_link, db_link.cursor() as db_cursor:
            db_cursor.execute(
                "SET LOCAL TIME ZONE INTERVAL '%sS';select errcode, errmsg, result from api_report_get(%s)",
                [api_query_tz_offset, edge_db.DbJsonArg(dbapi_query_jnode)])

            errcode, errmsg, api_reply_jnode = db_cursor.fetchone()
            if not api_reply_jnode or errcode != 0:
                eventlog.log_d('fail: api_report_get (%s)' % errmsg)
                raise ReportError

            report_jnode = api_reply_jnode
            return report_jnode

    try:
        return edge_db.with_db(db_args, db_handler)

    except edge_db.DbError:
        eventlog.log_d('fail: db')
        raise ReportError


def _get_limits():
    """
    :rtype: dict 
    """
    origins_jnode = []
    limits_jnode = {
        'origins': origins_jnode
    }

    root_info = ph.RootInfoUiconf.get_info()
    term_infos = ph.InfoUiconf.map_units_r(root_info, unit_types=[ph.UI_TYPE_TERMINAL])
    for term_info in term_infos:
        term_uimod = ph.UiMod.create(term_info['ref'])
        term_uiconf = TerminalUiconf(term_uimod.conf_path)
        term_uuid = term_uiconf.get_uuid()
        origin_jnode = {
            'uuid': term_uuid,
            'name': term_uiconf.get_name(default=term_uimod.pid),
            'lanes': _get_lanes(term_uimod.pid)  # TODO: fix this function
        }

        origins_jnode.append(origin_jnode)

    return limits_jnode


def _get_lanes(pid):
    """
    :param pid: str 
    :rtype: list[int]
    """
    ui_roadway = ph.UiMod(pid, ph.UI_TYPE_ROADWAY, '0')
    ui_jnode = ui_roadway.load_conf()
    lanes_jnode = ui_jnode['rw']['args']['lanes']
    lane_list = map(lambda lane: lane['name'], lanes_jnode)
    lane_list.sort()

    return lane_list


def _run(http_method, query_string, path_info):
    """
    :type http_method: str
    :type query_string: str
    :type path_info: list of str
    :rtype: cgi.CgiReply
    """

    # validate url-path
    valid_url_path_list = [
        [],
        ['limits']
    ]

    if path_info not in valid_url_path_list:
        eventlog.log_d("fail: not-found")
        return cgi.CgiReply(cgi.R404_NOT_FOUND)

    # validate http-method
    if http_method != 'GET':
        eventlog.log_d('fail: method-not-allowed (%s)' % http_method)
        return cgi.CgiReply(cgi.R405_METHOD_NOT_ALLOWED)

    # reply report-limits
    if len(path_info) == 1:
        try:
            reply_jnode = _get_limits()
            return cgi.CgiReply(cgi.R200_SUCCESS, reply_jnode)

        except ReportError:
            eventlog.log_d('fail: get report-limits')
            raise

    # define report-args
    try:
        limits_jnode, api_query_jnode, api_query_tz_offset \
            = _deserialize_transit_rate2_args(query_string)

    except ReportError:
        eventlog.log_d('fail: bad-request (deserialize query-string)')
        return cgi.CgiReply(cgi.R400_BAD_REQUEST)

    # reply report
    try:
        report_jnode = _get_report(api_query_jnode, api_query_tz_offset)

        for report_item in report_jnode.get('items', []):
            # count = 0
            for lane_item in report_item.get('items', []):
                # count += lane_item['count']
                if lane_item.get('speed') is None:
                    lane_item.pop('speed', None)
            # report_item['count'] = count

        report_jnode['limits'] = limits_jnode
        return cgi.CgiReply(cgi.R200_SUCCESS, report_jnode)

    except ReportError:
        eventlog.log_d('fail: get-report')
        raise


# noinspection PyBroadException
def _main():
    reply = cgi.CgiError500().reply
    try:
        http_method = os.environ['REQUEST_METHOD']
        raw_path_info = os.environ['PATH_INFO']
        query_string = os.environ['QUERY_STRING']
        eventlog.log_d("%s|%s|%s" % (http_method, raw_path_info, query_string))

        path_info = raw_path_info.split('/')[1:]
        reply = _run(http_method, query_string, path_info)

    except ReportError:
        pass

    except ParamsError:
        reply = cgi.CgiError400().reply

    except NotFoundError:
        reply = cgi.CgiError404().reply

    except BusyError:
        reply = cgi.CgiError409().reply

    except CamServiceError:
        reply = cgi.CgiCamError().reply

    except LicServiceError:
        reply = cgi.CgiLicError().reply

    except ServiceError:
        reply = cgi.CgiError503().reply

    except cgi.CgiError as x:
        reply = x.reply

    except EdgeError:
        pass

    except Exception:
        eventlog.log_x()

    finally:
        reply.send()

    sys.exit(0)


if __name__ == '__main__':
    eventlog.open_log('transit-rate2-cgi', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
