import os
import errno
import shutil
import fcntl
import ujson
import semver
import time

from edge import eventlog
from edge.exceptions import FsError, BusyError, InternalError, ParamsError

_export_pil_image_fn = None


def _export_pil_image(rgba_pil_image, filepath):
    """
    :type rgba_pil_image: PIL.Image.Image
    :type filepath: str
    """
    def fn(rgba, jpeg_path):
        """
        :type rgba: PIL.Image.Image
        :type jpeg_path: str
        """
        rgba.save(jpeg_path, 'JPEG', quality=60)

    def fn4(rgba, jpeg_path):
        """
        :type rgba: PIL.Image.Image
        :type jpeg_path: str
        """
        rgb_pil_image = rgba.convert('RGB')
        rgb_pil_image.save(jpeg_path, 'JPEG', quality=60)

    global _export_pil_image_fn
    if not _export_pil_image_fn:
        import PIL
        if semver.compare(PIL.PILLOW_VERSION, "4.2.0") == -1:
            _export_pil_image_fn = fn

        else:
            _export_pil_image_fn = fn4

    return _export_pil_image_fn(rgba_pil_image, filepath)


def silent_mkdir_r(target_path):
    """
    :type target_path: str
    :rtype: None
    """
    if target_path:
        try:
            os.makedirs(target_path)
        except os.error as x:
            if x.errno != errno.EEXIST:
                eventlog.log_w('fail: mkdir {} ({})'.format(
                    target_path, errno.errorcode[x.errno]))


def silent_rm_r(target_path):
    """
    :type target_path: str | unicode
    :rtype: None
    """
    if target_path:
        try:
            os.remove(target_path)
            return
        except EnvironmentError:
            pass

        try:
            shutil.rmtree(target_path, ignore_errors=True)
        except EnvironmentError:
            eventlog.log_x()


def copy_file(src, dst):
    """
    :type src: str
    :type dst: str
    :rtype: None
    """
    shutil.copy2(src, dst)


def move_files(src, dst):
    """
    Moves file or list of files from <src> to <dst>
    :type src: str | list of str
    :type dst: str
    """

    if type(src) == list:
        for src_file in src:
            shutil.move(src_file, dst)
    else:
        shutil.move(src, dst)


def silent_mv(src, dst):
    """
    :type src: str
    :type dst: str

    :raise EnvironmentError:
    """
    try:
        shutil.move(src, dst)
    except EnvironmentError as x:
        if x.errno == errno.EEXIST:
            silent_rm_r(dst)
            silent_mv(src, dst)

        elif x.errno != errno.ENOENT:
            raise


def silent_link(src, dst):
    """
    :type src: str
    :type dst: str

    :raise EnvironmentError:
    """
    for _ in [1, 2]:
        try:
            os.link(src, dst)
            break
        except OSError as x:
            if x.errno == errno.EEXIST:
                silent_rm_r(dst)
                pass

            elif x.errno == errno.ENOENT:
                silent_mkdir_r(os.path.dirname(dst))
                pass


def silent_cp(src, dst):
    """
    :type src: str
    :type dst: str

    :raise EnvironmentError:
    """
    for _ in [1, 2]:
        try:
            shutil.copy(src, dst)
            break
        except IOError as x:
            if x.errno == errno.ENOENT:
                silent_mkdir_r(os.path.dirname(dst))
                pass


def edit_dir(target_dir, fn, *fn_args, **fn_kwargs):
    try_count = 2
    dir_present = True
    while try_count > 0 and dir_present:
        # noinspection PyBroadException
        try:
            fn(*fn_args, **fn_kwargs)
            break

        except IOError as x:
            if x.errno == errno.ENOENT:
                dir_present = False

            else:
                eventlog.log_w('fail: fn (%s)' % x)

        except Exception:
            eventlog.log_x()

        if not dir_present:
            silent_mkdir_r(target_dir)
            dir_present = True

        try_count -= 1

    if not try_count:
        raise RuntimeError


def write_json_file(jnode, filepath, pretty=False):
    """
    :type jnode: dict | list
    :type filepath: str
    :type pretty: bool
    """
    indent = 2 if pretty else 0
    for _ in [1, 2]:
        try:
            with open(filepath, 'w') as json_file:
                ujson.dump(jnode, json_file, ensure_ascii=False, indent=indent)
                break

        except IOError as x:
            if x.errno != errno.ENOENT:
                eventlog.log_e('fail: save {} <{}>'.format(filepath, errno.errorcode[x.errno]))
                break

        silent_mkdir_r(os.path.dirname(filepath))


def write_jpeg_file(rgba_pil_image, filepath):
    """
    :type rgba_pil_image: PIL.Image.Image
    :type filepath: str
    """
    for _ in [1, 2]:
        try:
            _export_pil_image(rgba_pil_image, filepath)
            break

        except IOError as x:
            if x.errno != errno.ENOENT:
                eventlog.log_e('fail: save {} <{}>'.format(filepath, errno.errorcode[x.errno]))
                break

        silent_mkdir_r(os.path.dirname(filepath))


def write_rgba_file(rgba_pil_image, filepath):
    """
    :type rgba_pil_image: PIL.Image.Image
    :type filepath: str
    """
    for _ in [1, 2]:
        try:
            with open(filepath, 'wb') as f:
                f.write(rgba_pil_image.tobytes())
                break

        except IOError as x:
            if x.errno != errno.ENOENT:
                eventlog.log_e('fail: save {} <{}>'.format(filepath, errno.errorcode[x.errno]))
                break

        silent_mkdir_r(os.path.dirname(filepath))


def ln_schema(uischema_target_path, uiconf_path):
    """
    :type uischema_target_path: str
    :type uiconf_path: str
    """
    uiconf = load_json3(uiconf_path)
    uiconf_pri = uiconf['ro']  # type: dict

    uischema_source_path = uiconf_pri['schema-abspath']

    try:
        os.remove(uischema_target_path)
    except OSError as x:
        if x.errno != errno.ENOENT:
            eventlog.log_e('fail: rm {} ({})'.format(uischema_target_path, x.message))
            raise

    eventlog.log_i('exec: ln -s {} {}'.format(uischema_source_path, uischema_target_path))
    try:
        silent_mkdir_r(os.path.dirname(uischema_target_path))
        os.symlink(uischema_source_path, uischema_target_path)
        # eventlog.log_i('done: ln')

    except OSError as x:
        eventlog.log_e('fail: ln ({})'.format(x.message))
        raise


def create_jnode_from_default(defconf_path, conf_path):
    """
    :type defconf_path: str
    :type conf_path: str
    """
    if not defconf_path:
        eventlog.log_e_param('defconf_path')
        raise ValueError

    elif not conf_path:
        eventlog.log_e_param('conf_path')
        raise ValueError

    elif not path_exists(conf_path):
        silent_mkdir_r(os.path.dirname(conf_path))
        copy_file(defconf_path, conf_path)
        eventlog.log_i('done: cp {} {}'.format(defconf_path, conf_path))


def load_json3(filepath, **kwargs):
    """
    :type filepath: str
    :rtype: dict | list

    :raise FsError:
    """
    try:
        try:
            with open(filepath) as json_file:
                jnode = ujson.load(json_file)
                return jnode

        except IOError as x:
            eventlog.log_w('fail: IOError {} <{}>'.format(filepath, errno.errorcode[x.errno]))
            raise FsError

        except ValueError as x:
            eventlog.log_e('fail: ValueError {} <{}>'.format(filepath, x.message))
            raise FsError

    except FsError:
        if 'default' not in kwargs:
            raise

    return kwargs.get('default')


def validate_json(jnode, schema_path):
    """
    :type schema_path: str
    :type jnode: dict
    :rtype: None
    :raises: EdgeError
    """
    schema = load_json3(schema_path, default=None)
    if schema is None:
        eventlog.log_e('fail: InternalError (load schema)')
        raise InternalError

    import jsonschema
    try:
        jsonschema.validate(jnode, schema)

    except jsonschema.ValidationError as x:
        eventlog.log_d('fail: ParamsError ({})'.format(x.message))
        raise ParamsError

    except ValueError as x:
        eventlog.log_d('fail: ParamsError ({})'.format(x.message))
        raise ParamsError


def path_exists(target_path):
    """
    :type target_path: str
    :rtype: bool
    """
    return os.path.exists(target_path)


def join_path(fs_path, *fs_paths):
    return os.path.join(fs_path, *fs_paths)


def get_df(fs_path):
    """
    Query disk free statistics about the given path.

    :rtype: float
    :return: [0.0; 1.0]
    """
    st = os.statvfs(fs_path)
    free = st.f_bavail * st.f_frsize
    total = st.f_blocks * st.f_frsize
    # used = (st.f_blocks - st.f_bfree) * st.f_frsize

    return free / float(total)


# noinspection PyBroadException
def try_lock_pid_file(pid_path):
    """
    :type pid_path: str
    :rtype: bool
    """
    try:
        for retry in range(2):
            try:
                # Request exclusive (EX) non-blocking (NB) advisory lock.
                fd = os.open(pid_path, os.O_CREAT | os.O_WRONLY)
                fcntl.lockf(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                os.write(fd, str(os.getpid()))
                return True

            except os.error as x:
                if x.errno != errno.ENOENT:
                    raise

            try:
                os.makedirs(os.path.dirname(pid_path))
            except Exception:
                pass

    except Exception:
        pass

    return False


def try_rm_pid_file(pid_path):
    """
    :type pid_path: str
    :rtype: bool
    """
    try:
        os.remove(pid_path)
        return True
    except os.error:
        pass

    return False


class Lock(object):
    _RETRY_TOUT = 0.25  # seconds

    def __init__(self, lock_file, try_count=1):
        """
        :type lock_file: str
        :type try_count: int
        """
        self._lock_file = lock_file
        self._try_count = try_count

    def __call__(self, fn, *extra_args, **extra_kwargs):
        return lambda *fn_args, **fn_kwargs: self._exec(fn, *fn_args, **fn_kwargs)

    # noinspection PyBroadException
    def _exec(self, fn, *fn_args, **fn_kwargs):
        retry_count = 1
        while retry_count <= self._try_count:
            if not try_lock_pid_file(self._lock_file):
                eventlog.log_d('fail: lock <{}> (attempt={})'.format(self._lock_file, retry_count))
                time.sleep(self._RETRY_TOUT)
                retry_count += 1
                continue

            try:
                answer = fn(*fn_args, **fn_kwargs)
                return answer

            except Exception:
                eventlog.log_x()
                retry_count += 1
                continue

            finally:
                try_rm_pid_file(self._lock_file)

        eventlog.log_w('fail: lock <{}>'.format(self._lock_file))
        raise BusyError
