import os
import stat
import socket
import ujson

from glob import glob
from edge import eventlog

import commands
UI_NETWORK_JNODE_FILEPATH = '/etc/edge/setupstore/admin/actions/network.json'
MANAGEMENT_MAC_FILEPATH = '/etc/network/management-mac'


class NetError(Exception):
    pass


class NetToutError(NetError):
    pass


def get_es_uds_endpoint():
    """
    :rtype: str
    """
    return '/var/spool/edge/system/es-es-uds'


def get_ifname(mac_address):
    """
    :type mac_address: str
    :rtype: str | None
    """
    mac_files = glob('/sys/class/net/*/address')
    for mac_file in mac_files:
        if_name = mac_file.split('/')[-2]  # like eth0
        with open(mac_file, 'r') as text_file:
            if_mac = text_file.readlines()[0].strip()

        if if_mac == mac_address:
            return if_name


def get_ifmac(if_name):
    """
    :type if_name: str
    :rtype: str | None
    """
    mac_file = '/sys/class/net/{}/address'.format(if_name)
    for mac_file in glob(mac_file):
        with open(mac_file, 'r') as f:
            mac = f.readlines()[0].strip()
            return mac


def get_ra_uds_endpoint(ui_pid):
    """
    :type ui_pid: str
    :rtype: str
    """
    return '/var/spool/edge/pipeline-{}/ra0-ra0-uds'.format(ui_pid)
    # return '/var/spool/edge/pipeline-{}/ra_uds'.format(ui_pid)


def get_lr_uds_endpoint(ui_pid):
    """
    :type ui_pid: str
    :rtype: str
    """
    return '/var/spool/edge/pipeline-{}/lr0-lr0-uds'.format(ui_pid)


def query_uds(uds_endpoint, query_jnode, tout_sec=3):
    """
    :type uds_endpoint: str
    :type query_jnode: dict
    :type tout_sec: int

    :rtype: dict
    :raise NetError:
    """
    try:
        file_mode = os.stat(uds_endpoint).st_mode
        if not stat.S_ISSOCK(file_mode):
            eventlog.log_e('fail: file {} is not a socket')
            raise NetError

    except OSError as x:
        eventlog.log_e('fail: stat {} <{}>'.format(x.filename, x.strerror))
        raise NetError

    link = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        link.settimeout(tout_sec or 1)
        link.connect(uds_endpoint)

        link.settimeout(tout_sec or 1)

        link.send('{}\0'.format(ujson.dumps(query_jnode)))
        reply_utf8 = link.recv(1024).split('\0')[0]  # strip '\0' in the end
        reply_jnode = ujson.loads(reply_utf8)
        return reply_jnode

    except socket.timeout:
        if tout_sec:
            eventlog.log_w('fail: timeout')
            raise NetToutError
        else:
            eventlog.log_e('fail: timeout')
            raise NetError

    except socket.error as x:
        eventlog.log_e('fail: {}'.format(x))
        raise NetError

    finally:
        link.close()


def query_ra_alignment(ui_pid, alignment, min_rcs, speed_filter=0, tout_sec=3):
    """
    :type ui_pid: str
    :type alignment: bool
    :type min_rcs: int
    :type speed_filter: int
    :type tout_sec: int

    :raise NetError
    """
    endpoint = get_ra_uds_endpoint(ui_pid)
    query_jnode = {
        'method': 'set',
        'action': 'alignment',
        'args': {
            'sensor': {
                'mode': 'alignment' if alignment else 'monitoring',
                'mode-alignment': {
                    "speed-filter": speed_filter
                },
                'min-rcs': min_rcs
            }
        }
    }

    reply_jnode = query_uds(endpoint, query_jnode, tout_sec)
    status = reply_jnode.get('status')
    if status != 1:
        eventlog.log_e('fail: param reply#/status has invalid value <{}>'.format(status))
        raise NetError


def query_es_test(session_ref, test_ref, origin_uuid, bbox4_jnode=None, tout_sec=3):
    """
    :type session_ref: str
    :type test_ref: (str, str)
    :type origin_uuid: str
    :type bbox4_jnode: list[int] | None
    :type tout_sec: int

    :raise NetError
    """
    es_endpoint = get_es_uds_endpoint()
    jnode = {
        'action': 'store-conf',
        'args': {
            'tout': tout_sec,
            'session-ref': session_ref,
            'test-ref': test_ref,
            'origin': origin_uuid
        }
    }

    if bbox4_jnode:
        jnode['args']['bbox4'] = bbox4_jnode

    reply_jnode = query_uds(es_endpoint, jnode, tout_sec)
    code = reply_jnode.get('code')
    if code != 200:
        eventlog.log_e('fail: param reply#/code has invalid value <{}>'.format(code))
        raise NetError


def query_es_config(endpoints, tout_sec=3):
    """
    :type endpoints: list[str]
    :type tout_sec: int

    :raise NetError
    """
    es_endpoint = get_es_uds_endpoint()
    jnode = {
        'action': 'react-conf',
        'args': endpoints
    }

    reply_jnode = query_uds(es_endpoint, jnode, tout_sec)
    code = reply_jnode.get('code')
    if code != 200:
        eventlog.log_e('fail: param reply#/code has invalid value <{}>'.format(code))
        raise NetError


def query_gps_count():
    """
    :rtype: int
    """
    import gps
    session = gps.gps(host='localhost', port='2947', mode=gps.WATCH_ENABLE)
    try:
        for _ in range(5):
            msg = session.next()
            if 'DEVICES' == msg.get('class'):
                return len(msg.get('devices', []))
    finally:
        del session

    return 0


def query_lr_add_alarm(ui_pid, alarm_name, tout_sec=3):
    """
    :type ui_pid: str
    :type alarm_name: str
    :type tout_sec: int

    :raise NetError
    """
    endpoint = get_lr_uds_endpoint(ui_pid)
    query_jnode = {
        "action": "add-alarm",
        "params": {
            "name": alarm_name
        }
    }

    reply_jnode = query_uds(endpoint, query_jnode, tout_sec)
    status = reply_jnode.get('status')
    if status != 1:
        eventlog.log_e('fail: add-alarm (status:{})'.format(status))
        raise NetError


def query_lr_get_alarm(ui_pid, alarm_name, tout_sec=3):
    """
    :type ui_pid: str
    :type alarm_name: str
    :type tout_sec: int

    :rtype: bool
    :raise NetError
    """
    endpoint = get_lr_uds_endpoint(ui_pid)
    query_jnode = {
        "action": "get-alarm",
        "params": {
            "name": alarm_name
        }
    }

    reply_jnode = query_uds(endpoint, query_jnode, tout_sec)
    status = reply_jnode.get('status')
    if status == 1:
        return True
    elif status == 2:
        eventlog.log_e('fail: get-alarm (status:fail)')
        return False
    else:
        eventlog.log_e('fail: get-alarm (status:{})'.format(status))
        raise NetError


def _get_http_api_port():
    # TODO: parse /etc/nginx/sites-enabled/edge-site
    # https://github.com/fatiherikli/nginxparser
    # https://github.com/peakwinter/python-nginx
    return 8081


def get_http_api_endpoint():
    """
    :returns: uri
    :rtype: str | None
    """
    if os.path.exists(UI_NETWORK_JNODE_FILEPATH):
        with open(UI_NETWORK_JNODE_FILEPATH) as json_file:
            ui_network_jnode = ujson.load(json_file)
            if_mac = ui_network_jnode['ro']['mac']

    elif os.path.exists(MANAGEMENT_MAC_FILEPATH):
        with open(MANAGEMENT_MAC_FILEPATH, 'r') as text_file:
            if_mac = text_file.readlines()[0].strip()

    else:
        eventlog.log_e('fail: lookup management mac')
        raise NetError

    try:
        if_name = get_ifname(if_mac)
        if_ip = commands.getoutput(
            "ip addr show %s | grep -v '127.0.0.1' | grep -w inet | awk '{print $2}'"
            % if_name).split('/')[0]

        return 'http://{}:{}'.format(if_ip, _get_http_api_port())

    except Exception:
        eventlog.log_e('fail: define uri-host')
        raise NetError
