import pipes
import subprocess
from os import path

from edge.exceptions import ExternalError, InternalError, TimeoutError

DEV_NULL = open(path.devnull, 'w')


def run_silent(esc_cmd):
    """
    :param esc_cmd: Escaped shell-command. Use pipes.quote() to escape input form external sources
    :type esc_cmd: str
    :rtype: int
    """
    return subprocess.call(
        [esc_cmd],
        shell=True,
        stdout=DEV_NULL,
        stderr=subprocess.STDOUT)


def quote(data):
    """
    :type data: str
    :rtype: str
    """
    return pipes.quote(data)


def exec_shell(shell_cmd, tout_sec=10):
    """
    :type shell_cmd: str
    :type tout_sec: int | None

    :returns: (stdout+stderr, exist-status)
    :rtype: (str, int)
    """
    from edge import eventlog
    eventlog.log_d('exec')

    from pexpect.popen_spawn import PopenSpawn
    from pexpect.exceptions import EOF, TIMEOUT

    child_cmd = ' '.join(['/bin/bash', '-c', quote(shell_cmd)])
    child = PopenSpawn(child_cmd, timeout=tout_sec)
    try:
        child.expect_exact(EOF)
        child.wait()
    except TIMEOUT:
        eventlog.log_e('fail: TimeoutError (timeout:{}sec)'.format(tout_sec))
        raise TimeoutError

    if child.exitstatus is None:
        eventlog.log_e('fail: InternalError (signal-status:{})'.format(child.signalstatus))
        raise InternalError

    else:
        eventlog.log_d('done')
        return child.before, child.exitstatus


def exec_cli(cmd, args, tout_sec=10, encoding='ascii'):
    """
    :type cmd: str
    :type args: list | tuple
    :type tout_sec: int | None

    :param encoding: 'ascii' | 'utf-8'
    :type encoding: str

    :returns: stdout+stderr
    :rtype: str
    """
    from edge import eventlog
    eventlog.log_d('exec: <{}>'.format(cmd))

    from pexpect.popen_spawn import PopenSpawn
    from pexpect.exceptions import EOF, TIMEOUT

    child_cmd = ' '.join([cmd] + args)
    if encoding != 'ascii':
        child_cmd = child_cmd.encode(encoding)

    child = PopenSpawn(child_cmd, timeout=tout_sec, encoding=encoding)

    try:
        child.expect_exact(EOF)
        child.wait()
    except TIMEOUT:
        eventlog.log_e('fail: TimeoutError (timeout:{}sec)'.format(tout_sec))
        raise TimeoutError

    if child.exitstatus is None:
        eventlog.log_e('fail: InternalError (signal-status:{})'.format(child.signalstatus))
        raise InternalError

    elif child.exitstatus == 0:
        eventlog.log_d('done')
        return child.before

    elif child.exitstatus == 64:
        eventlog.log_e('fail: ExternalError (exit-status:EX_USAGE)')
        raise ExternalError

    elif child.exitstatus == 70:
        eventlog.log_e('fail: InternalError (exit-status:EX_SOFTWARE, before:<{}>)'.format(child.before))
        raise InternalError
