import datetime

from edge import eventlog
from edge.image.types import ContrastScheduleConf
from edge.model.geoposition import Geoposition
from edge.com.astral import Location


class IContrastSchedule:
    def __init__(self):
        pass

    def test_active(self, im_datetime_utc, im_geolocation):
        """
        :type im_datetime_utc: datetime.datetime
        :type im_geolocation: Geoposition
        :rtype: bool
        """
        raise NotImplementedError


class _ContrastSpan:
    def __init__(self, date_utc, min_dt_utc, max_dt_utc):
        """
        :type date_utc: datetime.date
        :type min_dt_utc: datetime.datetime
        :type max_dt_utc: datetime.datetime
        """
        self.date_utc = date_utc
        self.min_dt_utc = min_dt_utc
        self.max_dt_utc = max_dt_utc

    def __contains__(self, dt_utc):
        """
        :type dt_utc: datetime.datetime
        :rtype: bool
        """
        return self.min_dt_utc < dt_utc < self.max_dt_utc


class _TwilightContrastSchedule(IContrastSchedule):
    def __init__(self, conf):
        """
        :type conf: ContrastScheduleConf
        """
        IContrastSchedule.__init__(self)
        self._conf = conf

        self._last_geolocation = None
        self._last_datetime = None

        self._spans = []

    def _test_geolocation_changed(self, im_geolocation):
        """
        :type im_geolocation: Geoposition | None
        :rtype: bool
        """
        if im_geolocation is None:
            pass

        elif self._last_geolocation is None:
            self._last_geolocation = im_geolocation
            return True

        elif self._last_geolocation != im_geolocation:
            self._last_geolocation = im_geolocation
            return True

        return False

    def _test_date_changed(self, im_datetime_utc):
        """
        :type im_datetime_utc: datetime.datetime
        :rtype: bool
        """
        im_date_utc = im_datetime_utc.date()

        date_in_list = False
        for span in self._spans:
            if im_date_utc == span.date_utc:
                date_in_list = True
                break

        return not date_in_list

    def test_active(self, im_datetime_utc, im_geolocation):
        """
        :type im_datetime_utc: datetime.datetime
        :type im_geolocation: Geoposition | None
        :rtype: bool
        """
        day_changed = self._test_date_changed(im_datetime_utc)
        geo_changed = self._test_geolocation_changed(im_geolocation)

        if day_changed or geo_changed:
            if self._last_geolocation is None:
                return False

            location = Location(self._last_geolocation.latitude, self._last_geolocation.longitude)

            yesterday_utc = im_datetime_utc - datetime.timedelta(days=1)
            yesterday_dt_min = location.sunset(yesterday_utc)
            yesterday_dt_max = location.sunrise(im_datetime_utc)
            yesterday_span = _ContrastSpan(yesterday_utc.date(), yesterday_dt_min, yesterday_dt_max)

            today_dt_min = location.sunset(im_datetime_utc)
            today_dt_max = location.sunrise(im_datetime_utc + datetime.timedelta(days=1))
            today_span = _ContrastSpan(im_datetime_utc.date(), today_dt_min, today_dt_max)

            self._spans = [yesterday_span, today_span]

            eventlog.log_i('day-changed:{}, geo-changed:{}, twilight-y:({}, {}), twilight-t:({}, {})'.format(
                day_changed, geo_changed,
                self._spans[0].min_dt_utc.isoformat(), self._spans[0].max_dt_utc.isoformat(),
                self._spans[1].min_dt_utc.isoformat(), self._spans[1].max_dt_utc.isoformat()))

        for span in self._spans:
            if im_datetime_utc in span:
                return True

        return False


class _NeverContrastSchedule(IContrastSchedule):
    def __init__(self):
        IContrastSchedule.__init__(self)

    def test_active(self, im_datetime_utc, im_geolocation):
        """
        :type im_datetime_utc: datetime.datetime
        :type im_geolocation: Geoposition | None
        :rtype: bool
        """
        return False


class _AlwaysContrastSchedule(IContrastSchedule):
    def __init__(self):
        IContrastSchedule.__init__(self)

    def test_active(self, im_datetime_utc, im_geolocation):
        """
        :type im_datetime_utc: datetime.datetime
        :type im_geolocation: Geoposition | None
        :rtype: bool
        """
        return True


class _TimeContrastSchedule(IContrastSchedule):
    def __init__(self, conf):
        """
        :type conf: ContrastScheduleConf
        """
        IContrastSchedule.__init__(self)
        self._time_from = conf.time_from
        self._time_to = conf.time_to

    def test_active(self, im_datetime_utc, im_geolocation):
        """
        :type im_datetime_utc: datetime.datetime
        :type im_geolocation: Geoposition | None
        :rtype: bool
        """
        # TODO: works only if time_from from this day and time_to from next day
        return self._time_from < im_datetime_utc.time() or im_datetime_utc.time() < self._time_to


def create_schedule(conf):
    """
    :type conf: ContrastScheduleConf
    :rtype: IContrastSchedule
    """
    if conf.mode == conf.MODE_ALWAYS:
        return _AlwaysContrastSchedule()

    elif conf.mode == conf.MODE_NEVER:
        return _NeverContrastSchedule()

    elif conf.mode == conf.MODE_TIME:
        return _TimeContrastSchedule(conf)

    elif conf.mode == conf.MODE_NIGHT:
        return _TwilightContrastSchedule(conf)
