import ctypes
import cStringIO
import semver
import os
import io

import PIL
from PIL import Image, ImageDraw, ImageFont
from edge.image import contrast

from edge_bbox.types import RectF, mk_vehicle_bbox, mk_mmr_bbox, QuadI
from edge_rawutil import rawutil
from edge_improc import improc

from .types import (ImEngineArgs, PlateImArgs, ThumbImArgs, VehicleImArgs,
                    SceneImArgs, MemBlob, JsonBlobMeta, LiveImArgs,
                    ImFooterArgs, ImFooterVars)

from edge import eventlog
from ..model.types import Dir
from ..types import ImageRef
from ..model.track import ITrack

_FONT_PATH = '/usr/share/fonts/truetype/ttf-dejavu/DejaVuSans.ttf'


if semver.compare(PIL.PILLOW_VERSION, "4.2.0") == -1:
    def img_save(source_image, jpeg_stream):
        """
        :type source_image: PIL.Image.Image
        :type jpeg_stream: StringIO.OutputType
        """
        source_image.save(jpeg_stream, format="JPEG")

    img_save_fn = img_save
else:
    def img_save(source_image, jpeg_stream):
        """
        :type source_image: PIL.Image.Image
        :type jpeg_stream: StringIO.OutputType
        """
        rgb_image = source_image.convert('RGB')
        rgb_image.save(jpeg_stream, 'JPEG')

    img_save_fn = img_save


class ImError (RuntimeError):
    pass


class ImRgba:
    def __init__(self):
        self._w = 0
        self._h = 0
        self._dh = 0
        self._ts = 0

        self._buffer = ctypes.create_string_buffer(0)

    @property
    def ts(self):
        """
        :rtype: long
        """
        return self._ts

    @property
    def w(self):
        """
        :rtype: int
        """
        return self._w

    @property
    def h(self):
        """
        :rtype: int
        """
        return self._h

    @property
    def wh(self):
        return self._w, self._h

    @property
    def dh(self):
        """
        :rtype: int
        """
        return self._dh

    @property
    def buffer(self):
        """
        :rtype: ctypes.Array
        """
        return self._buffer

    @property
    def stride(self):
        """
        :rtype: int
        """
        return 4 * self._w

    def reset(self, w, h, dh, ts=14):
        """
        :type w: int
        :type h: int
        :type dh: int
        :type ts: int
        """
        self._w = w
        self._h = h
        self._dh = dh
        self._ts = ts

        if (h + dh) * self.stride > len(self._buffer):
            self._buffer = ctypes.create_string_buffer(self.stride * (h + dh))


class ImBuffer:
    """
    :type _buffer: ctypes.Array
    """
    def __init__(self):
        self._ts = 0
        self._w = 0
        self._h = 0

        self._buffer = ctypes.create_string_buffer(0)
        self._buffer_used = 0

    @property
    def ts(self):
        """
        :rtype: long
        """
        return self._ts

    @property
    def w(self):
        """
        :rtype: int
        """
        return self._w

    @property
    def h(self):
        """
        :rtype: int
        """
        return self._h

    @property
    def wh(self):
        return self._w, self._h

    @property
    def buffer_size(self):
        """
        :rtype: long
        """
        return len(self._buffer)

    @property
    def buffer_used(self):
        """
        :rtype: long
        """
        return self._buffer_used

    @property
    def buffer(self):
        """
        :rtype: ctypes.Array
        """
        return self._buffer

    @property
    def buffer_data(self):
        """
        :rtype: ctypes.Array
        """
        return self._buffer[0:self._buffer_used]

    def reset(self, w, h, buffer_used, ts=14):
        """
        :type ts: int
        :type w: int
        :type h: int
        :type buffer_used: int
        """
        self._ts = ts
        self._w = w
        self._h = h

        if buffer_used > self.buffer_size:
            self._buffer = ctypes.create_string_buffer(buffer_used)

        self._buffer_used = buffer_used


class ImTrans:

    #
    # (x', y') - scene-im point; (x'', y'') - recrop-image point.
    #
    # |x''|   | a  b  x |   |x'|
    # |y''| = | c  d  y | * |y'|
    # |1  |   | 0  0  1 |   |1 |
    #
    #
    # |x''|   | Ix Jx Ox |   |x'|
    # |y''| = | Iy Jy Oy | * |y'|
    # |1  |   | 0  0  1  |   |1 |
    #

    def __init__(self, ix=1.0, iy=0.0, jx=0.0, jy=1.0, ox=0.0, oy=0.0):
        """
        :type ix: float
        :type iy: float
        :type jx: float
        :type jy: float
        :type ox: float
        :type oy: float
        """
        self._ix = ix
        self._iy = iy
        self._jx = jx
        self._jy = jy
        self._ox = ox
        self._oy = oy

    def map_ltrb(self, source_ltrb):
        """
        :type source_ltrb: list[float]
        :rtype: list[float]
        """
        l, t, r, b = source_ltrb
        target_ltrb = [
            l * self._ix + t * self._jx + self._ox,
            l * self._iy + t * self._jy + self._oy,

            r * self._ix + b * self._jx + self._ox,
            r * self._iy + b * self._jy + self._oy
        ]

        return target_ltrb

    def map_point(self, l, t):
        """
        :type l:float
        :type t:float
        :rtype:
        """
        target_point = [
            l * self._ix + t * self._jx + self._ox,
            l * self._iy + t * self._jy + self._oy]

        return target_point

    def map_quad(self, source_quad):
        """
        :type source_quad: QuadI
        :rtype: QuadI
        """

        target_quad = QuadI(
            (source_quad.x0 * self._ix + source_quad.y0 * self._jx + self._ox),
            (source_quad.x0 * self._iy + source_quad.y0 * self._jy + self._oy),

            (source_quad.x1 * self._ix + source_quad.y1 * self._jx + self._ox),
            (source_quad.x1 * self._iy + source_quad.y1 * self._jy + self._oy),

            (source_quad.x2 * self._ix + source_quad.y2 * self._jx + self._ox),
            (source_quad.x2 * self._iy + source_quad.y2 * self._jy + self._oy),

            (source_quad.x3 * self._ix + source_quad.y3 * self._jx + self._ox),
            (source_quad.x3 * self._iy + source_quad.y3 * self._jy + self._oy)
        )
        return target_quad

    def serialize(self):
        """
        :rtype: dict
        """
        return {
            'o': [self._ox, self._oy],
            'i': [self._ix, self._iy],
            'j': [self._jx, self._jy]
        }


class ImLoader:
    MAX_WIDTH = 65535
    MAX_HEIGHT = 65535

    def __init__(self, max_target_width=MAX_WIDTH, max_target_height=MAX_HEIGHT):
        """
        :type max_target_width: int
        :type max_target_height: int
        """
        self._rawutil = rawutil.RawUtil.create(max_target_width, max_target_height, log_level='debug')

    def destroy(self):
        self._rawutil.destroy()

    def load_rgba(self, dh, im_ref, dh_data):
        """
        :type dh: int
        :type im_ref: ImageRef
        :type dh_data: ImRgba

        :rtype: None
        :raise ValueError:
        """

        if im_ref.codec == 'JPEG':
            raw_pil_im = Image.open(im_ref.url)

            dh_data.reset(raw_pil_im.size[0], raw_pil_im.size[1], dh, im_ref.ts)
            with Image.frombuffer(
                    'RGBA', dh_data.wh, dh_data.buffer,
                    'raw', 'RGBA', 0, 1) as rgba_pil_image:

                rgba_pil_image.readonly = 0
                rgba_pil_image.paste(raw_pil_im)

        elif im_ref.codec in ['MONO8T', 'I420PLN', 'NV12PLN']:
            dh_data.reset(im_ref.width, im_ref.height, dh, im_ref.ts)
            success = self._rawutil.handle_image3(
                im_ref.url, im_ref.codec, im_ref.width, im_ref.height, dh_data.buffer)

            if not success:
                eventlog.log_d('fail: conv image %s -> RGBA' % im_ref.codec)
                raise ValueError

            _, w, h = success
            dh_data.reset(w, h, dh, im_ref.ts)

        else:
            raise NotImplementedError

    def load_jpeg(self, im_ref, im_data):
        """
        :type im_ref: ImageRef
        :type im_data: ImBuffer

        :rtype: None
        """
        if im_ref.codec == 'JPEG' and im_ref.url is not None:
            wh = ImHelpers.get_jpeg_size(im_ref.url)
            im_data.reset(wh[0], wh[1], wh[0] * wh[1], im_ref.ts)

            with open(im_ref.url, 'rb') as bin_file:
                file_size = bin_file.readinto(im_data.buffer)

            im_data.reset(wh[0], wh[1], file_size, im_ref.ts)

        else:
            raise NotImplementedError


class ImSizeFitter:
    """
    :type _last_source_wh: (int, int) | None
    :type _last_target_wh: (int, int) | None
    """

    def __init__(self, max_wh=None):
        """
        :type max_wh: (int, int) | None
        """
        self._max_wh = max_wh

        self._last_source_wh = None
        self._last_target_wh = None

    def _get_target_wh(self, source_wh):
        """
        :type source_wh: (int, int)
        :rtype: (int, int) | None
        """
        # resolution constraints are absent
        if self._max_wh is None:
            eventlog.log_i("scale images from %s to %s" % (source_wh, source_wh))
            return None

        # resolution constraints are met
        source_w, source_h = source_wh
        max_w, max_h = self._max_wh
        if source_w <= max_w and source_h <= max_h:
            eventlog.log_i("scale images from %s to %s" % (source_wh, source_wh))
            return None

        # define resize-ratio to meet resolution constraints
        exact_width_ratio = float(max_w) / source_w
        exact_height_ratio = float(max_h) / source_h
        exact_resize_ratio = exact_height_ratio if exact_height_ratio < exact_width_ratio else exact_width_ratio

        # align target image width to 16 px. it's to be more friendy with quicksync-h264-encoder.
        _ALIGN_BOUNDARY_PX = 16
        _ALIGN_BOUNDARY_SEARCH_DEPTH = 3 * _ALIGN_BOUNDARY_PX

        source_wh_ratio = float(source_w) / source_h
        final_target_wigth = int(source_w * exact_resize_ratio) / _ALIGN_BOUNDARY_PX * _ALIGN_BOUNDARY_PX
        final_target_height = final_target_wigth / source_wh_ratio

        aligned_target_wigth = final_target_wigth
        while aligned_target_wigth >= (final_target_wigth - _ALIGN_BOUNDARY_SEARCH_DEPTH):
            aligned_target_height = aligned_target_wigth / source_wh_ratio
            if (aligned_target_height % 1) == 0:
                final_target_wigth = aligned_target_wigth
                final_target_height = int(aligned_target_height)
                break

            aligned_target_wigth -= _ALIGN_BOUNDARY_PX

        if (final_target_height % 1) == 0:
            target_size = final_target_wigth, int(final_target_height)
        else:
            target_size = final_target_wigth, int(final_target_height) + 1

        eventlog.log_d(
            "source=%s -> max-target=%s -> pre-target=%s -> target=%s" %
            (source_wh, self._max_wh, (final_target_wigth, final_target_height), target_size))

        eventlog.log_i("scale images from %s to %s" % (source_wh, target_size))

        return target_size

    def scaledown(self, source_wh):
        """
        :type source_wh: (int, int)
        :rtype: (int, int)
        """
        if source_wh != self._last_source_wh:
            self._last_target_wh = self._get_target_wh(source_wh)
            self._last_source_wh = source_wh

        if self._last_target_wh:
            return self._last_target_wh

        return source_wh


class ImHelpers:
    def __init__(self):
        pass

    @staticmethod
    def get_jpeg_size(jpeg_url):
        """
        :type jpeg_url: str
        :rtype: (int, int)
        """
        return Image.open(jpeg_url).size

    @staticmethod
    def get_jpeg_size2(jpeg_data):
        """
        :type jpeg_data: ctypes.Array
        :rtype: (int, int)
        """
        return Image.open(cStringIO.StringIO(jpeg_data)).size

    @staticmethod
    def mk_jpeg(source_image):
        """
        :type source_image: PIL.Image.Image
        :rtype: str
        """
        jpeg_stream = cStringIO.StringIO()
        try:
            img_save_fn(source_image, jpeg_stream)
            return jpeg_stream.getvalue()
        finally:
            jpeg_stream.close()

    @staticmethod
    def mk_jpeg2(blob_data):
        """
        :type blob_data: ImRgba
        :rtype: str
        """
        with Image.frombuffer(
                'RGBA', blob_data.wh, blob_data.buffer,
                'raw', 'RGBA', blob_data.stride, 1) as scene_im:

            return ImHelpers.mk_jpeg(scene_im)

    @staticmethod
    def mk_resize_image(source_data, target_w, target_h, resample=Image.NEAREST):
        """
        :type source_data: ImRgba
        :type target_w: int | float
        :type target_h: int | float
        :type resample: int
        :rtype: PIL.Image.Image
        """
        with Image.frombuffer(
                'RGBA', source_data.wh, source_data.buffer,
                'raw', 'RGBA', source_data.stride, 1) as crop_im:

            return crop_im.resize((int(target_w), int(target_h)), resample)

    @staticmethod
    def mk_resize_jpeg(blob_data, target_w, target_h, resample=Image.NEAREST):
        """
        :type blob_data: ImRgba
        :type target_w: int | float
        :type target_h: int | float
        :type resample: int
        :rtype: str
        """
        with ImHelpers.mk_resize_image(blob_data, target_w, target_h, resample) as pil_im:
            return ImHelpers.mk_jpeg(pil_im)

    @staticmethod
    def mk_resize_trans(source_data, target_w, target_h):
        """
        :type source_data: ImRgba
        :type target_w: int | float
        :type target_h: int | float
        :rtype: ImTrans
        """
        #
        # (x', y') - scene-im point; (x'', y'') - recrop-image point.
        #
        # |x''|   | Ix Jx Ox |   |x'|
        # |y''| = | Iy Jy Oy | * |y'|
        # |1  |   | 0  0  1  |   |1 |
        #
        #     scale           move              trans
        # | Ix  0   0 |   | 1   0   0  |   | Ix  0   0  |
        # | 0   Jy  0 | * | 0   1   0  | = | 0   Jy  0  |
        # | 0   0   1 |   | 0   0   1  |   | 0   0   1  |
        #

        target_trans = ImTrans(
            ix=target_w / float(source_data.w),
            jy=target_h / float(source_data.h))

        return target_trans

    @staticmethod
    def mk_crop_resize_image(source_data, crop_rect, target_w, target_h, resample=Image.NEAREST):
        """
        :type source_data: ImRgba
        :type crop_rect: RectF
        :type target_w: int | float
        :type target_h: int | float
        :type resample: int
        :rtype: PIL.Image.Image
        """
        scene_mv = memoryview(source_data.buffer)
        crop_offset = 4 * (int(crop_rect.ft) * source_data.w + int(crop_rect.fl))

        with Image.frombuffer(
                'RGBA', (int(crop_rect.fw), int(crop_rect.fh)), scene_mv[crop_offset:],
                'raw', 'RGBA', source_data.stride, 1) as crop_im:

            return crop_im.resize((int(target_w), int(target_h)), resample)

    @staticmethod
    def mk_crop_resize_jpeg(blob_data, crop_rect, target_w, target_h, resample=Image.NEAREST):
        """
        :type blob_data: ImRgba
        :type crop_rect: RectF
        :type target_w: int | float
        :type target_h: int | float
        :type resample: int
        :rtype: str
        """
        with ImHelpers.mk_crop_resize_image(blob_data, crop_rect, target_w, target_h, resample) as pil_im:
            return ImHelpers.mk_jpeg(pil_im)

    @staticmethod
    def mk_crop_resize_trans(crop_rect, target_w, target_h):
        """
        :type crop_rect: RectF
        :type target_w: int | float
        :type target_h: int | float
        :rtype: ImTrans
        """

        #
        # (x', y') - scene-im point; (x'', y'') - recrop-image point.
        #
        # |x''|   | Ix Jx Ox |   |x'|
        # |y''| = | Iy Jy Oy | * |y'|
        # |1  |   | 0  0  1  |   |1 |
        #
        #     scale           move              trans
        # | Ix  0   0 |   | 1   0   Ox |   | Ix  0   Ox*Ix |
        # | 0   Jy  0 | * | 0   1   Oy | = | 0   Jy  Oy*Jy |
        # | 0   0   1 |   | 0   0   1  |   | 0   0   1     |
        #

        ix = target_w / crop_rect.fw
        jy = target_h / crop_rect.fh
        return ImTrans(ix=ix, jy=jy, ox=-crop_rect.fl * ix, oy=-crop_rect.ft * jy)

    @staticmethod
    def draw_rect(pil_image, rect, rect_trans=None):
        """
        :type pil_image: PIL.Image.Image
        :type rect: RectF
        :type rect_trans: ImTrans | None
        """
        ltrb = [rect.fl, rect.ft, rect.fr, rect.fb]
        if rect_trans:
            ltrb = rect_trans.map_ltrb(ltrb)

        ImageDraw.Draw(pil_image).rectangle(ltrb, outline='red')

    @staticmethod
    def draw_poly(pil_image, quad, rect_trans=None):
        """
        :type pil_image: PIL.Image.Image
        :type quad: LrQuad
        :type rect_trans: ImTrans | None
        """
        if rect_trans:
            poly = rect_trans.map_quad(quad).serialize()
        else:
            poly = quad.serialize()

        ImageDraw.Draw(pil_image).polygon(poly, outline='red')

    @staticmethod
    def draw_footer(pil_image, footer_args, font, font_size):
        """
        :type pil_image: PIL.Image.Image
        :type footer_args: ImFooterArgs
        :type font: PIL.ImageFont.FreeTypeFont
        :type font_size: int
        """
        draw = ImageDraw.Draw(pil_image)
        draw.rectangle([(0, 0), pil_image.size], fill='black')
        # font_size = font.size

        y_delta = 0.4 * font_size  # top-margin
        start_y = int(y_delta / 2)

        leter_width = font.getsize('M')[0]

        for row in footer_args.rows:
            start_x = leter_width / 2  # left-margin = 0.5em
            for column in row.cols:
                column_width = column.width
                column_text = column.template
                column_align = column.align

                if column_align == -1:
                    x_delta = 0

                elif column_align == 0:
                    text_width, text_height = font.getsize(column_text)
                    x_delta = (column_width - text_width) / 2 - leter_width / 2

                elif column_align == 1:
                    text_width, text_height = font.getsize(column_text)
                    x_delta = column_width - text_width - leter_width

                else:
                    raise NotImplementedError

                draw.text((start_x + x_delta, start_y), column_text, font=font)
                start_x = start_x + column_width

            start_y = start_y + font_size + y_delta


class TrainerImProfile:
    _NAME = 'trainer'

    def __init__(self):
        pass

    def process(self, scene_dh_data, vehicle_bbox):
        """
        :type scene_dh_data: ImRgba
        :type vehicle_bbox: RectF | None
        :rtype: MemBlob
        """
        if vehicle_bbox is None:
            blob_data = ImHelpers.mk_jpeg2(scene_dh_data)
        else:
            blob_data = ImHelpers.mk_crop_resize_jpeg(scene_dh_data, vehicle_bbox, vehicle_bbox.fw, vehicle_bbox.fh,
                                                      Image.BICUBIC)

        return MemBlob(self._NAME, scene_dh_data.ts, blob_data)


class PlateImProfile:
    _NAME = 'plate'

    def __init__(self, args):
        """
        :type args: PlateImArgs
        """
        self._w = args.width  # type: int
        self._h = args.height  # type: int

    def process(self, scene_dh_data, plate_bbox):
        """
        :type scene_dh_data: ImRgba
        :type plate_bbox: RectF
        :rtype: list[MemBlob]
        """
        blob_data = ImHelpers.mk_crop_resize_jpeg(
            scene_dh_data, plate_bbox, self._w, self._h, Image.BICUBIC)
        return [MemBlob(self._NAME, scene_dh_data.ts, blob_data)]


class ThumbImProfile:
    _NAME = 'thumb'

    def __init__(self, args):
        """
        :type args: ThumbImArgs
        """
        self._w = float(args.width)
        self._h = float(args.height)
        self._wh_ratio = self._w / self._h

    def process(self, scene_dh_data, plate_bbox, vehicle_bbox):
        """
        :type scene_dh_data: ImRgba
        :type plate_bbox: RectF
        :type vehicle_bbox: RectF | None
        :rtype: list[MemBlob]
        """
        min_visible_factor = 0.0  # assume that all mmr bboxes are suitable
        crop_rect = mk_mmr_bbox(plate_bbox, scene_dh_data.wh, min_visible_factor)
        if not crop_rect:
            crop_rect = self._get_stub_bbox(plate_bbox, *scene_dh_data.wh)
            crop_rect.strip_invisible(*scene_dh_data.wh)

        blob_data = ImHelpers.mk_crop_resize_jpeg(
            scene_dh_data, crop_rect, self._w, self._h)
        return [MemBlob(self._NAME, scene_dh_data.ts, blob_data)]

    def _get_stub_bbox(self, plate_bbox, scene_iw, scene_ih):
        """
        :type plate_bbox: RectF
        :type scene_iw: int
        :type scene_ih: int
        :rtype: RectF
        """
        # full scene-h
        car_fw = scene_ih * self._wh_ratio
        if car_fw < scene_iw:
            car_center_fl = car_fw / 2.0
            car_center_fl_min = 0.0 + car_center_fl
            car_center_fl_max = scene_iw - car_center_fl

            plate_center_fl = (plate_bbox.fl + plate_bbox.fr) / 2.0

            if plate_center_fl < car_center_fl_min:
                car_fl = car_center_fl_min - car_center_fl
            elif plate_center_fl > car_center_fl_max:
                car_fl = car_center_fl_max - car_center_fl
            else:
                car_fl = plate_center_fl - car_center_fl

            return RectF(car_fl, 0.0, car_fl + car_fw, float(scene_ih))

        # full scene-w
        car_fh = scene_iw / self._wh_ratio
        if car_fh < scene_ih:
            car_center_ft = car_fh / 2.0
            car_center_ft_min = 0.0 + car_center_ft
            car_center_ft_max = scene_ih - car_center_ft

            plate_center_ft = (plate_bbox.ft + plate_bbox.fb) / 2.0

            if plate_center_ft < car_center_ft_min:
                car_ft = car_center_ft_min - car_center_ft
            elif plate_center_ft > car_center_ft_max:
                car_ft = car_center_ft_max - car_center_ft
            else:
                car_ft = plate_center_ft - car_center_ft

            return RectF(0.0, car_ft, float(scene_iw), car_ft + car_fh)

        raise NotImplementedError


class VehicleImProfile:
    _NAME = 'vehicle'

    def __init__(self, args):
        """
        :type args: VehicleImArgs
        """
        self._args = args
        self._dh_data = ImRgba()

        self._font = ImageFont.truetype(
            _FONT_PATH, args.footer.font_size, encoding='unic')

    def process(self, scene_dh_data, vehicle_bbox, footer_vars, quad, edge_improc):
        """
        :type scene_dh_data: ImRgba
        :type vehicle_bbox: RectF | None
        :type footer_vars: ImFooterVars
        :type quad: QuadI
        :type edge_improc: improc.Improc | None
        :rtype: list[MemBlob]
        """
        if not vehicle_bbox:
            return []

        footer_args = self._args.footer.eval(footer_vars)

        vehicle_im = ImHelpers.mk_crop_resize_image(
            scene_dh_data, vehicle_bbox, self._args.width, self._args.height)

        if edge_improc is not None:
            vehicle_im = edge_improc.improve_contrast_rgba(vehicle_im)

        vehicle_trans = ImHelpers.mk_crop_resize_trans(
            vehicle_bbox, self._args.width, self._args.height)

        dh_data = self._dh_data
        dh_data.reset(vehicle_im.size[0], vehicle_im.size[1], self._args.footer.height, scene_dh_data.ts)

        # render footer
        vehicle_dh_mv = memoryview(dh_data.buffer)
        footer_offset = int(dh_data.stride * dh_data.h)
        with Image.frombuffer(
                'RGBA', (dh_data.w, dh_data.dh), vehicle_dh_mv[footer_offset:],
                'raw', 'RGBA', dh_data.stride, 1) as footer_im:

            footer_im.readonly = 0
            ImHelpers.draw_footer(footer_im, footer_args, self._font, self._args.footer.font_size)

        # render image and annotations
        with Image.frombuffer(
                'RGBA', (dh_data.w, dh_data.h + dh_data.dh), vehicle_dh_mv,
                'raw', 'RGBA', dh_data.stride, 1) as vehicle_dh_im:

            vehicle_dh_im.readonly = 0
            vehicle_dh_im.paste(vehicle_im, (0, 0))
            if self._args.anno.plate:
                ImHelpers.draw_poly(vehicle_dh_im, quad, vehicle_trans)

            blob_data = ImHelpers.mk_jpeg(vehicle_dh_im)

        blob_meta = JsonBlobMeta({
            'trans': vehicle_trans.serialize(),
            'size': [vehicle_dh_im.size[0], vehicle_dh_im.size[1], len(blob_data)]
        })

        return [MemBlob(self._NAME, scene_dh_data.ts, blob_data, blob_meta)]


class SceneImProfile:
    _NAME = 'scene'

    def __init__(self, args):
        """
        :type args: SceneImArgs
        """
        self._args = args
        self._dh_data = ImRgba()

        self._font = ImageFont.truetype(
            _FONT_PATH, args.footer.font_size, encoding='unic')

    def process(self, scene_dh_data, plate_bbox, vehicle_bbox, footer_vars, track, quad):
        """
        :type scene_dh_data: ImRgba
        :type plate_bbox: RectF
        :type vehicle_bbox: RectF | None
        :type footer_vars: ImFooterVars
        :type track: ITrack
        :type quad: QuadI
        :rtype: list[MemBlob]
        """
        footer_args = self._args.footer.eval(footer_vars)

        if scene_dh_data.w != self._args.width:
            target_w = min(self._args.width, scene_dh_data.w)
            target_h = float(target_w) / scene_dh_data.w * scene_dh_data.h
            alt_im = ImHelpers.mk_resize_image(scene_dh_data, target_w, target_h)
            alt_trans = ImHelpers.mk_resize_trans(scene_dh_data, target_w, target_h)

            dh_data = self._dh_data
            dh_data.reset(alt_im.size[0], alt_im.size[1], self._args.footer.height, scene_dh_data.ts)

            # render scene
            with Image.frombuffer(
                    'RGBA', (dh_data.w, dh_data.h + dh_data.dh), dh_data.buffer,
                    'raw', 'RGBA', dh_data.stride, 1) as alt_dh_im:

                alt_dh_im.readonly = 0
                alt_dh_im.paste(alt_im, (0, 0))

            # render annotations and footer
            blob_data = self._render(dh_data, alt_trans, plate_bbox, vehicle_bbox, footer_args, track, quad)

            blob_meta = JsonBlobMeta({
                'trans': alt_trans.serialize(),
                'size': [alt_dh_im.size[0], alt_dh_im.size[1], len(blob_data)]
            })

        else:
            # render annotations and footer
            alt_trans = None
            blob_data = self._render(scene_dh_data, alt_trans, plate_bbox, vehicle_bbox, footer_args, track, quad)

            blob_meta = JsonBlobMeta({
                'trans': ImTrans().serialize(),
                'size': [scene_dh_data.w, scene_dh_data.h + scene_dh_data.dh, len(blob_data)]
            })

        return [MemBlob(self._NAME, scene_dh_data.ts, blob_data, blob_meta)]

    def _render(self, scene_dh_blob, scene_dh_trans, plate_bbox, vehicle_bbox, footer_args, track, quad):
        """
        :type scene_dh_blob: ImRgba
        :type scene_dh_trans: ImTrans | None
        :type plate_bbox: RectF
        :type vehicle_bbox: RectF | None
        :type footer_args: ImFooterArgs
        :type track: ITrack
        :rtype: str
        """

        # render footer
        scene_dh_mv = memoryview(scene_dh_blob.buffer)
        footer_offset = int(scene_dh_blob.stride * scene_dh_blob.h)
        with Image.frombuffer(
                'RGBA', (scene_dh_blob.w, scene_dh_blob.dh), scene_dh_mv[footer_offset:],
                'raw', 'RGBA', scene_dh_blob.stride, 1) as footer_im:

            footer_im.readonly = 0
            ImHelpers.draw_footer(footer_im, footer_args, self._font, self._args.footer.font_size)

        # render annotations
        with Image.frombuffer(
                'RGBA', (scene_dh_blob.w, scene_dh_blob.h + scene_dh_blob.dh), scene_dh_mv,
                'raw', 'RGBA', scene_dh_blob.stride, 1) as scene_alt_dh_im:

            scene_alt_dh_im.readonly = 0

            if self._args.anno.plate:
                ImHelpers.draw_poly(scene_alt_dh_im, quad, rect_trans=scene_dh_trans)

            if self._args.anno.vehicle and vehicle_bbox:
                ImHelpers.draw_rect(scene_alt_dh_im, vehicle_bbox, scene_dh_trans)

            if self._args.anno.track and track:
                self._draw_track(scene_alt_dh_im, plate_bbox, track, scene_dh_trans)

            return ImHelpers.mk_jpeg(scene_alt_dh_im)

    @staticmethod
    def _draw_track(pil_image, raw_plate_bbox, lr_track, rect_trans=None):
        """
        :type pil_image: PIL.Image.Image
        :type raw_plate_bbox: RectF
        :type lr_track: ITrack
        :type rect_trans: ImTrans | None
        """

        if rect_trans:
            ltrb = [raw_plate_bbox.fl, raw_plate_bbox.ft, raw_plate_bbox.fr, raw_plate_bbox.fb]
            ltrb = rect_trans.map_ltrb(ltrb)
            plate_bbox = RectF(*ltrb)
        else:
            plate_bbox = raw_plate_bbox

        trace = lr_track.trace
        im_draw = ImageDraw.Draw(pil_image)
        for index in range(0, len(trace) - 3, 2):
            try:
                l_image_x, l_image_y = trace[index], trace[index + 1]
                r_image_x, r_image_y = trace[index + 2], trace[index + 3]

                # scale point
                if rect_trans:
                    l_image_x, l_image_y = rect_trans.map_point(l_image_x, l_image_y)
                    r_image_x, r_image_y = rect_trans.map_point(r_image_x, r_image_y)

                l_in_plate = plate_bbox.ft < l_image_y < plate_bbox.fb and plate_bbox.fl < l_image_x < plate_bbox.fr
                r_in_plate = plate_bbox.ft < r_image_y < plate_bbox.fb and plate_bbox.fl < r_image_x < plate_bbox.fr

                both_ok = False
                if l_in_plate and r_in_plate:
                    continue

                elif l_in_plate and not r_in_plate:
                    # tweak left point
                    k = float(r_image_x - l_image_x) / (r_image_y - l_image_y)
                    b = float(l_image_x * r_image_y - r_image_x * l_image_y) / (r_image_y - l_image_y)
                    # x = k * y + b
                    if lr_track.direction == Dir.APPROACHING:
                        # tweak down
                        l_image_y = plate_bbox.fb + 0.5 * (plate_bbox.fb - plate_bbox.ft)
                    else:
                        # tweak up
                        l_image_y = plate_bbox.ft - 0.5 * (plate_bbox.fb - plate_bbox.ft)

                    l_image_x = k * l_image_y + b

                elif not l_in_plate and r_in_plate:
                    # tweak right point
                    k = float(r_image_x - l_image_x) / (r_image_y - l_image_y)
                    b = float(l_image_x * r_image_y - r_image_x * l_image_y) / (r_image_y - l_image_y)

                    # x = k * y + b
                    if lr_track.direction == Dir.APPROACHING:
                        # tweak up
                        r_image_y = plate_bbox.ft - 0.5 * (plate_bbox.fb - plate_bbox.ft)
                    else:
                        # tweak down
                        r_image_y = plate_bbox.fb + 0.5 * (plate_bbox.fb - plate_bbox.ft)

                    r_image_x = k * r_image_y + b

                else:
                    # both points OK
                    both_ok = True

                _dir = 1 if lr_track.direction == Dir.APPROACHING else -1

                if ((r_image_y - l_image_y) * _dir > 0 and (l_in_plate or r_in_plate)) or both_ok:
                    im_draw.line([(l_image_x, l_image_y), (r_image_x, r_image_y)], fill='white')
                    im_draw.rectangle((l_image_x - 2, l_image_y - 2, l_image_x + 2, l_image_y + 2), fill='red')
                    im_draw.rectangle((r_image_x - 2, r_image_y - 2, r_image_x + 2, r_image_y + 2), fill='red')

            except Exception as x:
                continue


class OverviewImProfile:
    _NAME = 'overview'

    def __init__(self):
        pass

    def process(self, scene_dh_data):
        """
        :type scene_dh_data: ImBuffer
        :rtype: MemBlob
        """
        wh = ImHelpers.get_jpeg_size2(scene_dh_data.buffer_data)
        blob_meta = JsonBlobMeta({
            'size': [wh[0], wh[1], scene_dh_data.buffer_used]
        })

        return MemBlob(self._NAME, scene_dh_data.ts, scene_dh_data.buffer_data, blob_meta)


class LiveImProfile:
    _NAME = 'live'
    _FONT_PT_SIZE = 20
    _FONT_ABS_PATH = os.path.join('/opt/edge/libexec/overlay/', 'ttf', 'Cousine-Bold.ttf')

    def __init__(self, args):
        """
        :type args: LiveImArgs
        """
        self._args = args
        self._dh_data = ImRgba()

        self._font = ImageFont.truetype(
            self._FONT_ABS_PATH, args.footer.font_size, encoding='utf-8')

    def _render(self, scene_dh_blob, footer_args):
        """
        :type scene_dh_blob: ImRgba
        :type footer_args: ImFooterArgs
        :rtype: str
        """

        # render footer
        scene_dh_mv = memoryview(scene_dh_blob.buffer)
        footer_offset = int(scene_dh_blob.stride * scene_dh_blob.h)
        with Image.frombuffer(
                'RGBA', (scene_dh_blob.w, scene_dh_blob.dh), scene_dh_mv[footer_offset:],
                'raw', 'RGBA', scene_dh_blob.stride, 1) as footer_im:

            footer_im.readonly = 0
            ImHelpers.draw_footer(footer_im, footer_args, self._font, self._args.footer.font_size)

        # render annotations
        with Image.frombuffer(
                'RGBA', (scene_dh_blob.w, scene_dh_blob.h + scene_dh_blob.dh), scene_dh_mv,
                'raw', 'RGBA', scene_dh_blob.stride, 1) as scene_alt_dh_im:

            scene_alt_dh_im.readonly = 0
            return ImHelpers.mk_jpeg(scene_alt_dh_im)

    def process(self, scene_dh_data, footer_vars):
        """
        :type scene_dh_data: ImRgba
        :type footer_vars: ImFooterVars
        :rtype: MemBlob
        """
        footer_args = self._args.footer.eval(footer_vars)
        blob_data = self._render(scene_dh_data, footer_args)

        return MemBlob(self._NAME, scene_dh_data.ts, blob_data)


class LiveImEngine:
    def __init__(self, args):
        """
        :type args: LiveImArgs
        """
        self._live_profile = LiveImProfile(args)

    def mk_live(self, scene_dh_data, footer_vars):
        """
        :type scene_dh_data: ImRgba
        :type footer_vars: ImFooterVars
        :rtype: MemBlob
        """
        return self._live_profile.process(scene_dh_data, footer_vars)


class OverviewImEngine:
    def __init__(self):
        self._overview_profile = OverviewImProfile()

    def mk_overview(self, scene_dh_data):
        """
        :type scene_dh_data: ImBuffer
        :rtype: MemBlob
        """
        return self._overview_profile.process(scene_dh_data)


class ImEngine:
    def __init__(self, args):
        """
        :type args: ImEngineArgs
        """
        self._vehicle_wh_ratio = args.vehicle.wh_ratio if args.vehicle else 4.0 / 3.0

        self._plate_profile = PlateImProfile(args.plate)
        self._thumb_profile = ThumbImProfile(args.thumb)
        self._vehicle_profile = VehicleImProfile(args.vehicle) if args.vehicle else None
        self._vehicle_schedule = contrast.create_schedule(args.contrast_args)
        self._scene_profile = SceneImProfile(args.scene)
        self._trainer_profile = TrainerImProfile()
        self._improc = improc.Improc()

    @staticmethod
    def mk_trainer_scene_blob(scene_dh_data):
        """
        :type scene_dh_data: ImRgba
        :rtype: MemBlob
        """
        return MemBlob('', scene_dh_data.ts, ImHelpers.mk_jpeg2(scene_dh_data))

    def mk_trainer_blob(self, scene_dh_data, plate_bbox):
        """
        :type scene_dh_data: ImRgba
        :type plate_bbox: RectF
        :rtype: MemBlob
        """
        try:
            vehicle_bbox = mk_vehicle_bbox(plate_bbox, scene_dh_data.wh, self._vehicle_wh_ratio)
        except ValueError as x:
            eventlog.log_d('fail: define vehicle-rect (%s)' % x)
            vehicle_bbox = None

        return self._trainer_profile.process(scene_dh_data, vehicle_bbox)

    def mk_blobs(self, scene_dh_data, plate_bbox, footer_vars, track, quad,
                 vehicle=False, scene=False, plate=False, thumb=False):
        """
        :type scene_dh_data: ImRgba
        :type plate_bbox: RectF
        :type footer_vars: ImFooterVars
        :type track: ITrack
        :type quad: QuadI
        :type vehicle: bool
        :type scene: bool
        :type plate: bool
        :type thumb: bool
        :rtype: list[MemBlob]
        """
        try:
            vehicle_bbox = mk_vehicle_bbox(plate_bbox, scene_dh_data.wh, self._vehicle_wh_ratio)
        except ValueError as x:
            eventlog.log_d('fail: define vehicle-rect (%s)' % x)
            vehicle_bbox = None

        # strip plate invisible parts
        visible_plate_bbox = plate_bbox
        if not plate_bbox.get_visible_100(scene_dh_data.w, scene_dh_data.h):
            visible_plate_bbox = RectF(plate_bbox.fl, plate_bbox.ft, plate_bbox.fr, plate_bbox.fb)
            visible_plate_bbox.strip_invisible(scene_dh_data.w, scene_dh_data.h)

        # The order of calls is significant
        blobs = []

        # plate
        if plate:
            blobs.extend(self._plate_profile.process(
                scene_dh_data, visible_plate_bbox))

        # thumb
        if thumb:
            blobs.extend(self._thumb_profile.process(
                scene_dh_data, visible_plate_bbox, vehicle_bbox))

        # vehicle
        if vehicle:
            improve_contrast = self._vehicle_schedule.test_active(footer_vars.dt_utc, footer_vars.geo)
            blobs.extend(self._vehicle_profile.process(
                scene_dh_data, vehicle_bbox, footer_vars, quad, self._improc if improve_contrast else None))

        # scene
        if scene:
            blobs.extend(self._scene_profile.process(
                scene_dh_data, visible_plate_bbox, vehicle_bbox, footer_vars, track, quad))

        return blobs


class ImEnv:
    def __init__(self, im_engine, footer_h, footer_def_vars):
        """
        :type im_engine: ImEngine
        :type footer_h: int
        :type footer_def_vars: dict
        """
        self.engine = im_engine
        self.scene_rgba = ImRgba()

        self.footer_h = footer_h
        self.footer_def_vars = footer_def_vars
