import ctypes
from glob import glob
from .types import LandmarkServiceArgs
import math
from edge_bbox.types import QuadI


class QuadArgs(ctypes.Structure):
    _fields_ = [
        ("x0", ctypes.c_int), ("y0", ctypes.c_int),
        ("x1", ctypes.c_int), ("y1", ctypes.c_int),
        ("x2", ctypes.c_int), ("y2", ctypes.c_int),
        ("x3", ctypes.c_int), ("y3", ctypes.c_int)
    ]


class _Api:
    _lib = ctypes.CDLL(glob('/opt/edge/lib/liblandmark*.so')[0])

    # bool
    # LlmCamCreate(
    #     double aPosL, double aPosU, double aPosF,
    #     double aRotUf, double aRotFl, double aRotLu,
    #     double aNearClipF, double aFarClipF,
    #     double aSensorWmm, double aSensorHmm,
    #     int aSensorWpx, int aSensorHpx,
    #     int aFocalDist,
    #     void **aSelf);
    camCreate = _lib.LlmCamCreate
    camCreate.argtypes = (
        ctypes.c_double, ctypes.c_double, ctypes.c_double,
        ctypes.c_double, ctypes.c_double, ctypes.c_double,
        ctypes.c_double, ctypes.c_double,
        ctypes.c_double, ctypes.c_double,
        ctypes.c_int, ctypes.c_int,
        ctypes.c_int,
        ctypes.POINTER(ctypes.c_void_p))

    camCreate.restype = ctypes.c_bool

    # void LlmCamDestroy(void* aSelf);
    camDestroy = _lib.LlmCamDestroy
    camDestroy.argtypes = (ctypes.c_void_p,)

    # bool
    # LlmCamMkImagePoint(
    #     void *aSelf,
    #     double aSceneL,
    #     double aSceneU,
    #     double aSceneF,
    #     double *anImageL,
    #     double *anImageT);
    mkImagePoint = _lib.LlmCamMkImagePoint
    mkImagePoint.argtypes = (
        ctypes.c_void_p, ctypes.c_double, ctypes.c_double, ctypes.c_double,
        ctypes.POINTER(ctypes.c_double),
        ctypes.POINTER(ctypes.c_double))
    mkImagePoint.restype = ctypes.c_bool

    # bool
    # LlmCamEstimateSpeedU(void *aSelf,
    #     struct quad *aQuad1, struct quad *aQuad2,
    #     double anUp1, double anUp2,
    #     uint64_t aTs1, uint64_t aTs2,
    #     double *aSpeed);

    measureSpeed = _lib.LlmCamEstimateSpeedU
    measureSpeed.argtypes = (
        ctypes.c_void_p, ctypes.POINTER(QuadArgs), ctypes.POINTER(QuadArgs),
        ctypes.c_double, ctypes.c_double, ctypes.c_uint64, ctypes.c_uint64,
        ctypes.POINTER(ctypes.c_double))

    measureSpeed.restype = ctypes.c_bool

    # LlmCamEstimateU(llm_cam_t aSelf, struct llm_quad *aQuad1, struct llm_quad *aQuad2,
    #     uint64_t aTs1, uint64_t aTs2, double aSpeed,
    #     double aMinUp, double aMaxUp, double *anUp);

    estimateUp = _lib.LlmCamEstimateU
    estimateUp.argtypes = (
        ctypes.c_void_p, ctypes.POINTER(QuadArgs), ctypes.POINTER(QuadArgs),
        ctypes.c_uint64, ctypes.c_uint64, ctypes.c_double, ctypes.c_double, ctypes.c_double,
        ctypes.POINTER(ctypes.c_double))

    estimateUp.restype = ctypes.c_bool
    # bool
    # LlmCamMkScenePointU(void *aSelf,
    #     double aFrameL, double aFrameT, double aSceneU,
    #     double *aSceneL, double *aSceneF);
    mkScenePoint = _lib.LlmCamMkScenePointU
    mkScenePoint.argtypes = (
        ctypes.c_void_p, ctypes.c_double, ctypes.c_double, ctypes.c_double,
        ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double))

    mkScenePoint.restype = ctypes.c_bool

    def __init__(self):
        pass


class ICameraService:
    def __init__(self):
        pass

    def mk_image_point(self, scene_l, scene_u, scene_f):
        """
        :type scene_l: float
        :type scene_u: float
        :type scene_f: float
        :rtype: (int, int)
        """
        raise NotImplementedError

    def get_speed(self, up, quad1, quad2, ts1, ts2):
        """
        :type up: float
        :type quad1: QuadI
        :type quad2: QuadI
        :type ts1: int
        :type ts2: int
        :rtype: (bool, float)
        """
        raise NotImplementedError

    def estimate_up(self, speed, quad1, quad2, ts1, ts2):
        """
        :type speed: float
        :type quad1: QuadI
        :type quad2: QuadI
        :type ts1: int
        :type ts2: int
        :rtype: (bool, float)
        """
        raise NotImplementedError

    def mk_world_point(self, image_x, image_y, world_t):
        """
        :type image_x: int
        :type image_y: int
        :type world_t: float
        :rtype: (float, float)
        """
        raise NotImplementedError


class _CameraService(ICameraService):
    def __init__(self, conf):
        """
        :type conf: LandmarkServiceArgs
        """
        ICameraService.__init__(self)
        self.handle = ctypes.c_void_p(0)
        answer = _Api.camCreate(
            conf.left, conf.up, conf.forward,
            math.radians(conf.elevation), math.radians(conf.azimuth), math.radians(conf.pitch),
            conf.clip_near, conf.clip_far,
            conf.sensor_w_mm, conf.sensor_h_mm,
            conf.sensor_w_px, conf.sensor_h_px,
            conf.focal_dist, self.handle)

    def mk_image_point(self, scene_l, scene_u, scene_f):
        """
        :type scene_l: float
        :type scene_u: float
        :type scene_f: float
        :rtype: (int, int)
        """
        image_l = ctypes.c_double()
        image_t = ctypes.c_double()
        answer = _Api.mkImagePoint(self.handle, scene_l, scene_u, scene_f, image_l, image_t)
        return int(image_l.value), int(image_t.value)

    def get_speed(self, up, quad1, quad2,  ts1, ts2):
        """
        :type up: float
        :type quad1: QuadI
        :type quad2: QuadI
        :type ts1: int
        :type ts2: int
        :rtype: (bool, float)
        """
        q1 = QuadArgs(quad1.x0, quad1.y0, quad1.x1, quad1.y1, quad1.x2, quad1.y2, quad1.x3, quad1.y3)
        q2 = QuadArgs(quad2.x0, quad2.y0, quad2.x1, quad2.y1, quad2.x2, quad2.y2, quad2.x3, quad2.y3)
        speed = ctypes.c_double()
        rc = _Api.measureSpeed(
            self.handle, ctypes.byref(q1), ctypes.byref(q2),
            up, up, ts1, ts2, ctypes.byref(speed))

        return rc, speed.value

    def mk_world_point(self, image_x, image_y, world_t):
        """
        :type image_x: int
        :type image_y: int
        :type world_t: float
        :rtype: (float, float)
        """
        scene_l = ctypes.c_double()
        scene_f = ctypes.c_double()
        rc = _Api.mkScenePoint(self.handle, image_x, image_y, world_t, scene_l, scene_f)
        return scene_f.value, scene_l.value

    def estimate_up(self, speed, quad1, quad2, ts1, ts2):
        """
        :type speed: float
        :type quad1: QuadI
        :type quad2: QuadI
        :type ts1: int
        :type ts2: int
        :rtype: (bool, float)
        """
        q1 = QuadArgs(quad1.x0, quad1.y0, quad1.x1, quad1.y1, quad1.x2, quad1.y2, quad1.x3, quad1.y3)
        q2 = QuadArgs(quad2.x0, quad2.y0, quad2.x1, quad2.y1, quad2.x2, quad2.y2, quad2.x3, quad2.y3)
        up = ctypes.c_double()
        rc = _Api.estimateUp(
            self.handle, ctypes.byref(q1), ctypes.byref(q2),
            ts1, ts2, speed, -1.0, +1.0, ctypes.byref(up))

        return rc, up.value

    def destroy(self):
        """
        :rtype: None
        """
        _Api.camDestroy(self.handle)


def init_service(conf):
    """
    :type conf: LandmarkServiceArgs
    :rtype: ICameraService
    """
    return _CameraService(conf)


# noinspection PyUnusedLocal
def fini_service(service):
    """
    :type service: ICameraService | _CameraService | None
    """
    if service:
        service.destroy()
