from edge_bbox.types import RectI
from edge.model.track import LrTrackItem
from edge.model.types import TimeSpan


class ImPlace:
    def __init__(self, r, x, y):
        """
        :type x: int
        :type y: int
        :type r: inte
        """
        self._r = r
        self._min_x = x - r
        self._max_x = x + r
        self._min_y = y - r
        self._max_y = y + r

    def __contains__(self, point_xy):
        """
        :type point_xy: (int, int)
        :rtype: bool
        """
        same_x = self._min_x < point_xy[0] < self._max_x
        same_y = self._min_y < point_xy[1] < self._max_y
        return same_x and same_y

    @property
    def xy(self):
        """
        :rtype: (int, int)
        """
        return self._min_x + self._r, self._min_y + self._r

    @property
    def r(self):
        return self._r

    @staticmethod
    def create(jnode):
        """
        :type jnode: list[int]
        :rtype: ImPlace
        """
        if len(jnode) != 3:
            raise ValueError

        return ImPlace(*jnode)

    def serialize(self):
        return [self._r, self._min_x + self._r, self._min_y + self._r]


class LrTrendItem:
    def __init__(self, place, span, event=False):
        """
        :type place: ImPlace
        :type span: TimeSpan
        :type event: bool
        """
        self.place = place
        self.range = span
        self.event = event

    @staticmethod
    def unpack(jnode):
        """
        :type jnode: dict
        :rtype: LrTrendItem
        """
        place = ImPlace.create(jnode['place'])
        span = TimeSpan.create(jnode['range'])
        return LrTrendItem(place, span, bool(jnode.get('event')))

    def serialize(self):
        """
        :rtype: dict
        """
        jnode = {
            'place': self.place.serialize(),
            'range': [self.range.min_ts, self.range.max_ts]
        }

        if self.event:
            jnode['event'] = True

        return jnode

    def set_event(self):
        self.event = True

    def clone(self):
        """
        :rtype: LrTrendItem
        """
        return LrTrendItem(self.place, TimeSpan(self.range.min_ts, self.range.max_ts), self.event)


class LrTrend(list):
    def __init__(self, plate_wh=(0, 0), items=()):
        """
        :type plate_wh: (int, int)
        :type items: list[LrTrendItem]
        """
        self.plate_w, self.plate_h = plate_wh
        list.__init__(self, items)

    @staticmethod
    def deserialize(jnode):
        """
        :type jnode: dict
        :rtype: LrTrend
        """
        plate_wh = jnode['plate-wh']
        items = map(lambda x: LrTrendItem.unpack(x), jnode['timeline'])
        return LrTrend(plate_wh, items)

    def pack(self):
        """
        :rtype: dict | None
        """
        if len(self):
            return {
                'plate-wh': [self.plate_w, self.plate_h],
                'timeline': map(lambda item: item.serialize(), self)
            }

    @property
    def min_ts(self):
        """
        :rtype: long
        """
        return self[0].range.min_ts if len(self) else 0

    @property
    def max_ts(self):
        """
        :rtype: long
        """
        return self[-1].range.max_ts if len(self) else 0

    def set_event(self, event_ts):
        """
        :type event_ts: int
        """
        for item in self:  # type: LrTrendItem
            if event_ts in item.range:
                item.set_event()
                break

    def prolong(self, lr_items, range_tout, lost_tout):
        """
        :type lr_items: list[LrTrackItem]
        :type range_tout: int
        :type lost_tout: int
        """
        entry = None
        for lr_item in lr_items:  # type: LrTrackItem
            lr_bbox = RectI(*lr_item.plate_quad.get_abs_bbox())
            if entry is None:
                # cleanup
                if len(self) and lr_item.ts > (self[-1].range.max_ts + lost_tout):
                    del self[:]  # self.clear()

                if len(self):
                    entry = self[-1]

            if entry:
                # upscale plate-wh
                if lr_bbox.iw > self.plate_w and lr_bbox.ih > self.plate_h:
                    self.plate_w = lr_bbox.iw
                    self.plate_h = lr_bbox.ih

                if lr_bbox.center in entry.place:
                    if lr_item.ts < entry.range.max_ts + range_tout:
                        entry.range.max_ts = lr_item.ts
                        continue

            entry = LrTrendItem(ImPlace(lr_bbox.ih, *lr_bbox.center), TimeSpan(lr_item.ts, lr_item.ts))
            self.append(entry)

    def retrend(self, lost_tout=0):
        """
        :type lost_tout: int
        :rtype: LrTrend[LrTrendItem]
        """
        trend_item = self[0].clone()  # type: LrTrendItem
        trend_items = [trend_item]  # type: list[LrTrendItem]

        for item in self:  # type: LrTrendItem
            if item.place.xy in trend_item.place:
                if item.range.min_ts < trend_items[-1].range.max_ts + lost_tout:
                    trend_item.range.max_ts = item.range.max_ts
                    if item.event:
                        trend_item.set_event()

                    continue

            trend_item = item.clone()
            trend_items.append(trend_item)

        return LrTrend((self.plate_w, self.plate_h), trend_items)
