import psycopg2
import psycopg2.pool
import psycopg2.extras

from edge import eventlog
from edge.com.db import DbArgs, DbJsonArg, DbError


class DbValueApi:
    def __init__(self, link_pool):
        """
        :type link_pool: psycopg2.pool.SimpleConnectionPool
        """
        self._link_pool = link_pool
        self._link = None

    def __enter__(self, *args, **kwargs):
        try:
            self._link = self._link_pool.getconn()
            self._link.__enter__(*args, **kwargs)
            return self

        except psycopg2.pool.PoolError as x:
            eventlog.log_e('fail: db-link <{}>'.format(x.message))
            raise DbError

    def __exit__(self, *args, **kwargs):
        self._link.__exit__(*args, **kwargs)
        self._link_pool.putconn(self._link)
        self._link = None

    def destroy(self):
        self._link_pool.closeall()

    def set_value(self, section, key, value_jnode):
        """
        :type section: str
        :type key: str
        :type value_jnode: dict
        """
        if not self._link:
            eventlog.log_e('fail: transaction is absent (use context manager)')
            raise DbError

        query_jnode = {
            'section': section,
            'key': key,
            'value': value_jnode
        }

        with self._link.cursor() as db_cursor:
            db_cursor.execute(
                'SELECT errcode,errmsg,result FROM api_value_set(%s)',
                [DbJsonArg(query_jnode)])

            errcode, errmsg, reply_jnode = db_cursor.fetchone()
            if errcode != 0:
                eventlog.log_e('fail: api_value_set (code={}, text={})'.format(errcode, errmsg))
                raise DbError

    def get_value(self, section, key=None):
        """
        :type section: str
        :type key: str | None
        :rtype: dict | None
        """
        if not self._link:
            eventlog.log_e('fail: transaction is absent (use context manager)')
            raise DbError

        query_jnode = {'section': section}
        if key:
            query_jnode['key'] = key

        with self._link.cursor() as db_cursor:
            db_cursor.execute(
                "SELECT errcode,errmsg,result FROM api_value_get(%s)",
                [DbJsonArg(query_jnode)])

            errcode, errmsg, reply_jnode = db_cursor.fetchone()
            if errcode != 0:
                eventlog.log_e('fail: api_value_set (code={}, text={})'.format(errcode, errmsg))
                raise DbError

            return reply_jnode

    def drop_value(self, section, key=None):
        """
        :type section: str
        :type key: str | None
        :rtype: dict
        """
        if not self._link:
            eventlog.log_e('fail: transaction is absent (use context manager)')
            raise DbError

        jnode = {'section': section}
        if key:
            jnode['key'] = key

        with self._link.cursor() as db_cursor:
            db_cursor.execute(
                "SELECT errcode,errmsg,result FROM api_value_drop(%s)",
                [DbJsonArg(jnode)])

            errcode, errmsg, reply_jnode = db_cursor.fetchone()
            if errcode != 0:
                eventlog.log_e('fail: api_value_set (code={}, text={})'.format(errcode, errmsg))
                raise DbError

            return reply_jnode


def init_service(args):
    """
    :type args: DbArgs
    :rtype: DbValueApi
    :raise DbError:
    """
    try:
        link_pool = psycopg2.pool.SimpleConnectionPool(
            minconn=1,
            maxconn=1,
            user=args.db_user,
            password=args.db_pass,
            database=args.db_name,
            host=args.db_host,
            port=args.db_port)

    except psycopg2.Error as x:
        eventlog.log_e('fail: create db-link pool <{}>'.format(x.message))
        raise DbError

    return DbValueApi(link_pool)


# noinspection PyUnusedLocal
def fini_service(service):
    """
    :type service: DbValueApi
    """
    service.destroy()
