import os
import pwd
import grp
import datetime
import ujson

import psycopg2
import psycopg2.pool
import psycopg2.extras
from psycopg2.extensions import connection as db_link, cursor

from edge import eventlog
from edge.com import db

DEF_LIST_OWNER = 'local'
DEF_LIST_NAME = 'local'

DEF_PROF_NAME = 'default'
DEF_PROF_OLD = 'none'
DEF_PROF_NEW = 'set'
DEF_PROF_SAME = 'set'

DEF_PROF_TMPL = {
    'rowsep': '[\n\r]+',
    'colsep': ',',
    'fields': {
        'plate': {
            'format': '{0}'
        }
    }
}


class ProfileInfo:
    def __init__(self, name):
        """
        :type name: str
        """
        self.name = name


class Profile:
    def __init__(self, name, desc, tmpl):
        """
        :type name: str
        :type desc: str
        :type tmpl: str
        """
        self.name = name
        self.desc = desc
        self.tmpl = tmpl


class ListInfo:
    def __init__(self, owner, name, ver_dt, ver_desc):
        """
        :type owner: str
        :type name: str
        :type ver_dt: datetime.datetime
        :type ver_desc: str
        """
        self.owner = owner
        self.ident = name
        self.utime = ver_dt
        self.comment = ver_desc

    def serialize(self):
        """
        :rtype: dict
        """
        jnode = {
            'owner': self.owner,
            'name': self.ident,
            'time': self.utime,
        }

        if len(self.comment):
            jnode['comment'] = self.comment

        return jnode


class ListVersion:
    def __init__(self, ver_dt, ver_rev, ver_desc):
        """
        :type ver_dt: datetime.datetime
        :type ver_rev: str
        :type ver_desc: str
        """
        self.utime = ver_dt
        self.revision = ver_rev
        self.comment = ver_desc


class ListMatch:
    def __init__(self, plate_list, plate_text, plate_dist, plate_extra):
        """
        :type plate_list: ListInfo
        :type plate_text: str
        :type plate_dist: int
        :type plate_extra: dict
        """
        self.list = plate_list
        self.plate = plate_text
        self.dist = plate_dist
        self.extra = plate_extra

    @staticmethod
    def unpack(jnode):
        """
        :type jnode: dict
        :rtype: ListMatch
        """
        owner = jnode['list-owner']
        name = jnode['list-ident']
        ver_dt = jnode['list-date']
        ver_desc = jnode.get('list-description', '')
        list_info = ListInfo(owner, name, ver_dt, ver_desc)

        plate_text = jnode['plate']
        plate_dist = jnode['plate-distance']

        keys_to_delete = ['list-owner', 'list-ident', 'list-date', 'list-description', 'plate', 'plate-distance']

        for key in keys_to_delete:
            if jnode.get(key) is not None:
                del jnode[key]

        return ListMatch(list_info, plate_text, plate_dist, jnode)

    def serialize(self):
        """
        :rtype: dict
        """
        jnode = {
            'plate-list': self.list.serialize(),
            'plate-text': self.plate,
            'plate-dist': self.dist,
            'plate-extra': self.extra
        }

        return jnode


class ListHistoryItem:
    def __init__(self, moment, revision, sha1, template, description, operator, inserted, updated, deleted):
        """
        :type moment:
        :type revision: str
        :type sha1: str
        :type template: str
        :type description: str | None
        :type operator: str | None
        :type inserted: int
        :type updated: int
        :type deleted: int
        """
        self.moment = moment
        self.revision = revision
        self.sha1 = sha1
        self.template = template
        self.description = description
        self.operator = operator
        self.inserted = inserted
        self.updated = updated
        self.deleted = deleted

    @staticmethod
    def unpack(jnode):
        """
        :type jnode: dict
        :rtype: ListHistoryItem
        """
        moment = jnode['moment']
        revision = jnode['revision']
        sha1 = jnode['sha1']

        template = jnode['template']
        description = jnode.get('description')
        operator = jnode.get('operator')

        inserted = jnode['inserted']
        updated = jnode['updated']
        deleted = jnode['deleted']

        return ListHistoryItem(moment, revision, sha1, template, description, operator, inserted, updated, deleted)


class PlateHistoryItem:
    def __init__(self, moment, revision, op, sha1, operator, description):
        """
        :type moment:
        :type revision: str
        :type op: str
        :type sha1: str
        :type operator: str | None
        :type description: str | None
        """
        self.moment = moment
        self.revision = revision
        self.sha1 = sha1
        self.description = description
        self.operator = operator
        self.op = op

    @staticmethod
    def unpack(jnode):
        """
        :type jnode: dict
        :rtype: PlateHistoryItem
        """
        moment = jnode['moment']
        revision = jnode['revision']
        sha1 = jnode['sha1']

        op = jnode['op']
        description = jnode.get('description')
        operator = jnode.get('operator')

        return PlateHistoryItem(moment, revision, op, sha1, operator, description)


def link_guard(fn):
    def proxy_fn(self, *args, **kwargs):
        """
        :type self: DbWantedApi
        """
        if not self._link:
            eventlog.log_e('fail: db-link (use context manager)')
            raise db.DbError

        return fn(self, *args, **kwargs)
    return proxy_fn


class DbWantedApi:
    """
    :type _link: db_link | None
    """
    def __init__(self, link_pool):
        """
        :type link_pool: psycopg2.pool.SimpleConnectionPool
        """
        self._link_pool = link_pool
        self._link = None

    def __enter__(self, *args, **kwargs):
        try:
            self._link = self._link_pool.getconn()
            self._link.__enter__(*args, **kwargs)
            return self

        except psycopg2.pool.PoolError as x:
            eventlog.log_e('fail: db-link ({})'.format(x.message))
            raise db.DbError

    def __exit__(self, *args, **kwargs):
        self._link.__exit__(*args, **kwargs)
        self._link_pool.putconn(self._link)
        self._link = None

    def destroy(self):
        self._link_pool.closeall()

    @link_guard
    def load_profile_infos(self):
        """
        :rtype: list[ProfileInfo]
        """
        with self._link.cursor() as db_cursor:
            db_cursor.execute('SELECT ident FROM hwp.watchlistpl;')
            names = db_cursor.fetchall()
            return map(ProfileInfo, names)

    @link_guard
    def load_profile(self, name):
        """
        :type name: str
        :rtype: Profile | None
        """
        try:
            with self._link.cursor() as db_cursor:  # type: cursor
                db_cursor.execute("SELECT * FROM hwp.watchlistpl('list',%s,'',%s);", [name, None])
                answer_row = db_cursor.fetchone()
                answer_jnode = answer_row[0]  # type: [dict]
                if answer_jnode is None:
                    return None
                elif len(answer_jnode) == 0:
                    return None
                elif len(answer_jnode) > 1:
                    raise NotImplementedError

                p_jnode = answer_jnode[0]

                desc = p_jnode.get('description', '')
                tmpl = p_jnode['template']
                return Profile(name, desc, tmpl)

        except psycopg2.Error as x:
            eventlog.log_e('fail: db (pgcode:%s, pgerror:%s)' % (x.pgcode, x.pgerror))
            raise NotImplementedError

    @link_guard
    def save_profile(self, name, desc, tmpl):
        """
        :type name: str
        :type desc: str
        :type tmpl: dict
        """
        try:
            with self._link.cursor() as db_cursor:  # type: cursor
                db_cursor.execute(
                    "SELECT * FROM hwp.watchlistpl('set',%s,%s,%s);",
                    [name, desc, db.DbJsonArg(tmpl)])

        except psycopg2.Error as x:
            db.trap_hint_error(x)
            eventlog.log_e('fail: db (pgcode:%s, pgerror:%s)' % (x.pgcode, x.pgerror))
            raise db.DbError

    @link_guard
    def drop_profile(self, name):
        """
        :type name: str
        """
        try:
            with self._link.cursor() as db_cursor:  # type: cursor
                db_cursor.execute("SELECT * FROM hwp.watchlistpl('delete',%s);", [name])
        except psycopg2.Error as x:
            db.trap_hint_error(x)
            eventlog.log_e('fail: db (pgcode:%s, pgerror:%s)' % (x.pgcode, x.pgerror))
            raise db.DbError

    @link_guard
    def load_list_infos(self):
        """
        :rtype: list[ListInfo]
        """
        with self._link.cursor() as db_cursor:
            db_cursor.execute('SELECT owner,ident,utime,description FROM hwp.watchlist;')
            answer_rows = db_cursor.fetchall()

            list_infos = []
            for owner, name, ver_dt, ver_desc in answer_rows:
                list_infos.append(ListInfo(owner, name, ver_dt, ver_desc))

            return list_infos

    @link_guard
    def load_list_versions(self, owner, name):
        """
        :type owner: str
        :type name: str
        :rtype: list[ListVersion]
        """
        with self._link.cursor() as db_cursor:
            db_cursor.execute(
                'SELECT moment,revision,description FROM hwp.watchlistlog '
                'WHERE owner=%s and ident=%s ORDER BY moment DESC;',
                [owner, name])
            answer_rows = db_cursor.fetchall()

        list_versions = []
        for ver_dt, ver_rev, ver_desc in answer_rows:
            list_versions.append(ListVersion(ver_dt, ver_rev, ver_desc))

        return list_versions

    @link_guard
    def drop_list(self, owner, name):
        """
        :type owner: str
        :type name: str

        :return: Affected row count.
        :rtype: int
        """
        with self._link.cursor() as db_cursor:  # type: cursor
            db_cursor.execute(
                'SELECT affected FROM hwp.watchlist_delete(%s,%s);',
                [owner, name])

            answer_row = db_cursor.fetchone()
            return answer_row[0]

    @link_guard
    def save_list(self, list_owner, list_name, list_file, list_sha1, list_desc,
                  prof_tmpl, prof_name, prof_old, prof_new, prof_same, operator='dbuser'):
        """
        :type list_owner: str
        :type list_name: str

        :param list_file: list-file-path or list-file-data
        :type list_file: str

        :type list_sha1: str | None
        :type list_desc: str | None

        :param prof_tmpl: Json. Has higher priority then prof_name.
        :type prof_tmpl: str | None

        :type prof_name: str | None
        :type prof_old: str
        :type prof_new: str
        :type prof_same: str
        :type operator: str
        """
        if os.path.isfile(list_file):
            uid = pwd.getpwnam('postgres').pw_uid
            gid = grp.getgrnam('postgres').gr_gid
            os.chown(list_file, uid, gid)

        if prof_tmpl:
            prof_tmpl = ujson.dumps(prof_tmpl)

        with self._link.cursor() as db_cursor:
            try:
                db_cursor.execute(
                    'SELECT total,inserted,deleted,updated,errors '
                    'FROM hwp.watchload(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)',
                    [list_owner, list_name, prof_tmpl, prof_name, list_file, list_sha1,
                     list_desc, operator, prof_old, prof_same, prof_new])

            except psycopg2.Error as x:
                db.trap_hint_error(x)
                eventlog.log_e('fail: db (pgcode:%s, pgerror:%s)' % (x.pgcode, x.pgerror))
                raise db.DbError

            total, inserted, deleted, updated, errors = db_cursor.fetchone()
            eventlog.log_i(
                'done: db (total:%s, inserted:%s, deleted:%s, updated:%s, errors:%s)' %
                (total, inserted, deleted, updated, len(errors)))

            if len(errors) != 0:
                for error in errors:
                    line, raw_data, error = error
                    data = raw_data.encode('utf-8')
                    eventlog.log_e('line:{}, data:<{}>, error:<{}>'.format(line, data, error))
                raise db.DbError

    @link_guard
    def cleanup_list(self, owner, name, ver_max_rev=None, ver_max_hash=None, ver_max_time=None):
        """
        :type owner: str
        :type name: str

        :type ver_max_rev: str | None
        :type ver_max_hash: str | None
        :type ver_max_time: str | None

        :return: Affected rows count.
        :rtype: int
        """
        try:
            eventlog.log_i(
                'exec: SELECT affected FROM hwp.watchlist_cleanup(%s,%s,%s,%s,%s);' %
                (owner, name, ver_max_time, ver_max_rev, ver_max_hash))

            with self._link.cursor() as db_cursor:  # type: cursor
                db_cursor.execute(
                    'SELECT affected FROM hwp.watchlist_cleanup(%s,%s,%s,%s,%s);',
                    [owner, name, ver_max_time, ver_max_rev, ver_max_hash])

                answer_row = db_cursor.fetchone()
                return answer_row[0]

        except psycopg2.Error as x:
            db.trap_hint_error(x)
            eventlog.log_e('fail: db (pgcode:%s, pgerror:%s)' % (x.pgcode, x.pgerror))
            raise db.DbError

    @link_guard
    def revert_list(self, owner, name, ver_min_rev=None, ver_count=None):
        """
        :type owner: str
        :type name: str
        :type ver_min_rev: str | None
        :type ver_count: int | None
        :rtype: int
        """
        try:
            with self._link.cursor() as db_cursor:  # type: cursor
                db_cursor.execute(
                    'SELECT inserted,updated,deleted FROM hwp.watchlist_revert(%s,%s,%s,%s);',
                    [owner, name, ver_count, ver_min_rev])

                inserted, updated, deleted = db_cursor.fetchone()
                eventlog.log_i('done: inserted:%s, updated:%s, deleted:%s' % (inserted, updated, deleted))

        except psycopg2.Error as x:
            db.trap_hint_error(x)
            eventlog.log_e('fail: pgcode:%s, pgerror:%s' % (x.pgcode, x.pgerror))
            raise db.DbError

    @link_guard
    def check_plate(self, plate_text, max_dist=0):
        """
        :type plate_text: str
        :type max_dist: int
        :rtype: list[ListMatch]
        """
        with self._link.cursor() as db_cursor:  # type: cursor
            db_cursor.execute(
                'SELECT watchlist FROM hwp.watchlist(%s,%s);',
                [plate_text, max_dist])

            answer_row = db_cursor.fetchone()
            # todo :check
            if len(answer_row):
                return map(ListMatch.unpack, answer_row[0])
            else:
                return []

    @link_guard
    def get_list_history(self, list_owner, list_name, limit=1, offset=0):
        """
        :type list_owner: str
        :type list_name: str
        :type limit:int
        :type offset: int
        :rtype list[ListHistoryItem]
        """
        with self._link.cursor() as db_cursor:
            try:
                db_cursor.execute(
                    'SELECT listwatchlog '
                    'FROM hwp.listwatchlog(%s,%s,%s,%s)',
                    [list_owner, list_name, limit, offset])

            except psycopg2.Error as x:
                db.trap_hint_error(x)
                eventlog.log_e('fail: db (pgcode:%s, pgerror:%s)' % (x.pgcode, x.pgerror))
                raise db.DbError

            answer_row = db_cursor.fetchone()

        if len(answer_row):
            return map(ListHistoryItem.unpack, answer_row[0])
        else:
            return []

    @link_guard
    def get_plate_history(self, list_owner, list_name, plate_text, limit=1, offset=0):
        """
        :type list_owner: str
        :type list_name: str
        :type plate_text: str
        :type limit:int
        :type offset: int
        :rtype list[PlateHistoryItem]
        """
        with self._link.cursor() as db_cursor:
            try:
                db_cursor.execute(
                    'SELECT listwatchplatehistory '
                    'FROM hwp.listwatchplatehistory(%s,%s,%s,%s,%s)',
                    [list_owner, list_name, plate_text, limit, offset])

            except psycopg2.Error as x:
                db.trap_hint_error(x)
                eventlog.log_e('fail: db (pgcode:%s, pgerror:%s)' % (x.pgcode, x.pgerror))
                raise db.DbError

            answer_row = db_cursor.fetchone()

        if len(answer_row):
            return map(PlateHistoryItem.unpack, answer_row[0])
        else:
            return []


def init_service(args):
    """
    :type args: db.DbArgs
    :rtype: DbWantedApi
    :raise db.DbError:
    """
    try:
        link_pool = psycopg2.pool.SimpleConnectionPool(
            minconn=1,
            maxconn=1,
            user=args.db_user,
            password=args.db_pass,
            database=args.db_name,
            host=args.db_host,
            port=args.db_port)

    except psycopg2.Error as x:
        eventlog.log_e('fail: create db-link pool <{}>'.format(x.message))
        raise db.DbError

    return DbWantedApi(link_pool)


# noinspection PyUnusedLocal
def fini_service(service):
    """
    :type service: DbWantedApi
    """
    service.destroy()
