from twisted.internet.protocol import ReconnectingClientFactory, connectionDone
from twisted.protocols.basic import LineOnlyReceiver

import time
import sortedcontainers
import gps.misc
import ujson

from edge.com import astral
from edge.service import tz
from edge import eventlog
from edge.model.geoposition import Geoposition
from edge.model.stats_provider import IStatsProvider
from edge.service.types import GpsdServiceArgs


_TPV_MODE_NONE = 0
_TPV_MODE_NO_FIX = 1
_TPV_MODE_2D_FIX = 2
_TPV_MODE_3D_FIX = 3


class _GeodataCounters:
    """
    :type mts: long | None
    :type tpv_ts: long | None
    :type tpv_lat: float | None
    :type tpv_lon: float | None
    :type tpv_alt: float | None
    """
    def __init__(self):
        self.mts = None

        self.tpv_ts = None
        self.tpv_lat = None
        self.tpv_lon = None
        self.tpv_alt = None

    def inc(self, tpv_ts, tpv_lat, tpv_lon, tpv_alt):
        """
        :type tpv_ts: long | None
        :type tpv_lat: float | None
        :type tpv_lon: float | None
        :type tpv_alt: float | None
        """
        self.mts = int(time.time())

        self.tpv_ts = tpv_ts
        self.tpv_lat = tpv_lat
        self.tpv_lon = tpv_lon
        self.tpv_alt = tpv_alt


class _GpsdCounters:
    """
    :type mode: int
    :type mode_mts: int | None
    :type store_count: long
    """

    MODE_0D_FIX = 0
    MODE_1D_FIX = 1
    MODE_2D_FIX = 2
    MODE_3D_FIX = 3

    def __init__(self):
        self.mode = 0
        self.mode_mts = None

        self.store_count = long()
        self.geodata = _GeodataCounters()

    def inc_mode(self, mode):
        if self.mode != mode:
            self.mode = mode
            self.mode_mts = int(time.time())

    def inc_store_count(self, count):
        self.store_count += count


class IGeolocationService(IStatsProvider):
    def __init__(self):
        IStatsProvider.__init__(self)

    def get_geoposition(self, ts):
        """
        :type ts: int
        :rtype: Geoposition | None
        """
        raise NotImplementedError


class _GeopositionLog:
    def __init__(self, max_entry_count, max_age_delta_usec, max_lookup_dist_usec):
        """
        :type max_entry_count: int
        :type max_age_delta_usec: int
        :type max_lookup_dist_usec: int
        """
        self._max_entry_count = max_entry_count
        self._max_age_delta_usec = max_age_delta_usec
        self._max_lookup_dist_usec = max_lookup_dist_usec

        self._ts_to_geodata_map = sortedcontainers.SortedDict()

    def add_record(self, ts, geoposition):
        """
        :type ts: int
        :type geoposition: Geoposition
        """
        self._ts_to_geodata_map[ts] = geoposition

        while len(self._ts_to_geodata_map) > self._max_entry_count:
            self._ts_to_geodata_map.popitem(last=False)

        while (self._ts_to_geodata_map.peekitem()[0] - self._ts_to_geodata_map.peekitem(0)[0]) \
                > self._max_age_delta_usec:
            self._ts_to_geodata_map.popitem(last=False)

    def lookup_nearest(self, lookup_ts):
        """
        :type lookup_ts: int
        :rtype: Geoposition | None
        """
        entry_count = len(self._ts_to_geodata_map)
        if not entry_count:
            return None

        # insert lookup_ts
        bisect_index = self._ts_to_geodata_map.bisect_right(lookup_ts)
        if 0 < bisect_index < entry_count:
            # exact-match with lookup_ts
            less_equal_ts_dist = lookup_ts - self._ts_to_geodata_map.iloc[bisect_index - 1]
            if less_equal_ts_dist == 0:
                return self._ts_to_geodata_map.peekitem(bisect_index - 1)[1]

            # greater_ts is better
            greater_ts_dist = self._ts_to_geodata_map.iloc[bisect_index] - lookup_ts
            if less_equal_ts_dist < greater_ts_dist and less_equal_ts_dist <= self._max_lookup_dist_usec:
                return self._ts_to_geodata_map.peekitem(bisect_index - 1)[1]

            # less_equal_ts is better
            return self._ts_to_geodata_map.peekitem(bisect_index)[1] \
                if greater_ts_dist <= self._max_lookup_dist_usec else None

        # push-back lookup_ts
        if bisect_index == entry_count:
            less_equal_ts_dist = lookup_ts - self._ts_to_geodata_map.iloc[entry_count - 1]
            return self._ts_to_geodata_map.peekitem(bisect_index - 1)[1] \
                if less_equal_ts_dist <= self._max_lookup_dist_usec else None

        # push-front lookup_ts
        greater_ts_dist = self._ts_to_geodata_map.iloc[bisect_index] - lookup_ts
        if greater_ts_dist <= self._max_lookup_dist_usec:
            return self._ts_to_geodata_map.peekitem(bisect_index)[1]
        else:
            eventlog.log_w('fail: target-ts:{}usec, min-ts:{}usec, dt:{}sec'.format(
                lookup_ts, self._ts_to_geodata_map.peekitem(0)[0],
                (self._ts_to_geodata_map.peekitem(0)[0] - lookup_ts) / 1e6))
            return None

    def get_count(self):
        return len(self._ts_to_geodata_map)


class _GpsdProtocol(LineOnlyReceiver):
    delimiter = b'\n'

    def __init__(self, geodata_store, stat_counters=None):
        """
        :type geodata_store: _GeopositionLog
        :type stat_counters: _GpsdCounters | None
        """
        self._geodata_store = geodata_store
        self._stat_counters = stat_counters

    def connectionMade(self):
        # noinspection PyBroadException
        try:
            self.sendLine('?WATCH={"enable":true,"json":true}')
        except Exception:
            eventlog.log_x()

    # Called when an established connection is lost.
    def connectionLost(self, reason=connectionDone):
        pass

    def lineReceived(self, raw_reply):
        try:
            self._handle(raw_reply)

        except ValueError:
            eventlog.log_x()

    def _handle(self, raw_reply):
        """
        :type raw_reply: str
        """
        reply_jnode = ujson.loads(raw_reply)
        if reply_jnode['class'] == 'TPV':
            # Time/date stamp in ISO8601 format, UTC.
            # May have a fractional part of up to .001sec precision. May be absent if mode is not 2 or 3.
            raw_tpv_time = reply_jnode.get('time')
            tpv_time_usec = None
            tpv_time_sec = None
            if raw_tpv_time:
                tpv_iso_time = gps.isotime(raw_tpv_time.encode("ascii"))
                tpv_time_usec = int(tpv_iso_time * 1e6)
                tpv_time_sec = int(tpv_iso_time)

            # Latitude in degrees: +/- signifies North/South. Present when mode is 2 or 3.
            tpv_lat = reply_jnode.get('lat')

            # Longitude in degrees: +/- signifies East/West. Present when mode is 2 or 3.
            tpv_lon = reply_jnode.get('lon')

            # Altitude in meters. Present if mode is 3.
            tpv_alt = reply_jnode.get('alt')

            # Speed over ground, meters per second.
            tpv_speed = reply_jnode.get('speed')

            # TODO: implement stub stat counters
            if self._stat_counters:
                self._stat_counters.geodata.inc(tpv_time_sec, tpv_lat, tpv_lon, tpv_alt)

            # NMEA mode: %d, 0=no mode value yet seen, 1=no fix, 2=2D, 3=3D.
            tpv_mode = reply_jnode['mode']
            if self._stat_counters:
                self._stat_counters.inc_mode(tpv_mode)

            if tpv_time_usec is not None and tpv_mode in [_TPV_MODE_2D_FIX, _TPV_MODE_3D_FIX]:
                self._geodata_store.add_record(tpv_time_usec, Geoposition(tpv_lat, tpv_lon, tpv_alt, tpv_speed))
                if self._stat_counters:
                    self._stat_counters.inc_store_count(self._geodata_store.get_count())


class _GpsdClientFactory(ReconnectingClientFactory):
    initialDelay = 1
    maxDelay = 15
    factor = 1.6180339887498948

    def __init__(self, geodata_store, stat_counters=None):
        """
        :type geodata_store: _GeopositionLog
        :type stat_counters: _GpsdCounters | None
        """
        self._geodata_store = geodata_store
        self._stat_counters = stat_counters

    def buildProtocol(self, addr):
        self.resetDelay()
        protocol = _GpsdProtocol(self._geodata_store, self._stat_counters)
        protocol.factory = self
        return protocol

    def clientConnectionFailed(self, connector, reason):
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)

    def clientConnectionLost(self, connector, unused_reason):
        ReconnectingClientFactory.clientConnectionLost(self, connector, unused_reason)


class _BaseGeolocationService(IGeolocationService):
    def __init__(self, twisted_reactor, endpoint, geodata_store, stat_counters=None, azimuth=None):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorTCP
        :type endpoint: (str, int)
        :type geodata_store: _GeopositionLog
        :type stat_counters: _GpsdCounters | None
        :type azimuth: float | None
        """
        IGeolocationService.__init__(self)

        self._azimuth = azimuth
        self._geodata_store = geodata_store

        gpsd_host, gpsd_port = endpoint
        client_factory = _GpsdClientFactory(self._geodata_store, stat_counters)
        twisted_reactor.connectTCP(gpsd_host, gpsd_port, client_factory)

    def get_geoposition(self, ts):
        """
        :type ts: int
        :rtype: Geoposition | None
        """
        geo = self._geodata_store.lookup_nearest(ts)
        if geo is not None:
            geo.azimuth = self._azimuth
        return geo


class _GeolocationService(_BaseGeolocationService):
    def __init__(self, twisted_reactor, tz_service, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorTCP
        :type tz_service: tz.ITzService
        :type args: GpsdServiceArgs
        """
        self._azimut = args.azimuth
        self._counters = _GpsdCounters() if bool(args.endpoint) else None
        self._tz_service = tz_service
        geodata_store = _GeopositionLog(args.max_entry_count, args.max_age_delta_usec, args.max_lookup_dist_usec)
        _BaseGeolocationService.__init__(self, twisted_reactor, args.endpoint, geodata_store, self._counters, args.azimuth)

    def collect_stats(self, json_handler):
        """
        :type json_handler: (dict) -> None
        :rtype: None
        """
        if self._counters is None:
            return

        stat_jnode = {
            'last-geoposition': {
                'azimuth': self._azimut,
            }
        }

        if self._counters.mode_mts is not None:
            stat_jnode['last-status'] = {
                'mtime': [
                    self._tz_service.get_datetime_isostr(self._counters.mode_mts),
                    int(self._counters.mode_mts)
                ],
                'fix-mode': self._counters.mode
            }

        if self._counters.geodata.mts is not None:
            geo_jnode = {
                'mtime': [
                    self._tz_service.get_datetime_isostr(self._counters.geodata.mts),
                    int(self._counters.geodata.mts)
                ]
            }

            if self._counters.geodata.tpv_ts is not None:
                geo_jnode['time'] = [
                    self._tz_service.get_datetime_isostr(self._counters.geodata.tpv_ts),
                    int(self._counters.geodata.tpv_ts)
                ]

            if None not in [self._counters.geodata.tpv_lat, self._counters.geodata.tpv_lon]:
                geo = Geoposition(self._counters.geodata.tpv_lat, self._counters.geodata.tpv_lon)
                astral_loc = astral.Location(geo.latitude, geo.longitude)
                # <latChar><DDD.uuuuuu><longChar><ddd.uuuuuu>
                coordinates = '{}{:.6f}{}{:.6f}'.format(
                    'N' if geo.latitude >= 0 else 'S', abs(geo.latitude),
                    'E' if geo.longitude >= 0 else 'W', abs(geo.longitude))

                geo_jnode['latitude'] = geo.latitude
                geo_jnode['longitude'] = geo.longitude
                geo_jnode['coordinates'] = coordinates
                geo_jnode['sunset'] = self._tz_service.datetime_to_isostr(astral_loc.sunset())
                geo_jnode['sunrise'] = self._tz_service.datetime_to_isostr(astral_loc.sunrise())

            # add altitude
            if self._counters.geodata.tpv_alt is not None:
                geo_jnode['altitude'] = self._counters.geodata.tpv_alt

            stat_jnode['last-geoposition'] = geo_jnode

        json_handler({
            'type': 'geolocation',
            'name': 'geolocation.0',
            'stat': stat_jnode
        })


class _StubGeolocationService(IGeolocationService):
    def __init__(self, geoposition, tz_service):
        """
        :type geoposition: Geoposition | None
        :type tz_service: tz.ITzService
        """
        IGeolocationService.__init__(self)

        # http://catb.org/gpsd/gpsd_json.html
        self._geoposition = geoposition
        self._astral_loc = astral.Location(self._geoposition.latitude, self._geoposition.longitude) if geoposition else None
        self._tz_service = tz_service

    def get_geoposition(self, ts):
        """
        :type ts: int
        :rtype: Geoposition | None
        """
        return self._geoposition

    def collect_stats(self, json_handler):
        """
        :type json_handler: (dict) -> None
        :rtype: None
        """
        if self._geoposition is None:
            return

        # <latChar><DDD.uuuuuu><longChar><ddd.uuuuuu>
        coordinates = '{}{:.6f}{}{:.6f}'.format(
            'N' if self._geoposition.latitude >= 0 else 'S', abs(self._geoposition.latitude),
            'E' if self._geoposition.longitude >= 0 else 'W', abs(self._geoposition.longitude))

        jnode = {
            'type': 'geolocation',
            'name': 'geolocation.0',
            'stat': {
                'last-geoposition': {
                    'mtime': self._tz_service.get_datetime_now_isostr_seconds(),
                    'coordinates': coordinates,
                    'latitude': self._geoposition.latitude,
                    'longitude': self._geoposition.longitude,
                    'sunset': self._tz_service.datetime_to_isostr(self._astral_loc.sunset()),
                    'sunrise': self._tz_service.datetime_to_isostr(self._astral_loc.sunrise())
                }
            }
        }

        if self._geoposition.altitude is not None:
            jnode['stat']['last-geoposition']['altitude'] = self._geoposition.altitude

        if self._geoposition.azimuth is not None:
            jnode['stat']['last-geoposition']['azimuth'] = self._geoposition.azimuth

        json_handler(jnode)


def init_service(twisted_reactor, args, tz_service):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorTCP
    :type args: GpsdServiceArgs
    :type tz_service: tz.ITzService
    :rtype: IGeolocationService
    """
    if args.provider == args.PROVIDER_STATIC:
        service = _StubGeolocationService(args.geodata, tz_service)
    else:
        service = _GeolocationService(twisted_reactor, tz_service, args)

    return service


# noinspection PyUnusedLocal
def fini_service(service):
    """
    :type service: IGeolocationService | _GeolocationService
    """
    pass
