import pytz

from edge.service.types import GpsdServiceArgs, TzServiceArgs

from edge import eventlog
from edge.com import fs
from edge.types import JnodeError
from edge.model.geoposition import Geoposition
from edge.webserver.sysconf.uiconf.time import TimeUiconf


class JsonGpsdServiceArgs(GpsdServiceArgs):
    def __init__(self, *args, **kwargs):
        GpsdServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def default():
        """
        :rtype: JsonGpsdServiceArgs
        """
        return GpsdServiceArgs(endpoint=('localhost', 2947))

    @staticmethod
    def deserialzie(raw_jnode):
        """
        :type raw_jnode: dict

        :rtype: JsonGpsdServiceArgs
        :raises JnodeError:
        """
        rt_jnode = raw_jnode.get('rt')
        if rt_jnode is None:
            return JsonGpsdServiceArgs.default()

        az = rt_jnode.get('azimuth')
        jnode = rt_jnode.get('source')
        if jnode is None:
            eventlog.log_e_param('#/')
            raise JnodeError

        if not (type(jnode) in [tuple, list] and len(jnode) == 3):
            eventlog.log_e_param('#/')
            raise JnodeError

        # deserialize geodata
        if jnode[0] == 'static':
            _, latitude, longitude = jnode
            if not (type(latitude) in [int, float] and -90.0 <= latitude <= +90.0):
                eventlog.log_e_param('#/1')
                raise JnodeError

            if not (type(longitude) in [int, float] and -180.0 <= longitude <= +180.0):
                eventlog.log_e_param('#/2')
                raise JnodeError

            return JsonGpsdServiceArgs(provider=JsonGpsdServiceArgs.PROVIDER_STATIC,
                                       geodata=Geoposition(latitude, longitude, azimuth=az))

        if jnode[0] == 'gpsd':
            _, gpsd_host, gpsd_port = jnode
            if not (type(gpsd_host) in [str, unicode] and len(gpsd_host) <= 32):
                eventlog.log_e_param('#/1')
                raise JnodeError

            if not (type(gpsd_port) in [int] and 0 < gpsd_port <= 65535):
                eventlog.log_e_param('#/2')
                raise JnodeError

            return JsonGpsdServiceArgs(provider=JsonGpsdServiceArgs.PROVIDER_GPSD, endpoint=(gpsd_host, gpsd_port))

        eventlog.log_e_param('#/0')
        raise JnodeError

    @classmethod
    def create(cls, conf_path):
        """
        :type conf_path: str

        :rtype: JsonGpsdServiceArgs
        :raise JnodeError:
        """
        conf_jnode = fs.load_json3(conf_path, default=None)
        if not conf_jnode:
            eventlog.log_e_param('#/')
            raise JnodeError

        return cls.deserialzie(conf_jnode)


class JsonTzServiceArgs(TzServiceArgs):
    def __init__(self, *args, **kwargs):
        TzServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(jnode):
        """
        :type jnode: dict

        :rtype: JsonTzServiceArgs
        :raise JnodeError:
        """
        # deserialize #/rw
        rw_jnode = jnode.get('rw')
        if dict != type(rw_jnode):
            eventlog.log_e_param('#/rw')
            raise JnodeError

        # deserialize #/rw/tz
        tz_name = rw_jnode.get('args', {}).get('tz')
        if not (type(tz_name) in [unicode, str] and 3 <= len(tz_name) <= 32):
            eventlog.log_e_param('#/rw/tz')
            raise JnodeError

        if tz_name not in pytz.all_timezones:
            eventlog.log_e_param('#/rw/tz')
            raise JnodeError

        tz_info = pytz.timezone(tz_name)
        return JsonTzServiceArgs(tz_info)

    @classmethod
    def create(cls, conf_path=TimeUiconf.CONF_PATH):
        """
        :type conf_path: str

        :rtype: JsonTzServiceArgs
        :raise JnodeError:
        """
        conf_jnode = fs.load_json3(conf_path, default=None)
        if not conf_jnode:
            eventlog.log_e_param('#/')
            raise JnodeError

        return cls.deserialize(conf_jnode)
