# noinspection PyProtectedMember
from twisted.internet.inotify import (INotify, INotifyError,
                                      IN_MOVED_TO, IN_CLOSE_WRITE,
                                      IN_DELETE_SELF, IN_MOVE_SELF)
from twisted.python.filepath import FilePath

from edge import eventlog
from edge.exceptions import *
from edge.com import fs


class ReactError(Exception):
    pass


class IReactService:
    def __init__(self):
        pass

    def set_handler(self, handler):
        """
        :type handler: (str, dict) -> None
        """
        raise NotImplementedError

    def set_endpoints(self, endpoints):
        """
        :type endpoints: list[str]
        """
        raise NotImplementedError


def init_service(twisted_reactor, endpoints):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
    :type endpoints: list[str]
    :rtype: IReactService
    """
    service = _ReactService(twisted_reactor)
    service.set_endpoints(endpoints)
    return service


def fini_service(service):
    """
    :type service: IReactService | _ReactService
    """
    service.destroy()


class _ReactService(IReactService):
    _WATCH_MASK = IN_MOVED_TO | IN_CLOSE_WRITE | IN_DELETE_SELF | IN_MOVE_SELF

    def __init__(self, twisted_reactor):

        IReactService.__init__(self)

        self._reactor = twisted_reactor
        self._endpoints = set()

        self._inotify = INotify(twisted_reactor)
        self._inotify.startReading()

        self._handler = None

    def destroy(self):
        map(self._unwatch, self._endpoints)

    def set_handler(self, handler):
        """
        :type handler: (str, dict) -> None
        """
        self._handler = handler

    def set_endpoints(self, endpoints):
        """
        :type endpoints: list[str]
        """
        last_set = self._endpoints
        next_set = set(endpoints)
        same_set = last_set & next_set

        rm_set = last_set - same_set
        mk_set = next_set - same_set

        self._endpoints = next_set
        map(self._unwatch, rm_set)
        map(self._rewatch, mk_set)

    def _rewatch(self, endpoint):
        """
        :type endpoint: str
        """

        # noinspection PyBroadException
        try:
            try:
                self._inotify.ignore(FilePath(endpoint))
                eventlog.log_i('done: ignore {}'.format(endpoint))
            except KeyError:
                pass

            fs.silent_mkdir_r(endpoint)
            self._inotify.watch(FilePath(endpoint), self._WATCH_MASK, callbacks=[self._handle_react])
            eventlog.log_i('done: watch {}'.format(endpoint))

        except INotifyError as x:
            eventlog.log_e('fail: watch {} <{}>'.format(endpoint, x.message))

        except Exception:
            eventlog.log_x()

    def _unwatch(self, endpoint):
        """
        :type endpoint: str
        """
        try:
            self._inotify.ignore(FilePath(endpoint))
            eventlog.log_i('done: ignore {}'.format(endpoint))
        except KeyError:
            pass

    def _handle_react(self, _, filepath, mask):
        """
        :type filepath: FilePath
        :type mask: int
        """
        if mask & (IN_DELETE_SELF | IN_MOVE_SELF):
            self._reactor.callLater(0, self._rewatch, filepath.path)

        else:
            # noinspection PyBroadException
            try:
                react_jnode = fs.load_json3(filepath.path)
                self._handler(filepath.basename(), react_jnode)

            except FsError:
                pass

            except Exception:
                eventlog.log_x()

            finally:
                fs.silent_rm_r(filepath.path)
