import os
import shutil
import time

from edge import eventlog
from edge.com import fs
from edge.webserver.cgi import CgiError500
from edge.webserver.sysconf import ph


class UimodBackup:
    def __init__(self, source_filepath, backup_filepath):
        """
        :type backup_filepath: str
        :type source_filepath: str
        """
        self._source_filepath = source_filepath
        self._backup_filepath = backup_filepath

    @staticmethod
    def create(backup_dir, uimod):
        """
        :type backup_dir: str
        :type uimod: ph.UiMod | ph.UiMod2
        :rtype: UimodBackup | None
        """
        backup_uiconf_name = 'ui-{}-old.json'.format(str(uimod))
        backup_uiconf_path = fs.join_path(backup_dir, backup_uiconf_name)
        try:
            eventlog.log_d('exec: backup {} to {}'.format(uimod.conf_path, backup_uiconf_path))
            shutil.copy2(uimod.conf_path, backup_uiconf_path)
            eventlog.log_d('done')

            return UimodBackup(uimod.conf_path, backup_uiconf_path)

        except (IOError, os.error, shutil.Error) as x:
            eventlog.log_e('fail: cp {} to {} ({})'.format(uimod.conf_path, backup_uiconf_path, x.message))

    def restore(self):
        """
         :rtype: bool
         """
        try:
            eventlog.log_d('exec: restore {} to {}'.format(self._backup_filepath, self._source_filepath))
            shutil.copy2(self._backup_filepath, self._source_filepath)
            eventlog.log_d('done')
            return True

        except (os.error, shutil.Error) as x:
            eventlog.log_e('fail: cp {} to {} ({})'.format(self._backup_filepath, self._source_filepath, x.message))

        return False

    def remove(self):
        eventlog.log_d('exec: delete {}'.format(self._backup_filepath))
        os.remove(self._backup_filepath)


class DgmodBackup:
    def __init__(self, source_dirpath, backup_dirpath):
        """
        :type source_dirpath: str
        :type backup_dirpath: str
        """
        self._source_dirpath = source_dirpath
        self._backup_dirpath = backup_dirpath

    @staticmethod
    def create(dir_path, dgmod):
        """
        :type dir_path: str
        :type dgmod: dg.DgMod
        :rtype: DgmodBackup | None
        """
        source_dirpath = dgmod.conf_dir
        backup_dirpath = fs.join_path(dir_path, 'edge-{}.{}'.format(dgmod.type, dgmod.id))

        try:
            eventlog.log_d('exec: backup {} to {}'.format(source_dirpath, backup_dirpath))
            shutil.copytree(source_dirpath, backup_dirpath)
            eventlog.log_d('done')

            return DgmodBackup(source_dirpath, backup_dirpath)

        except (IOError, os.error, shutil.Error) as x:
            eventlog.log_e('fail: backup {} to {} ({})'.format(source_dirpath, backup_dirpath, x.message))

    def restore(self):
        """
         :rtype: bool
         """
        def onerror_fn(error_func, error_path, exc_info):
            eventlog.log_w('fail: rm {} ({})'.format(error_path, exc_info[1].message))

        try:
            eventlog.log_d('exec: restore {} to {}'.format(self._backup_dirpath, self._source_dirpath))
            shutil.rmtree(self._source_dirpath, ignore_errors=False, onerror=onerror_fn)
            shutil.copytree(self._backup_dirpath, self._source_dirpath)
            eventlog.log_d('done')
            return True

        except shutil.Error as x:
            for srcname, dstname, why in x.args[0]:
                eventlog.log_e('fail: cp {} to {} ({})'.format(srcname, dstname, why))

        return False

    def remove(self):
        def onerror_fn(error_func, error_path, exc_info):
            eventlog.log_w('fail: rm {} ({})'.format(error_path, exc_info[1].message))

        eventlog.log_d('exec: delete {}'.format(self._backup_dirpath))
        shutil.rmtree(self._backup_dirpath, ignore_errors=False, onerror=onerror_fn)


class NewFileBackup:
    def __init__(self, backup_filepath):
        """
        :type backup_filepath: str
        """
        self._backup_filepath = backup_filepath

    @staticmethod
    def create(backup_dir, uimod, new_uiconf):
        """
        :type backup_dir: str
        :type uimod: ph.UiMod | ph.UiMod2
        :type new_uiconf: dict
        :rtype: NewFileBackup | None
        """
        # backup configuration files
        new_name = 'ui-{}-new.json'.format(str(uimod))
        new_filepath = fs.join_path(backup_dir, new_name)
        # noinspection PyBroadException
        try:
            eventlog.log_d('exec: create {}'.format(new_filepath))
            fs.write_json_file(new_uiconf, new_filepath, pretty=True)

        except Exception:
            eventlog.log_x()

        # handle backup succcess
        jnode_backup = NewFileBackup(new_filepath)
        eventlog.log_d('done')
        return jnode_backup

    @staticmethod
    def restore():
        """
         :rtype: bool
         """
        return True

    def remove(self):
        os.remove(self._backup_filepath)


class CgiBackup(object):
    def __init__(self, bak_dir, bak_files):
        """
        :type bak_dir: str
        :type bak_files: dict

        """
        self._bak_dir = bak_dir
        self._bak_files = bak_files

    def __call__(self, fn, *extra_args, **extra_kwargs):
        return lambda *fn_args, **fn_kwargs: self._exec(fn, *fn_args, **fn_kwargs)

    # noinspection PyBroadException
    def _exec(self, fn, *fn_args, **fn_kwargs):
        """
        :raise CgiError:
        """
        bak_dir = '%s.%s' % (self._bak_dir, int(time.time() * 1000000))

        self._create_backup(bak_dir)
        try:
            answer = fn(*fn_args, **fn_kwargs)
            self._remove_backup(bak_dir)
            return answer

        except Exception:
            self._restore_backup(bak_dir)
            raise

    def _create_backup(self, bak_dir):
        """
        :type bak_dir: str
        :raise CgiError:
        """
        try:
            os.makedirs(bak_dir)
            for (name, file_path) in self._bak_files.iteritems():
                shutil.copy(file_path, os.path.join(bak_dir, name))

        except Exception:
            eventlog.log_e('fail: http500')
            eventlog.log_x()
            raise CgiError500

    def _restore_backup(self, bak_dir):
        """
        :type bak_dir: str
        :raise CgiError:
        """
        # noinspection PyBroadException
        try:
            for (name, file_path) in self._bak_files.iteritems():
                shutil.copy(os.path.join(bak_dir, name), file_path)

        except Exception:
            eventlog.log_x()

        self._remove_backup(bak_dir)

    @staticmethod
    def _remove_backup(bak_dir):
        """
        :type bak_dir: str
        """
        # noinspection PyBroadException
        try:
            shutil.rmtree(bak_dir)
        except Exception:
            eventlog.log_x()
