import sys
import os
import ujson

import fcntl

from edge import eventlog
from edge.exceptions import *


R200_SUCCESS = 200
R400_BAD_REQUEST = 400
R404_NOT_FOUND = 404
R405_METHOD_NOT_ALLOWED = 405
R409_CONFLICT = 409
R500_INTERNAL_SERVER_ERROR = 500
R503_SERVICE_UNAVAILABLE = 503


class CgiError(EdgeError):
    def __init__(self, reply):
        """
        :type reply: CgiReply
        """
        self.reply = reply


class CgiError400(CgiError):
    def __init__(self, body_jnode=None):
        CgiError.__init__(self, CgiReply(R400_BAD_REQUEST, body_jnode))


class CgiError404(CgiError):
    def __init__(self, body_jnode=None):
        CgiError.__init__(self, CgiReply(R404_NOT_FOUND, body_jnode))


class CgiError405(CgiError):
    def __init__(self, body_jnode=None):
        CgiError.__init__(self, CgiReply(R405_METHOD_NOT_ALLOWED, body_jnode))


class CgiError409(CgiError):
    def __init__(self, body_jnode=None):
        CgiError.__init__(self, CgiReply(R409_CONFLICT, body_jnode))


class CgiError500(CgiError):
    def __init__(self, body_jnode=None):
        CgiError.__init__(self, CgiReply(R500_INTERNAL_SERVER_ERROR, body_jnode))


class CgiError503(CgiError):
    def __init__(self, body_jnode=None):
        CgiError.__init__(self, CgiReply(R503_SERVICE_UNAVAILABLE, body_jnode))


class CgiLicError(CgiError503):
    def __init__(self):
        CgiError503.__init__(self, {
            'status': R503_SERVICE_UNAVAILABLE,
            'errors': [
                {
                    'code': 'lpr-licence-absent',
                    'message': 'free lpr licence is not available'
                }
            ]
        })


class CgiCamError(CgiError503):
    def __init__(self):
        CgiError503.__init__(self, {
            'status': R503_SERVICE_UNAVAILABLE,
            'errors': [
                {
                    'code': 'videosource-error',
                    'message': 'something wrong with videosource'
                }
            ]
        })


class CgiPathInfo(list):
    def __init__(self, raw_path_info):
        """
        :type raw_path_info: str
        """
        list.__init__(self, raw_path_info.split('/')[1:])


class CgiRequest:
    """
    :type _body_jnode: dict | None
    """
    def __init__(self, method, path_info, query_string, body):
        """
        :type method: str
        :type path_info: CgiPathInfo
        :type query_string: str | None
        :type body: str | None
        """
        self.method = method
        self.path_info = path_info
        self.query_string = query_string
        self.body = body

        self._body_jnode = None

    @property
    def query_params(self):
        """
        :rtype: list[str]
        :raise CgiError:
        """
        if self.query_string is None:
            return []

        self.assert_query_present()
        return self.query_string[1:].split('&')

    @property
    def body_jnode(self):
        """
        :rtype: dict
        :raise CgiError:
        """
        if self.body is None:
            eventlog.log_e('fail: http400 (invalid body)')
            raise CgiError400

        elif self._body_jnode is None:
            try:
                self._body_jnode = ujson.loads(self.body)
                if self.body_jnode is None:
                    raise ValueError

            except ValueError:
                eventlog.log_e('fail: http400 (loads)')
                raise CgiError400

        return self._body_jnode

    @staticmethod
    def _read_stdin():
        """
        :rtype: str | None
        :raises CgiError:
        """
        # make stdin a non-blocking file
        stdin_fd = sys.stdin.fileno()
        old_flags = fcntl.fcntl(stdin_fd, fcntl.F_GETFL)
        fcntl.fcntl(stdin_fd, fcntl.F_SETFL, old_flags | os.O_NONBLOCK)
        try:
            return sys.stdin.read() or None

        except Exception as x:
            eventlog.log_e('fail: http500 ({})'.format(x))
            raise CgiError500

        finally:
            fcntl.fcntl(stdin_fd, fcntl.F_SETFL, old_flags)

    @classmethod
    def unpack(cls, path_info_factory=CgiPathInfo):
        """
        :type path_info_factory: (str) -> CgiPathInfo
        :rtype: CgiRequest
        :raises CgiError:
        """
        try:
            # read request
            env_method = os.environ['REQUEST_METHOD']
            env_path_info = os.getenv('PATH_INFO') or None
            env_query_string = os.getenv('QUERY_STRING') or None
            body = cls._read_stdin() or None

            eventlog.log_i('%s|%s|%s|%s' % (env_method, env_path_info, env_query_string, body))

            # unpack request
            path_info = path_info_factory(env_path_info or '')
            return cls(env_method, path_info, env_query_string, body)

        except CgiError:
            eventlog.log_e('fail: unpack query-args')
            raise

        except Exception as x:
            eventlog.log_e('fail: {}'.format(x))
            raise CgiError500

    def assert_query_present(self):
        if self.query_string is None:
            eventlog.log_d('fail: http400')
            raise CgiError400

        elif not self.query_string.startswith('?'):
            eventlog.log_d('fail: http400')
            raise CgiError400

        elif len(self.query_string) <= 1:
            eventlog.log_d('fail: http400')
            raise CgiError400

    def assert_query_absent(self):
        if self.query_string:
            eventlog.log_d('fail: http400')
            raise CgiError400

    def assert_body_present(self):
        if not self.body:
            eventlog.log_d('fail: http400')
            raise CgiError400

    def assert_body_absent(self):
        if self.body:
            eventlog.log_d('fail: http400')
            raise CgiError400


class CgiReply:
    def __init__(self, status=R200_SUCCESS, body=None):
        """
        :type status: int
        :type body: dict | list | None
        """
        self.status = str(status)
        self.body = body

    @classmethod
    def create(cls, reply_jnode):
        """
        :rtype: CgiReply
        """
        reply_status = reply_jnode['status']
        if reply_jnode.get('envelope', True):
            if 'envelope' in reply_jnode:
                del reply_jnode['envelope']
            reply_body = reply_jnode

        elif reply_status == R200_SUCCESS:
            reply_body = reply_jnode['result']

        else:
            reply_body = None

        return cls(reply_status, reply_body)

    @classmethod
    def create200(cls, result_jnode=None, envelope=True):
        """
        :rtype: CgiReply
        """
        reply_jnode = r200(result_jnode, envelope)
        return cls.create(reply_jnode)

    def send(self):
        sys.stderr.write(self.status)
        if self.body:
            body_utf8 = ujson.dumps(self.body, ensure_ascii=False)
            sys.stdout.write(body_utf8)


def r200(result_jnode=None, envelope=True):
    """
    :rtype: dict
    """
    if result_jnode is None:
        result_jnode = []

    jnode = {
        'envelope': envelope,
        'status': R200_SUCCESS,
        'result': result_jnode
    }

    return jnode


def _fail(http_code):
    """
    :type http_code: int
    :rtype: dict
    """
    return {
        'envelope': True,
        'status': http_code,
        'errors': [],
    }


class CgiModule:
    def __init__(self, handler):
        """
        :type handler: () -> CgiReply
        """
        self._handler = handler

    def run(self):
        """
        :rtype: CgiReply
        """
        return self._handler()


# noinspection PyBroadException
def run_cgi(module_factory, path_info_factory=CgiPathInfo):
    """
    :type module_factory: (CgiRequest) -> CgiModule
    :type path_info_factory: (str) -> CgiPathInfo
    """
    reply = CgiError500().reply
    try:
        request = CgiRequest.unpack(path_info_factory)
        module = module_factory(request)
        reply = module.run()

    except ParamsError:
        reply = CgiError400().reply

    except NotFoundError:
        reply = CgiError404().reply

    except BusyError:
        reply = CgiError409().reply

    except CamServiceError:
        reply = CgiCamError().reply

    except LicServiceError:
        reply = CgiLicError().reply

    except ServiceError:
        reply = CgiError503().reply

    except CgiError as x:
        reply = x.reply

    except EdgeError:
        pass

    except Exception:
        eventlog.log_x()

    finally:
        reply.send()

    sys.exit(0)
