import fcntl
import os

from edge import eventlog
from edge.com import fs
from edge.exceptions import *
from edge.webserver.sysconf import ph


def try_lock(lock_dir, lock_name):
    """
    :type lock_dir: str
    :type lock_name: str
    :rtype: int
    """
    import warnings
    warnings.warn('message', PendingDeprecationWarning, stacklevel=2)

    if not os.path.exists(lock_dir):
        os.makedirs(lock_dir)

    lock_path = fs.join_path(lock_dir, lock_name + '.lock')
    lock_fd = os.open(lock_path, os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
    try:
        # Request exclusive (EX) non-blocking (NB) advisory lock.
        fcntl.lockf(lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        return lock_fd

    except IOError as x:
        eventlog.log_e("fail: fcntl.lockf (%s)" % x.message)
        raise BusyError


def unlock(lock_fd):
    import warnings
    warnings.warn('message', PendingDeprecationWarning, stacklevel=2)
    try:
        fcntl.lockf(lock_fd, fcntl.LOCK_UN)
        os.close(lock_fd)
    except IOError as x:
        eventlog.log_d('fail: {}'.format(x.message))


class UimodsLocker:
    def __init__(self, uimods):
        """
        :type uimods: set[ph.UiMod | ph.UiMod2]
        """
        self._fd_list = []
        try:
            self._fd_list = map(lambda uimod: self._lock(uimod), uimods)
        except Exception:
            self.release()
            raise

    @staticmethod
    def _lock(uimod):
        """
        :type uimod: ph.UiMod | ph.UiMod2
        :rtype: int
        """
        dir_path = os.path.dirname(uimod.lock_path)
        if not os.path.dirname(dir_path):
            os.makedirs(dir_path)

        lock_fd = os.open(uimod.lock_path, os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
        try:
            # Request exclusive (EX) non-blocking (NB) advisory lock.
            fcntl.lockf(lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            return lock_fd

        except IOError as x:
            os.close(lock_fd)
            eventlog.log_d('fail: {}'.format(x.message))
            raise BusyError

    @staticmethod
    def _unlock(lock_fd):
        """
        :rtype: None
        """
        # Request exclusive (EX) non-blocking (NB) advisory lock.
        fcntl.lockf(lock_fd, fcntl.LOCK_UN)
        os.close(lock_fd)

    def release(self):
        """
        :rtype: None
        """
        # noinspection PyBroadException
        try:
            map(lambda lock_fd: self._unlock(lock_fd), self._fd_list)
        except Exception:
            eventlog.log_x()
            os.abort()


class CgiLock(object):
    def __init__(self, lock_file):
        """
        :type lock_file: str
        """
        self._lock_file = lock_file

    def __call__(self, fn, *extra_args, **extra_kwargs):
        return lambda *fn_args, **fn_kwargs: self._exec(fn, *fn_args, **fn_kwargs)

    # noinspection PyBroadException
    def _exec(self, fn, *fn_args, **fn_kwargs):
        if fs.try_lock_pid_file(self._lock_file):
            answer = fn(*fn_args, **fn_kwargs)
            if not fs.try_rm_pid_file(self._lock_file):
                eventlog.log_e('fail: rm pidfile <{}>'.format(self._lock_file))
            return answer

        else:
            eventlog.log_e('fail: BusyError (lock:<{}>)'.format(self._lock_file))
            raise BusyError
