import os

from edge import eventlog
from edge.com import fs
from edge.exceptions import *
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import blobstore as bs
from edge.webserver.sysconf.dgstats.blobstore import BsDgstats
from edge.webserver.sysconf.uiconf.time import TimeUiconf
from edge.webserver.sysconf.uimod.eventstore import EsUimod


def _get_disk_size():
    result = os.statvfs(ph.CTRL_ARCH_DIR)
    block_size = result.f_frsize
    total_blocks = result.f_blocks

    mega = 1024 * 1024
    total_size = total_blocks * block_size / mega

    return total_size


# noinspection PyClassHasNoInit
class EventstoreApi:
    @classmethod
    def _get_limits(cls):
        """
        :rtype: dict
        """
        total_size = _get_disk_size()

        limits = {
            'cleanup': {
                'min-df': {
                    'min': 0,
                    'max': total_size
                },
                'max-du': {
                    'min': 0,
                    'max': total_size
                },
                'max-abs-time': {
                    'min': 0,
                    'max': 4294967295
                },
                'max-rel-time': {
                    'min': 0,
                    'max': 4294967295
                }
            }
        }
        return limits

    @classmethod
    def load_unit(cls, uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = EsUimod.create(uimod_ref)
        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        stats_tz = TimeUiconf.get_tz()
        stats = BsDgstats.get_chan_stats_64k(uimod_ref.split('.')[-1], stats_tz)

        ext_uiconf_pub = uiconf['rw']  # type: dict
        reply = {
            'type': ph.UI_TYPE_EVENTSTORE,
            'conf': ext_uiconf_pub['args'],
            'limits': cls._get_limits()
        }
        if stats:
            reply['stats'] = stats

        return reply

    @classmethod
    def save_unit(cls, uimod_ref, new_uiconf):
        """
        :type uimod_ref: str
        :type new_uiconf: dict
        :rtype: dict
        """
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = EsUimod.create(uimod_ref)
        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        fs.validate_json({'args': new_uiconf_conf}, uimod.schema_path)

        uiconf_pub = uiconf['rw']  # type: dict
        if new_uiconf_conf['name'] != uiconf_pub['args']['name']:
            ph.RootInfoUiconf.set_unit_name(uimod_ref, new_uiconf_conf['name'])

        uiconf_pub['args'] = new_uiconf_conf
        uimod.save_conf(uiconf)

        bs.bs65535_reappend_chan(uiconf)

        stats_tz = TimeUiconf.get_tz()
        stats = BsDgstats.get_chan_stats_64k(uimod_ref.split('.')[-1], stats_tz)

        ext_uiconf_pub = uiconf_pub
        reply = {
            'type': ph.UI_TYPE_EVENTSTORE,
            'conf': ext_uiconf_pub['args'],
            'limits': cls._get_limits()
        }
        if stats:
            reply['stats'] = stats

        return reply

    @classmethod
    def make_unit(cls, factory_conf):
        """
        :type factory_conf: dict
        :rtype: dict
        """
        root_info, root_links = ph.RootInfoUiconf.get_info_links()
        ctrl_info = ph.InfoUiconf.find_unit_r(root_info, unit_type=ph.UI_TYPE_CONTROLLER)
        _, _, es_uimod_id = ph.RootInfoUiconf.new_uimod_pid_type_id(ctrl_info, ph.UI_TYPE_EVENTSTORE)  # TODO es_uimod_id

        # add store_jnode to /etc/edge/setupstore/admin/action/eventstore/<chan-ref>/0.json
        es_uimod_ref = '{}.{}'.format(ph.UI_TYPE_EVENTSTORE, es_uimod_id)
        uimod = EsUimod.create(es_uimod_ref)
        uiconf = uimod.install(factory_conf['name'], es_uimod_id)

        bs.bs65535_append_chan(uiconf)

        ctrl_units = ctrl_info['units']  # type: list
        ctrl_units.append({
            'internal': False,
            'role': ph.UI_ROLE_ARCHIVE,
            'type': ph.UI_TYPE_EVENTSTORE,
            'name': factory_conf['name'],
            'ref': es_uimod_ref
        })

        ph.RootInfoUiconf.set_info_links(root_info, root_links)

        root_info = ph.RootInfoUiconf.get_ext_root_info(root_info)
        return {'info': [root_info], 'created': [es_uimod_ref]}

    @classmethod
    def delete_unit(cls, uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        EsUimod.create(uimod_ref).uninstall()
        bs.bs65535_remove_chan(uimod_ref)

        root_info = ph.RootInfoUiconf.del_unit_r(uimod_ref)
        reply = {
            'info': [root_info]
        }
        return reply
