import gps
import time
import os

from edge import eventlog
from edge.com import fs
from edge.exceptions import *
from edge.webserver.sysconf import ph

import edge_creator as creator


def _geodata_rt_conf(new_jnode):
    """
    :type new_jnode: dict
    :rtype: dict
    """
    az = new_jnode.get('azimuth', None)
    provider_jnode = new_jnode['providers'][0]

    if provider_jnode['provider'] == 'manual':
        jnode = {
            'azimuth': az,
            'source': ['static', provider_jnode['latitude'], provider_jnode['longitude']]
        }
    elif provider_jnode['provider'] == 'controller':
        jnode = {
            'azimuth': az,
            'source': ['gpsd', 'localhost', 2947]
        }
    else:
        jnode = dict()

    return jnode


def _test_uiconf_conf_changed(uiconf_conf, new_uiconf_conf):
    """
    :type uiconf_conf: dict
    :type new_uiconf_conf: dict
    :rtype:
    """
    if uiconf_conf.get('azimuth') != new_uiconf_conf.get('azimuth'):
        return True

    provider_info = uiconf_conf['providers'][0]
    new_provider_info = new_uiconf_conf['providers'][0]

    if provider_info['provider'] != new_provider_info['provider']:
        return True

    # old_mode == new_mode
    if provider_info['provider'] == 'manual':
        if provider_info['latitude'] != new_provider_info['latitude']:
            return True

        if provider_info['longitude'] != new_provider_info['longitude']:
            return True

    return False


def _test_gps_device_configured():
    """
    :rtype: bool
    """
    _GPSD_CONF_PATH = '/etc/default/gpsd'

    # gps_present = net.query_gps_count() > 0
    try:
        if os.path.isfile(_GPSD_CONF_PATH):
            with open(_GPSD_CONF_PATH, 'r') as gpsd_conf:
                gpsd_lines = gpsd_conf.readlines()

            for line in gpsd_lines:
                if line.startswith('DEVICES="/'):
                    return True

    except Exception as x:
        eventlog.log_e('fail: {}'.format(x.message))

    return False


def _get_gps_device_lat_long(localhost='localhost', port='2947'):
    """
    :type localhost: str
    :type port: str
    :rtype: (float, float) | (None, None)
    """
    session = gps.gps(localhost, port)
    session.stream(gps.WATCH_ENABLE | gps.WATCH_NEWSTYLE)

    start_time = int(time.time())
    latitude = longitude = None
    for _ in range(5):
        try:
            if not session.waiting(1):  # 1 second
                continue
            report = session.next()
            if report['class'] == 'DEVICES' and len(report['devices']) == 0:
                return latitude, longitude

            if report['class'] == 'TPV':
                latitude = report['lat']
                longitude = report['lon']
                return latitude, longitude

            end_time = int(time.time())
            diff_time = end_time - start_time
            if diff_time > 2:  # 2 seconds
                return latitude, longitude

        except KeyError:
            pass
        except StopIteration:
            break

    return latitude, longitude


def _add_limits_stats(uiconf_conf):
    """
    :type uiconf_conf: dict
    :rtype: dict
    """
    if _test_gps_device_configured():
        unit_ext_jnode = {
            'type': ph.UI_TYPE_GNSS,
            'conf': uiconf_conf,
            'limits': {
                'providers': ['manual', 'controller']
            }
        }

    else:
        unit_ext_jnode = {
            'type': ph.UI_TYPE_GNSS,
            'conf': uiconf_conf,
            'limits': {
                'providers': ['manual']
            }
        }

    uiconf_providers = uiconf_conf['providers']
    if len(uiconf_providers) != 1:
        raise NotImplementedError

    for provider in uiconf_providers:
        if provider['provider'] == 'manual':
            unit_ext_jnode['stats'] = {
                'provider': 'manual',
                'latitude': uiconf_providers[0]['latitude'],
                'longitude': uiconf_providers[0]['longitude']
            }

        elif provider['provider'] == 'controller':
            lat, lon = _get_gps_device_lat_long()
            if None not in [lat, lon]:
                unit_ext_jnode['stats'] = {
                    'provider': 'controller',
                    'latitude': lat,
                    'longitude': lon
                }

    return unit_ext_jnode


# noinspection PyClassHasNoInit
class ApiGnss:
    @staticmethod
    def load_unit(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_GNSS)
        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        unit_ext_jnode = _add_limits_stats(uiconf['rw']['conf'])
        return unit_ext_jnode

    @staticmethod
    def save_unit(uimod_ref, unit_jnode):
        """
        :type uimod_ref: str
        :type unit_jnode: dict
        :rtype: dict
        """
        new_uiconf_conf = unit_jnode.get('conf')

        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_GNSS)
        fs.validate_json(new_uiconf_conf, uimod.schema_path)

        uiconf = uimod.load_conf()
        uiconf_pub = uiconf['rw']
        if _test_uiconf_conf_changed(uiconf_pub['conf'], new_uiconf_conf):
            uiconf_pub['conf'] = new_uiconf_conf
            uiconf['rt'] = _geodata_rt_conf(new_uiconf_conf)
            uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
            creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})

        unit_ext_jnode = _add_limits_stats(new_uiconf_conf)
        return unit_ext_jnode
