# coding=utf-8

import os
import datetime
import ujson

from edge import eventlog
from edge.com import fs, sh

from edge.exceptions import *

from edge.webserver.sysconf import ph, edge_creator as creator
from edge.webserver.sysconf.uimod.sign import SignUimod
from edge.webserver.sysconf.uiconf.iit_ca import IitCaUiconf, IitCaEndUserUiconf


def _iit_cli(*args):
    """
    :rtype: str
    """
    return sh.exec_cli('/opt/edge/bin/iit-cli', list(args), tout_sec=10, encoding='utf-8')


def _list_media(cxt=None):
    """
    :rtype: list[dict]
    """
    import EUSignCP

    iit_api = cxt
    if iit_api is None:
        EUSignCP.EULoad()
        iit_api = EUSignCP.EUGetInterface()  # type:EUSignCP.EU_INTERFACE
        # ctx.SetUIMode(1)
        iit_api.Initialize()

    media_list = []

    type_index = 1
    type_desc = list()
    while iit_api.EnumKeyMediaTypes(type_index, type_desc):
        if type_desc[0] != u'е.ключ ІІТ Алмаз-1К':
            type_index += 1
            continue

        dev_index = 0
        dev_desc = list()
        while iit_api.EnumKeyMediaDevices(type_index, dev_index, dev_desc):
            media_list.append({
                'model': type_desc[0],
                'serial': dev_desc[0]
            })

            dev_index += 1

        type_index += 1

    return media_list


def _edit_iit_conf(new_uiconf_conf):
    """
    :type new_uiconf_conf: dict
    :rtype: list[dict]
    """
    import EUSignCP

    EUSignCP.EULoad()
    ctx = EUSignCP.EUGetInterface()  # type:EUSignCP.EU_INTERFACE
    # ctx.SetUIMode(1)
    ctx.Initialize()

    mode_config = {
        'bOfflineMode': False
    }
    ctx.SetModeSettings(mode_config)

    jnode_fs = new_uiconf_conf['crl']

    file_store_settings = dict()
    try:
        ctx.GetFileStoreSettings(file_store_settings)
    except RuntimeError:
        eventlog.log_d('fail: GetFileStoreSettings(). Creating settings')

    file_store_settings['szPath'] = '/data/certificates'
    file_store_settings['bCheckCRLs'] = jnode_fs['enable']
    file_store_settings['bAutoRefresh'] = False
    file_store_settings['bOwnCRLsOnly'] = jnode_fs['only-own-crls']
    file_store_settings['bFullAndDeltaCRLs'] = jnode_fs['full-and-delta-crls']
    file_store_settings['bAutoDownloadCRLs'] = True
    file_store_settings['bSaveLoadedCerts'] = True
    file_store_settings['dwExpireTime'] = 3600

    ctx.SetFileStoreSettings(file_store_settings)

    # print 'file_store_settings', file_store_settings

    jnode_proxy = new_uiconf_conf['proxy']
    jnode_proxy_auth_type = jnode_proxy['auth-type']
    proxy_setting = dict()
    try:
        ctx.GetProxySettings(proxy_setting)
    except RuntimeError:
        eventlog.log_i('fail: GetProxySettings(). Creating settings')

    proxy_setting['bUseProxy'] = jnode_proxy.get('enable', False)
    proxy_setting['szAddress'] = u'{}'.format(jnode_proxy['host'])
    proxy_setting['szPort'] = u'{}'.format(jnode_proxy['port'])
    if jnode_proxy_auth_type == 'anonymous':
        proxy_setting['bAnonymous'] = True
        proxy_setting['szUser'] = u''
        proxy_setting['szPassword'] = u''
        proxy_setting['bSavePassword'] = False

    elif jnode_proxy_auth_type == 'credentials':
        proxy_setting['bAnonymous'] = False
        proxy_setting['szUser'] = u'{}'.format(jnode_proxy['auth-user'])
        proxy_setting['szPassword'] = u'{}'.format(jnode_proxy['auth-pass'])
        proxy_setting['bSavePassword'] = True

    ctx.SetProxySettings(proxy_setting)

    # print 'proxy_setting', proxy_setting

    jnode_cmp = new_uiconf_conf['cmp']
    cmp_settings = dict()
    try:
        ctx.GetCMPSettings(cmp_settings)
    except RuntimeError:
        eventlog.log_i('fail: GetCMPSettings(). Creating settings')

    cmp_settings['szPort'] = u'{}'.format(jnode_cmp['port'])
    cmp_settings['szAddress'] = u'{}'.format(jnode_cmp['host'])
    # cmp_settings['szCommonName'] = u'name'
    cmp_settings['bUseCMP'] = True
    ctx.SetCMPSettings(cmp_settings)

    # print 'cmp_settings', cmp_settings

    jnode_oscp = new_uiconf_conf['ocsp']
    oscp_settings = dict()
    try:
        ctx.GetOCSPSettings(oscp_settings)
    except RuntimeError:
        eventlog.log_i('fail: GetOCSPSettings(). Creating settings')
    oscp_settings['bUseOCSP'] = jnode_oscp.get('enable', False)
    oscp_settings['szAddress'] = u'{}'.format(jnode_oscp['host'])
    oscp_settings['szPort'] = u'{}'.format(jnode_oscp['port'])
    oscp_settings['bBeforeStore'] = False
    ctx.SetOCSPSettings(oscp_settings)
    # print 'oscp_settig', oscp_settig

    jnode_tsp = new_uiconf_conf['tsp']
    tsp_settings = dict()
    try:
        ctx.GetTSPSettings(tsp_settings)
    except RuntimeError:
        eventlog.log_i('fail: GetTSPSettings(). Creating settings')
    tsp_settings['szAddress'] = u'{}'.format(jnode_tsp['host'])
    tsp_settings['szPort'] = u'{}'.format(jnode_tsp['port'])
    tsp_settings['bGetStamps'] = True
    ctx.SetTSPSettings(tsp_settings)
    # print 'tsp_settings', tsp_settings

    # IIT library doesn't work if LDAP settings are not set
    ldap_settings = dict()
    try:
        ctx.GetLDAPSettings(ldap_settings)
    except RuntimeError:
        eventlog.log_i('fail: GetLDAPSettings(). Creating stub settings')
    ldap_settings['bUseLDAP'] = False
    ldap_settings['szAddress'] = u''
    ldap_settings['szPort'] = u''
    ldap_settings['bAnonymous'] = True
    ldap_settings['szUser'] = u''
    ldap_settings['szPassword'] = u''
    ctx.SetLDAPSettings(ldap_settings)

    keys = _list_media(ctx)
    return keys
    # ctx.GetPrivateKeyMedia()

    # cmp_settings2 = dict()
    # ctx.GetCMPSettings(cmp_settings2)
    # print cmp_settings2


class IitCaApi:
    def __init__(self):
        pass

    @staticmethod
    def create_service_from_sign(ui_sign_jnode):
        """
        :type ui_sign_jnode: dict
        :rtype: dict
        """
        root_info, root_links = ph.RootInfoUiconf.get_info_links()
        controller_info = ph.RootInfoUiconf.find_unit_r(root_info, unit_type=ph.UI_TYPE_CONTROLLER)
        if controller_info is None:
            eventlog.log_d('fail: NotFoundError ({})'.format(ph.UI_TYPE_TERMINAL))
            raise NotFoundError

        tr_uirefs = ph.RootInfoUiconf.find_unit_refs_r(root_info, unit_types=[ph.UI_TYPE_TRAFFIC_RULES])

        iit_ca_unit_name = u'ІІТ ЦСК'
        # parse old conf
        ui_sign_conf = ui_sign_jnode['rw']['conf']
        ui_sign_conf_key = ui_sign_conf['key']
        media_type = ui_sign_conf_key['model']
        media_serial = ui_sign_conf_key['serial']
        media_password = ui_sign_conf_key['password']
        del ui_sign_conf['key']
        sign_active = ui_sign_conf['active']
        del ui_sign_conf['active']
        ui_sign_conf['name'] = iit_ca_unit_name

        _, _, iit_ca_id = ph.RootInfoUiconf.new_uimod_pid_type_id(controller_info, ph.UI_TYPE_IIT_CA)
        iit_ca_uiref = '{}.{}'.format(ph.UI_TYPE_IIT_CA, iit_ca_id)
        iit_ca_uimod = ph.UiMod2(iit_ca_uiref)
        iit_ca_uiconf = IitCaUiconf.create({'ref': iit_ca_uiref})
        iit_ca_uiconf.set_pub_conf(ui_sign_conf)

        iit_ca_uimod.install(iit_ca_uiconf, iit_ca_uiref)
        iit_ca_units_info = []

        # add enduser
        if 0 not in [len(media_type), len(media_serial), len(media_password)]:
            end_user_uimod = iit_ca_uimod.child(ph.UI_TYPE_IIT_CA_END_USER, child_chan=0)
            enduser_factory = {
                'conf': {
                    'name': u'ІІТ Користувач ЦСК',
                    'active': sign_active,
                    'media': {
                        'model': media_type,
                        'serial': media_serial,
                        'password': media_password
                    },
                    'parent-ref': iit_ca_uiref,
                    'zip-name': None
                }
            }
            end_user_uiconf = IitCaEndUserUiconf.create(enduser_factory)

            end_user_uimod.install(end_user_uiconf, iit_ca_uiref)

            # install uimod-info
            iit_ca_units_info.append({
                'ref': end_user_uimod.ref,
                'type': end_user_uimod.type,
                'name': end_user_uiconf.pub_name,
                'role': ph.UI_ROLE_SERVICE,
                'internal': True,
                'factory': [{
                    'type': ph.UI_TYPE_IIT_CA_END_USER,
                    'role': ph.UI_ROLE_UNIT
                }]
            })

            # add links
            for tr_uiref in tr_uirefs:
                root_links.append([tr_uiref, end_user_uimod.ref])

        if controller_info.get('units') is None:
            controller_info['units'] = []

        controller_info['units'].append({
            'ref': iit_ca_uimod.ref,
            'type': iit_ca_uimod.type,
            'name': iit_ca_unit_name,
            'role': ph.UI_ROLE_UNIT,
            'internal': True,
            'units': iit_ca_units_info
        })

        # add links
        for tr_uiref in tr_uirefs:
            root_links.append([tr_uiref, iit_ca_uimod.ref])

        ph.RootInfoUiconf.set_info_links(root_info, root_links)
        return iit_ca_id, root_info

    @staticmethod
    def make_service(ctrl_uiref):
        """
        :type ctrl_uiref: str
        :rtype: dict
        """
        root_info, root_links = ph.RootInfoUiconf.get_info_links()

        controller_info = ph.RootInfoUiconf.find_unit_r(root_info, unit_ref=ctrl_uiref)
        if controller_info is None:
            eventlog.log_d('fail: NotFoundError ({})'.format(ph.UI_TYPE_TERMINAL))
            raise NotFoundError

        _, _, iit_ca_id = ph.RootInfoUiconf.new_uimod_pid_type_id(controller_info, ph.UI_TYPE_IIT_CA)
        service_uiref = '{}.{}'.format(ph.UI_TYPE_IIT_CA, iit_ca_id)
        service_uimod = ph.UiMod2(service_uiref)

        service_uiconf = IitCaUiconf.create({'ref': service_uiref})
        service_uimod.install(service_uiconf, service_uiref)

        _edit_iit_conf(service_uiconf.pub_conf)

        if controller_info.get('units') is None:
            controller_info['units'] = []

        controller_info['units'].append({
            'ref': service_uimod.ref,
            'type': service_uimod.type,
            'name': service_uiconf.pub_name,
            'role': ph.UI_ROLE_SERVICE,
            'internal': True,
            'units': [],
            'factory': [{
                'type': ph.UI_TYPE_IIT_CA_END_USER,
                'role': ph.UI_ROLE_UNIT
            }]
        })

        # add links
        tr_uirefs = ph.RootInfoUiconf.find_unit_refs_r(root_info, unit_types=[ph.UI_TYPE_TRAFFIC_RULES])
        for tr_uiref in tr_uirefs:
            root_links.append([tr_uiref, service_uimod.ref])

        ph.RootInfoUiconf.set_info_links(root_info, root_links)

    @staticmethod
    def drop_service(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_IIT_CA)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_d('fail: NotFoundError')
            raise

        root_info, root_links = ph.RootInfoUiconf.get_info_links()
        unit_info, _ = ph.RootInfoUiconf.find_unit_parent_r(root_info, uimod_ref)

        end_user_refs = ph.RootInfoUiconf.find_unit_refs_r(
            unit_info,
            [ph.UI_TYPE_IIT_CA_END_USER])

        for end_user_ref in end_user_refs:
            IitCaApi.drop_end_user(end_user_ref)

        refs_to_drop = ph.RootInfoUiconf.find_unit_refs_r(unit_info, [ph.UI_TYPE_IIT_CA])
        for ref_to_delete in refs_to_drop:
            uimod = ph.UiMod2(ref_to_delete)
            root_info = ph.RootInfoUiconf.del_unit_r(uimod.ref)
            uimod.uninstall()

        ext_root_info = ph.RootInfoUiconf.get_ext_root_info(root_info)
        return {
            'info': [ext_root_info]
        }

    @staticmethod
    def load_service(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_IIT_CA)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_d('fail: NotFoundError')
            raise

        uiconf = IitCaUiconf(uimod.uiconf)
        stats_jnode = {
            'end-users': []
        }

        def add_user(user_info):
            """
            :type user_info: dict
            :rtype: None
            """
            stats_jnode['end-users'].append({'ref': user_info['ref']})

        root_info, root_links = ph.RootInfoUiconf.get_info_links()
        unit_info = ph.RootInfoUiconf.find_unit_r(root_info, unit_ref=uimod.ref)
        ph.RootInfoUiconf.map_units_r(unit_info, add_user, [ph.UI_TYPE_IIT_CA_END_USER])

        return {
            'type': ph.UI_TYPE_IIT_CA,
            'conf': uiconf.pub_conf,
            'stats': stats_jnode
        }

    @staticmethod
    def load_service_factory(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_IIT_CA)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_d('fail: NotFoundError')
            raise

        return {
            'limits': {
                'media': _list_media()
            }
        }

    @staticmethod
    def save_service(uimod_ref, new_uiconf):
        """
        :type uimod_ref: str
        :type new_uiconf: dict
        :rtype: dict
        """
        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_IIT_CA)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_d('fail: NotFoundError')
            raise

        uiconf = IitCaUiconf(uimod.uiconf)

        new_uiconf_conf = new_uiconf.get('conf')
        if new_uiconf_conf is None:
            eventlog.log_d('ParamsError')
            raise ParamsError

        new_proxy_pass = new_uiconf_conf['proxy'].get('auth-pass')

        if new_proxy_pass is None:
            new_uiconf_conf['proxy']['auth-pass'] = uiconf.proxy_pass

        name_changed = uiconf.set_pub_name(new_uiconf_conf['name'])
        conf_changed = uiconf.set_pub_conf(new_uiconf_conf) or name_changed
        if conf_changed:
            uiconf.validate()
            uimod.save_uiconf(uiconf)

        # edit iit config
        _edit_iit_conf(new_uiconf_conf)

        # generate
        uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
        creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf.raw})

        end_users = list()

        def end_user_handler(user_info):
            """
            :type user_info: dict
            :rtype: None
            """
            end_users.append({'ref': user_info['ref']})

        root_info, root_links = ph.RootInfoUiconf.get_info_links()
        unit_info = ph.RootInfoUiconf.find_unit_r(root_info, unit_ref=uimod.ref)
        ph.RootInfoUiconf.map_units_r(unit_info, end_user_handler, [ph.UI_TYPE_IIT_CA_END_USER])

        reply = {
            'type': ph.UI_TYPE_IIT_CA,
            'conf': new_uiconf_conf,
            'stats': {
                'end-users': end_users
            }
        }
        return reply

    @staticmethod
    def load_end_user(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_IIT_CA_END_USER)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_d('fail: NotFoundError')
            raise

        uiconf = IitCaEndUserUiconf(uimod.uiconf)

        crt_stats = uiconf.crt_stats
        if crt_stats is None:
            try:
                crt_stats = ujson.loads(_iit_cli('sign', 'crt-info', uiconf.serial, uiconf.password))
                uiconf.set_crt_stats(crt_stats)
                uimod.save_uiconf(uiconf)
            except InternalError:
                pass

        jnode = {
            'type': uimod.type,
            'conf': uiconf.pub_conf,
            'stats': {
                'media': {
                    'model': uiconf.model,
                    'serial': uiconf.serial
                }
            }
        }

        csr_name = uiconf.csr_name
        if csr_name:
            jnode['stats']['csr'] = csr_name

        crt_stats = uiconf.crt_stats
        if crt_stats:
            jnode['stats']['crt'] = crt_stats
        return jnode

    @staticmethod
    def save_end_user(uimod_ref, new_uiconf):
        """
        :type uimod_ref: str
        :type new_uiconf: dict
        :rtype: dict
        """
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_IIT_CA_END_USER)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        uiconf = IitCaEndUserUiconf(uimod.uiconf)

        name_changed = uiconf.set_pub_name(new_uiconf_conf['name'])
        conf_changed = uiconf.set_pub_conf(new_uiconf_conf) or name_changed
        if conf_changed:
            uiconf.validate()
            uimod.save_uiconf(uiconf)

        uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
        creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf.raw})

        jnode = {
            'type': uimod.type,
            'conf': uiconf.pub_conf,
            'stats': {
                'media': {
                    'model': uiconf.model,
                    'serial': uiconf.serial
                }
            }
        }

        csr_name = uiconf.csr_name
        if csr_name:
            jnode['stats']['csr'] = csr_name

        crt_stats = uiconf.crt_stats
        if crt_stats:
            jnode['stats']['crt'] = crt_stats

        return jnode

    @staticmethod
    def make_end_user(parent_uiref, factory_conf):
        """
        :type parent_uiref: str
        :type factory_conf: dict
        :rtype: dict
        """
        factory_conf_conf = factory_conf['conf']

        format_media = factory_conf_conf.get('format-media')
        if format_media is None or type(format_media) != bool:
            eventlog.log_d('fail: ParamsError ({}) <format-media> is not provided'.format(ph.UI_TYPE_IIT_CA_END_USER))
            raise ParamsError

        root_info, root_links = ph.RootInfoUiconf.get_info_links()
        iit_ca_uimod = ph.UiMod2(parent_uiref)

        iit_ca_info = ph.RootInfoUiconf.find_unit_r(root_info, unit_ref=parent_uiref)
        if iit_ca_info is None:
            eventlog.log_d('fail: NotFoundError ({})'.format(ph.UI_TYPE_IIT_CA))
            raise NotFoundError

        end_user_info = ph.RootInfoUiconf.find_unit_r(iit_ca_info, unit_type=ph.UI_TYPE_IIT_CA_END_USER)
        if end_user_info is not None:
            eventlog.log_d('fail: ParamsError ({}) already exist'.format(ph.UI_TYPE_IIT_CA_END_USER))
            raise ParamsError

        from edge.service import about
        hw_serial = about.get_hw_serial()
        if hw_serial is None:
            eventlog.log_e('fail: InternalError (about.json#/hw is not provided)')
            raise InternalError

        _, _, end_user_id = ph.RootInfoUiconf.new_uimod_pid_type_id(iit_ca_info, ph.UI_TYPE_IIT_CA_END_USER)
        end_user_uimod = iit_ca_uimod.child(ph.UI_TYPE_IIT_CA_END_USER, end_user_id)

        factory_conf_conf['parent-ref'] = iit_ca_uimod.ref

        if format_media:
            # u'EU-15-1219-ЕП-120520.p10'
            # u'EU-KEP-15-1219-ЕП-120520.p10'
            dt = datetime.datetime.now()

            csr_ua_ds_name = u'EU-{}-ЕП-{:02}{:02}{}.p10'.format(hw_serial, dt.day, dt.month, dt.year)
            csr_ua_kep_name = u'EU-KEP-{}-ЕП-{:02}{:02}{}.p10'.format(hw_serial, dt.day, dt.month, dt.year)

            zip_file_name = u'EU-{}-ЕП-{:02}{:02}{}.zip'.format(hw_serial, dt.day, dt.month, dt.year)
            csr_ua_zip_path = u'/etc/edge/setupstore/admin/actions/{}/{}'.format(end_user_uimod.ref, zip_file_name)

            factory_conf_conf['zip-name'] = zip_file_name
            end_user_uiconf = IitCaEndUserUiconf.create(factory_conf)

            fs.silent_mkdir_r(os.path.dirname(csr_ua_zip_path))
            _iit_cli('key', 'create', end_user_uiconf.serial, end_user_uiconf.password,
                     csr_ua_ds_name, csr_ua_kep_name, csr_ua_zip_path)
        else:
            # check pass
            end_user_uiconf = IitCaEndUserUiconf.create(factory_conf)
            reply = _iit_cli('key', 'test-access', end_user_uiconf.serial, end_user_uiconf.password)
            if 'PASSED' not in reply:
                eventlog.log_d('fail: ParamsError ({}) media test-access failed'.format(ph.UI_TYPE_IIT_CA_END_USER))
                raise ParamsError

        end_user_uimod.install(end_user_uiconf, iit_ca_uimod.ref)

        iit_ca_info['units'].append({
            'ref': end_user_uimod.ref,
            'type': end_user_uimod.type,
            'name': end_user_uiconf.pub_name,
            'role': ph.UI_ROLE_UNIT,
            'internal': True,
            'units': []
        })

        tr_uirefs = ph.RootInfoUiconf.find_unit_refs_r(root_info, unit_types=[ph.UI_TYPE_TRAFFIC_RULES])
        for tr_uiref in tr_uirefs:
            root_links.append([tr_uiref, end_user_uimod.ref])

        ph.RootInfoUiconf.set_info_links(root_info, root_links)

        ext_root_info = ph.RootInfoUiconf.get_ext_root_info(root_info)
        reply = {
            'info': [ext_root_info],
            'created': [end_user_uimod.ref]
        }
        return reply

    @staticmethod
    def drop_end_user(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_IIT_CA_END_USER)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        uiconf = IitCaEndUserUiconf(uimod.uiconf)
        csr_file_path = u'/etc/edge/setupstore/admin/actions/{}/{}'.format(uimod.ref, uiconf.csr_name)
        fs.silent_rm_r(csr_file_path)

        uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
        uirefs.discard(uimod.ref)

        root_info = ph.RootInfoUiconf.del_unit_r(uimod.ref)
        uimod.uninstall()

        creator.generate(uimod.create_backup_dirpath(), uirefs, forced=True)

        ext_root_info = ph.RootInfoUiconf.get_ext_root_info(root_info)
        reply = {
            'info': [ext_root_info]
        }
        return reply

    @staticmethod
    def testpack_end_user(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_IIT_CA_END_USER)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        uiconf = IitCaEndUserUiconf(uimod.uiconf)

        media_serial = uiconf.serial
        media_password = uiconf.password

        from edge.webserver import file_lock
        lock = file_lock.UimodsLocker({uimod})
        try:
            return ujson.loads(_iit_cli('key', 'test-all', media_serial, media_password))
        finally:
            lock.release()


if __name__ == '__main__':
    eventlog.open_log('api-iit-ca', 'debug')
    try:
        sign_uiconf = SignUimod.load_conf(default=None)
        IitCaApi.create_service_from_sign(sign_uiconf)
    finally:
        eventlog.close_log()
