from edge import eventlog
from edge.com import fs
from edge.exceptions import *
from edge.webserver.sysconf import ph

from psuctl import Endpoint, PsuCtlError


# noinspection PyClassHasNoInit
class LensApi:
    @staticmethod
    def _get_rotate_limits(endpoint):
        """
        :type endpoint: Endpoint
        :rtype: dict | None
        """
        script, action_conf = None, None
        try:
            script, action_conf = 'rotate-iris', {'action': 'get-limits'}
            min_iris, max_iris = endpoint.run_action(script, action_conf)
            if max_iris is None:
                raise PsuCtlError

            script, action_conf = 'rotate-focus', {'action': 'get-limits'}
            min_focus, max_focus = endpoint.run_action(script, action_conf)
            if max_focus is None:
                raise PsuCtlError

        except PsuCtlError:
            eventlog.log_e('fail: InternalError (script:{}, conf:{})'.format(script, action_conf))
            raise InternalError

        return min_iris, max_iris, min_focus, max_focus

    @classmethod
    def _rotate(cls, endpoint, old_iris, new_iris, old_focus, new_focus):
        """
        :type endpoint: Endpoint
        :type old_iris: int
        :type new_iris: int
        :type old_focus: int
        :type new_focus: int
        :rtype: dict
        """
        script, action_conf = None, None
        try:
            min_iris, max_iris, min_focus, max_focus = cls._get_rotate_limits(endpoint)

            if old_iris != new_iris:
                if new_iris == 0:
                    iris_steps = -(max_iris - min_iris)
                else:
                    iris_steps = int(new_iris - old_iris)

                script, action_conf = 'rotate-iris', {'steps': iris_steps}
                endpoint.run_action(script, action_conf)

            if old_focus != new_focus:
                if new_focus == 0:
                    focus_steps = -(max_focus - min_focus)
                else:
                    focus_steps = int(new_focus - old_focus)

                script, action_conf = 'rotate-focus', {'steps': focus_steps}
                endpoint.run_action(script, action_conf)

            return min_iris, max_iris, min_focus, max_focus

        except PsuCtlError:
            eventlog.log_e('fail: InternalError (script:{}, conf:{})'.format(script, action_conf))
            raise InternalError

    @classmethod
    def load_unit(cls, uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_LENS)
        cam_dgstats = ph.CamDgStats.create(uimod.pid)
        cam_dgstats.test_availability()

        with Endpoint() as endpoint:
            _, max_iris, _, max_focus = cls._get_rotate_limits(endpoint)

        uiconf = uimod.load_conf()
        ext_uiconf_pub = uiconf['rw']

        reply = {
            'type': ph.UI_TYPE_LENS,
            'conf': ext_uiconf_pub,
            'limits': {
                'iris': max_iris,
                'focus': max_focus
            },
            'stats': cam_dgstats.get_stats(uimod.pid)
        }
        return reply

    @classmethod
    def save_unit(cls, uimod_ref, new_uiconf):
        """
        :type uimod_ref: str
        :type new_uiconf: dict
        :rtype: dict
        """
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_LENS)
        cam_dgstats = ph.CamDgStats.create(uimod.pid)
        cam_dgstats.test_availability()

        fs.validate_json(new_uiconf_conf, uimod.schema_path)

        uiconf = uimod.load_conf()
        uiconf_pub = uiconf['rw']
        with Endpoint() as endpoint:
            if uiconf_pub != new_uiconf_conf:
                old_iris = uiconf_pub['iris']
                old_focus = uiconf_pub['focus']

                new_iris = new_uiconf_conf['iris']
                new_focus = new_uiconf_conf['focus']

                _, max_iris, _, max_focus = cls._rotate(endpoint, old_iris, new_iris, old_focus, new_focus)

                uiconf['rw'] = new_uiconf_conf
                uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
                from edge.webserver.sysconf.edge_creator import generate
                generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})

            else:
                _, max_iris, _, max_focus = cls._get_rotate_limits(endpoint)

        reply = {
            'type': ph.UI_TYPE_LENS,
            'conf': new_uiconf_conf,
            'limits': {
                'iris': max_iris,
                'focus': max_focus
            },
            'stats': cam_dgstats.get_stats(uimod.pid)
        }
        return reply
