import base64
import glob
import os
import pexpect
import subprocess
import ujson

from edge import eventlog
from edge.com import fs
from edge.exceptions import *

DEMO_KEYREF = 'embedded-demo'

_FPS_TO_TITLE_MAP = {
    25: 'license.lpr.group.25',
    6: 'license.lpr.group.6',
    0: 'license.lpr.group.0'
}


class _Lsvpwc:
    _HASP_FEATURE_HAS_EXPIRED_TEXT = 'Feature has expired'

    _HASP_FEATURE_NOT_FOUND_TEXT = 'Feature not found'
    _HASP_FEATURE_NOT_FOUND_CODE = 'H0031'

    _HASP_KEY_NOT_FOUND_TEXT = 'Sentinel key not found'
    _HASP_KEY_NOT_FOUND_CODE = 'H0007'

    def __init__(self):
        pass

    @classmethod
    def lsvpwc_key(cls, key_ref):
        """
        if licgroup present but expired, the lsvpwc returns licgroup with
        max_fps==0 and nthreads_max==0 and templates==''

        :type key_ref: str
        :rtype: list[dict]
        """

        """
        # /opt/edge/bin/lsvpwc -m /opt/edge/libexec/lpr/aorp/modules/ -t json
        [
            {
                "key_id": "1148957907760805623",
                "key_series": "107392",
                "fps_max": 0,
                "nthreads_max": 0,
                "templates": "",
                "flags": "_"
            },
            {
                "key_id": "1087110020766043669",
                "key_series": "107392",
                "fps_max": 25,
                "nthreads_max": 8,
                "templates": "*",
                "flags": "_"
            }
        ]        
        """

        vpw_licgroups = cls._lsvpwc(by_group=False)
        vpw_licgroups = filter(lambda x: x['key_id'] == key_ref, vpw_licgroups)
        return vpw_licgroups

    @classmethod
    def lsvpwc_key_all(cls):
        """
        :rtype: list[dict]
        """

        """
        # /opt/edge/bin/lsvpwc -m /opt/edge/libexec/lpr/aorp/modules/ -t json -g
        [
            {
                "key_id": "----",
                "key_series": "----",
                "fps_max": 0,
                "nthreads_max": 0,
                "templates": "",
                "flags": "_"
            },
            {
                "key_id": "----",
                "key_series": "----",
                "fps_max": 25,
                "nthreads_max": 8,
                "templates": "*", // space-separated-list or *
                "flags": "_"
            }
        ]
        """

        vpw_licgroups = cls._lsvpwc(by_group=True)
        return vpw_licgroups

    @classmethod
    def _lsvpwc(cls, by_group):
        """
        :type by_group: bool
        :rtype: list[dict]
        """
        command_line = '/opt/edge/bin/lsvpwc -m /opt/edge/libexec/lpr/aorp/modules/ -t json'
        if by_group:
            command_line += ' -g'

        lsvpwc_proc = pexpect.spawn(command_line)
        try:
            lsvpwc_proc.expect(pexpect.EOF)
            output = lsvpwc_proc.before
            if not output:
                eventlog.log_e('fail: lsvpwc (stdout is empty)')
                raise InternalError

        finally:
            lsvpwc_proc.close()

        try:
            vpw_licgroups = ujson.loads(output)
            return vpw_licgroups

        except ValueError:
            if cls._HASP_FEATURE_HAS_EXPIRED_TEXT in output:
                eventlog.log_w(cls._HASP_FEATURE_HAS_EXPIRED_TEXT)

            elif cls._HASP_FEATURE_NOT_FOUND_TEXT in output or cls._HASP_FEATURE_NOT_FOUND_CODE in output:
                eventlog.log_w(cls._HASP_FEATURE_NOT_FOUND_TEXT)

            elif cls._HASP_KEY_NOT_FOUND_TEXT in output or cls._HASP_KEY_NOT_FOUND_CODE in output:
                eventlog.log_w(cls._HASP_KEY_NOT_FOUND_TEXT)

            else:
                eventlog.log_e('fail: loads ({})'.format(output))
                raise InternalError

        return []

    @staticmethod
    def get_ui_countries_jnode(vpw_license_countries):
        """
        :type vpw_license_countries: str
        :rtype: dict | None
        """
        # if lsvpwc returns '' (empty templates), license has expired
        if not vpw_license_countries:
            return []

        # if lsvpwc returns * activate all world
        if '*' in vpw_license_countries:
            vpw_license_countries = ' '

        stdout, stderr = subprocess.Popen(
            '/opt/edge/bin/lsvpwi -m /opt/edge/libexec/lpr/aorp/modules/ -h -t cut -s id %s' % vpw_license_countries,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True).communicate()

        if stderr:
            eventlog.log_w('fail: lsvpwi (%s)' % stderr)
            return None

        template_refs = stdout.split('\n')[:-1]

        from collections import defaultdict
        country_to_region_to_templates_map = defaultdict(lambda: defaultdict(list))
        for template_ref in template_refs:
            items = template_ref.split('.')
            item_count = len(items)
            if item_count == 2:
                country_a2 = items[0]
                region_a2 = country_a2

            elif item_count == 3:
                country_a2 = items[0]
                region_a2 = items[1]

            else:
                eventlog.log_e('fail: unsupported (templater-ref:{})'.format(template_ref))
                raise NotImplementedError

            country_to_region_to_templates_map[country_a2][region_a2].append(template_ref)

        child = pexpect.spawn('/opt/edge/bin/lpr-iso3166', template_refs)
        try:
            child.expect(pexpect.EOF)
            stdout = child.before
            if not stdout:
                eventlog.log_e('fail: lpr-iso3166 (stdout is empty)')
                raise InternalError

        finally:
            child.close()

        country_region_to_titles_map = ujson.loads(stdout)

        countries_jnode = []
        for country_a2, region_to_templates_map in country_to_region_to_templates_map.iteritems():
            country_title = None
            regions_jnode = []
            for region_a2, template_refs in region_to_templates_map.iteritems():
                region_jnode = {
                    'templates': template_refs
                }

                country_title = country_region_to_titles_map[country_a2][region_a2]['country-title']
                region_title = country_region_to_titles_map[country_a2][region_a2].get('region-title')
                if region_title:
                    region_jnode['title'] = region_title

                regions_jnode.append(region_jnode)

            if 'title' in regions_jnode[0]:
                regions_jnode = sorted(regions_jnode, key=lambda region_item: region_item['title'])

            countries_jnode.append({
                'title': country_title,
                'regions': regions_jnode
            })

        countries_jnode.sort(key=lambda country_item: country_item['title'])
        return countries_jnode


class _Hasplm:
    _LIB_NAME = 'libsntl_adminapi_linux_x86_64.so'

    ATTACHED_KEYS_DIR = '/var/hasplm/installed/107392/'
    DETACHED_KEYS_DIR = '/etc/edge/hasplm/detached/'
    BASE_KEY_NAME_SUFFIX = '_base.v2c'
    UPDATED_KEY_NAME_SUFFIX = '_update*.v2c'

    def __init__(self):
        pass

    @staticmethod
    def _silent_mkdir_r(dirpath):
        """
        :type dirpath: str
        :rtype: bool
        """
        if not fs.path_exists(dirpath):
            fs.silent_mkdir_r(dirpath)
            return True

        return False

    @classmethod
    def test_key_attached(cls, keyref):
        """
        :type keyref: str
        :rtype: bool
        """
        # mkdir for detached keys when called first time
        cls._silent_mkdir_r(cls.DETACHED_KEYS_DIR)

        path_list = glob.glob(cls.DETACHED_KEYS_DIR + cls.BASE_KEY_NAME_SUFFIX)
        detached_keyrefs = map(lambda x: os.path.basename(x.strip(cls.BASE_KEY_NAME_SUFFIX)), path_list)

        return keyref not in detached_keyrefs

    @classmethod
    def list_detached_keyrefs(cls):
        """
        :rtype: list[str]
        """
        if cls._silent_mkdir_r(cls.DETACHED_KEYS_DIR):
            return []

        path_list = glob.glob(cls.DETACHED_KEYS_DIR + '*.v2c')
        detached_keyrefs = map(lambda x: os.path.basename(x.split('_')[0]), path_list)
        detached_keyrefs = list(set(detached_keyrefs))
        return detached_keyrefs

    @staticmethod
    def remove_key(keyref):
        """
        :type keyref: str
        :rtype: None
        """
        key_to_delete_filepath_set = []
        attached_key = False

        # attached "base" keys
        key_to_delete_filepath = os.path.join(_Hasplm.ATTACHED_KEYS_DIR, keyref + _Hasplm.BASE_KEY_NAME_SUFFIX)
        if os.path.exists(key_to_delete_filepath):
            key_to_delete_filepath_set.append(key_to_delete_filepath)
            attached_key = True

        # attached "updated" keys
        updated_key_to_delete_filepath_set = glob.glob(os.path.join(_Hasplm.ATTACHED_KEYS_DIR, keyref)
                                                       + _Hasplm.UPDATED_KEY_NAME_SUFFIX)

        for updated_key_to_delete in updated_key_to_delete_filepath_set:
            if os.path.exists(updated_key_to_delete):
                key_to_delete_filepath_set.append(updated_key_to_delete)
                attached_key = True

        # detached "base" key
        key_to_delete_filepath = os.path.join(_Hasplm.DETACHED_KEYS_DIR, keyref + _Hasplm.BASE_KEY_NAME_SUFFIX)
        if os.path.exists(key_to_delete_filepath):
            key_to_delete_filepath_set.append(key_to_delete_filepath)

        # detached "updated" keys
        updated_key_to_delete_filepath_set = glob.glob(os.path.join(_Hasplm.DETACHED_KEYS_DIR, keyref)
                                                       + _Hasplm.UPDATED_KEY_NAME_SUFFIX)
        for updated_key_to_delete in updated_key_to_delete_filepath_set:
            if os.path.exists(updated_key_to_delete):
                key_to_delete_filepath_set.append(updated_key_to_delete)

        # delete keys
        for filepath_to_delete in key_to_delete_filepath_set:
            if os.path.exists(filepath_to_delete):
                os.remove(filepath_to_delete)

            else:
                eventlog.log_e('fail: InternalError (file absent)')
                raise InternalError

        if attached_key:
            _Hasplm.restart()

    @classmethod
    def restart(cls):
        """
        :rtype: None
        """
        import platform
        if platform.machine() != 'x86_64':
            eventlog.log_w('fail: unsupported (machine is not x86)')
            return

        os.system('service aksusbd restart')

        # wait until license-manager is available
        retry_count = 0
        while retry_count < 4:
            try:
                cls._get_info()
                break
            except RuntimeError:
                retry_count += 1

        if retry_count >= 4:
            raise RuntimeError('fail: HASP license-manager unavailable')

    @classmethod
    def _get_info(cls):
        """
        :return: hasp_license_manager_info
        :rtype: dict
        """
        query_text = ''.join([
            '<admin>',
            '<license_manager format="json">',
            '<element name="*"/>',
            '</license_manager>',
            '</admin>'
        ])

        reply_text = cls._call_admin_api_so(query_text)
        hasp_license_manager_info = ujson.loads(reply_text)[:-1]  # strip summary
        return hasp_license_manager_info

    @classmethod
    def _call_admin_api_so(cls, query_text, vendor_id=None):
        """
        :type query_text: str
        :type vendor_id: str | None
        :rtype: str
        """
        import ctypes

        lib_admin_api = ctypes.CDLL(cls._LIB_NAME)

        session_self = ctypes.POINTER(ctypes.c_void_p)()
        session_scope = '<haspscope><host>localhost</host></haspscope>'
        answer = lib_admin_api.sntl_admin_context_new_scope(ctypes.byref(session_self), session_scope)
        if answer:
            eventlog.log_e('fail: sntl_admin_context_new_scope ({})'.format(answer))
            raise RuntimeError(answer)

        query_scope = ''
        if vendor_id:
            query_scope = query_scope.join([
                '<haspscope>',
                '<vendor><id>{}</id></vendor>'.format(vendor_id),
                '</haspscope>'
            ])

        try:
            reply_buffer = ctypes.POINTER(ctypes.c_char_p)()
            answer = lib_admin_api.sntl_admin_get(session_self, query_scope, query_text, ctypes.byref(reply_buffer))
            if answer:
                eventlog.log_e('fail: sntl_admin_get ({})'.format(answer))
                raise RuntimeError(answer)

            try:
                reply_text = ctypes.cast(reply_buffer, ctypes.c_char_p).value
            finally:
                lib_admin_api.sntl_admin_free(reply_buffer)

        finally:
            lib_admin_api.sntl_admin_context_delete(session_self)

        return cls._strip_reply_comments(reply_text)

    @staticmethod
    def _strip_reply_comments(raw_reply_text):
        """
        Strip comments in JSON-string and make string a JSON-array
        :type raw_reply_text: str
        :rtype: str
        """
        begin_comment = 0
        end_comment = 0

        while begin_comment != -1 and end_comment != -1:
            begin_comment = raw_reply_text.find('/*')
            end_comment = raw_reply_text.find('*/')
            if begin_comment == -1 and end_comment == -1:
                continue
            s = raw_reply_text[:begin_comment] + '' + raw_reply_text[end_comment + 2:]
            raw_reply_text = s

        return '[' + raw_reply_text + ']'


class HaspRus:
    _TEMP_V2C_PATH = '/tmp/license'

    _BIN_PATH = '/opt/edge/bin/hasp_update'
    _BIN_CMD_GET_DEMO_KEYREF = 'get_demo'
    _BIN_CMD_GET_KEYS_INFO = 'l'
    _BIN_CMD_GET_HOST_C2V = 'f'
    _BIN_CMD_GET_KEY_C2V = 'i'
    _BIN_CMD_INSTALL_V2C = 'u'

    def __init__(self):
        pass

    @classmethod
    def run(cls, params):
        """
        :type params: list[str]

        :return: (exitcode, stdout, stderr)
        :rtype: (str, str, int)
        """
        cmdline = ' '.join([cls._BIN_PATH] + params)
        eventlog.log_d('exec: {}'.format(cmdline))

        hasprus = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = hasprus.communicate()
        exitcode = hasprus.returncode
        return exitcode, stdout, stderr

    @classmethod
    def list_attached_keys_info(cls):
        """
        :rtype: dict
        """

        """
        # /opt/edge/bin/hasp_update l
        {"1096745228098839639":"HASP-SL","1136738148337368145":"HASP-SL"}
        #         
        """
        exitcode, stdout, stderr = cls.run([cls._BIN_CMD_GET_KEYS_INFO])
        if exitcode != 0:
            eventlog.log_e('fail: InternalError ({}, {}, {})'.format(exitcode, stdout, stderr))
            raise InternalError

        keyref_to_type_map = ujson.loads(stdout)
        return keyref_to_type_map

    @classmethod
    def install_v2c_file(cls, v2c_path):
        """
        :type v2c_path: str
        :rtype: None
        """
        exitcode, stdout, stderr = cls.run([cls._BIN_CMD_INSTALL_V2C, v2c_path])
        if exitcode != 0 or stdout:
            eventlog.log_e('fail: InternalError ({}, {}, {})'.format(exitcode, stdout, stderr))
            raise InternalError

    @classmethod
    def install_v2c_demo(cls):
        """
        :rtype: None
        """
        if cls.get_demo_keyref():
            eventlog.log_i('done: already activated')

        else:
            cls.install_v2c_file('/opt/edge/share/edge-embedded-demo.v2c')
            eventlog.log_i('done')

    @classmethod
    def install_v2c_base64(cls, v2c_base64):
        """
        :type v2c_base64: str
        :rtype: None
        """
        try:
            v2c_text = base64.b64decode(v2c_base64)
        except TypeError:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        with open(cls._TEMP_V2C_PATH, 'w') as text_file:
            text_file.write(v2c_text)

        cls.install_v2c_file(cls._TEMP_V2C_PATH)

    @classmethod
    def get_host_c2v_base64(cls):
        """
        :rtype: str
        """
        exitcode, stdout, stderr = cls.run([cls._BIN_CMD_GET_HOST_C2V])
        if exitcode == 0 and stdout:
            return base64.b64encode(stdout)

        eventlog.log_e('fail: InternalError ({}, {}, {})'.format(exitcode, stdout, stderr))
        raise InternalError

    @classmethod
    def get_key_c2v_base64(cls, key_ref):
        """
        :type key_ref: str
        :rtype: str
        """
        exitcode, stdout, stderr = cls.run([cls._BIN_CMD_GET_KEY_C2V, key_ref])
        if exitcode == 0 and stdout:
            return base64.b64encode(stdout)

        eventlog.log_e('fail: InternalError ({}, {}, {})'.format(exitcode, stdout, stderr))
        raise InternalError

    @classmethod
    def get_demo_keyref(cls):
        """
        Returns haspid of demokey if it is installed, returns None if demo-key not found
        :rtype: str | None
        """
        exitcode, stdout, stderr = cls.run([cls._BIN_CMD_GET_DEMO_KEYREF])
        if exitcode != 0:
            eventlog.log_e('fail: InternalError ({}, {})'.format(stdout, stderr))
            raise InternalError

        demo_keyrefs = ujson.loads(stdout)
        if len(demo_keyrefs) > 0:
            return demo_keyrefs[0]


# noinspection PyClassHasNoInit
class LicenseApi:
    @classmethod
    def get_keys_info(cls):
        """
        :rtype: list[dict]
        """
        attached_keys_info = cls.get_attached_uikeys_info()
        detached_keys_info = cls.get_detached_uikeys_info()
        return attached_keys_info + detached_keys_info

    @staticmethod
    def get_used_license_count(license_group=None):
        """
        :type license_group: str | None
        :rtype: int
        """
        lr_uiconf_path_list = glob.glob('/etc/edge/setupstore/admin/actions/pipeline/*/lpr/*.json')

        used_license_count = 0
        for uiconf_path in lr_uiconf_path_list:
            uiconf = fs.load_json3(uiconf_path, default=None)
            if uiconf is None:
                eventlog.log_e('fail: InternalError')
                raise InternalError

            uiconf_pub = uiconf['rw']  # type: dict
            if license_group is None:
                used_license_count += uiconf_pub['license-count']

            elif license_group == uiconf_pub['license-group']:
                used_license_count += uiconf_pub['license-count']

        return used_license_count

    @staticmethod
    def get_total_license_count():
        """
        :rtype: int
        """
        total_license_count = 0
        for license_key in _Lsvpwc.lsvpwc_key_all():
            total_license_count += license_key['nthreads_max']

        return total_license_count

    @classmethod
    def test_lpr_license_available(cls):
        """
        :rtype: None
        """
        total_count = cls.get_total_license_count()
        used_count = cls.get_used_license_count()
        eventlog.log_d('license total_thread_count: {}, used_thread_count: {}'.format(
            total_count, used_count))

        if used_count >= total_count:
            eventlog.log_d('fail: free lpr licence is not available')
            raise LicServiceError

        eventlog.log_d('done')

    @staticmethod
    def get_attached_uikeys_info():
        """
        :rtype: list[dict]
        """
        # TODO removable / detachable
        keyref_to_type_map = HaspRus.list_attached_keys_info()

        attached_uikeys_info = []
        for keyref, keytype in keyref_to_type_map.iteritems():
            sw_key = keytype != 'HASP_HL'
            attached_uikeys_info.append({
                'id': keyref,
                'removable': False,#sw_key,
                'detachable': False,#sw_key,
                'attached': _Hasplm.test_key_attached(keyref)
            })

        demo_keyref = HaspRus.get_demo_keyref()
        if demo_keyref:
            # replace demo-key jnode with stub
            first_demo_ui_key_jnode = filter(
                lambda x: x['id'] == demo_keyref,
                attached_uikeys_info)[0]

            attached_uikeys_info.remove(first_demo_ui_key_jnode)
            attached_uikeys_info.append({
                'id': DEMO_KEYREF,
                'removable': False,
                'detachable': False,
                'attached': True
            })

        else:
            attached_uikeys_info.append({
                'id': DEMO_KEYREF,
                'removable': False,
                'detachable': True,
                'attached': False
            })

        return attached_uikeys_info

    @staticmethod
    def get_detached_uikeys_info():
        """
        :rtype: list[dict]
        """
        # TODO removable / detachable
        detached_keyrefs = _Hasplm.list_detached_keyrefs()
        detached_uikeys_info = map(lambda keyref: {
            'id': keyref,
            'removable': False,#True,
            'detachable': False,#True,
            'attached': False
        }, detached_keyrefs)

        return detached_uikeys_info

    @staticmethod
    def get_host_c2v_base64():
        """
        :rtype: str
        """
        return HaspRus.get_host_c2v_base64()

    @staticmethod
    def get_key_c2v_base64(keyref):
        """
        :type keyref: str
        :rtype: str
        """
        # check if key-id valid
        vpw_licgroups = _Lsvpwc.lsvpwc_key(keyref)
        if not vpw_licgroups:
            eventlog.log_e('fail: NotFoundError (key-ref:{})'.format(keyref))
            raise NotFoundError

        return HaspRus.get_key_c2v_base64(keyref)

    @classmethod
    def install_v2c_base64(cls, v2c_base64):
        """
        :type v2c_base64: str
        :rtype: None
        """
        HaspRus.install_v2c_base64(v2c_base64)
        eventlog.log_d('done')

    @staticmethod
    def attach_key(keyref):
        """
        :type keyref: str
        :rtype: None
        """
        if keyref == 'embedded-demo':
            HaspRus.install_v2c_demo()
            return

        # try attach other then 'embedded-demo' key
        if keyref not in _Hasplm.list_detached_keyrefs():
            eventlog.log_e('fail: NotFoundError (requested key %s is absent)' % keyref)
            raise NotFoundError

        _silent_mkdir_r(_Hasplm.DETACHED_KEYS_DIR)

        key_to_attach_filepath_set = []

        key_to_attach_filepath = os.path.join(_Hasplm.DETACHED_KEYS_DIR, keyref + _Hasplm.BASE_KEY_NAME_SUFFIX)
        if os.path.exists(key_to_attach_filepath):
            key_to_attach_filepath_set.append(key_to_attach_filepath)

        updated_key_to_attach_filepath_set = glob.glob(os.path.join(_Hasplm.DETACHED_KEYS_DIR, keyref)
                                                       + _Hasplm.UPDATED_KEY_NAME_SUFFIX)
        for updated_key_to_attach in updated_key_to_attach_filepath_set:
            if os.path.exists(updated_key_to_attach):
                key_to_attach_filepath_set.append(updated_key_to_attach)

        if key_to_attach_filepath_set:
            fs.move_files(key_to_attach_filepath_set, _Hasplm.ATTACHED_KEYS_DIR)
            _Hasplm.restart()

    @staticmethod
    def detach_key(keyref):
        """
        :type keyref: str
        :rtype: None
        """
        if keyref not in HaspRus.list_attached_keys_info().keys():
            eventlog.log_e('fail: NotFoundError (key %s is absent)' % keyref)
            raise NotFoundError

        _silent_mkdir_r(_Hasplm.DETACHED_KEYS_DIR)

        key_to_detach_filepath_set = []

        key_to_detach_filepath = os.path.join(_Hasplm.ATTACHED_KEYS_DIR, keyref + _Hasplm.BASE_KEY_NAME_SUFFIX)
        if os.path.exists(key_to_detach_filepath):
            key_to_detach_filepath_set.append(key_to_detach_filepath)

        updated_key_to_detach_filepath_set = glob.glob(os.path.join(_Hasplm.ATTACHED_KEYS_DIR, keyref)
                                                       + _Hasplm.UPDATED_KEY_NAME_SUFFIX)
        for updated_key_to_detach in updated_key_to_detach_filepath_set:
            if os.path.exists(updated_key_to_detach):
                key_to_detach_filepath_set.append(updated_key_to_detach)

        if key_to_detach_filepath_set:
            fs.move_files(key_to_detach_filepath_set, _Hasplm.DETACHED_KEYS_DIR)
            _Hasplm.restart()
            return

        eventlog.log_e('fail: detach key')
        raise InternalError

    @staticmethod
    def remove_key(keyref):
        """
        :type keyref: str
        :rtype: dict
        """
        _Hasplm.remove_key(keyref)

    @staticmethod
    def get_licinfo_key_keyref(keyref):
        """
        :type keyref: str
        :rtype: list[dict] | None
        """
        if keyref == DEMO_KEYREF:
            keyref = HaspRus.get_demo_keyref()

        vpw_licgroups = _Lsvpwc.lsvpwc_key(keyref)
        if not vpw_licgroups:
            eventlog.log_e('fail: NotFoundError (keyref:{})'.format(keyref))
            raise NotFoundError

        from collections import defaultdict
        acc = defaultdict(lambda: defaultdict(set))

        for licgroup in vpw_licgroups:
            group_ref = licgroup['fps_max']
            if acc[group_ref].get('templates', None) is None:
                acc[group_ref]['templates'] = licgroup['templates'].split(' ')

            else:
                acc[group_ref]['templates'] += licgroup['templates'].split(' ')

            if acc[group_ref].get('nthreads_max', None) is None:
                acc[group_ref]['nthreads_max'] = licgroup['nthreads_max']

            else:
                acc[group_ref]['nthreads_max'] += licgroup['nthreads_max']

        ui_groups_info = []
        for licgroup, value in acc.iteritems():
            ui_groups_info.append({
                'id': str(licgroup),
                'title': _FPS_TO_TITLE_MAP[licgroup],
                'total-license-count': value['nthreads_max'],
                'fps-limit': licgroup,
                'countries': _Lsvpwc.get_ui_countries_jnode(' '.join(value['templates']))
            })

        if ui_groups_info:
            return {'lpr': ui_groups_info}

    @staticmethod
    def get_licinfo_key_all(key='lpr'):
        """
        :rtype: dict
        """
        vpw_licgroups = _Lsvpwc.lsvpwc_key_all()

        def get_jnode(lic_group):
            """
            :type lic_group: dict
            :rtype: dict
            """
            templates = lic_group['templates'] if lic_group['fps_max'] != 0 else '*'
            used_count = LicenseApi.get_used_license_count(str(lic_group['fps_max']))
            jnode = {
                'id': str(lic_group['fps_max']),
                'title': _FPS_TO_TITLE_MAP[lic_group['fps_max']],
                'total-license-count': lic_group['nthreads_max'],
                'used-license-count': used_count,
                'free-license-count': lic_group['nthreads_max'] - used_count,
                'fps-limit': lic_group['fps_max'],
                'countries': _Lsvpwc.get_ui_countries_jnode(templates)
            }
            return jnode

        ui_groups_info = map(get_jnode, vpw_licgroups)

        # drop expired licenses from list
        # expired_license_group_list = filter(lambda x: x['fps-limit'] == 0, ui_groups_info)
        # for expired in expired_license_group_list:
        #     ui_groups_info.remove(expired)

        return {key: ui_groups_info}


def _silent_mkdir_r(dirpath):
    """
    :type dirpath: str
    :rtype: bool
    """
    if not os.path.exists(dirpath):
        fs.silent_mkdir_r(dirpath)
        return True

    return False
