import subprocess

import ujson
import os

from edge import eventlog
from edge.com import fs
from edge.exceptions import *
from edge.webserver.sysconf import ph

import edge_creator as creator
import math
from edge.com.sh import exec_shell


# noinspection PyClassHasNoInit
def run_rlandmark_cmd(query_params, sensor_w_mm, sensor_h_mm, focal_dist_mm):
    """
    :type query_params: list[str]
    :type sensor_w_mm: float
    :type sensor_h_mm: float
    :type focal_dist_mm: float
    :rtype: str | None
    """

    cam_fov_h = math.degrees(2 * math.atan2(sensor_w_mm, 2 * focal_dist_mm))
    cam_fov_v = math.degrees(2 * math.atan2(sensor_h_mm, 2 * focal_dist_mm))

    args = {
        '--fov-vertical': cam_fov_v,
        '--fov-horizontal': cam_fov_h,

        '--position-forward': None,
        '--position-left': None,
        '--position-top': None,

        '--rotation-landmark-image': None,
        '--rotation-landmark-forward': None,
        '--rotation-landmark-left': None
    }

    for name_value in query_params:
        name, value = name_value.split('=', 1)

        if name == 'camera-forward':
            args['--position-forward'] = float(value)

        elif name == 'camera-left':
            args['--position-left'] = float(value)

        elif name == 'camera-up':
            args['--position-top'] = float(value)

        elif name == 'landmark-image':
            if value in ['bottom-left', 'bottom-right']:
                args['--rotation-landmark-image'] = value

        elif name == 'landmark-forward':
            args['--rotation-landmark-forward'] = float(value)

        elif name == 'landmark-left':
            args['--rotation-landmark-left'] = float(value)

    invalid_args = filter(lambda (arg_name, arg_value): arg_value is None,
                          args.iteritems())
    if invalid_args:
        for invalid_arg_name, _ in invalid_args:
            eventlog.log_d('fail: param <{}> is absent'.format(invalid_arg_name))
        raise ParamsError

    else:
        # tweak landmark coordinates relative to camera position
        args['--rotation-landmark-forward'] -= args['--position-forward']
        args['--rotation-landmark-left'] -= args['--position-left']

        command_line = '/opt/edge/bin/landmark_v2 --mode-rlandmark'
        for name, value in args.iteritems():
            command_line += ' {}={}'.format(name, value)

        try:
            stdout, exitcode = exec_shell(command_line)
            return ujson.loads(stdout)
        except InternalError:
            eventlog.log_e('fail: landmark')
            raise InternalError


# noinspection PyClassHasNoInit
class _HgRepo:
    _HG_DIR_PATH = '/var/opt/edge/hg'

    @classmethod
    def commit_location(cls, term_pid, location_uiconf, location_jpeg_path):
        """
        :type term_pid: str
        :type location_uiconf: dict
        :type location_jpeg_path: str
        """
        hgjpeg_path = fs.join_path(cls._HG_DIR_PATH, term_pid, '0.jpeg')
        hgconf_path = fs.join_path(cls._HG_DIR_PATH, term_pid, '{}.0.json'.format(ph.UI_TYPE_LOCATION))

        first_time = not fs.path_exists(hgconf_path)
        fs.silent_cp(location_jpeg_path, hgjpeg_path)
        fs.write_json_file(location_uiconf, hgconf_path, pretty=True)

        import hglib
        hgrepo = hglib.open(cls._HG_DIR_PATH)
        if first_time:
            hgrepo.add([hgconf_path, hgjpeg_path])

        hgrepo.commit(message='edit({}/location.0)'.format(term_pid), user='admin')


# noinspection PyClassHasNoInit
class LocationApi:
    @staticmethod
    def load_unit(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_LOCATION)
        uiconf = uimod.load_conf(default=None)
        cam_dgstats = ph.CamDgStats.create(uimod.pid)
        cam_stats_jnode = cam_dgstats.get_stats(uimod.pid)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        uiconf_pub = uiconf['rw']
        reply = {
            'type': ph.UI_TYPE_LOCATION,
            'conf': uiconf_pub['args'],
            'stats': cam_stats_jnode
        }
        return reply

    @staticmethod
    def save_unit(uimod_ref, new_uiconf):
        """
        :type uimod_ref: str
        :type new_uiconf: dict
        :rtype: dict
        """
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_LOCATION)
        fs.validate_json({'args': new_uiconf_conf}, uimod.schema_path)

        uiconf = uimod.load_conf()
        if uiconf['rw']['args'] != new_uiconf_conf:
            live_jpeg_path = _get_live_image(uimod.pid)
            _HgRepo.commit_location(uimod.pid, uiconf, live_jpeg_path)

            uiconf['rw']['args'] = new_uiconf_conf
            uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
            creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})

        cam_dgstats = ph.CamDgStats.create(uimod.pid)
        cam_stats_jnode = cam_dgstats.get_stats(uimod.pid)
        reply = {
            'type': ph.UI_TYPE_LOCATION,
            'conf': new_uiconf_conf,
            'stats': cam_stats_jnode
        }
        return reply

    @staticmethod
    def estimate_rotation(uimod_ref, query_params):
        """
        :type uimod_ref: str
        :type query_params: list[str]
        :rtype: dict
        """
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_LOCATION)

        whf = ph.CamUiStats.create(uimod.pid).get_whf()
        if whf is None:
            eventlog.log_d('fail: NotImplementedError')
            raise NotImplementedError

        landmark_jnode = run_rlandmark_cmd(query_params, *whf)

        landmark_rot = landmark_jnode['unit-rotation-auto']
        cam_rotation = {
            'camera': {
                'rotation': {
                    'elevation': landmark_rot['elevation'],
                    'azimuth': landmark_rot['azimuth']
                }
            }
        }

        return cam_rotation


def _get_live_image(ui_pid):
    """
    :type ui_pid: str
    :rtype: str
    """
    # get last overlay imagepath
    or_id = str(int(ui_pid) * 10)
    or_store = '/var/spool/edge/pipeline-{0}/or{1}-or{1}-store/'.format(ui_pid, or_id)
    file_names = os.listdir(or_store)

    def unpack_ts(file_name):
        """
        :type file_name: str
        :rtype: int
        """
        return int(file_name.split('-')[-1])

    blob_name = max(file_names, key=unpack_ts)
    blob_path = os.path.join(or_store, blob_name)
    return blob_path
