from shapely.geometry import polygon

from edge import eventlog
from edge.com import fs
from edge.exceptions import *
from edge.webserver.sysconf import ph
import edge_creator as creator
from edge.webserver.sysconf.api_license import LicenseApi


# noinspection PyClassHasNoInit
class LprApi:
    @staticmethod
    def load_unit(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_LPR)
        cam_dgstats = ph.CamDgStats.create(uimod.pid)
        cam_dgstats.test_availability()
        cam_stats_jnode = cam_dgstats.get_stats(uimod.pid)

        uiconf = uimod.load_conf()
        ext_uiconf_pub = uiconf['rw']

        lic_limits = LicenseApi.get_licinfo_key_all('license')

        selected_group = filter(lambda x: x['id'] == ext_uiconf_pub['license-group'], lic_limits['license'])
        if not selected_group:
            ext_uiconf_pub['license-group'] = '0'
            ext_uiconf_pub['license-count'] = 0

        reply_jnode = {
            'type': ph.UI_TYPE_LPR,
            'conf': ext_uiconf_pub,
            'limits': lic_limits,
            'stats': cam_stats_jnode
        }
        return reply_jnode

    @classmethod
    def save_unit(cls, uimod_ref, new_uiconf):
        """
        :type uimod_ref: str
        :type new_uiconf: dict
        :rtype: dict
        """
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_LPR)
        cam_dgstats = ph.CamDgStats.create(uimod.pid)
        cam_dgstats.test_availability()

        fs.validate_json(new_uiconf_conf, uimod.schema_path)

        lic_limits = LicenseApi.get_licinfo_key_all('license')

        def validate_azone():
            azone_jnode = new_uiconf_conf['plate-recognizer']['image-azone']
            try:
                # validate self-intersection of include-mas
                include_jnode = azone_jnode['include-mask']

                polymask = polygon.Polygon(include_jnode)
                if not polymask.is_valid:
                    eventlog.log_d('fail: http400 (invalid include-mask)')
                    raise ParamsError

            except (ValueError, IndexError):
                eventlog.log_d('include-mask contains less then 3 vertexes. using whole image as include-mask')
                azone_jnode['include-mask'] = []

            try:
                # validate self-intersection of exclude-mask
                exclude_jnode = azone_jnode['exclude-mask']

                for exclude_mask in exclude_jnode:
                    polymask = polygon.Polygon(exclude_mask)
                    if not polymask.is_valid:
                        eventlog.log_d('fail: http400 (invalid exclude-mask)')
                        raise ParamsError

            except (ValueError, IndexError):
                eventlog.log_d('at least one of poly-regions in exclude-mask contains less then 3 vertexes.'
                               ' ignoring exclude-mask')
                azone_jnode['exclude-mask'] = []

        def validate_lic():
            lic_group = new_uiconf_conf['license-group']
            selected_group = filter(lambda x: x['id'] == lic_group, lic_limits['license'])
            if lic_group == '0' or len(selected_group) == 0:
                eventlog.log_e('fail: http400 (invalid license group)')
                raise ParamsError

        validate_azone()
        validate_lic()

        uiconf = uimod.load_conf()
        if uiconf['rw'] != new_uiconf_conf:
            uiconf['rw'] = new_uiconf_conf
            uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
            creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})

        ext_uiconf_pub = new_uiconf_conf

        cam_stats_jnode = cam_dgstats.get_stats(uimod.pid)
        reply_jnode = {
            'type': ph.UI_TYPE_LPR,
            'conf': ext_uiconf_pub,
            'limits': lic_limits,
            'stats': cam_stats_jnode
        }
        return reply_jnode
