
from edge import eventlog
from edge.com import fs
from edge.exceptions import *
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
import edge_creator as creator


# noinspection PyClassHasNoInit
class MvCameraApi:
    @staticmethod
    def load_unit(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_MV_CAMERA)
        uiconf = uimod.load_conf()
        ext_uiconf_pub = uiconf['rw']
        stats_jnode = ph.CamDgStats.create(uimod.pid).get_stats(uimod.pid)
        stats_jnode['vendor'] = ext_uiconf_pub.pop('vendor')
        stats_jnode['model'] = ext_uiconf_pub.pop('model')
        stats_jnode['serial'] = ext_uiconf_pub.pop('serial')
        ext_uiconf_pub.pop('type', None)
        reply = {
            'type': ph.UI_TYPE_MV_CAMERA,
            'conf': ext_uiconf_pub,
            'stats': stats_jnode,
            'limits': uimod.schema_limits
        }
        return reply

    @classmethod
    def save_unit(cls, uimod_ref, new_uiconf):
        """
        :type uimod_ref: str
        :type new_uiconf: dict
        :rtype: dict
        """
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_MV_CAMERA)

        uiconf = uimod.load_conf()
        uiconf_pub = uiconf['rw']

        new_uiconf_conf['vendor'] = uiconf_pub.get('vendor')
        new_uiconf_conf['model'] = uiconf_pub.get('model')
        new_uiconf_conf['serial'] = uiconf_pub.get('serial')

        def validate_uiconf_pub():
            fs.validate_json(new_uiconf_conf, uimod.schema_path)

            # ROI checks - (width_offset + ROI_width) < max_width
            from copy import deepcopy
            roi_test = deepcopy(new_uiconf_conf)
            roi_test['videostream']['resolution']['w'] += roi_test['videostream']['resolution']['x']
            roi_test['videostream']['resolution']['h'] += roi_test['videostream']['resolution']['y']
            fs.validate_json(roi_test, uimod.schema_path)

            control = new_uiconf_conf['videostream']['control']
            if control['profile'] in ['auto-127', 'auto-70-150']:
                min_gain = control.get('gain-auto-min')
                max_gain = control.get('gain-auto-max')

                min_exposure = control.get('exposure-auto-min')
                max_exposure = control.get('exposure-auto-max')

                if None in [min_gain, max_gain, min_exposure, max_exposure]:
                    eventlog.log_d(
                        'fail: ParamsError (gain-auto-min:{}, gain-auto-max:{}, '
                        'exposure-auto-min:{}, exposure-auto-max:{})'.format(
                            min_gain, max_gain, min_exposure, max_exposure))
                    raise ParamsError

                if min_gain > max_gain:
                    # GrabControl works bad when min_value == max_value
                    eventlog.log_d('fail: ParamsError (gain-auto-min:{}, gain-auto-max:{})'.format(
                        min_gain, max_gain))
                    raise ParamsError

                if min_exposure > max_exposure:
                    # GrabControl works bad when min_value == max_value
                    eventlog.log_d('fail: ParamsError (exposure-auto-min:{}, exposure-auto-max:{})'.format(
                        min_exposure, max_exposure))
                    raise ParamsError

            elif control['profile'] == 'manual':
                gain = control.get('gain-manual')
                exposure = control.get('exposure-manual')
                if None in [gain, exposure]:
                    eventlog.log_d('fail: ParamsError (gain:{}, exposure:{})'.format(gain, exposure))
                    raise ParamsError

        validate_uiconf_pub()

        restart_setupstore = []
        if uiconf_pub['videostream'] != new_uiconf_conf['videostream']:
            if cls._test_gc_mode_changed(uiconf_pub, new_uiconf_conf):
                restart_setupstore.append(True)

            uiconf['rw'] = new_uiconf_conf
            uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
            creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})

        if restart_setupstore:
            dg.LifeCycle.sighup_ss_dgmod()

        stats_jnode = ph.CamDgStats.create(uimod.pid).get_stats(uimod.pid)
        stats_jnode['vendor'] = new_uiconf_conf.pop('vendor')
        stats_jnode['model'] = new_uiconf_conf.pop('model')
        stats_jnode['serial'] = new_uiconf_conf.pop('serial')
        new_uiconf_conf.pop('type', None)
        reply = {
            'type': ph.UI_TYPE_MV_CAMERA,
            'conf': new_uiconf_conf,
            'stats': stats_jnode,
            'limits': uimod.schema_limits
        }
        return reply

    @staticmethod
    def _test_gc_mode_changed(uiconf_pub, new_uiconf_pub):
        """
        :type uiconf_pub: dict
        :type new_uiconf_pub: dict
        :rtype: bool
        """
        old_gc_mode = uiconf_pub['videostream']['control']['profile'] == 'auto-70-150'
        new_gc_mode = new_uiconf_pub['videostream']['control']['profile'] == 'auto-70-150'
        return old_gc_mode != new_gc_mode


# noinspection PyClassHasNoInit
class MvCameraSearchApi:
    CAM_STATUS_USED = 'used'
    CAM_STATUS_UNUSED = 'unused'
    CAM_STATUS_UNDEFINED = 'undefined'

    VENDOR_POINTGRAY = 'pointgrey'
    VENDOR_BASLER = 'basler'
    VENDOR_DALSA = 'dalsa'
    VENDOR_JAI = 'jai'

    _VENDORS = (
        VENDOR_POINTGRAY,
        VENDOR_BASLER,
        VENDOR_DALSA,
        VENDOR_JAI)

    @staticmethod
    def _test_cam_used(serial):
        """
        :type serial: str
        :rtype: bool
        """
        root_info = ph.RootInfoUiconf.get_info()
        cam_uimods = ph.InfoUiconf.map_units_r(
            root_info, lambda x: ph.UiMod.create(x['ref']),
            unit_types=[ph.UI_TYPE_MV_CAMERA])

        for uimod in cam_uimods:
            uiconf = uimod.load_conf(default=None)
            if uiconf is None:
                pass

            elif serial == uiconf['rw']['serial']:
                return True

        return False

    @classmethod
    def _test_cam_unused(cls, vendor, serial):
        """
        :type vendor: str
        :type serial: str
        :rtype: bool
        """
        if vendor not in cls._VENDORS:
            eventlog.log_d('fail: ParamsError (invalid vendor)')
            raise ParamsError

        import subprocess
        sh_cmd = '. /opt/edge/rc.setenv; /opt/edge/bin/mediagrabber-info --vendor={} --serial={}'.format(vendor, serial)
        sh = subprocess.Popen([sh_cmd], stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stderr = sh.stderr.read()
        if stderr:
            eventlog.log_e('fail: mediagrabber-info ({})'.format(stderr))
            raise InternalError

        stdout = sh.stdout.read()
        try:
            # workaround for JAI cameras, strip anything before JSON
            stdout = stdout[stdout.index('{'):]
        except ValueError:
            pass

        import ujson
        try:
            ujson.loads(stdout)
            return True

        except ValueError:
            # camera status is 'undefined' if:
            # - it is not found in mv-camera-uiconf
            # - camera-info is not a valid JSON
            pass

        return False

    @classmethod
    def _get_cam_status(cls, vendor, serial):
        """
        :type vendor: str
        :type serial: str
        :rtype: str
        """
        if cls._test_cam_used(serial):
            return cls.CAM_STATUS_USED

        elif cls._test_cam_unused(vendor, serial):
            return cls.CAM_STATUS_UNUSED

        else:
            return cls.CAM_STATUS_UNDEFINED

    @classmethod
    def _get_cams_info(cls, vendor):
        """
        :type vendor: str
        :rtype: list[dict]
        """

        """
        # . /opt/edge/rc.setenv; /opt/edge/bin/mediagrabber-info --vendor=pointgrey
        Printing all PointGrey devices information found so far:
          == 0. PointGrey Device Information ==
          ===================================
            DeviceSerial     : 16050102
            DeviceInterface  : GigE
            DeviceModel      : Blackfly BFLY-PGE-23S6C
        # 
        """
        import subprocess
        sh_cmd = '. /opt/edge/rc.setenv; /opt/edge/bin/mediagrabber-info --vendor={}'.format(vendor)
        sh = subprocess.Popen([sh_cmd],  stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stderr = sh.stderr.read()
        if stderr:
            eventlog.log_e('fail: mediagrabber-info ({})'.format(stderr))
            raise InternalError

        stdout = sh.stdout.read()

        def unpack_stdout(mginfo_stdout):
            """
            :type mginfo_stdout: str
            :rtype: list[dict]
            """
            cam_sections = mginfo_stdout.split('===================================')[1:]

            answer = []
            for section in cam_sections:
                section_lines = section.split('\n')[1:]

                cam_info = dict()
                for line in section_lines:
                    if line:
                        param, value = line.split(':', 1)  # type: str, str
                        cam_info[param.strip()] = value.strip()

                answer.append(cam_info)

            return answer

        cams_info = []
        for camera in unpack_stdout(stdout):
            model = camera['DeviceModel']
            serial = camera['DeviceSerial']
            status = cls._get_cam_status(vendor, serial)

            cams_info.append({
                'vendor': vendor,
                'model': model,
                'serial': serial,
                'status': status
            })

        return cams_info

    @classmethod
    def list_cams(cls):
        cams_info = []
        for vendor in cls._VENDORS:
            cams_info += cls._get_cams_info(vendor)

        return cams_info
