import commands
import os
import shutil
import time

import debinterface.interfaces
from edge import eventlog
from edge.exceptions import *
from edge.com import net, fs
from edge.webserver.sysconf import ph


UI_NETWORK_SCHEMA_FILEPATH = '/etc/edge/setupstore/schemas/network.json'
_UICONF_PATH = '/etc/edge/setupstore/admin/actions/network.json'
_DEF_UICONF_PATH = '/opt/edge/share/setupstore/ui/network-jnode.json'

MANAGEMENT_MAC_FILEPATH = '/etc/network/management-mac'

LOCK_FILEPATH = '/var/run/edge/network'

INTERFACES_PATH = '/etc/network/interfaces'
ETC_HOSTNAME_FILEPATH = '/etc/hostname'
ETC_HOSTS_FILEPATH = '/etc/hosts'

RESTART_NONE = 0
RESTART_HOSTNAME = 1
RESTART_NETWORKING = 2

BACKUP_DIRPATH = '/etc/edge/setupstore/backup/network.' + str(round(time.time() * 1000000)) + '/'


def _try_create_backup():
    # noinspection PyBroadException
    try:
        os.makedirs(BACKUP_DIRPATH)
        shutil.copy(ETC_HOSTNAME_FILEPATH, BACKUP_DIRPATH)
        shutil.copy(ETC_HOSTS_FILEPATH, BACKUP_DIRPATH)
        shutil.copy(_UICONF_PATH, BACKUP_DIRPATH)
        return True

    except Exception:
        return False


def _try_restore_backup():
    # noinspection PyBroadException
    try:
        shutil.copy(BACKUP_DIRPATH + 'hostname', ETC_HOSTNAME_FILEPATH)
        shutil.copy(BACKUP_DIRPATH + 'hosts', ETC_HOSTS_FILEPATH)
        shutil.copy(BACKUP_DIRPATH + 'network.json', _UICONF_PATH)
        os.rmdir(BACKUP_DIRPATH)
        return True

    except Exception:
        return False


def _remove_backup():
    # noinspection PyBroadException
    try:
        shutil.rmtree(BACKUP_DIRPATH)
    except Exception:
        pass


def _edit_hostname(old_hostname, new_hostname):

    if old_hostname == new_hostname:
        return RESTART_NONE

    with open(ETC_HOSTS_FILEPATH, 'r') as old_hosts_file:
        old_hosts_lines = old_hosts_file.readlines()

    line_found = False
    new_host_lines = []
    for old_line in old_hosts_lines:
        stripped_old_line = old_line.strip()
        if not stripped_old_line or stripped_old_line.startswith('#'):
            new_host_lines.append(stripped_old_line)
            continue

        param_set = stripped_old_line.split()
        if len(param_set) < 2:
            new_host_lines.append(stripped_old_line)
            continue

        hostname = param_set[1]
        if hostname == old_hostname:
            param_set[1] = new_hostname
            new_host_lines.append('\t'.join(param_set))
            line_found = True

        else:
            new_host_lines.append(stripped_old_line)

    if not line_found:
        new_host_lines.append('\t'.join(['127.0.0.1', new_hostname]))

    with open(ETC_HOSTS_FILEPATH, 'w') as new_hosts_file:
        new_hosts_file.write('\n'.join(new_host_lines))

    with open(ETC_HOSTNAME_FILEPATH, 'w') as hostname_file:
        hostname_file.write(new_hostname)

    return RESTART_HOSTNAME


def _check_iface_changed(old_jnode, new_jnode):
    if old_jnode['ip-mode'] != new_jnode['ip-mode']:
        return True

    if old_jnode['ip-mode'] == 'static':
        if old_jnode['ip-address'] != new_jnode['ip-address']:
            return True

        if old_jnode['ip-mask'] != new_jnode['ip-mask']:
            return True

        if old_jnode['ip-gateway'] != new_jnode['ip-gateway']:
            return True
    else:
        pass

    if old_jnode['dns-mode'] != new_jnode['dns-mode']:
        return True

    if old_jnode['dns-mode'] == 'static':
        if old_jnode['dns-1'] != new_jnode['dns-1']:
            return True

        if old_jnode['dns-2'] != new_jnode['dns-2']:
            return True
    else:
        pass

    return False


def _edit_iface(old_jnode, new_jnode):
    if not _check_iface_changed(old_jnode, new_jnode):
        return RESTART_NONE

    mac_address = new_jnode['mac']

    ifname = net.get_ifname(mac_address)

    iface_conf = debinterface.interfaces.Interfaces(interfaces_path=INTERFACES_PATH)

    adapter_list = filter(lambda item: item.export(options_list=['name'])['name'] == ifname, iface_conf.adapters)

    if len(adapter_list) != 1:
        raise ValueError
    adapter = adapter_list[0]

    adapter.reset()
    if new_jnode['ip-mode'] == 'static':
        adapter.setName(ifname)
        adapter.setAuto(True)
        adapter.setAddrFam('inet')
        adapter.setAddressSource(new_jnode['ip-mode'])
        adapter.setAddress(new_jnode['ip-address'])
        adapter.setNetmask(new_jnode['ip-mask'])
        adapter.setGateway(new_jnode['ip-gateway'])

    elif new_jnode['ip-mode'] == 'dhcp':
        adapter.setName(ifname)
        adapter.setAuto(True)
        adapter.setAddrFam('inet')
        adapter.setAddressSource(new_jnode['ip-mode'])

    else:
        raise NotImplementedError('unsupported ip-mode %s' % new_jnode['ip-mode'])

    if new_jnode['dns-mode'] == 'static':
        dns_1 = new_jnode['dns-1']
        if dns_1:
            adapter.setDnsNameservers(dns_1)

        dns_2 = new_jnode['dns-2']
        if dns_2:
            adapter.setDnsNameservers(dns_1 + ' ' + dns_2)

    elif new_jnode['dns-mode'] == 'dhcp':
        pass

    else:
        raise NotImplementedError('unsupported dns-mode %s' % new_jnode['dns-mode'])

    iface_conf.writeInterfaces()

    return RESTART_NETWORKING


# noinspection PyClassHasNoInit
class NetworkApi:
    @staticmethod
    def load_unit(unit_uiref):
        """
        :type unit_uiref: str
        :rtype: dict
        """
        def install_uiconf(def_uiconf_path, uiconf_path):
            """
            :type def_uiconf_path: str
            :type uiconf_path: str
            """
            if not os.path.exists(uiconf_path):
                try:
                    with open(MANAGEMENT_MAC_FILEPATH, 'r') as text_file:
                        mac_addr = text_file.readline().strip()

                except IOError:
                    eventlog.log_w('management-mac file not found, using eth0 as management interface')
                    mac_addr = net.get_ifmac('eth0')
                    if not mac_addr:
                        eventlog.log_d('fail: InternalError (get ifmac)')
                        raise InternalError

                def_uiconf = fs.load_json3(def_uiconf_path)
                def_uiconf_pri = def_uiconf['ro']  # type: dict
                def_uiconf_pub = def_uiconf['rw']  # type: dict

                def_uiconf_pri['mac'] = mac_addr
                def_uiconf_pub['hostname'] = commands.getoutput('hostname')

                fs.write_json_file(def_uiconf, uiconf_path, pretty=True)

        def install_uischema(uiconf_path, uischema_path):
            """
            :type uiconf_path: str
            :type uischema_path: str
            """
            if not os.path.islink(uischema_path):
                fs.ln_schema(uischema_path, uiconf_path)

            elif not os.path.exists(os.readlink(uischema_path)):
                fs.ln_schema(uischema_path, uiconf_path)

        install_uiconf(_DEF_UICONF_PATH, _UICONF_PATH)
        install_uischema(_UICONF_PATH, UI_NETWORK_SCHEMA_FILEPATH)

        uiconf = fs.load_json3(_UICONF_PATH, default=None)
        if uiconf is None:
            eventlog.log_e('fail: InternalError')
            raise InternalError

        uiconf_pub = uiconf['rw']  # type: dict
        uiconf_pri = uiconf['ro']  # type: dict
        uiconf_pub['mac'] = uiconf_pri['mac']
        unit_ext_jnode = {
            'type': ph.UI_TYPE_NETWORK,
            'conf': uiconf['rw']
        }

        return unit_ext_jnode

    @staticmethod
    def save_unit(unit_uiref, unit_jnode):
        """
        :type unit_uiref: str
        :type unit_jnode:  dict
        :rtype: dict
        """
        uiconf = fs.load_json3(_UICONF_PATH, default=None)
        if uiconf is None:
            eventlog.log_e('fail: NotFoundError')
            raise NotFoundError

        new_uiconf_pub = unit_jnode['conf']
        fs.validate_json(new_uiconf_pub, UI_NETWORK_SCHEMA_FILEPATH)

        if not fs.try_lock_pid_file(LOCK_FILEPATH):
            eventlog.log_d('fail: BusyError')
            raise BusyError

        if not _try_create_backup():
            eventlog.log_e('fail: InternalError')
            raise InternalError

        uiconf_pub = uiconf['rw']
        try:
            restart_mask = 0
            restart_mask |= _edit_iface(uiconf_pub, new_uiconf_pub) or restart_mask
            restart_mask |= _edit_hostname(uiconf_pub['hostname'], new_uiconf_pub['hostname']) or restart_mask

            uiconf['rw'] = new_uiconf_pub.copy()
            uiconf['rw'].pop('mac')
            fs.write_json_file(uiconf, _UICONF_PATH, pretty=True)

            if restart_mask & RESTART_HOSTNAME:
                os.system("hostnamectl set-hostname %s" % new_uiconf_pub['hostname'])

            if restart_mask & RESTART_NETWORKING:
                os.system("service networking restart")

            _remove_backup()
            unit_ext_jnode = {
                'type': ph.UI_TYPE_NETWORK,
                'conf': new_uiconf_pub
            }

            return unit_ext_jnode

        except KeyError:
            eventlog.log_d('fail: ParamsError (mac-address)')
            raise ParamsError

        except Exception:
            eventlog.log_x()
            if _try_restore_backup():
                _remove_backup()

            raise InternalError
