
from edge import eventlog
from edge.com import fs, sh
from edge.exceptions import *

_UICONF_PATH = '/etc/edge/setupstore/admin/actions/openvpn.json'
_DGCONF_PATH = '/etc/default/openvpn'


def _get_new_ref(uiconf_profiles):
    """
    :type uiconf_profiles: list[dict]
    :rtype: str
    """
    used_refs = set(map(lambda x: int(x['ref']), uiconf_profiles))

    ref = None
    for candidate in xrange(0, 32):
        if candidate not in used_refs:
            return str(candidate)

    if ref is None:
        eventlog.log_e('fail: get new-ref')
        raise InternalError


def _openvpn_reload_restart():
    sh.run_silent('service openvpn stop')
    sh.run_silent('systemctl daemon-reload')
    sh.run_silent('service openvpn start')


def _mk_profile_path(ref):
    """
    :type ref: str
    :rtype: str
    """
    return '/etc/openvpn/{}.conf'.format(ref)


def _mk_user_path(ref):
    """
    :type ref: str
    :rtype: str
    """
    return '/etc/openvpn/{}.user'.format(ref)


def _get_active_profiles(jnode):
    """
    :type jnode: dict
    :rtype: set[str]
    """
    active_ref_set = set()
    for profile_jnode in jnode['rw']['conf']['profiles']:
        if profile_jnode['stats']['active']:
            active_ref_set.add(profile_jnode['ref'])

    return active_ref_set


def _set_autostart_profile(active_profiles):
    """
    :type active_profiles: set[str]
    :rtype: int
    """
    # Edit profile autostart.
    with open(_DGCONF_PATH, 'r') as old_default_vpn_file:
        old_defaults_vpn_lines = old_default_vpn_file.readlines()

    new_defaults_vpn_lines = []
    for old_line in old_defaults_vpn_lines:
        if not old_line.startswith('AUTOSTART='):
            new_defaults_vpn_lines.append(old_line)

    if len(active_profiles) == 0:
        autostart_line = 'AUTOSTART="none"\n'

    else:
        autostart_line = 'AUTOSTART="%s"\n' % ' '.join(active_profiles)

    new_defaults_vpn_lines.append(autostart_line)

    with open(_DGCONF_PATH, 'w') as new_default_vpn_file:
        new_default_vpn_file.write(''.join(new_defaults_vpn_lines))


def _add_auth_to_profile(conf_path, user_path, prof_user=None, prof_pass=None):
    """
    :type conf_path: str
    :type prof_user: str | None
    :type prof_pass: str | None
    :raises: BadRequestError
    """
    with open(conf_path, 'r') as old_default_vpn_file:
        old_defaults_vpn_lines = old_default_vpn_file.readlines()

    new_defaults_vpn_lines = []
    auth_user_path = False
    for old_line in old_defaults_vpn_lines:
        if old_line == 'auth-user-pass\n':
            auth_user_path = True
            if None in [prof_user, prof_pass]:
                eventlog.log_e('fail: user/pass not defined')
                raise ParamsError

            auth_conf = 'auth-user-pass {}\n'.format(user_path)
            new_defaults_vpn_lines.append(auth_conf)

        else:
            new_defaults_vpn_lines.append(old_line)

    if auth_user_path:
        with open(conf_path, 'w') as new_default_vpn_file:
            new_default_vpn_file.write(''.join(new_defaults_vpn_lines))

        with open(user_path, 'w') as user_vpn_file:
            user_vpn_file.write('{}\n{}\n'.format(prof_user, prof_pass))


# noinspection PyClassHasNoInit,PyUnusedLocal
class OpenVpnApi:
    @staticmethod
    def load_unit(unit_uiref):
        """
        :type unit_uiref: str
        :rtype: dict | None
        """
        uiconf = fs.load_json3(_UICONF_PATH, default=None)
        if uiconf is None:
            eventlog.log_e('fail: InternalError')
            raise InternalError

        return uiconf['rw']

    @staticmethod
    def make_profile(unit_uiref, prof_name, conf_path, prof_user=None, prof_pass=None):
        """
        :type unit_uiref: str
        :type prof_name: str
        :type conf_path: str
        :type prof_user: str | None
        :type prof_pass: str | None
        :rtype: dict
        """
        # TODO: validate name is unique
        uiconf = fs.load_json3(_UICONF_PATH, default=None)
        uiconf_pub = uiconf['rw']  # type: dict
        uiconf_conf = uiconf_pub['conf']  # type: dict
        uiconf_profiles = uiconf_conf['profiles']  # type: list

        if len(filter(lambda x: x['name'] == prof_name, uiconf_profiles)):
            eventlog.log_d('fail: ParamsError (name <{}> already exist)'.format(prof_name))
            raise ParamsError

        new_profile_ref = _get_new_ref(uiconf_profiles)

        user_path = _mk_user_path(new_profile_ref)
        # check if auth-user-path and user + pass not None
        _add_auth_to_profile(conf_path, user_path, prof_user, prof_pass)

        new_profile_path = _mk_profile_path(new_profile_ref)
        fs.silent_cp(conf_path, new_profile_path)

        uiconf_profiles.append({
            'ref': new_profile_ref,
            'name': prof_name,
            'internal': False,
            'stats': {
                'active': False
            }
        })

        fs.write_json_file(uiconf, _UICONF_PATH, pretty=True)
        return uiconf_pub

    @staticmethod
    def drop_profile(unit_uiref, profile_ref):
        """
        :type unit_uiref: str
        :type profile_ref: str
        :rtype: dict
        """
        # TODO: check if profile exist
        # TODO: check if profile internal
        profile_path = _mk_profile_path(profile_ref)
        user_path = _mk_user_path(profile_ref)

        fs.silent_rm_r(profile_path)
        fs.silent_rm_r(user_path)

        uiconf = fs.load_json3(_UICONF_PATH, default=None)
        uiconf_pub = uiconf['rw']  # type: dict
        uiconf_conf = uiconf_pub['conf']  # type: dict

        new_profiles = filter(lambda x: x['ref'] != profile_ref, uiconf_conf['profiles'])

        if len(new_profiles) == len(uiconf_conf['profiles']):
            eventlog.log_e('fail: NotFoundError (profile-ref:{})'.format(profile_ref))
            raise NotFoundError

        rm_profile_info = filter(lambda x: x['ref'] == profile_ref, uiconf_conf['profiles'])[0]  # type: dict
        if rm_profile_info['internal']:
            eventlog.log_e('fail: ParamsError (profile-ref:{}, internal:{}'.format(profile_ref, True))
            raise ParamsError

        uiconf_conf['profiles'] = new_profiles

        active_profiles = _get_active_profiles(uiconf)
        _set_autostart_profile(active_profiles)
        _openvpn_reload_restart()

        fs.write_json_file(uiconf, _UICONF_PATH, pretty=True)
        return uiconf_pub

    @staticmethod
    def make_profile_active(unit_uiref, profile_ref, active):
        """
        :type unit_uiref: str
        :type profile_ref: str
        :type active: bool
        :rtype: dict
        """
        uiconf = fs.load_json3(_UICONF_PATH, default=None)
        uiconf_pub = uiconf['rw']  # type: dict
        uiconf_conf = uiconf_pub['conf']  # type: dict

        profiles_info = filter(lambda x: x['ref'] == profile_ref, uiconf_conf['profiles'])
        if len(profiles_info) == 0:
            eventlog.log_e('fail: NotFoundError (profile-ref:{})'.format(profile_ref))
            raise NotFoundError

        for profile_info in profiles_info:  # type: dict
            profile_info['stats']['active'] = active

            active_profiles = _get_active_profiles(uiconf)
            _set_autostart_profile(active_profiles)
            _openvpn_reload_restart()

            fs.write_json_file(uiconf, _UICONF_PATH, pretty=True)
            return uiconf_pub

        eventlog.log_e('fail: NotFoundError (profile-ref:{})'.format(profile_ref))
        raise NotFoundError
