
from edge import eventlog
from edge.com import fs
from edge.exceptions import *
from edge.types import FormatTemplate
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf
from edge.webserver.sysconf.uiconf.rtsp_streaming import RtspStreamingUiconf

import edge_creator as creator


# noinspection PyClassHasNoInit
class RtspStreamingApi:
    @classmethod
    def load_unit(cls, uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_RTSP_STREAMING)

        uiconf = uimod.load_conf(default=None)
        if not uiconf:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        ph.CamDgStats.create(uimod.pid).test_availability()

        uiconf_pub = uiconf['rw']
        try:
            rtsp_url = RtspStreamingUiconf(uiconf_pub).rtsp_url
        except (ValueError, RuntimeError, IOError):
            eventlog.log_e('fail: InternalError')
            raise InternalError

        uiconf_pub['streaming']['url'] = rtsp_url
        return {
            'type': ph.UI_TYPE_RTSP_STREAMING,
            'conf': uiconf_pub
        }

    @classmethod
    def save_unit(cls, uimod_ref, new_uiconf):
        """
        :type uimod_ref: str
        :type new_uiconf: dict
        :rtype: dict
        """
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(ph.UI_TYPE_RTSP_STREAMING)

        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        ph.CamDgStats.create(uimod.pid).test_availability()

        uiconf_pub = uiconf['rw']

        def validate_uimod_pub():
            fs.validate_json(new_uiconf_conf, uimod.schema_path)

            new_streaming = new_uiconf_conf['streaming']
            if new_streaming['transport'] == 'rtp-over-udp-or-tcp':
                if new_streaming['min-udp-port'] == 0 and new_streaming['max-udp-port'] != 0:
                    eventlog.log_d('fail: ParamsError')
                    raise ParamsError

                if new_streaming['max-udp-port'] == 0 and new_streaming['min-udp-port'] != 0:
                    eventlog.log_d('fail: ParamsError')
                    raise ParamsError

            # validate 'format' jnodes from json
            footer_rows = new_uiconf_conf['live']['footer']['rows']
            if cls._validate_format(footer_rows) is False:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError

            old_rtsp_url = RtspStreamingUiconf(uiconf_pub).rtsp_url
            new_rtsp_url = new_uiconf_conf['streaming']['url']
            if old_rtsp_url != new_rtsp_url:
                eventlog.log_d('fail: ParamsError (rtsp-url is read-only)')
                raise ParamsError

        validate_uimod_pub()

        new_uiconf_conf['streaming']['url'] = RtspStreamingUiconf(new_uiconf_conf).rtsp_url

        if uiconf_pub == new_uiconf_conf:
            return new_uiconf_conf

        restart_setupstore = []

        if uiconf['rw']['enabled'] != new_uiconf_conf['enabled']:
            restart_setupstore.append(True)

        uiconf['rw'] = new_uiconf_conf
        uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
        creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})

        if restart_setupstore:
            dg.LifeCycle.sighup_ss_dgmod()

        return {
            'type': ph.UI_TYPE_RTSP_STREAMING,
            'conf': new_uiconf_conf
        }

    @staticmethod
    def _validate_format(jnode):
        """
        :type jnode: list
        :rtype: bool
        """
        key_to_true_map = {
            'datetime': True,
            'datetime-ms': True,
            'post-name': True,
            'geolocation': True,
            'vendor': True,
            'model': True,
            'serial': True
        }

        for row in jnode:
            for col in row['cols']:
                try:
                    FormatTemplate(template=col['format']).substitute(key_to_true_map)
                except KeyError as x:
                    eventlog.log_d('fail: invalid key (%s)' % x.message)
                    return False

        return True
