import os
import subprocess
from tempfile import NamedTemporaryFile

from edge import eventlog
from edge.com import fs
from edge.exceptions import *
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.uimod.sign import SignUimod


ENV = os.environ.copy()
upd = {'OPENSSL_CONF': '/etc/edge/openssl1.0/openssl.cnf'}
ENV.update(upd)


def run_silent(esc_cmd, known_errors=None):
    """
    :param esc_cmd: Escaped shell-command. Use pipes.quote() to escape input form external sources
    :type esc_cmd: str,
    :type known_errors: list[str]
    :rtype: (int, str)| None
    """
    if known_errors is None:
        known_errors = []

    try:
        # eventlog.log_d("exec <%s>" % esc_cmd)
        p = subprocess.Popen([esc_cmd], env=ENV, stderr=subprocess.PIPE, stdout=subprocess.PIPE, shell=True)
        out, err = p.communicate()
        rc = p.returncode

        # eventlog.log_d("<%s> <%s>" % (out, err))
        if len(err) == 0 or err in known_errors:
            return rc, out
        else:
            return None

    except Exception as x:
        eventlog.log_e('fail: <%s>' % x.message)

    return None


def _test_crt_key(crt_filepath, key_filepath, key_password):
    """
    :type crt_filepath: str
    :type key_filepath: str
    :type key_password: str
    :rtype: bool
    """

    valid = False

    for keytype in ['pk8', 'dat', 'pfx', 'pem']:
        if valid:
            break

        key_public_key = 'random1'

        if keytype == 'pk8':
            tmp_key = NamedTemporaryFile()
            reply = run_silent('openssl1.0 pkcs8 -in %s -inform DER -out %s -passin pass:%s' %
                               (key_filepath, tmp_key.name, key_password))
            if reply is None:
                continue

            rc, out = reply
            if rc < 0:
                continue

            reply = run_silent('openssl1.0 pkey -pubout -in %s' % tmp_key.name)
            if reply is None:
                continue

            rc, key_public_key = reply
            if rc < 0:
                continue

        elif keytype == 'dat':
            continue

        elif keytype == 'pfx':
            tmp_key = NamedTemporaryFile()
            reply = run_silent('openssl1.0 pkcs12 -in %s -nocerts -out %s -passin pass:%s -nodes' %
                               (key_filepath, tmp_key.name, key_password),
                               ['MAC verified OK\n'])
            if reply is None:
                continue

            rc, out = reply
            if rc < 0:
                continue

            reply = run_silent('openssl1.0 pkey -pubout -in %s' % tmp_key.name)
            if reply is None:
                continue

            rc, key_public_key = reply
            if rc < 0:
                continue

        elif keytype == 'pem':
            reply = run_silent('openssl1.0 pkey -pubout -in %s' % key_filepath)
            if reply is None:
                continue

            rc, key_public_key = reply
            if rc < 0:
                continue

        for crttype in ['crt', 'cer']:
            crt_public_key = 'random2'

            if crttype == 'crt':
                reply = run_silent('openssl1.0 x509 -pubkey -in %s -noout' % crt_filepath)
                if reply is None:
                    continue

                rc, crt_public_key = reply
                if rc < 0:
                    continue

            elif crttype == 'cer':
                tmp_crt = NamedTemporaryFile()
                reply = run_silent('openssl1.0 x509 -inform DER -in %s -out %s' % (crt_filepath, tmp_crt.name))
                if reply is None:
                    continue

                rc, stdout = reply
                if rc < 0:
                    continue

                reply = run_silent('openssl1.0 x509 -pubkey -in %s -noout' % tmp_crt.name)

                if reply is None:
                    continue

                rc, crt_public_key = reply
                if rc < 0:
                    continue

            if crt_public_key == key_public_key:
                eventlog.log_d("valid for key <%s> certificate <%s>" % (keytype, crttype))
                valid = True
                break

            else:
                eventlog.log_e("fail: private key and certificate have different public keys")
                # eventlog.log_d("key <%s> certificate <%s>" % (key_public_key, crt_public_key))

    return valid


class Result:
    def __init__(self, key_file, crt_file, key_pass):
        """
        :type key_file: file
        :type crt_file: file
        :type key_pass: str
        """
        self.key = key_file
        self.crt = crt_file
        self.key_pass = key_pass


def check2(crt_filepath, key_filepath, key_password):
    """
    :type crt_filepath: str
    :type key_filepath: str
    :type key_password: str
    :rtype: Result | None
    """

    pem_key_file = NamedTemporaryFile(prefix='tmpkey', suffix='.key', dir='/tmp/edge')
    cer_certificate_file = NamedTemporaryFile(prefix='tmpcrt', suffix='.crt', dir='/tmp/edge')

    for keytype in ['pk8', 'dat', 'pfx', 'pem']:

        key_public_key = 'random1'

        if keytype == 'pk8':
            tmp_key = pem_key_file.name
            reply = run_silent('openssl1.0 pkcs8 -in %s -inform DER -out %s -passin pass:%s' %
                               (key_filepath, tmp_key, key_password))
            if reply is None:
                continue

            rc, out = reply
            if rc < 0:
                continue

            reply = run_silent('openssl1.0 pkey -pubout -in %s' % tmp_key)
            if reply is None:
                continue

            rc, key_public_key = reply
            if rc < 0:
                continue

        elif keytype == 'dat':
            continue

        elif keytype == 'pfx':
            tmp_key = pem_key_file.name
            reply = run_silent('openssl1.0 pkcs12 -in %s -nocerts -out %s -passin pass:%s -nodes' %
                               (key_filepath, tmp_key, key_password),
                               ['MAC verified OK\n'])
            if reply is None:
                continue

            rc, out = reply
            if rc < 0:
                continue

            reply = run_silent('openssl1.0 pkey -pubout -in %s' % tmp_key)
            if reply is None:
                continue

            rc, key_public_key = reply
            if rc < 0:
                continue

            fs.silent_cp(key_filepath, pem_key_file.name)

        elif keytype == 'pem':
            reply = run_silent('openssl1.0 pkey -pubout -in %s' % key_filepath)
            if reply is None:
                continue

            rc, key_public_key = reply
            if rc < 0:
                continue

            fs.silent_cp(key_filepath, pem_key_file.name)

        for crttype in ['crt', 'cer']:
            crt_public_key = 'random2'

            if crttype == 'crt':
                reply = run_silent('openssl1.0 x509 -pubkey -in %s -noout' % crt_filepath)
                if reply is None:
                    continue

                rc, crt_public_key = reply
                if rc < 0:
                    continue

                fs.silent_cp(crt_filepath, cer_certificate_file.name)

            elif crttype == 'cer':
                tmp_crt = cer_certificate_file.name
                reply = run_silent('openssl1.0 x509 -inform DER -in %s -out %s' % (crt_filepath, tmp_crt))
                if reply is None:
                    continue

                rc, stdout = reply
                if rc < 0:
                    continue

                reply = run_silent('openssl1.0 x509 -pubkey -in %s -noout' % tmp_crt)

                if reply is None:
                    continue

                rc, crt_public_key = reply
                if rc < 0:
                    continue

            if crt_public_key == key_public_key:
                eventlog.log_d("valid for key <%s> certificate <%s>" % (keytype, crttype))
                return Result(pem_key_file, cer_certificate_file, key_password)

            else:
                eventlog.log_e("fail: private key and certificate have different public keys")
                eventlog.log_d("key <%s> certificate <%s>" % (key_public_key, crt_public_key))

    return None


def _get_limits_jnode(ui_jnode):
    """
    :type ui_jnode: dict
    :rtype: dict
    """
    ro_jnode = ui_jnode.get('ro')

    crt_file = ro_jnode.get('crt') or False
    key_file = ro_jnode.get('key') or False

    if crt_file and key_file:
        return {'activatable': True}

    return {'activatable': False}


def load_unit():
    """
    :rtype: dict
    """
    SignUimod.install()
    uiconf = SignUimod.load_conf()

    ext_uiconf_pub = uiconf['rw']
    reply = {
        'type': ph.UI_TYPE_SIGN,
        'conf': ext_uiconf_pub['args'],
        'limits': _get_limits_jnode(uiconf)
    }
    return reply


def save_unit(new_uiconf):
    """
    :type new_uiconf: dict
    :rtype: dict
    """
    new_uiconf_conf = new_uiconf.get('conf')

    uiconf = SignUimod.load_conf()
    uiconf_pub = uiconf.get('rw')
    uiconf_pub['args']['active'] = new_uiconf_conf['active']
    SignUimod.save_conf(uiconf)

    # regenerate speedcontrol configs
    from edge.webserver.sysconf.api_traffic_rules import ApiTrafficRules

    root_info = ph.RootInfoUiconf.get_info()
    uimods = ph.InfoUiconf.map_units_r(
        root_info, lambda x: ph.UiMod.create(x['ref']),
        unit_types=[ph.UI_TYPE_TRAFFIC_RULES])

    for uimod in uimods:
        ApiTrafficRules.save_unit(uimod.ref, dict())

    reply = {
        'type': ph.UI_TYPE_SIGN,
        'conf': new_uiconf_conf,
        'limits': _get_limits_jnode(uiconf)
    }
    return reply


def install_crt_key(crt_path, key_path, key_pass):
    """
    :type crt_path: str
    :type key_path: str
    :type key_pass: str | None
    :rtype: dict
    """
    eventlog.log_d('exec: put crt %s key %s pass %s' % (crt_path, key_path, key_pass))
    if not _test_crt_key(crt_path, key_path, key_pass):
        eventlog.log_e('fail: ParamsError (key and cert match)')
        raise ParamsError

    uiconf = SignUimod.install_key_crt(crt_path, key_path, key_pass)
    dg.LifeCycle.restart_dgmod(dg.EDGE_TYPE_SPEEDCONTROL)

    ext_uiconf_pub = uiconf.get('rw')
    reply = {
        'type': ph.UI_TYPE_SIGN,
        'conf': ext_uiconf_pub['args'],
        'limits': _get_limits_jnode(uiconf)
    }
    return reply
