from edge import eventlog
from edge.com import fs
from edge.exceptions import *
from edge.webserver.sysconf import ph
from edge.types import FormatTemplate

import edge.webserver.sysconf.edge_creator as creator


# noinspection PyClassHasNoInit
class SpotSpeedApi:
    UIMOD_TYPE = ph.UI_TYPE_SPOT_SPEED

    @classmethod
    def make_detector(cls, term_pid, factory_conf):
        """
        :type term_pid: str
        :type factory_conf: dict
        :rtype: (str, dict)
        """
        if not ph.RootInfoUiconf.validate_term(term_pid):
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        det_uimod_pid = term_pid
        det_uimod_type = factory_conf.get('type')
        det_name = factory_conf.get('name')

        if det_uimod_type != cls.UIMOD_TYPE or det_name is None:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        answer = ph.RootInfoUiconf.add_detector(det_uimod_pid, det_uimod_type, det_name)
        if answer is None:
            eventlog.log_e('fail: InternalError')
            raise InternalError

        det_uimod_id, root_info = answer

        uimod = ph.UiMod(det_uimod_pid, det_uimod_type, det_uimod_id)
        uimod.install()

        uiconf = uimod.load_conf()
        uiconf['rw']['args']['name'] = det_name
        uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
        creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})
        return uimod.ref, root_info

    @classmethod
    def drop_detector(cls, uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(cls.UIMOD_TYPE)
        if uimod.load_conf(default=None) is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        root_info, root_links = ph.RootInfoUiconf.get_info_links()
        _, parent_info = ph.RootInfoUiconf.find_unit_parent_r(root_info, uimod_ref)
        parent_uimod = ph.UiMod.create(parent_info['ref'])
        parent_uiconf = parent_uimod.load_conf(default=True)

        uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref, root_links)
        uirefs.remove(uimod.ref)

        root_info = ph.RootInfoUiconf.del_unit_r(uimod.ref)
        creator.generate(uimod.create_backup_dirpath(), uirefs, {parent_uimod: parent_uiconf})
        uimod.uninstall()

        ext_root_info = ph.RootInfoUiconf.get_ext_root_info(root_info)
        reply = {
            'info': [ext_root_info]
        }
        return reply

    @classmethod
    def load_detector(cls, uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(cls.UIMOD_TYPE)

        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        uiconf_pub = uiconf['rw']
        reply = {
            'type': cls.UIMOD_TYPE,
            'conf': uiconf_pub['args']
        }
        reply.update(ph.get_detector_limits_stats(uimod.pid, lanes=True, classes=True))
        return reply

    @staticmethod
    def _validate_format(jnode):
        """
        :type jnode: list
        :rtype: bool
        """
        key_to_true_map = {
            'speed': True,
            'datetime': True,
            'datetime-ms': True,
            'post-name': True,
            'geolocation': True,
            'plate': True,
            'lane': True,
            'dir': True,
            'video-speed': True,
            'video-speed2': True,
            'radar-speed': True,
            'radar-speed2': True,
            'max-speed': True,
            'speed-limit': True,
            'radar-overspeed': True,
            'vendor': True,
            'model': True,
            'serial': True,
            'main-speed': True,
            'control-speed': True,
            'main-overspeed': True,
            'wim-speed': True, # TODO: move to 2-nd class ?
            'wim-overspeed': True
        }

        for row in jnode:
            for col in row['cols']:
                try:
                    FormatTemplate(template=col['format']).substitute(key_to_true_map)
                except KeyError as x:
                    eventlog.log_d('fail: invalid key (%s)' % x.message)
                    return False

        return True

    @classmethod
    def save_detector(cls, uimod_ref, new_uiconf):
        """
        :type uimod_ref: str
        :type new_uiconf: dict
        :rtype: dict
        """
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(cls.UIMOD_TYPE)

        if uimod.load_conf(default=None) is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        def validate_body():
            fs.validate_json({'args': new_uiconf_conf}, uimod.schema_path)

            # validate 'format' jnodes from json
            entry_images = new_uiconf_conf['images']['violation']

            entry_v_rows = entry_images['vehicle']['footer']['rows']
            if cls._validate_format(entry_v_rows) is False:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError

            entry_s_rows = entry_images['scene']['footer']['rows']
            if cls._validate_format(entry_s_rows) is False:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError

        validate_body()

        uiconf = uimod.load_conf()
        if uiconf['rw']['args'] != new_uiconf_conf:
            if new_uiconf_conf['name'] != uiconf['rw']['args']['name']:
                ph.RootInfoUiconf.set_unit_name(uimod.ref, new_uiconf_conf['name'])

            uiconf['rw']['args'] = new_uiconf_conf
            uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
            creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})

        reply = {
            'type': cls.UIMOD_TYPE,
            'conf': new_uiconf_conf
        }
        reply.update(ph.get_detector_limits_stats(uimod.pid, lanes=True, classes=True))
        return reply


# noinspection PyClassHasNoInit
class SpotSpeedWagaApi(SpotSpeedApi):
    UIMOD_TYPE = ph.UI_TYPE_SPOT_SPEED_WAGA
