import uuid

from edge import eventlog
from edge.exceptions import *
from edge.com import fs
from edge.exceptions import FsError
from edge.webserver.sysconf import ph, edge_creator as creator
from edge.webserver.sysconf.api_license import LicenseApi
from edge.webserver.sysconf.api_tcp_recipient import TcpRecipientApi
from edge.webserver.sysconf.api_leddar import LeddarApi
from edge.webserver.sysconf.dgconf.setupstore import SsDgconf
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf
from edge.webserver.sysconf.uiconf.time import TimeUiconf


# noinspection PyClassHasNoInit
class TerminalApi:
    @classmethod
    def _create_term(cls, ctrl_uiref, factory_conf, factory_info, factory_links, check_lpr_license=True):
        """
        :type ctrl_uiref: str
        :type factory_conf: dict
        :type factory_info: dict
        :type factory_links: list
        :type check_lpr_license: bool
        :rtype: str dict
        """
        from edge.com import db

        if check_lpr_license:
            LicenseApi.test_lpr_license_available()

        internal = factory_conf.get('internal', False)
        root_info, term_info = ph.RootInfoUiconf.add_term(ctrl_uiref, factory_info, factory_links, internal)

        term_ref = term_info['ref']
        term_pid = ph.UiMod.create(term_ref).pid
        term_uuid = factory_conf.get('uuid') or str(uuid.uuid4())

        factory_conf['term-pid'] = term_pid
        factory_conf['term-uuid'] = term_uuid

        def install_unit_fn(unit_info):
            """
            :type unit_info: dict
            :rtype: None
            """
            unit_uimod = ph.UiMod.create(unit_info['ref'])
            unit_uimod.install(factory_conf)

        # noinspection PyBroadException
        try:
            # install uimods
            TimeUiconf.install()
            factory_uimod = ph.UiMod(term_pid, ph.UI_TYPE_FACTORY, '0')
            ph.InfoUiconf.map_units_r(term_info, install_unit_fn)
            ph.TermAboutUiconf.save_conf(term_pid, term_uuid)

            # install dgmods
            uirefs = ph.InfoUiconf.find_unit_refs_r(term_info)
            uirefs.append('webserver.0')  # edit ws conf while addind pipeline
            all_used_dgmods = creator.generate(
                factory_uimod.create_backup_dirpath(),
                uirefs, forced=True)

            all_used_dgmods = filter(lambda x: x.type not in [dg.EDGE_TYPE_WEBSERVER, dg.EDGE_TYPE_EVENTSTORE],
                                     all_used_dgmods)

            SsDgconf.append_ss_group(factory_uimod.pid, map(lambda x: x.conf_dir, all_used_dgmods))
            db.edit_term(term_uuid, term_name=term_pid, term_place=term_pid, manual=True)

            dg.LifeCycle.restart_ws_dgmod()
            dg.LifeCycle.sighup_ss_dgmod()

            ext_root_info = ph.RootInfoUiconf.get_ext_root_info(root_info)
            return term_ref, ext_root_info

        except NotImplementedError:
            eventlog.log_d('fail: ParamsError (unsupported camera)')
            raise ParamsError

        except BusyError:
            raise

        except Exception:
            eventlog.log_x()
            TerminalApi.drop_terminal2(term_pid)
            raise InternalError

    @classmethod
    def create_radar_term(cls, ctrl_uiref, factory_conf):
        """
        :type ctrl_uiref: str
        :type factory_conf: dict
        :rtype: dict
        """
        factory_uiconf_path = fs.join_path(ph.CTRL_DEF_UICONF_DIR, 'radar-terminal-info.json')
        factory_info, factory_links = ph.InfoUiconf.get_info_links(factory_uiconf_path)
        term_ref, root_info = cls._create_term(ctrl_uiref, factory_conf or dict(), factory_info, factory_links, False)
        return {'info': [root_info], 'created': [term_ref]}

    @classmethod
    def create_mv_term(cls, ctrl_uiref, factory_conf):
        """
        :type ctrl_uiref: str
        :type factory_conf: dict
        :rtype: dict
        """
        factory_uiconf_path = fs.join_path(ph.CTRL_DEF_UICONF_DIR, 'mv-terminal-info.json')
        factory_info, factory_links = ph.InfoUiconf.get_info_links(factory_uiconf_path)
        term_ref, root_info = cls._create_term(ctrl_uiref, factory_conf or dict(), factory_info, factory_links)
        return {'info': [root_info], 'created': [term_ref]}

    @classmethod
    def create_rtsp_term(cls, ctrl_uiref, factory_conf):
        """
        :type ctrl_uiref: str
        :type factory_conf: dict | None
        :rtype: dict
        """
        factory_uiconf_path = fs.join_path(ph.CTRL_DEF_UICONF_DIR, 'rtsp-terminal-info.json')
        factory_info, factory_links = ph.InfoUiconf.get_info_links(factory_uiconf_path)
        term_ref, root_info = cls._create_term(ctrl_uiref, factory_conf or dict(), factory_info, factory_links)
        return {'info': [root_info], 'created': [term_ref]}

    @classmethod
    def create_waga_rtsp_term(cls, ctrl_uiref, factory_conf):
        """
        :type ctrl_uiref: str
        :type factory_conf: dict | None
        :rtype: dict
        """
        factory_uiconf_path = fs.join_path(ph.CTRL_DEF_UICONF_DIR, 'rtsp-waga-terminal-info.json')
        factory_info, factory_links = ph.InfoUiconf.get_info_links(factory_uiconf_path)
        term_ref, root_info = cls._create_term(ctrl_uiref, factory_conf or dict(), factory_info, factory_links)
        return {'info': [root_info], 'created': [term_ref]}

    @staticmethod
    def make_unit(term_uiref, factory_conf):
        """
        :type term_uiref: str
        :type factory_conf: dict
        :rtype: dict
        """
        uimod_type = factory_conf.get('type')
        if uimod_type is None:
            eventlog.log_d('fail: ParamsError (type)')
            raise ParamsError

        from .api_parking import ParkingApi
        from .api_wrong_way import WrongWayApi
        from .api_bus_lane import BusLaneApi
        from .api_red_light import RedLightApi
        from .api_wanted import WantedApi
        from .api_section_speed import SectionSpeedApi
        from .api_spot_speed import SpotSpeedApi
        from .api_railroad_crossing import RailroadCrossingApi

        uitype_to_factory_map = {
            ph.UI_TYPE_PARKING: ParkingApi.make_detector,
            ph.UI_TYPE_RED_LIGHT: RedLightApi.make_detector,
            ph.UI_TYPE_SECTION_SPEED: SectionSpeedApi.make_detector,
            ph.UI_TYPE_SPOT_SPEED: SpotSpeedApi.make_detector,
            ph.UI_TYPE_WANTED: WantedApi.make_detector,
            ph.UI_TYPE_WRONG_WAY: WrongWayApi.make_detector,
            ph.UI_TYPE_BUS_LANE: BusLaneApi.make_detector,
            ph.UI_TYPE_TCP_RECIPIENT: TcpRecipientApi.make_service,
            ph.UI_TYPE_LEDDAR: LeddarApi.make_service,
            ph.UI_TYPE_RAILROAD_CROSSING: RailroadCrossingApi.make_detector
        }

        factory_fn = uitype_to_factory_map.get(uimod_type)
        if not factory_fn:
            eventlog.log_d('fail: ParamsError (type)')
            raise ParamsError

        term_uimod = ph.UiMod.create(term_uiref)
        unit_uiref, terms_info = factory_fn(term_uimod.pid, factory_conf)
        ext_terms_info = ph.RootInfoUiconf.get_ext_root_info(terms_info)

        reply = {
            'info': [ext_terms_info],
            'created': [unit_uiref]
        }
        return reply

    @staticmethod
    def get_factory_limits(term_uiref, roles):
        """
        :type term_uiref: str
        :type roles: list(str)
        :rtype: dict
        """
        uimod = ph.UiMod.create(term_uiref)
        uimod.assert_type(ph.UI_TYPE_TERMINAL)

        service_list = [
            {'role': ph.UI_ROLE_SERVICE, 'type': ph.UI_TYPE_TCP_RECIPIENT},
            {'role': ph.UI_ROLE_SERVICE, 'type': ph.UI_TYPE_LEDDAR}
        ]

        detector_list = [
            {'role': ph.UI_ROLE_DETECTOR, 'type': ph.UI_TYPE_RED_LIGHT},
            {'role': ph.UI_ROLE_DETECTOR, 'type': ph.UI_TYPE_RAILROAD_CROSSING},
            {'role': ph.UI_ROLE_DETECTOR, 'type': ph.UI_TYPE_WANTED},
            {'role': ph.UI_ROLE_DETECTOR, 'type': ph.UI_TYPE_WRONG_WAY},
            {'role': ph.UI_ROLE_DETECTOR, 'type': ph.UI_TYPE_BUS_LANE},
            {'role': ph.UI_ROLE_DETECTOR, 'type': ph.UI_TYPE_SECTION_SPEED},
            {'role': ph.UI_ROLE_DETECTOR, 'type': ph.UI_TYPE_PARKING}
        ]

        if ph.RootInfoUiconf.get_raterm_pid() == uimod.pid:
            detector_list.append(
                {'role': ph.UI_ROLE_DETECTOR, 'type': ph.UI_TYPE_SPOT_SPEED})

        if not roles:
            roles = [ph.UI_ROLE_DETECTOR, ph.UI_ROLE_SERVICE]

        reply = []
        if ph.UI_ROLE_DETECTOR in roles:
            reply.extend(detector_list)

        if ph.UI_ROLE_SERVICE in roles:
            reply.extend(service_list)

        return reply

    @staticmethod
    def load_terminal(term_uiref):
        """
        :type term_uiref: str
        :rtype: dict
        """
        uimod = ph.UiMod.create(term_uiref)
        uimod.assert_type(ph.UI_TYPE_TERMINAL)

        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        uiconf_pri = uiconf['ro']
        ext_uiconf_pub = uiconf['rw']
        ext_uiconf_pub['origin-guid'] = uiconf_pri['uuid']
        reply = {
            'type': ph.UI_TYPE_TERMINAL,
            'conf': ext_uiconf_pub
        }
        return reply

    @staticmethod
    def save_terminal(term_uiref, new_uiconf):
        """
        :type term_uiref: str
        :type new_uiconf: dict
        :rtype: dict
        """
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod.create(term_uiref)
        uimod.assert_type(ph.UI_TYPE_TERMINAL)

        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        fs.validate_json(new_uiconf_conf, uimod.schema_path)

        uiconf_pri = uiconf['ro']  # type: dict
        uiconf_pub = uiconf['rw']  # type: dict

        if uiconf_pri['uuid'] != new_uiconf_conf['origin-guid']:
            eventlog.log_d('fail: ParamsError (#/origin-guid is read-only)')
            raise ParamsError

        new_origin_name = new_uiconf_conf['origin-name']
        new_origin_place = new_uiconf_conf['origin-place']

        if uiconf_pub['origin-name'] != new_origin_name or uiconf_pub['origin-place'] != new_origin_place:
            term_uuid = TerminalUiconf(uimod.conf_path).get_uuid()

            from edge.com import db
            db.edit_term(term_uuid, new_origin_name, new_origin_place, manual=True)

            ph.TermAboutUiconf.save_conf(uimod.pid, term_uuid, term_place=new_origin_place, term_name=new_origin_name)

            uiconf_pub['origin-name'] = new_origin_name
            uiconf_pub['origin-place'] = new_origin_place

            uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
            creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod: uiconf})

        reply = {
            'type': ph.UI_TYPE_TERMINAL,
            'conf': new_uiconf_conf
        }
        return reply

    @classmethod
    def drop_terminal(cls, term_uiref):
        """
        :type term_uiref: str
        :rtype: dict
        """
        root_info = ph.RootInfoUiconf.get_info()
        term_info = ph.InfoUiconf.find_unit_r(root_info, term_uiref)
        if term_info is None:
            eventlog.log_e('fail: ParamsError')
            raise ParamsError

        elif term_info.get('internal', True):
            eventlog.log_d('fail: ParamsError (terminal is internal)')
            raise ParamsError

        uimod = ph.UiMod.create(term_uiref)
        uimod.assert_type(ph.UI_TYPE_TERMINAL)

        ext_terms_info = cls.drop_terminal2(uimod.pid)

        # restarting edgeserver with PCD
        eventlog.log_i('exec: restarting modules')
        dg.LifeCycle.sighup_ss_dgmod()
        dg.LifeCycle.restart_ws_dgmod()

        return ext_terms_info

    @classmethod
    def drop_terminal2(cls, term_pid):
        """
        :type term_pid: str
        :rtype: dict
        """
        from edge.com import db
        from edge.webserver.sysconf.dgmod.eventstore import EsDeleter
        from edge.webserver.sysconf.dgmod.grabcontrol import GcDeleter
        from edge.webserver.sysconf.dgmod.lpr import LrDeleter
        from edge.webserver.sysconf.dgmod.rawcodec import RcDeleter
        from edge.webserver.sysconf.dgmod.mediagrabber import MgDeleter
        from edge.webserver.sysconf.dgmod.radar import RaDeleter
        from edge.webserver.sysconf.dgmod.renderer import OrDeleter
        from edge.webserver.sysconf.dgmod.rtspgrabber import RgDeleter
        from edge.webserver.sysconf.dgmod.rtspserver import RsDeleter
        from edge.webserver.sysconf.dgmod.speedcontrol import ScDeleter
        from edge.webserver.sysconf.dgmod.tracker import TrDeleter
        from edge.webserver.sysconf.dgmod.webserver import WsDeleter
        from edge.webserver.sysconf.dgmod.mediacodecgst import McDeleter
        from edge.webserver.sysconf.dgmod.blobstore import VideoBsDeleter
        from edge.webserver.sysconf.dgmod.trafficlight import TlDeleter
        from edge.webserver.sysconf.dgmod.lbsapp import LbsDeleter

        fn_list = [
            MgDeleter.create, GcDeleter.create, RgDeleter.create,
            OrDeleter.create, RsDeleter.create, RaDeleter.create,
            LrDeleter.create, TrDeleter.create, ScDeleter.create,
            EsDeleter.create, WsDeleter.create, RcDeleter.create,
            McDeleter.create, VideoBsDeleter.create,
            TlDeleter.create, LbsDeleter.create
        ]

        factory_uimod = ph.UiMod(term_pid, ph.UI_TYPE_FACTORY, '0') # TODO
        entry_uiref = ph.UiMod.create_term(term_pid).ref
        root_info = ph.RootInfoUiconf.get_info()
        term_info = ph.InfoUiconf.find_unit_r(root_info, unit_ref=entry_uiref)
        if term_info is None:
            eventlog.log_i('fail find entry_uiref {}'.format(entry_uiref))
            raise NotImplementedError

        entry_uirefs = ph.InfoUiconf.find_unit_refs_r(term_info)

        all_used_deleters = filter(bool, map(lambda fn: fn(factory_uimod), fn_list))
        for uiref in entry_uirefs:
            uimod = ph.RootInfoUiconf.get_uimods({uiref}, root_info).pop()  # TODO
            all_used_deleters += filter(bool, map(lambda fn: fn(uimod), fn_list))

        map(lambda deleter: deleter.run(), all_used_deleters)

        # rm configs of ui-modules
        term_uimod = ph.UiMod.create_term(term_pid)

        shemas_dirpath = fs.join_path(ph.TERM_UISCHEMA_DIR, term_pid)
        ui_dirpath = fs.join_path(ph.TERM_UICONF_DIR, term_pid)
        map(fs.silent_rm_r, [shemas_dirpath, ui_dirpath])

        terms_info = ph.RootInfoUiconf.del_unit_r(term_uimod.ref)
        ext_terms_info = ph.RootInfoUiconf.get_ext_root_info(terms_info)

        from edge.webserver.sysconf.dgconf.setupstore import SsDgconf
        SsDgconf.remove_ss_group(factory_uimod.pid)

        reply = {
            'info': [ext_terms_info]
        }
        return reply

    @staticmethod
    def get_ra_type(term_pid):
        """
        :rtype: str | None
        """
        location_uimod = ph.UiMod(term_pid, ph.UI_TYPE_LOCATION, '0')
        try:
            conf = location_uimod.load_conf()
            conf_ro = conf.get('ro') or dict()
            return conf_ro.get('type')

        except FsError:
            pass


def _main():
    # ph.RootInfoUiconf._CONF_PATH = '/home/user/Documents/BitBucket/edgeserver/_share/edge-python-tools/edge-python-tools/share/misc/root.json'
    # factory_uiconf = fs.load_json3('/home/user/Documents/BitBucket/edgeserver/_share/edge-python-tools/edge-python-tools/share/misc/mv_term.json')

    ph.RootInfoUiconf._CONF_PATH = (
        '/Users/d.feskov/CLionProjects/edgeserver/'
        '_share/edge-python-tools/edge-python-tools/share/misc/root.json')
    factory_uiconf = fs.load_json3(
        '/Users/d.feskov/CLionProjects/edgeserver/'
        '_share/edge-python-tools/edge-python-tools/share/misc/mv_term.json')  # type: dict

    factory_info = factory_uiconf['pub']['info']
    factory_links = factory_uiconf['pri']['links']
    TerminalApi._create_term('controller.0', dict(), factory_info, factory_links)


if __name__ == '__main__':
    _main()
