from edge import eventlog
from edge.com import fs, net, sh
from edge.exceptions import InternalError, ParamsError
from edge.webserver.backup import CgiBackup
from edge.webserver.file_lock import CgiLock
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.uiconf.time import TimeUiconf


# noinspection PyClassHasNoInit
class _NtpDgconf:
    CONF_PATH = '/etc/ntp.conf'

    @classmethod
    def set_autosync(cls, time_uiconf_pri, next_ntp_source, next_gnss_source):
        """
        :type time_uiconf_pri: dict
        :type next_ntp_source: dict
        :type next_gnss_source: dict
        :rtype: bool
        """
        ntp_lines = []

        autosync = False
        if next_ntp_source.get('autosync', False):
            autosync = True
            for server_conf in next_ntp_source['servers']:
                address = server_conf['address']
                if server_conf['prefered']:
                    ntp_lines.append('server {} prefer'.format(address))
                else:
                    ntp_lines.append('server {}'.format(address))

            for server_conf in next_ntp_source['servers']:
                ntp_lines.append('restrict {} noquery notrap'.format(server_conf['address']))

        if next_gnss_source.get('autosync', False):
            autosync = True
            gnss_conf_lines = time_uiconf_pri.get('ntp-conf-gnss-vit', [])
            if not gnss_conf_lines:
                gnss_conf_lines = time_uiconf_pri.get('ntp-conf-gnss')
                if gnss_conf_lines is None:
                    eventlog.log_e_param('ui#/ro/ntp-conf-gnss')
                    raise InternalError

            ntp_lines.extend(gnss_conf_lines)

        ntp_lines.extend(time_uiconf_pri.get('ntp-conf-footer', []))

        with open(cls.CONF_PATH, 'w') as text_file:
            text_file.write('\n'.join(ntp_lines))

        return autosync


# noinspection PyClassHasNoInit
class _NtpdateDgconf:
    CONF_PATH = '/etc/default/ntpdate'

    @classmethod
    def set_autosnc(cls, ntp_autosync):
        """
        :type ntp_autosync: bool
        """
        with open(cls.CONF_PATH, 'r') as text_file:
            last_lines = text_file.readlines()

        next_lines = filter(lambda line: not line.startswith('exit 0'), last_lines)
        if ntp_autosync:
            # do sync time using ntpdate on system startup
            pass
        else:
            # do not sync time using ntpdate on system startup
            next_lines.insert(0, 'exit 0\n')

        with open(cls.CONF_PATH, 'w') as text_file:
            text_file.write(''.join(next_lines))


# noinspection PyClassHasNoInit
class TimeApi:
    _GNSS_SOURCE_LIMITS = {
        'type': 'gnss',
        'sync-auto': True,
        'sync-manual': False
    }

    _GNSS_SOURCE_CONF = {
        'type': 'gnss',
        'autosync': False
    }

    @classmethod
    def load_unit(cls):
        """
        :rtype: dict
        """
        TimeUiconf.install()

        uiconf = TimeUiconf.load_conf()
        uiconf_pub = uiconf['rw']  # type: dict

        ext_uiconf_pub = uiconf_pub
        ext_sources = ext_uiconf_pub['args']['sources']

        # patch config if GNSS device detected
        if len(filter(lambda x: x['type'] == 'gnss', ext_sources)) == 0:
            if net.query_gps_count() > 0:
                ext_sources.append(cls._GNSS_SOURCE_CONF)

        if len(filter(lambda x: x['type'] == 'gnss', ext_uiconf_pub['limits']['sources'])) == 0:
            ext_uiconf_pub['limits']['sources'].append(cls._GNSS_SOURCE_LIMITS)

        TimeUiconf.save_conf(uiconf)

        unit_ext_jnode = {
            'type': ph.UI_TYPE_TIME,
            'conf': ext_uiconf_pub['args'],
            'limits': ext_uiconf_pub['limits']
        }

        return unit_ext_jnode

    @classmethod
    @CgiLock('/var/run/edge/datetime')
    @CgiBackup('/etc/edge/setupstore/backup/datetime', {
        'time.json': TimeUiconf.CONF_PATH,
        'ntp.conf': _NtpDgconf.CONF_PATH,
        'ntpdate': _NtpdateDgconf.CONF_PATH
    })
    def save_unit(cls, new_uiconf_pub):
        """
        :type new_uiconf_pub: dict
        :rtype: dict
        """
        fs.validate_json({'args': new_uiconf_pub['conf']}, TimeUiconf.SCHEMA_PATH)  # TODO
        new_uiconf_conf = new_uiconf_pub['conf']

        uiconf = TimeUiconf.load_conf()
        uiconf_pub = uiconf['rw']  # type: dict
        if new_uiconf_conf['tz'] not in uiconf_pub['limits']['tz']:
            eventlog.log_d('fail: ParamsError (#/args/tz)')
            raise ParamsError

        def get_gnss_sync_source(sync_sources):
            """
            :param sync_sources:
            :rtype: dict
            """
            gnss_sources = filter(lambda source: source['type'] == 'gnss', sync_sources)
            gnss_sources_count = len(gnss_sources)
            if gnss_sources_count == 1:
                return gnss_sources[0]

            elif gnss_sources_count == 0:
                return {
                    'autosync': False,
                    'type': 'gnss'
                }

            else:
                eventlog.log_d('fail: ParamsError (#/args/sources)')
                raise ParamsError

        def get_ntp_sync_source(sync_sources):
            """
            :type sync_sources: list[dict]
            :rtype: dict
            """
            ntp_sources = filter(lambda source: source['type'] == 'ntp', sync_sources)
            ntp_sources_count = len(ntp_sources)
            if ntp_sources_count == 1:
                return ntp_sources[0]

            elif ntp_sources_count == 0:
                return {
                    'autosync': False,
                    'type': 'ntp',
                    'servers': []
                }

            else:
                eventlog.log_d('fail: ParamsError (#/args/sources)')
                raise ParamsError

        next_gnss_source = get_gnss_sync_source(new_uiconf_conf.get('sources', []))
        next_ntp_source = get_ntp_sync_source(new_uiconf_conf.get('sources', []))

        _NtpdateDgconf.set_autosnc(next_ntp_source.get('autosync', False))
        autosync = _NtpDgconf.set_autosync(uiconf['ro'], next_ntp_source, next_gnss_source)

        # restart ntpd
        sh.run_silent('service ntp stop')
        sh.run_silent('update-rc.d ntp disable')
        if autosync:
            sh.run_silent('update-rc.d ntp enable')
            sh.run_silent('service ntp start')

        # if tz changed restart
        if uiconf_pub['args']['tz'] != new_uiconf_conf['tz']:
            eventlog.log_i('exec: restart overlayrender, speedcontrol, webserver')
            dg.LifeCycle.restart_dgmod(dg.EDGE_TYPE_RENDERER)
            dg.LifeCycle.restart_dgmod(dg.EDGE_TYPE_SPEEDCONTROL)
            dg.LifeCycle.restart_ws_dgmod()
            # TODO: restart edgeserver if next_time < current_time

        # write new jnode to disk
        uiconf_pub['args'] = new_uiconf_conf
        TimeUiconf.save_conf(uiconf)

        return cls.load_unit()  # TODO: strip limits from uiconf

    @classmethod
    def sync_manual(cls, sync_source):
        """
        :type sync_source: dict
        :rtype: dict
        """
        import datetime

        fs.validate_json(sync_source, TimeUiconf.SCHEMA_MANUAL_PATH)
        source_type = sync_source.get('type')
        if source_type == 'ntp':
            cls._sync_manual_ntp(sync_source)
            return {
                'type': 'ntp',
                'time': datetime.datetime.now().isoformat()
            }

        elif source_type == 'user':
            cls._sync_manual_user(sync_source)
            return {
                'type': 'user',
                'time': datetime.datetime.now().isoformat()
            }

        else:
            eventlog.log_i('fail: ParamsError (#/type)')
            raise ParamsError

    @staticmethod
    def _sync_manual_ntp(sync_source):
        """
        :type sync_source: dict
        :rtype: None
        """
        # set OS time
        servers = map(lambda server: sh.quote(server['address']), sync_source.get('servers', []))
        if not servers:
            eventlog.log_e('fail: InternalError')
            raise InternalError

        sh_cmd = 'ntpdate -u {}'.format(' '.join(servers))
        answer = sh.run_silent(sh_cmd)
        if answer == 0:
            eventlog.log_i('done: sh=<{}>'.format(sh_cmd))
        else:
            eventlog.log_e('fail: sh=<{}> rc=<{}>'.format(sh_cmd, answer))
            raise InternalError

        # sync BIOS time to OS time
        sh_cmd = 'hwclock -w'
        answer = sh.run_silent(sh_cmd)
        if answer == 0:
            eventlog.log_i('done: sh=<{}>'.format(sh_cmd))
        else:
            eventlog.log_e('fail: sh=<{}> rc=<{}>'.format(sh_cmd, answer))
            raise InternalError

    @staticmethod
    def _sync_manual_user(sync_source):
        """
        :type sync_source: dict
        :rtype: None
        """
        import iso8601
        tz_datetime = iso8601.parse_date(sync_source['time'])
        utc_datetime = tz_datetime - tz_datetime.utcoffset()

        # set OS time
        sh_cmd = 'date --set="{:%Y-%m-%d %H:%M:%S}"'.format(utc_datetime)
        answer = sh.run_silent(sh_cmd)
        if answer == 0:
            eventlog.log_i('done: sh=<{}>'.format(sh_cmd))
        else:
            eventlog.log_e('fail: http500. sh=<{}> rc=<{}>'.format(sh_cmd, answer))
            raise InternalError

        # sync BIOS time to OS time
        sh_cmd = 'hwclock -w'
        answer = sh.run_silent(sh_cmd)
        if answer == 0:
            eventlog.log_i('done: sh=<{}>'.format(sh_cmd))
        else:
            eventlog.log_e('fail: http500. sh=<{}> rc=<{}>'.format(sh_cmd, answer))
            raise InternalError
