import time

from edge import eventlog
from edge.com import fs
from edge.exceptions import *
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.uiconf.wiegand_pcb import *

from pymodbus.client.sync import ModbusSerialClient
from pymodbus.exceptions import ModbusException
from pymodbus.bit_write_message import WriteSingleCoilResponse
from pymodbus.bit_read_message import ReadCoilsResponse


class _BoardUnit:
    def __init__(self, board_node, conf_uiref, conf_bit_address, conf_bit_active):
        """
        :type board_node: _BoardNode
        :type conf_uiref: str
        :type conf_bit_address: int
        :type conf_bit_active: bool
        """
        self._board_node = board_node
        self._conf_uiref = conf_uiref
        self._conf_bit_address = conf_bit_address
        self._conf_bit_active = conf_bit_active

    def __enter__(self):
        self._board_node.__enter__()
        return self

    def __exit__(self, x_class, x_object, traceback):
        return self._board_node.__exit__(x_class, x_object, traceback)

    def get_stats(self):
        """
        :rtype: dict
        """
        bit_value = self._board_node.get_port_state(self._conf_bit_address)
        return {
            'ref': self._conf_uiref,
            'active': bit_value == self._conf_bit_active
        }


class _BoardInputUnit(_BoardUnit):
    @classmethod
    def create(cls, board_node, uiref):
        """
        :type board_node: _BoardNode
        :type uiref: str

        :rtype: _BoardUnit
        """
        uiconf = WiegandPcbInputPortUiconf(ph.UiMod2(uiref).uiconf)
        return cls(board_node, uiref, uiconf.ro_bit_address, uiconf.ro_bit_active_state)


class _BoardRelayUnit(_BoardUnit):
    @classmethod
    def create(cls, board_node, uiref):
        """
        :type board_node: _BoardNode
        :type uiref: str

        :rtype: _BoardUnit
        """
        uiconf = WiegandPcbRelayPortUiconf(ph.UiMod2(uiref).uiconf)
        return cls(board_node, uiref, uiconf.rw_bit_address, uiconf.rw_bit_active_state)

    def set_state(self, active):
        """
        :type active: bool

        :rtype: None
        :raises InternalError:
        """
        bit_value = self._conf_bit_active if active else not self._conf_bit_active
        self._board_node.set_port_state(self._conf_bit_address, bit_value)


class _BoardNode:
    _SLAVE_ADDRESS = 0x31
    _RX_TO_TX_TOUT_SEC = 0.025
    _RETRY_TOUT = 0.25  # seconds

    def __init__(self, endpoint, lock_path):
        """
        :type endpoint: str
        :type lock_path: str
        """
        self._modbus = ModbusSerialClient(
            method='rtu', port=endpoint, baudrate=19200,
            bytesize=8, stopbits=1, parity='N')

        self._lock_path = lock_path
        self._ports_state = []

        self._input_ports = []
        self._relay_ports = []

    def __enter__(self):
        for attempt in range(1, 4):
            if not fs.try_lock_pid_file(self._lock_path):
                eventlog.log_w('fail: lock <{}> (attempt={})'.format(self._lock_path, attempt))
                time.sleep(self._RETRY_TOUT)
                continue

            return self

        eventlog.log_e('fail: lock <{}>'.format(self._lock_path))
        raise BusyError

    def __exit__(self, x_class, x_object, traceback):
        # unlock
        fs.try_rm_pid_file(self._lock_path)

    @classmethod
    def create(cls, uiref):
        """
        :param uiref: wiegand-board.x
        :type uiref: str

        :rtype: _BoardNode
        """
        uiconf = WiegandPcbUiconf(ph.UiMod2(uiref).uiconf)
        return cls(uiconf.endpoint, uiconf.lock_path)

    def create_input(self, uiref):
        """
        :param uiref: wiegand-board.x~input-port.x
        :type uiref: str

        :rtype: _BoardInputUnit
        """
        board_unit = _BoardInputUnit.create(self, uiref)
        self._input_ports.append(board_unit)
        return board_unit

    def create_relay(self, uiref):
        """
        :param uiref: wiegand-board.x~relay-port.x
        :type uiref: str

        :rtype: _BoardRelayUnit
        """
        board_unit = _BoardRelayUnit.create(self, uiref)
        self._relay_ports.append(board_unit)
        return board_unit

    def get_stats(self):
        """
        :rtype: dict
        """
        return {
            'relay-ports': map(lambda x: x.get_stats(), self._relay_ports),
            'input-ports': map(lambda x: x.get_stats(), self._input_ports)
        }

    def _get_ports_state(self):
        """
        :returns: (OUT1, OUT2, IN1, IN2, IN3, IN4)
        :rtype: (bool, bool, bool, bool, bool, bool)

        :raises InternalError:
        """
        try:
            with self._modbus as modbus:
                response = modbus.read_coils(0x0000, count=6, unit=self._SLAVE_ADDRESS)
                time.sleep(self._RX_TO_TX_TOUT_SEC)

            if not isinstance(response, ReadCoilsResponse):
                eventlog.log_e('fail: read_coils ({})'.format(str(response or 'timeout')))
                raise InternalError

            eventlog.log_d('done: read_coils ([OUT1, OUT2, IN1, IN2, IN3, IN4]:{})'.format(response.bits))
            return response.bits

        except ModbusException as x:
            eventlog.log_e('fail: ModbusException ({})'.format(str(x)))
            raise InternalError

    def get_port_state(self, bit_address):
        """
        :type bit_address: int
        :rtype: bool

        :raises InternalError:
        """
        if not self._ports_state:
            self._ports_state = self._get_ports_state()

        return self._ports_state[bit_address]

    def set_port_state(self, bit_address, bit_value, ack_tout_sec=1):
        """
        :type bit_address: int
        :type bit_value: bool
        :type ack_tout_sec: float

        :rtype: None
        :raises InternalError:
        """
        modbus_timeout_back = self._modbus.timeout
        self._modbus.timeout = ack_tout_sec
        try:
            with self._modbus as modbus:
                response = modbus.write_coil(bit_address, bit_value, unit=self._SLAVE_ADDRESS)
                time.sleep(self._RX_TO_TX_TOUT_SEC)

            if not isinstance(response, WriteSingleCoilResponse):
                eventlog.log_e('fail: write_coil ({})'.format(str(response or 'timeout')))
                raise InternalError

            elif response.value != bit_value:
                eventlog.log_e('fail: write_coil (address:{}, query-value:{}, reply-value:{})'
                               .format(hex(bit_address), hex(bit_value), hex(response.value)))
                raise InternalError

            else:
                eventlog.log_d('done: write_coil (address:{}, value:{})'
                               .format(hex(bit_address), hex(bit_value)))

        except ModbusException as x:
            eventlog.log_e('fail: ModbusException ({})'.format(str(x)))
            raise InternalError

        finally:
            self._modbus.timeout = modbus_timeout_back


class WiegandPcbApi:
    def __init__(self):
        pass

    @classmethod
    def make_service(cls, ctrl_uiref, endpoint):
        """
        :type ctrl_uiref: str
        :type endpoint: endpoint
        :rtype: dict
        """
        root_info, root_links = ph.RootInfoUiconf.get_info_links()

        # controller_uimod = UiMod.create_controller(controller_id)
        controller_info = ph.RootInfoUiconf.find_unit_r(root_info, unit_ref=ctrl_uiref)
        if controller_info is None:
            eventlog.log_d('fail: NotFoundError ({})'.format(ph.UI_TYPE_TERMINAL))
            raise NotFoundError

        _, _, wiegand_pcb_id = ph.RootInfoUiconf.new_uimod_pid_type_id(controller_info, ph.UI_TYPE_WIEGAND_PCB)
        service_uiref = '{}.{}'.format(ph.UI_TYPE_WIEGAND_PCB, wiegand_pcb_id)

        lock_path = '/var/run/edge/{}.lock'.format(service_uiref)
        service_uiconf = WiegandPcbUiconf.create({'lock-path': lock_path, 'endpoint': endpoint})
        service_uimod = ph.UiMod2(service_uiref)
        service_uimod.install(service_uiconf, service_uiref)

        #
        service_units_info = []

        for unit_chan in range(2):
            # install uimod
            unit_uiconf = WiegandPcbRelayPortUiconf.create(unit_chan)
            unit_uimod = service_uimod.child(ph.UI_TYPE_WIEGAND_PCB_RELAY_PORT, unit_chan)
            unit_uimod.install(unit_uiconf, service_uimod.ref)

            # install uimod-info
            service_units_info.append({
                'ref': unit_uimod.ref,
                'type': unit_uimod.type,
                'name': 'OUT{}'.format(unit_chan),
                'role': ph.UI_ROLE_UNIT,
                'internal': True,
            })

        for unit_chan in range(4):
            # install uimod
            unit_uiconf = WiegandPcbInputPortUiconf.create(unit_chan)
            unit_uimod = service_uimod.child(ph.UI_TYPE_WIEGAND_PCB_INPUT_PORT, unit_chan)
            unit_uimod.install(unit_uiconf, service_uimod.ref)

            # install uimod-info
            service_units_info.append({
                'ref': unit_uimod.ref,
                'type': unit_uimod.type,
                'name': 'IN{}'.format(unit_chan),
                'role': ph.UI_ROLE_UNIT,
                'internal': True,
            })

        #
        if controller_info.get('units') is None:
            controller_info['units'] = []

        controller_info['units'].append({
            'ref': service_uimod.ref,
            'type': service_uimod.type,
            'name': 'Wiegand PCB',
            'role': ph.UI_ROLE_UNIT,
            'internal': True,
            'units': service_units_info
        })

        ph.RootInfoUiconf.set_info_links(root_info, root_links)
        return wiegand_pcb_id, root_info

    @staticmethod
    def drop_service(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_WIEGAND_PCB)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        root_info, root_links = ph.RootInfoUiconf.get_info_links()
        unit_info, _ = ph.RootInfoUiconf.find_unit_parent_r(root_info, uimod_ref)
        refs_to_delete = ph.RootInfoUiconf.find_unit_refs_r(
            unit_info,
            [ph.UI_TYPE_WIEGAND_PCB, ph.UI_TYPE_WIEGAND_PCB_RELAY_PORT, ph.UI_TYPE_WIEGAND_PCB_INPUT_PORT])

        for ref_to_delete in refs_to_delete:
            uimod = ph.UiMod2(ref_to_delete)
            root_info = ph.RootInfoUiconf.del_unit_r(uimod.ref)
            uimod.uninstall()

        ext_root_info = ph.RootInfoUiconf.get_ext_root_info(root_info)
        reply = {
            'info': [ext_root_info]
        }
        return reply

    @staticmethod
    def load_service(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_WIEGAND_PCB)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        return uimod.uiconf.pub

    @staticmethod
    def load_service_stats(uimod_ref):
        """
        :param uimod_ref: wiegand-pcb.x
        :type uimod_ref: str

        :rtype: dict
        """
        node_uimod = ph.UiMod2(uimod_ref)
        node_uimod.assert_type(ph.UI_TYPE_WIEGAND_PCB)
        node_uimod.assert_ref()

        root_uimod_info = ph.RootInfoUiconf.get_info()
        node_uimod_info = ph.RootInfoUiconf.find_unit_r(root_uimod_info, unit_ref=uimod_ref)

        board_node = _BoardNode.create(uimod_ref)

        def map_fn(unit_uimod_info):
            unit_uimod_ref = unit_uimod_info['ref']
            unit_uimod_type = unit_uimod_info['type']
            if unit_uimod_type == ph.UI_TYPE_WIEGAND_PCB_INPUT_PORT:
                board_node.create_input(unit_uimod_ref)

            elif unit_uimod_type == ph.UI_TYPE_WIEGAND_PCB_RELAY_PORT:
                board_node.create_relay(unit_uimod_ref)

            else:
                raise NotImplementedError

        ph.RootInfoUiconf.map_units_r(node_uimod_info, map_fn, unit_types=[
            ph.UI_TYPE_WIEGAND_PCB_RELAY_PORT, ph.UI_TYPE_WIEGAND_PCB_INPUT_PORT])

        with board_node:
            stats = board_node.get_stats()
        return stats

    @staticmethod
    def save_service(uimod_ref, new_uiconf):
        """
        :type uimod_ref: str
        :type new_uiconf: dict
        :rtype: dict
        """
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_WIEGAND_PCB)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        uiconf = uimod.uiconf
        name_changed = uiconf.set_pub_name(new_uiconf_conf['name'])
        conf_changed = uiconf.set_pub_conf(new_uiconf_conf) or name_changed
        if conf_changed:
            uiconf.validate()
            uimod.save_uiconf(uiconf)

        if name_changed:
            ph.RootInfoUiconf.set_unit_name(uimod.ref, uiconf.pub_name)

        return uiconf.pub

    @staticmethod
    def load_relay_unit(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_WIEGAND_PCB_RELAY_PORT)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        return uimod.uiconf.pub

    @staticmethod
    def save_relay_unit(uimod_ref, new_uiconf):
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_WIEGAND_PCB_RELAY_PORT)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        uiconf = uimod.uiconf
        name_changed = uiconf.set_pub_name(new_uiconf_conf['name'])
        conf_changed = uiconf.set_pub_conf(new_uiconf_conf) or name_changed

        if conf_changed:
            uiconf.validate()
            uimod.save_uiconf(uiconf)

        if name_changed:
            ph.RootInfoUiconf.set_unit_name(uimod.ref, uiconf.pub_name)

        return uiconf.pub

    @staticmethod
    def load_input_unit(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_WIEGAND_PCB_INPUT_PORT)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        return uimod.uiconf.pub

    @staticmethod
    def save_input_unit(uimod_ref, new_uiconf):
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod2(uimod_ref)
        try:
            uimod.assert_type(ph.UI_TYPE_WIEGAND_PCB_INPUT_PORT)
            uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        uiconf = uimod.uiconf
        name_changed = uiconf.set_pub_name(new_uiconf_conf['name'])
        conf_changed = uiconf.set_pub_conf(new_uiconf_conf) or name_changed

        if conf_changed:
            uiconf.validate()
            uimod.save_uiconf(uiconf)

        if name_changed:
            ph.RootInfoUiconf.set_unit_name(uimod.ref, uiconf.pub_name)

        return uiconf.pub

    @staticmethod
    def load_input_unit_stats(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        input_uimod = ph.UiMod2(uimod_ref)
        try:
            input_uimod.assert_type(ph.UI_TYPE_WIEGAND_PCB_INPUT_PORT)
            input_uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        board_node = _BoardNode.create(uiref=input_uimod.parent_ref)
        board_unit = board_node.create_input(input_uimod.ref)
        with board_node:
            stats = board_unit.get_stats()

        return stats

    @staticmethod
    def load_relay_unit_stats(uimod_ref):
        """
        :param uimod_ref: wiegand-pcb.x~relay-port.x
        :type uimod_ref: str

        :rtype: dict
        """
        relay_uimod = ph.UiMod2(uimod_ref)
        try:
            relay_uimod.assert_type(ph.UI_TYPE_WIEGAND_PCB_RELAY_PORT)
            relay_uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        board_node = _BoardNode.create(uiref=relay_uimod.parent_ref)
        board_unit = board_node.create_relay(relay_uimod.ref)
        with board_node:
            stats = board_unit.get_stats()
        return stats

    @staticmethod
    def activate_relay_unit(uimod_ref, timeout_ms=None):
        """
        :type uimod_ref: str
        :type timeout_ms: int | None
        :rtype: dict
        """
        relay_uimod = ph.UiMod2(uimod_ref)
        try:
            relay_uimod.assert_type(ph.UI_TYPE_WIEGAND_PCB_RELAY_PORT)
            relay_uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        board_node = _BoardNode.create(uiref=relay_uimod.parent_ref)
        board_unit = board_node.create_relay(relay_uimod.ref)
        with board_node:
            board_unit.set_state(True)
            if timeout_ms is None:
                return {
                    'stats': {
                        'active': True
                    }
                }

            else:
                time.sleep(timeout_ms / 1000.0)
                board_unit.set_state(False)
                return {
                    'stats': {
                        'active': False
                    }
                }

    @staticmethod
    def deactive_relay_unit(uimod_ref):
        """
        :type uimod_ref: str
        :rtype: dict
        """
        relay_uimod = ph.UiMod2(uimod_ref)
        try:
            relay_uimod.assert_type(ph.UI_TYPE_WIEGAND_PCB_RELAY_PORT)
            relay_uimod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        board_node = _BoardNode.create(uiref=relay_uimod.parent_ref)
        board_unit = board_node.create_relay(relay_uimod.ref)
        with board_node:
            board_unit.set_state(False)

        return {
            'stats': {
                'active': False
            }
        }


def _main():
    eventlog.open_log('wiegand-test', 'debug')
    # noinspection PyBroadException
    try:
        node = _BoardNode.create('node_ref')
        # unit = WiegandPcbUnit.create(node, 'unit_ref')
        node.set_port_state(0, True)
        print node.get_port_state(0)
        print node.get_port_state(1)

    except Exception:
        eventlog.log_x()
        exit(1)

    finally:
        eventlog.close_log()


if __name__ == '__main__':
    _main()
