import os

from edge import eventlog
from edge.com import fs, sh
from edge.exceptions import *
from edge.webserver.backup import CgiBackup
from edge.webserver.file_lock import CgiLock
from edge.webserver.sysconf import ph


# noinspection PyClassHasNoInit
class _Assets:
    DEF_UICONF_PATH = '/opt/edge/share/setupstore/ui/zabbix-jnode.json'
    UISCHEMA_PATH = '/etc/edge/setupstore/schemas/zabbix.json'
    UICONF_PATH = '/etc/edge/setupstore/admin/actions/zabbix.json'

    WEB_CONF = '/etc/zabbix/zabbix_agentd.conf.d/web.conf'
    VARUSR_CONF = '/opt/varusr/local/etc/zabbix/varusr.conf'


# noinspection PyClassHasNoInit
class ZabbixApi:
    @staticmethod
    def load_unit():
        """
        :rtype: dict
        """
        def install_uiconf(def_uiconf_path, uiconf_path):
            """
            :type def_uiconf_path: str
            :type uiconf_path: str
            """
            if not fs.path_exists(uiconf_path):
                fs.create_jnode_from_default(def_uiconf_path, uiconf_path)

        def install_uischema(uiconf_path, uischema_path):
            """
            :type uiconf_path: str
            :type uischema_path: str
            """
            if os.path.islink(uischema_path):
                if fs.path_exists(os.readlink(uischema_path)):
                    return

            fs.ln_schema(uischema_path, uiconf_path)

        install_uiconf(_Assets.DEF_UICONF_PATH, _Assets.UICONF_PATH)
        install_uischema(_Assets.UICONF_PATH, _Assets.UISCHEMA_PATH)

        # load ui-jnode
        uiconf = fs.load_json3(_Assets.UICONF_PATH, default=None)
        if uiconf is None:
            eventlog.log_d('fail: InternalError')
            raise InternalError

        uiconf['rw']['type'] = ph.UI_TYPE_ZABBIX_AGENT
        return uiconf['rw']

    @classmethod
    @CgiLock('/var/run/edge/zabbix')
    @CgiBackup('/etc/edge/setupstore/backup/zabbix', {
        'zabbix.json': _Assets.UICONF_PATH,
        'web.conf': _Assets.WEB_CONF,
    })
    def save_unit(cls, unit_jnode):
        """
        :type unit_jnode: dict
        :rtype: dict
        """
        fs.validate_json(unit_jnode, _Assets.UISCHEMA_PATH)

        # edit conf files
        uiconf = fs.load_json3(_Assets.UICONF_PATH)
        uiconf_pub = uiconf['rw']  # type: dict
        if uiconf_pub != unit_jnode:
            cls._edit_web_conf(uiconf_pub['conf'], unit_jnode['conf'])

            uiconf_pub['conf'] = unit_jnode['conf']
            fs.write_json_file(uiconf, _Assets.UICONF_PATH, pretty=True)

            # restart services
            if unit_jnode['conf']['enabled']:
                sh.run_silent('update-rc.d zabbix-agent enable')
                sh.run_silent('service zabbix-agent restart')
            else:
                sh.run_silent('service zabbix-agent stop')
                sh.run_silent('update-rc.d zabbix-agent disable')

        unit_ext_jnode = unit_jnode
        unit_ext_jnode['type'] = ph.UI_TYPE_ZABBIX_AGENT
        return unit_jnode

    @staticmethod
    def _edit_web_conf(last_rw_jnode_conf, next_rw_jnode_conf):
        def mk_header_lines():
            if fs.path_exists(_Assets.WEB_CONF):
                with open(_Assets.WEB_CONF, 'r') as conf:
                    lines = conf.readlines()
            else:
                lines = []

            header_lines = []
            for line in lines:
                if line.startswith('#'):
                    header_lines.append(line)

            return header_lines

        conf_lines = mk_header_lines()

        enabled = next_rw_jnode_conf['enabled']
        if enabled:
            conf_lines.append('\nListenPort={}\n'.format(next_rw_jnode_conf['agent-port']))

            hostname = next_rw_jnode_conf.get('agent-host')
            if hostname:
                conf_lines.append('\nHostname={}\n'.format(hostname))

            if not len(next_rw_jnode_conf['servers']):
                next_rw_jnode_conf['servers'] = last_rw_jnode_conf['servers']

            conf_lines.append('Server=')
            servers = map(lambda server: server['host'], next_rw_jnode_conf['servers'])
            conf_lines.append(', '.join(servers))

            conf_lines.append('\nServerActive=')
            servers = map(lambda server: ('%s:%d' % (server['host'], server['port'])), next_rw_jnode_conf['servers'])
            conf_lines.append(', '.join(servers))

            conf_lines.append('\nInclude=/etc/edge/zabbix_agentd.d/\n')
            if fs.path_exists(_Assets.VARUSR_CONF):
                conf_lines.append('\nInclude=%s\n' % _Assets.VARUSR_CONF)

        with open(_Assets.WEB_CONF, 'w') as web_conf:
            web_conf.write(''.join(conf_lines))
