import os
import signal
import subprocess

from edge import eventlog
from edge.com import fs, sh

EDGE_TYPE_SETUPSTORE = 'setupstore'
EDGE_TYPE_EVENTSTORE = 'eventstore'
EDGE_TYPE_LPR = 'lpr'
EDGE_TYPE_MEDIACODEC = 'mediacodec'
EDGE_TYPE_MEDIAGRABBER = 'mediagrabber'
EDGE_TYPE_GRABCONTROL = 'grabcontrol'
EDGE_TYPE_RADAR = 'radar'
EDGE_TYPE_RENDERER = 'overlayrender'
EDGE_TYPE_RTSP_GRABBER = 'rtspgrabber'
EDGE_TYPE_RTSP_SERVER = 'rtspserver'
EDGE_TYPE_SPEEDCONTROL = 'speedcontrol'
EDGE_TYPE_TRACKER = 'tracker'
EDGE_TYPE_WEBSERVER = 'webserver'
EDGE_TYPE_RAWCODEC = 'rawcodecgst'
EDGE_TYPE_BLOBSTORE = 'blobstore'
EDGE_TYPE_MEDIACODEC_GST = 'mediacodecgst'
EDGE_TYPE_TRAFFIC_LIGHT = 'traffic-light'
EDGE_TYPE_LBSAPP = 'lbsapp'
EDGE_TYPE_LEDDAR = 'leddar'


class DgMod:
    def __init__(self, dgmod_type, dgmod_id, uimod_pid=None):
        """
        :type dgmod_type: str
        :type dgmod_id: str
        :type uimod_pid: str | None
        """
        self._uimod_pid = uimod_pid
        self.type = dgmod_type
        self.id = dgmod_id

    @staticmethod
    def create(dgmod_type, uimod_pid, uimod_id):
        """
        :type dgmod_type: str
        :type uimod_pid: str
        :type uimod_id: str
        :rtype: DgMod
        """
        if dgmod_type in [EDGE_TYPE_WEBSERVER, EDGE_TYPE_SETUPSTORE]:
            dgmod_id = '0'

        elif dgmod_type == EDGE_TYPE_EVENTSTORE:
            dgmod_id = '65535'

        else:
            dgmod_id = uimod_id

        return DgMod(dgmod_type, dgmod_id, uimod_pid)

    def __hash__(self):
        return hash(self.type + self.id)

    def __str__(self):
        return '<{}|{}>'.format(self.type, self.id)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                getattr(other, 'type', None) == self.type and
                getattr(other, 'id', None) == self.id)

    @property
    def pid(self):
        """
        :rtype: str
        """
        if self._uimod_pid is None:
            eventlog.log_e('fail: unsupported')
            raise NotImplementedError

        return self._uimod_pid

    @property
    def conf_dir(self):
        """
        :rtype: str
        """
        if self.type == EDGE_TYPE_SETUPSTORE:
            return '/etc/edge/{}'.format(self.type)
        else:
            return '/etc/edge/{}.{}'.format(self.type, self.id)

    def get_conf_path(self, conf_name):
        """
        :type conf_name: str
        :rtype: str
        """
        return fs.join_path(self.conf_dir, conf_name)

    def load_json_conf(self, conf_name, **kwargs):
        """
        :type conf_name: str
        :rtype: dict | None
        """
        conf_path = self.get_conf_path(conf_name)
        return fs.load_json3(conf_path, **kwargs)

    def load_main_conf(self, **kwargs):
        """
        :rtype: dict | None
        """
        return self.load_json_conf('main.conf', **kwargs)

    def save_json_conf(self, conf_name, conf_jnode):
        """
        :type conf_name: str
        :type conf_jnode: dict
        :rtype: None
        """
        conf_path = self.get_conf_path(conf_name)
        fs.write_json_file(conf_jnode, conf_path, pretty=True)

    def save_main_conf(self, conf_jnode):
        """
        :type conf_jnode: dict
        :rtype: None
        """
        return self.save_json_conf('main.conf', conf_jnode)

    def save_misc_conf(self, conf_jnode):
        """
        :type conf_jnode: dict
        :rtype: None
        """
        return self.save_json_conf('misc.conf', conf_jnode)

    def save_text_conf(self, conf_name, conf_text):
        """
        :type conf_name: str
        :type conf_text: str
        :rtype: None
        """
        conf_path = self.get_conf_path(conf_name)
        with open(conf_path, 'w') as text_file:
            text_file.write(conf_text)

    def save_haltaction_conf(self, conf_text):
        """
        :type conf_text: str
        :rtype: None
        """
        self.save_text_conf('haltaction.conf', conf_text)

    def save_setupstore_conf(self, active):
        """
        :type active: bool
        :rtype: None
        """
        if active:
            conf_text = 'Active=true\nArgSet=--id={} --log-level=info\n'.format(self.id)
        else:
            conf_text = 'Active=false\nArgSet=--id={} --log-level=info\n'.format(self.id)

        self.save_text_conf('setupstore.conf', conf_text)

    def install(self):
        """
        :rtype None
        """
        fs.silent_mkdir_r(self.conf_dir)

    def uninstall(self):
        """
        :rtype: None
        """
        fs.silent_rm_r(self.conf_dir)


CTRL_SS_DGMOD = DgMod(EDGE_TYPE_SETUPSTORE, dgmod_id='0')
CTRL_BS_DGMOD = DgMod(EDGE_TYPE_BLOBSTORE, dgmod_id='65535')
CTRL_ES_DGMOD = DgMod(EDGE_TYPE_EVENTSTORE, dgmod_id='65535')
CTRL_WS_DGMOD = DgMod(EDGE_TYPE_WEBSERVER, dgmod_id='0')


# noinspection PyClassHasNoInit
class LifeCycle:
    _WS_RESTART_FILEPATH = '/var/spool/edge/ws0-ws0-restart'
    _SS_RESTART_FILEPATH = '/var/spool/edge/ss0-ss0-restart'

    @classmethod
    def restart_edgeserver(cls):
        # PCD restarts edgeserver when this file appears
        with open(cls._SS_RESTART_FILEPATH, 'w'):
            pass

    @staticmethod
    def restart_dgmod(dgmod_type, dgmod_id=None):
        """
        :type dgmod_type: str
        :type dgmod_id: str
        """
        if dgmod_id is None:
            sh_cmd = 'pkill -9 {}'.format(dgmod_type)
            eventlog.log_i('exec: {}'.format(sh_cmd))
            subprocess.Popen([sh_cmd], shell=True)

        else:
            eventlog.log_i('exec: {}.{}'.format(dgmod_type, dgmod_id))
            sh.run_silent(
                "kill -9 `ps ax |"
                "grep \"%s --id.%s\" |"
                "grep -v grep |"
                "awk '{print $1}'`" % (dgmod_type, dgmod_id))

    @classmethod
    def restart_ws_dgmod(cls):
        eventlog.log_i('exec')
        # PCD restarts webserver when this file appears
        fs.silent_mkdir_r(os.path.dirname(cls._WS_RESTART_FILEPATH))

        with open(cls._WS_RESTART_FILEPATH, 'w'):
            pass

    @staticmethod
    def sighup_dgmod(dgmod_type, dgmod_id):
        """
        :type dgmod_type: str
        :type dgmod_id: str
        """
        eventlog.log_i('exec: {}.{}'.format(dgmod_type, dgmod_id))
        sh.run_silent(
            "kill -1 `ps ax |"
            "grep \"%s --id.%s\" |"
            "grep -v grep |"
            "awk '{print $1}'`" % (dgmod_type, dgmod_id))

    @staticmethod
    def sighup_dgmods(dgmod_type):
        """
        :type dgmod_type: str
        """
        eventlog.log_i('exec: {}'.format(dgmod_type))
        sh_cmd = 'pkill -1 {}'.format(dgmod_type)
        subprocess.Popen([sh_cmd], shell=True)

    @staticmethod
    def sighup_ss_dgmod():
        eventlog.log_i('exec')
        os.kill(int(subprocess.check_output(['pidof', EDGE_TYPE_SETUPSTORE])), signal.SIGHUP)

    @staticmethod
    def sigint_dgmod(dgmod_type, dgmod_id):
        """
        :type dgmod_type: str
        :type dgmod_id: str
        """
        eventlog.log_i('exec: {}.{}'.format(dgmod_type, dgmod_id))
        sh.run_silent(
            "kill -2 `ps ax |"
            "grep \"%s --id.%s\" |"
            "grep -v grep |"
            "awk '{print $1}'`" % (dgmod_type, dgmod_id))
