from itertools import ifilter

from edge.com import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator


# noinspection PyUnusedLocal
def gc_get_creators(uimod, all_used_uimods, uimod_to_newconf_map, forced=False):
    """
    :type uimod: ph.UiMod | ph.UiMod2
    :type all_used_uimods: set[ph.UiMod, ph.UiMod2]
    :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
    :type uimod_to_newconf_map: dict
    :type forced: bool
    :rtype: list of BaseCreator
    """
    if uimod.type != ph.UI_TYPE_MV_CAMERA:
        return []

    try:
        cam_uimod = next(ifilter(lambda x: x.type == ph.UI_TYPE_MV_CAMERA, all_used_uimods))
        lpr_uimod = next(ifilter(lambda x: x.type == ph.UI_TYPE_LPR, all_used_uimods))

        activate = any([
            uimod_to_newconf_map.get(cam_uimod)
        ])
        if forced or activate:
            return [GcCreator(cam_uimod, lpr_uimod)]

    except StopIteration:
        pass

    return []


class GcCreator(BaseCreator):
    def __init__(self, cam_uimod, lpr_uimod):
        """
        :type cam_uimod: ph.UiMod
        :type lpr_uimod: ph.UiMod
        """
        BaseCreator.__init__(self)
        self._cam_uimod = cam_uimod
        self._lpr_uimod = lpr_uimod
        self._gc_dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_GRABCONTROL, cam_uimod.pid, cam_uimod.id)
        self._main_active = None

    @property
    def dgmod(self):
        """
        :rtype: dg.DgMod
        """
        return self._gc_dgmod

    def prepare(self, uimod_to_newconf_map):
        """
        :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
        :type uimod_to_newconf_map: dict
        """
        ui_pid = self._cam_uimod.pid
        spool_dir = ph.get_spool_dir(ui_pid)

        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()
        cam_uiconf_control = cam_uiconf['rw']['videostream']['control']

        cam_uiconf_pri = cam_uiconf['ro']
        #
        exp_time_conf = [0, 0, 0]
        gain_conf = [0, 0, 0]
        gain_step = 1.0
        exptime_step = 20
        if cam_uiconf_control['profile'] == 'auto-70-150':
            gain_conf = [cam_uiconf_control['gain-auto-min'], cam_uiconf_control['gain-auto-max'], gain_step]
            exp_time_conf = [
                cam_uiconf_control['exposure-auto-min'],
                cam_uiconf_control['exposure-auto-max'],
                exptime_step]
        #

        main_conf = {
            "ControlIR": cam_uiconf_pri.get('control-ir', True),
            "Analyzer": {
                "ParamSet": {
                    "Target": {
                        "Median": [70, 150],
                        "SteadyState": {
                            "Unstable": 5,
                            "PreStable": 15,
                            "Stable": 30
                        }
                    },
                    "Sweep": {
                        "Gain": gain_conf,
                        "ExpTime": exp_time_conf,
                        "Gamma": [0.4, 1.5, 0.1],
                        "Lut12": {
                            "Default": "/opt/edge/share/setupstore/edge/mediagrabber/lut12.default.conf",
                            "Evening": "",
                            "Night": "/opt/edge/share/setupstore/edge/mediagrabber/lut12.olvia.conf",
                        }
                    }
                }
            },
            "FileApi": {
                "AddSceneImgReact": {
                    "Url": fs.join_path(spool_dir, 'mg{}-gc{}-scene'.format(
                        self._cam_uimod.id, self._cam_uimod.id
                    )),
                },
                "AddSceneAoiReact": {
                    "Url": fs.join_path(spool_dir, 'lr{}-gc{}-scene'.format(
                        self._lpr_uimod.id, self._cam_uimod.id
                    )),
                },
                "SceneControlEvent": ["/etc/edge/mediagrabber.{}/dyn/".format(self._gc_dgmod.id)]
            }
        }

        ss_active = cam_uiconf_control['profile'] == 'auto-70-150'
        self._main_active = main_conf, ss_active

    def complete(self):
        main_conf, ss_active = self._main_active  # type: dict, bool

        edge_gc = self._gc_dgmod  # type: dg.DgMod
        edge_gc.install()
        edge_gc.save_main_conf(main_conf)
        edge_gc.save_setupstore_conf(ss_active)

    def rollback(self):
        pass

    def notify(self):
        """
        :rtype: None
        """
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class GcDeleter:
    def __init__(self, dgmod):
        """
        :type dgmod: dg.DgMod
        """
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod):
        """
        :type entry_uimod: ph.UiMod
        :rtype: GcDeleter | None
        """
        if entry_uimod.type == ph.UI_TYPE_MV_CAMERA:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_GRABCONTROL, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self):
        self._dgmod.uninstall()
