from itertools import ifilter

from edge.com import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator


# noinspection PyUnusedLocal
def leddar_get_creators(uimod, linked_uimods, uimod_to_newconf_map, forced=False):
    """
    :type uimod: ph.UiMod
    :type linked_uimods: list[ph.UiMod]
    :type uimod_to_newconf_map: dict
    :type forced: bool
    :rtype: list of BaseCreator
    """
    if uimod.type != ph.UI_TYPE_LEDDAR:
        return []

    try:
        ledar_uimod = next(ifilter(lambda x: x.type == ph.UI_TYPE_LEDDAR, linked_uimods))
        rules_uimod = next(ifilter(lambda x: x.type == ph.UI_TYPE_TRAFFIC_RULES, linked_uimods))

        activate = any([
            uimod_to_newconf_map.get(ledar_uimod)
        ])
        if forced or activate:
            return [LeddarCreator(ledar_uimod, rules_uimod)]

    except StopIteration:
        pass

    return []


class LeddarCreator(BaseCreator):
    def __init__(self, ledar_uimod, rules_uimod):
        """
        :type ledar_uimod: ph.UiMod
        :type rules_uimod: ph.UiMod
        """
        BaseCreator.__init__(self)
        self._ledar_uimod = ledar_uimod
        self._rules_uimod = rules_uimod
        self._leddar_dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_LEDDAR, rules_uimod.pid, rules_uimod.id)
        self._main_active = None

    @property
    def dgmod(self):
        """
        :rtype: dg.DgMod
        """
        return self._leddar_dgmod

    def prepare(self, uimod_to_conf_map):
        """
        :type uimod_to_conf_map: dict
        """
        ui_pid = self._ledar_uimod.pid
        spool_dir = ph.get_spool_dir(ui_pid)

        #
        main_conf = {
            "driver": {
                "connection": {
                    "type": "usb",
                    "address": 1,
                    "endpoint": "/dev/ttyUSB0"
                },
                "dist-thr": 5.1
            },
            "bg-detector": {
                "beam-angle": 45.1,
                "beam-seg-count": 16,
                "beam-depth-map": [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1 ]
            },
            "loop-detector": {
                "changelog": {
                    "entry-count": 50,
                    "entry-tout": 300,
                    "entry-age": 300
                },
                "loops": [{
                    "ref": "loop0",
                    "dist": [0.1, 0.5],
                    "amp": [0.0, 1000.0],
                    "seg": [7, 8]
                }]
            },
            "classifier": {
                "bike": [80, 220],
                "car": [220, 560]
            },
            "scene-sync": "/var/spool/edge/mg0-lg0-pipe",
            "scene-event": {
                "endpoints": [
                    "/tmp/lg0-tr0-scene"
                ]
            },
            "track-event": {
                "endpoints": [
                    fs.join_path(spool_dir, 'leddar{}-sc{}-track'.format(self._ledar_uimod.id, self._rules_uimod.id))
                ]
            }
        }

        self._main_active = main_conf, True

    def complete(self):
        main_conf, ss_active = self._main_active  # type: dict, bool

        edge_leddar = self._leddar_dgmod  # type: dg.DgMod
        edge_leddar.install()
        edge_leddar.save_main_conf(main_conf)
        edge_leddar.save_setupstore_conf(ss_active)

    def rollback(self):
        pass

    def notify(self):
        """
        :rtype: None
        """
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class LeddarDeleter:
    def __init__(self, dgmod):
        """
        :type dgmod: dg.DgMod
        """
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod):
        """
        :type entry_uimod: ph.UiMod
        :rtype: LeddarDeleter | None
        """
        if entry_uimod.type == ph.UI_TYPE_MV_CAMERA:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_LEDDAR, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self):
        self._dgmod.uninstall()
