from itertools import ifilter

from edge.com import fs, net
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator


# noinspection PyUnusedLocal
def lr_get_creators(entry_uimod, all_used_uimods, uimod_to_newconf_map, forced=False):
    """
    :type entry_uimod: ph.UiMod | ph.UiMod2
    :type all_used_uimods: set[ph.UiMod, ph.UiMod2]
    :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
    :type uimod_to_newconf_map: dict
    :type forced: bool
    :rtype: list of BaseCreator
    """
    if entry_uimod.type != ph.UI_TYPE_LPR:
        return []

    try:
        cam_uimod = next(ifilter(lambda x: x.type in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA], all_used_uimods))
        lpr_uimod = next(ifilter(lambda x: x.type == ph.UI_TYPE_LPR, all_used_uimods))
        fusion_uimod = next(ifilter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods))
        tcp_recipient_uimods = filter(lambda x: x.type == ph.UI_TYPE_TCP_RECIPIENT, all_used_uimods)
        ledar_uimods = filter(lambda x: x.type == ph.UI_TYPE_LEDDAR, all_used_uimods)
        location_uimods = filter(lambda x: x.type == ph.UI_TYPE_LOCATION, all_used_uimods)
        ipcom_waga_uimods = filter(lambda x: x.type == ph.UI_TYPE_IPCOM_WAGA, all_used_uimods)

        activate = any([
            uimod_to_newconf_map.get(lpr_uimod),
            uimod_to_newconf_map.get(cam_uimod),

            any(map(lambda x: uimod_to_newconf_map.get(x), tcp_recipient_uimods)),
            any(map(lambda x: uimod_to_newconf_map.get(x), ledar_uimods)),
            any(map(lambda x: uimod_to_newconf_map.get(x), location_uimods)),
            any(map(lambda x: uimod_to_newconf_map.get(x), ipcom_waga_uimods))
        ])

        if forced or activate:
            return [LrCreator(lpr_uimod, cam_uimod, fusion_uimod, tcp_recipient_uimods, ledar_uimods,
                              location_uimods, ipcom_waga_uimods)]

    except StopIteration:
        pass

    return []


class LrCreator(BaseCreator):
    def __init__(self, lpr_uimod, cam_uimod, fusion_uimod, tcp_recipient_uimods, ledar_uimods,
                 location_uimods, ipcom_waga_uimods):
        """
        :type lpr_uimod: ph.UiMod
        :type cam_uimod: ph.UiMod
        :type fusion_uimod: ph.UiMod
        :type tcp_recipient_uimods: list of ph.UiMod
        :type ledar_uimods: list of ph.UiMod
        :type location_uimods: list of ph.UiMod
        :type ipcom_waga_uimods: list of ph.UiMod
        """
        BaseCreator.__init__(self)
        self._lpr_uimod = lpr_uimod
        self._cam_uimod = cam_uimod
        self._fusion_uimod = fusion_uimod
        self._tcp_recipient_uimods = tcp_recipient_uimods
        self._ledar_uimods = ledar_uimods
        self._location_uimods = location_uimods
        self._ipcom_waga_uimods = ipcom_waga_uimods

        self._lpr_dgmod = dg.DgMod.create(dg.EDGE_TYPE_LPR, lpr_uimod.pid, lpr_uimod.id)
        self._main_conf = None

    @property
    def dgmod(self):
        """
        :rtype: dg.DgMod
        """
        return self._lpr_dgmod

    def prepare(self, uimod_to_newconf_map):
        """
        :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
        :type uimod_to_newconf_map: dict
        """
        spool_path = ph.get_spool_dir(self._lpr_uimod.pid)

        lpr_uiconf = uimod_to_newconf_map.get(self._lpr_uimod) or self._lpr_uimod.load_conf()
        lpr_pri = lpr_uiconf['ro']
        lpr_pub = lpr_uiconf['rw']
        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()
        cam_pri = cam_uiconf['ro']

        if self._cam_uimod.type == ph.UI_TYPE_MV_CAMERA:
            if cam_pri['raw-format'] == 'MONO8T':
                add_scene_react_dir = fs.join_path(spool_path, 'mg{}-lr{}-scene'.format(
                    self._cam_uimod.id, self._lpr_uimod.id))
            else:
                add_scene_react_dir = fs.join_path(spool_path, 'rc{}-lr{}-scene'.format(
                    self._cam_uimod.id, self._lpr_uimod.id))

        elif self._cam_uimod.type == ph.UI_TYPE_RTSP_CAMERA:
            add_scene_react_dir = fs.join_path(spool_path, 'rg{}-lr{}-scene'.format(
                self._cam_uimod.id, self._lpr_uimod.id))

        else:
            raise NotImplementedError

        image_wh = ph.CamDgStats.create(self._lpr_uimod.pid).get_stream_wh() or (640, 480)
        scene_event_jnode = [fs.join_path(spool_path, 'lr{}-tr{}-scene'.format(
            self._lpr_uimod.id, self._fusion_uimod.id))]

        # enable gc
        if self._cam_uimod.type == ph.UI_TYPE_MV_CAMERA:
            if cam_uiconf['rw']['videostream']['control']['profile'] == 'auto-70-150':
                scene_event_jnode.append(fs.join_path(spool_path, 'lr{}-gc{}-scene'.format(
                    self._lpr_uimod.id, self._cam_uimod.id)))

        plate_detector_jnode = lpr_pub['plate-detector']
        plate_detector_jnode["extra-ranges-analyse"] = lpr_pri.get('extra-ranges-analyse') or False
        plate_detector_jnode["extra-angle-analyse"] = lpr_pri.get('extra-angle-analyse') or False
        plate_detector_jnode["precise-analyse"] = lpr_pri.get('precise-analyse') or False

        main_conf = {
            "nodes": [
                {
                    "bundle": ["libunity_essentials", "1.0.0"],
                    "type": "essentials",
                    "name": "essentials",
                    "units": [
                        {
                            "type": "api",
                            "name": "api",
                            "conf": {
                                "endpoint": net.get_lr_uds_endpoint(self._lpr_uimod.pid)
                            }
                        },
                        {
                            "type": "counters",
                            "name": "counters",
                            "conf": {
                                "stat-path": fs.join_path(spool_path, 'zabbix', 'lr0'),
                                "stat-tout": 3
                            }
                        }
                    ]
                },
                {
                    "bundle": ["libunity_edge", "1.0.0"],
                    "type": "media-server",
                    "name": "media-server",
                    "units": [
                        {
                            "type": "media",
                            "name": "media",
                            "conf": {
                                "endpoint": add_scene_react_dir
                            }
                        }
                    ]
                },
                {
                    "bundle": ["libunity_edge", "1.0.0"],
                    "type": "event-emitter",
                    "name": "event-emitter",
                    "units": [
                        {
                            "type": "tx-track-bus",
                            "name": "tx-track-bus",
                            "conf": {
                                "store": {
                                    "endpoint": fs.join_path(spool_path, 'lr{0}-lr{0}-store'.format(
                                        self._lpr_uimod.id)),
                                    "min-file-count": 20,
                                    "max-file-count": 25
                                },
                                "track": [fs.join_path(spool_path, 'lr{}-tr{}-track'.format(
                                    self._lpr_uimod.id, self._fusion_uimod.id))]
                            }
                        },
                        {
                            "type": "tx-scene-bus",
                            "name": "tx-scene-bus",
                            "conf": {
                                "scene": scene_event_jnode
                            }
                        }
                    ]
                },
                {
                    "bundle": ["libunity_anpr", "1.0.0"],
                    "type": "anpr",
                    "name": "anpr",
                    "units": [
                        {
                            "type": "anpr",
                            "name": "anpr",
                            "conf": {
                                "engine": {
                                    "log-settings": False,
                                    "license-group": int(lpr_pub['license-group']),
                                    "license-count": lpr_pub['license-count'],
                                    "image-tout": lpr_pri.get('image-tout') or 250,
                                    "plate-detector": plate_detector_jnode,
                                    "plate-recognizer": lpr_pub['plate-recognizer'],
                                    "plate-tracker": {
                                        "lost-tout": lpr_pub['plate-tracker']['lost-tout'],
                                        "size-min": lpr_pub['plate-tracker'].get('min-size', 4)
                                    }
                                },
                                "stream": {
                                    "wh": image_wh,
                                },
                                "postprocess": lpr_pri.get('postprocess') or 1000,
                                "activation": lpr_pri.get('activation') or "image"
                            }
                        }
                    ]
                }
            ],
            "links": [
                ("anpr", "tx-scene-bus"),
                ("anpr", "media"),
                ("anpr", "api"),
                ("api", "anpr"),

                ("counters", "anpr"),
                ("counters", "tx-scene-bus"),
                ("counters", "tx-track-bus")
            ]
        }

        speedcam_node = None
        add_speedcam = lpr_pri.get('measure-lpr-speed') or False
        if add_speedcam:
            import numpy as np
            import math

            def eul2rotm(theta_xyz):
                """
                Calculates Rotation Matrix given euler angles.
                """
                rmat_x = np.array([
                    [1,     0,                        0                   ],
                    [0,     math.cos(theta_xyz[0]),  -math.sin(theta_xyz[0])],
                    [0,     math.sin(theta_xyz[0]),   math.cos(theta_xyz[0])]
                ])

                rmat_y = np.array([
                    [math.cos(theta_xyz[1]),  0,      math.sin(theta_xyz[1])],
                    [0,                       1,      0                   ],
                    [-math.sin(theta_xyz[1]), 0,      math.cos(theta_xyz[1])]
                ])

                rmat_z = np.array([
                    [math.cos(theta_xyz[2]),  -math.sin(theta_xyz[2]), 0],
                    [math.sin(theta_xyz[2]),   math.cos(theta_xyz[2]), 0],
                    [0,                        0,                      1]
                ])

                rmat = np.dot(rmat_z, np.dot(rmat_y, rmat_x))
                return rmat

            location_uimod = self._location_uimods[0]
            location_uiconf = uimod_to_newconf_map.get(location_uimod) or location_uimod.load_conf()

            location_rw_jnode = location_uiconf['rw']
            location_rw_dev_jnode = location_rw_jnode['args']['radar']
            rotation_jnode = location_rw_dev_jnode['rotation']

            rot_x_grad = float(rotation_jnode['elevation'])
            rot_y_grad = float(rotation_jnode['azimuth'])
            rot_z_grad = float(rotation_jnode['roll'])

            rot_x_rad = -rot_x_grad * math.pi / 180
            rot_y_rad = -rot_y_grad * math.pi / 180
            rot_z_rad = -rot_z_grad * math.pi / 180

            rmat = eul2rotm([rot_x_rad, rot_y_rad, rot_z_rad])
            rmat_1d = rmat.ravel()
            rmat_1d_list = list(rmat_1d.tolist())

            cam_uiconf_pri_frustrum = cam_pri['intrinsic']['frustum']
            cam_wh_mm = cam_uiconf_pri_frustrum['sensor-wh']
            cam_wh_px = cam_pri['intrinsic']['image-wh']
            cam_focal_dist = cam_uiconf_pri_frustrum['focal-dist']

            pixel_w_mm = cam_wh_mm[0] / cam_wh_px[0]
            pixel_h_mm = cam_wh_mm[1] / cam_wh_px[1]

            limits_jnode = lpr_pri.get('speedcam-limits') or {"min-track-size": 14, "max-stddev": [5.0, 5.0, 10.0]}
            speedcam_node = {
                "units": [
                    {
                        "type": "speedcam",
                        "name": "speedcam",
                        "conf": {
                            "camera": {
                                "intrinsics": {
                                    "projection-mat": [
                                        cam_focal_dist / pixel_w_mm, 0, cam_wh_px[0] / 2,
                                        0, cam_focal_dist / pixel_h_mm, cam_wh_px[1] / 2,
                                        0, 0, 1
                                    ]
                                },
                                "extrinsics": {
                                    "rotation-mat": rmat_1d_list
                                }
                            },
                            "limits": limits_jnode
                        }
                    }
                ],
                "type": "speedcam",
                "name": "speedcam",
                "bundle": [
                    "libunity_speedcam",
                    "1.0.0"
                ]
            }

        main_links = main_conf['links']  # type: list

        if speedcam_node:
            main_conf['nodes'].append(speedcam_node)
            main_links.append(("speedcam", "counters"))

        ipcom_node = None
        if len(self._ipcom_waga_uimods) == 1:
            waga_uimod = self._ipcom_waga_uimods[0]
            waga_conf = uimod_to_newconf_map.get(waga_uimod) or waga_uimod.load_conf()
            waga_pub_conf = waga_conf.get('rw', {}).get('conf')
            waga_ev_matcher = waga_pub_conf['event-matcher']
            waga_server = waga_pub_conf['server']
            ipcom_node = {
                "units": [
                    {
                        "type": "ipcom~waga",
                        "name": "ipcom~waga",
                        "conf": {
                            "endpoint": waga_server['host'] + ':{}'.format(waga_server['port']),
                            "lanes": waga_ev_matcher['lanes'],
                            "match-dist": waga_ev_matcher['max-vrp-dist'],
                            "tout-ms": waga_ev_matcher['match-tout'],
                            "vendor": waga_conf.get('ro', {}).get('vendor', '')
                        }
                    }
                ],
                "type": "ipcom",
                "name": "ipcom",
                "bundle": [
                    "libunity_ipcom",
                    "1.0.0"
                ]
            }

        if ipcom_node:
            main_conf['nodes'].append(ipcom_node)
            main_links.append(("ipcom~waga", "counters"))

        mmr_engine = lpr_pri.get('mmr-engine', None)
        if mmr_engine is not None and len(mmr_engine) > 0:
            main_links.append(("anpr", "mmr"))
            main_links.append(("mmr", "counters"))

            if speedcam_node and ipcom_node:
                main_links.append(("mmr", "speedcam"))
                main_links.append(("speedcam", "ipcom~waga"))
                main_links.append(("ipcom~waga", "tx-track-bus"))
            elif ipcom_node:
                main_links.append(("mmr", "ipcom~waga"))
                main_links.append(("ipcom~waga", "tx-track-bus"))
            else:
                main_links.append(("mmr", "tx-track-bus"))

            mmr_jnode = {
                "conf": {
                    "engines": mmr_engine,
                    "min-conf": lpr_pri.get('mmr-min-conf', 50)
                },
                "name": "mmr",
                "type": "mmr"
            }

            # workaround: xnn should be loaded befoare the anpr.
            main_conf['nodes'].insert(0, {
                "bundle": ["libunity_xnn", "1.0.0"],
                "type": "xnn",
                "name": "xnn",
                "units": [
                    mmr_jnode
                ]
            })

        else:
            if speedcam_node and ipcom_node:
                main_links.append(("anpr", "speedcam"))
                main_links.append(("speedcam", "ipcom~waga"))
                main_links.append(("ipcom~waga", "tx-track-bus"))
            elif ipcom_node:
                main_links.append(("anpr", "ipcom~waga"))
                main_links.append(("ipcom~waga", "tx-track-bus"))
            else:
                main_links.append(("anpr", "tx-track-bus"))

        # create wiegand unit
        wiegand_conf = lpr_pri.get('wiegand-conf', None)
        if wiegand_conf:
            wiegand_jnode = {
                "type": "board",
                "name": "board",
                "conf": wiegand_conf
            }

            main_conf['nodes'].append({
                "bundle": ["libunity_wiegand", "1.0.0"],
                "type": "board",
                "name": "board",
                "units": [
                    wiegand_jnode
                ]
            })

            main_links.append(("anpr", "board"))
            main_links.append(("board", "counters"))

        # create wiegand-udp unit
        wiegand_udp_conf = lpr_pri.get('wiegand-udp-conf', None)
        if wiegand_udp_conf:
            wiegand_udp_jnode = {
                "type": "board-udp",
                "name": "board-udp",
                "conf": wiegand_udp_conf
            }

            main_conf['nodes'].append({
                "bundle": ["libunity_wiegand", "1.0.0"],
                "type": "board-udp",
                "name": "board-udp",
                "units": [
                    wiegand_udp_jnode
                ]
            })

            main_links.append(("anpr", "board-udp"))
            main_links.append(("board-udp", "counters"))

        tcp_index = 0
        for tcp_recipient_uimod in self._tcp_recipient_uimods:
            tcp_index += 1
            tcp_recipient_conf = uimod_to_newconf_map.get(tcp_recipient_uimod) or tcp_recipient_uimod.load_conf()
            tcp_recipient_pub = tcp_recipient_conf['rw']

            tcp_jnode = {
                "type": "recipient",
                "name": tcp_recipient_uimod.ref,
                "conf": tcp_recipient_pub['conf']
            }

            main_conf['nodes'].append({
                "bundle": ["libunity_recipient", "1.0.0"],
                "type": "vast2",
                "name": "vast2-{}".format(tcp_index),
                "units": [
                    tcp_jnode
                ]
            })

            main_links.append(("anpr", tcp_recipient_uimod.ref))
            main_links.append((tcp_recipient_uimod.ref, "counters"))

        if len(self._ledar_uimods) == 1:
            leddar_uimod = self._ledar_uimods[0]
            leddar_conf = uimod_to_newconf_map.get(leddar_uimod) or leddar_uimod.load_conf()
            leddar_conn_info = leddar_conf['ro']['connection']
            leddar_conf_conf = leddar_conf['rw']['conf']
            leddar_aquisition = leddar_conf_conf['acquisition']
            leddar_loops = leddar_conf_conf['object-detectors']
            for leddar_loop in leddar_loops:
                leddar_loop['min-score'] = 1

            leddar_pub = {
                'conf': {
                    "driver": {
                        "connection": leddar_conn_info,
                        "dist-thr": 5.1,
                        "acquisition": leddar_aquisition,
                    },
                    "bg-detector": {
                        "beam-angle": 45.1,
                        "beam-seg-count": 16,
                        "beam-depth-map": [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1 ]
                    },
                    "loop-detector": {
                        "changelog": {
                            "entry-count": 1,
                            "entry-age": 300
                        },
                        "loops": leddar_loops
                    },
                    "classifier": {
                        "bike": [80, 220],
                        "car": [220, 560]
                    },
                    "scene-event": {
                        "endpoints": [
                            "/tmp/lg0-tr0-scene"
                        ]
                    },
                    "track-event": {
                        "endpoints": [
                            fs.join_path(spool_path, 'lg{0}-sc{0}-track'.format(self._lpr_uimod.id))
                        ]
                    }}
            }

            leddar_jnode = {
                "type": "leddar",
                "name": "leddar",
                "conf": leddar_pub['conf']
            }

            main_conf['nodes'].append({
                "bundle": ["libunity_leddar", "1.0.0"],
                "type": "leddar",
                "name": "leddar",
                "units": [
                    leddar_jnode
                ]
            })

            main_links.append(("media", "leddar"))
            main_links.append(("leddar", "counters"))

        self._main_conf = main_conf

    def complete(self):
        lr_dgmod = self._lpr_dgmod

        fs.silent_mkdir_r(fs.join_path(lr_dgmod.conf_dir, 'unit.conf.d'))
        conf_path = fs.join_path(lr_dgmod.conf_dir, 'unit.conf.d', 'main.conf')
        fs.write_json_file(self._main_conf, conf_path, pretty=True)
        lr_dgmod.save_setupstore_conf(active=True)

    def rollback(self):
        pass

    def notify(self):
        """
        :rtype: None
        """
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class LrDeleter:
    def __init__(self, dgmod):
        """
        :type dgmod: dg.DgMod
        """
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod):
        """
        :type entry_uimod: ph.UiMod
        :rtype: LrDeleter | None
        """
        if entry_uimod.type == ph.UI_TYPE_LPR:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_LPR, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self):
        self._dgmod.uninstall()
