from itertools import ifilter

from edge.com import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator


# noinspection PyUnusedLocal
def mc_get_creators(entry_uimod, all_used_uimods, uimod_to_newconf_map, forced=False):
    """
    :type entry_uimod: ph.UiMod | ph.UiMod2
    :type all_used_uimods: set[ph.UiMod, ph.UiMod2]
    :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
    :type uimod_to_newconf_map: dict
    :type forced: bool
    :rtype: list of BaseCreator
    """
    if entry_uimod.type not in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA]:
        return []

    try:
        cam_uimod = next(ifilter(lambda x: x.type in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA], all_used_uimods))

        activate = any([
            uimod_to_newconf_map.get(cam_uimod)
        ])

        if forced or activate:
            return [McGstCreator(cam_uimod)]

    except StopIteration:
        pass

    return []


class McGstCreator(BaseCreator):
    def __init__(self, cam_uimod):
        """
        :type cam_uimod: ph.UiMod
        """
        BaseCreator.__init__(self)
        self._cam_uimod = cam_uimod
        self._mc_dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_MEDIACODEC_GST, cam_uimod.pid, cam_uimod.id)
        self._main_ss = None

    @property
    def dgmod(self):
        """
        :rtype: dg.DgMod
        """
        return self._mc_dgmod

    def prepare(self, uimod_to_newconf_map):
        """
        :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
        :type uimod_to_newconf_map: dict
        """
        ui_pid = self._cam_uimod.pid
        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()
        spool_dir = ph.get_spool_dir(ui_pid)

        quality = 30000
        if cam_uiconf:
            quality = cam_uiconf['ro'].get('h264Quality', 30000)

        # create main.conf
        main_conf = {
            "Store": [self._mc_dgmod.id, '0'],
            "React": fs.join_path(spool_dir, 'mg{0}-mcgst{0}-scene'.format(self._cam_uimod.id)),
            "codec": {
                "Scale": {
                    "MaxWidth": 65535,
                    "MaxHeight": 65535,
                    "Interpolation": 1
                },
                "Encode": {
                    "Format": "H264",
                    "UsedThreadMax": 1,
                    "Quality": quality
                }
            }
        }

        rgbs = cam_uiconf['ro'].get("rgbs", False)
        mcgst_scene = cam_uiconf['ro'].get("mcgst-scene", True)

        ss_active = mcgst_scene and not rgbs
        self._main_ss = main_conf, ss_active

    def complete(self):
        main_conf, ss_active = self._main_ss  # type: dict, bool

        edge_mc = self._mc_dgmod
        edge_mc.install()
        edge_mc.save_main_conf(main_conf)
        edge_mc.save_setupstore_conf(ss_active)

    def rollback(self):
        pass

    def notify(self):
        """
        :rtype: None
        """
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class McDeleter:
    def __init__(self, dgmod):
        """
        :type dgmod: dg.DgMod
        """
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod):
        """
        :type entry_uimod: ph.UiMod
        :rtype: McDeleter | None
        """
        if entry_uimod.type in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA]:
            dgmod = dg.DgMod.create(
                dg.EDGE_TYPE_MEDIACODEC_GST, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self):
        self._dgmod.uninstall()
