from itertools import ifilter

from edge import eventlog
from edge.com import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator


# noinspection PyUnusedLocal
def mg_get_creators(entry_uimod, all_used_uimods, uimod_to_newconf_map, forced=False):
    """
    :type entry_uimod: ph.UiMod | ph.UiMod2
    :type all_used_uimods: set[ph.UiMod, ph.UiMod2]
    :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
    :type uimod_to_newconf_map: dict
    :type forced: bool
    :rtype: list of BaseCreator
    """
    if entry_uimod.type != ph.UI_TYPE_MV_CAMERA:
        return []

    try:
        cam_uimod, fusion_uimod, lpr_uimod, ra_uimod, rl_uimods , rc_uimods = (
            next(ifilter(lambda x: x.type == ph.UI_TYPE_MV_CAMERA, all_used_uimods)),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods)),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_LPR, all_used_uimods)),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_RADAR, all_used_uimods), None),
            filter(lambda x: x.type == ph.UI_TYPE_RED_LIGHT, all_used_uimods),
            filter(lambda x: x.type == ph.UI_TYPE_RAILROAD_CROSSING, all_used_uimods))

        activate = any([
            uimod_to_newconf_map.get(cam_uimod),
            any(map(lambda x: uimod_to_newconf_map.get(x), rl_uimods)),
            any(map(lambda x: uimod_to_newconf_map.get(x), rc_uimods))
        ])
        return [MgCreator(cam_uimod, lpr_uimod, fusion_uimod, rl_uimods, rc_uimods, ra_uimod)]

    except StopIteration:
        pass

    return []


class MgCreator(BaseCreator):
    def __init__(self, cam_uimod, lpr_uimod, fusion_uimod, rl_uimods, rc_uimods, ra_uimod):
        """
        :type cam_uimod: ph.UiMod
        :type lpr_uimod: ph.UiMod
        :type fusion_uimod: ph.UiMod
        :type rl_uimods: list[ph.UiMod]
        :type rc_uimods: list[ph.UiMod]
        :type ra_uimod: ph.UiMod | None
        """
        BaseCreator.__init__(self)
        self._cam_uimod = cam_uimod
        self._lpr_uimod = lpr_uimod
        self._fusion_uimod = fusion_uimod
        self._rl_uimods = rl_uimods
        self._rc_uimods = rc_uimods
        self._ra_uimod = ra_uimod
        self._edge_mg = dg.DgMod.create(
            dg.EDGE_TYPE_MEDIAGRABBER, cam_uimod.pid, cam_uimod.id)
        self._main_misc_halt = None

    @property
    def dgmod(self):
        """
        :rtype: dg.DgMod
        """
        return self._edge_mg

    def prepare(self, uimod_to_newconf_map):
        """
        :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
        :type uimod_to_newconf_map: dict
        """
        ui_pid = self._cam_uimod.pid
        spool_dir = ph.get_spool_dir(ui_pid)

        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()

        if self._ra_uimod:
            tx_pipes = [fs.join_path(spool_dir, 'mg{}-ra{}-pipe'.format(
                self._cam_uimod.id, self._ra_uimod.id
            ))]
        else:
            tx_pipes = []

        ui_acquisition_jnode = cam_uiconf['rw']['videostream']
        ui_sensor_jnode = ui_acquisition_jnode['resolution']
        ui_control_jnode = ui_acquisition_jnode['control']

        #
        exp_time_limits = [0, 0]
        gain_limits = [0, 0]
        gain = 0
        exp_time = 0
        mg_control_mode = 'Auto'
        if ui_control_jnode['profile'] == 'manual':
            mg_control_mode = 'Manual'

        if mg_control_mode == 'Auto':
            gain_limits = [ui_control_jnode['gain-auto-min'], ui_control_jnode['gain-auto-max']]
            exp_time_limits = [ui_control_jnode['exposure-auto-min'], ui_control_jnode['exposure-auto-max']]

        else:
            gain = ui_control_jnode['gain-manual']
            exp_time = ui_control_jnode['exposure-manual']

        lib_path = ''
        if cam_uiconf['rw']['vendor'] == 'Basler':
            lib_path = '/opt/edge/libexec/mediagrabber/libmgbasler.so'

        elif cam_uiconf['rw']['vendor'] == 'Teledyne DALSA':
            lib_path = '/opt/edge/libexec/mediagrabber/libdriver-dalsa.so'

        elif cam_uiconf['rw']['vendor'] == 'PointGrey':
            lib_path = '/opt/edge/libexec/mediagrabber/libdriver-pointgrey.so'

        elif cam_uiconf['rw']['vendor'] == 'NPK ES-Experts':
            lib_path = '/opt/edge/libexec/mediagrabber/libdriver-ese.so'

        elif cam_uiconf['rw']['vendor'] == 'JAI Corporation':
            lib_path = '/opt/edge/libexec/mediagrabber/libdriver-jai.so'

        fps = ui_acquisition_jnode['fps']
        main_conf = {
            "FileApi": {
                "Store": fs.join_path(spool_dir, 'mg{0}-mg{0}-store-raw'.format(self._cam_uimod.id)),
                "FileCountLow": fps * 2,
                "FileCountHigh": fps * 2 + 5,
                "SceneEvent": {
                    "Dirs": [],
                    "PixelFormat": cam_uiconf['ro']['raw-format']
                },
                "HaltEvent": {
                    "SignalAbsentTout": 3000,
                    "DeviceAbsentTout": 60000
                }
            },
            "TxPipes": tx_pipes,
            "Driver": {
                "Vendor": cam_uiconf['rw']['vendor'],
                "Model": cam_uiconf['rw']['model'],
                "Id": cam_uiconf['rw']['serial'],
                "File": lib_path,
                "ParamSet": {
                    "VideoSource": {
                        "AutoSetup": {
                            "ExpTimeLimits": exp_time_limits,
                            "Target": {
                                "Brightness": 0.50196
                            },
                            "GainLimits": gain_limits
                        },
                        "ManSetup": {
                            "Lut12": "/opt/edge/share/setupstore/edge/mediagrabber/lut12.default.conf",
                            "GainAll": gain,
                            "ExpTime": exp_time
                        },
                        "Sweep": {
                            "Profile": "MinGain",
                            "ExpTimeMode": mg_control_mode,
                            "GainMode": mg_control_mode,
                            "LutMode": "None"
                        },
                        "Acquisition": {
                            "Sensor": {
                                "Width": ui_sensor_jnode['w'],
                                "OffsetX": ui_sensor_jnode['x'],
                                "OffsetY": ui_sensor_jnode['y'],
                                "Height": ui_sensor_jnode['h']
                            },
                            "Gamma": -1,
                            "Fps": fps,
                            "Rotate180": ui_acquisition_jnode['rotate180']
                        }
                    },
                    "IoBox": [],
                    "Misc": {
                        "TransferDataSize": cam_uiconf['ro']['transfer-data-size']
                    }
                }
            },
            "Monitoring": {
                "StatTout": 5,
                "StatPath": fs.join_path(spool_dir, 'zabbix', 'vs0')
            }
        }

        if cam_uiconf['ro']['raw-format'] == 'MONO8T':
            main_conf['FileApi']['SceneEvent']['Dirs'] = [
                fs.join_path(spool_dir, 'mg{}-lr{}-scene'.format(
                    self._cam_uimod.id, self._lpr_uimod.id
                )),
                fs.join_path(spool_dir, 'mg{}-tr{}-scene'.format(
                    self._cam_uimod.id, self._fusion_uimod.id
                ))
            ]
            # enable mediacodecgst
            main_conf['FileApi']['SceneEvent']['Dirs'].append(
                fs.join_path(spool_dir, 'mg{0}-mcgst{0}-scene'.format(self._cam_uimod.id)))

            # enable gc
            if ui_control_jnode['profile'] == 'auto-70-150':
                main_conf['FileApi']['SceneEvent']['Dirs'].append(
                    fs.join_path(spool_dir, 'mg{0}-gc{0}-scene'.format(self._cam_uimod.id)))

        else:
            main_conf['FileApi']['SceneEvent']['Dirs'] = [
                fs.join_path(spool_dir, 'mg{0}-rc{0}-scene'.format(self._cam_uimod.id))
            ]

        # add mg to rl path
        if len(self._rl_uimods) == 1:
            rl_uimod = self._rl_uimods[0]
            red_light_jnode = uimod_to_newconf_map.get(rl_uimod) or rl_uimod.load_conf()
            vendor = red_light_jnode['ro'].get('vendor', 'zonechecker')
            if vendor == 'zonechecker':
                main_conf['FileApi']['SceneEvent']['Dirs'].append(
                    fs.join_path(spool_dir, 'mg{}-zc{}-scene'.format(
                        self._cam_uimod.id, rl_uimod.id
                    )))
        elif len(self._rc_uimods) == 1:
            rc_uimod = self._rc_uimods[0]
            railroad_crossing_jnode = uimod_to_newconf_map.get(rc_uimod) or rc_uimod.load_conf()
            vendor = railroad_crossing_jnode['ro'].get('vendor', 'zonechecker')
            if vendor == 'zonechecker':
                main_conf['FileApi']['SceneEvent']['Dirs'].append(
                    fs.join_path(spool_dir, 'mg{}-zc{}-scene'.format(
                        self._cam_uimod.id, rc_uimod.id
                    )))

    # event-dirs from 'ro'
        try:
            extra_dirs_jnode = cam_uiconf['ro']['main']['FileApi']['SceneEvent']['Dirs']
        except KeyError:
            extra_dirs_jnode = None

        if type(extra_dirs_jnode) == list:
            main_conf['FileApi']['SceneEvent']['Dirs'].extend(extra_dirs_jnode)

        elif type(extra_dirs_jnode) == unicode or type(extra_dirs_jnode) == str:
            main_conf['FileApi']['SceneEvent']['Dirs'].append(extra_dirs_jnode)

        elif extra_dirs_jnode is not None:
            eventlog.log_d('wrong ui-jnode#/ro/FileApi/SceneEvent/Dirs')

        ir_jnode = cam_uiconf['ro'].get('ir', None)
        if ir_jnode is not None:
            ir = {
                "Source": ir_jnode.get('source', "ExposureActive"),
                "Mode": "Output",
                "Pin": ir_jnode['pin'],
                "Inverter": ir_jnode['sync-inverter']
            }
            main_conf['Driver']['ParamSet']['IoBox'].append(ir)

        # create haltaction.conf

        halt_lines = cam_uiconf['ro'].get('haltaction')
        if halt_lines is None:
            halt_lines = [
                "0:python2 /opt/edge/bin/psuctl run led-camera color G blink 0HZ;",

                "1:python2 /opt/edge/bin/psuctl run led-camera color R blink 1HZ;"
                "python2 /opt/edge/bin/psuctl run camera-reboot",

                "2:python2 /opt/edge/bin/psuctl run led-camera color R blink 0HZ;"
                "python2 /opt/edge/bin/psuctl run camera-reboot"
            ]

        halt_conf = '\n'.join(halt_lines) if halt_lines else ''

        misc_conf = cam_uiconf['ro']['misc']
        if misc_conf.get('TsAdjNs') is not None:
            misc_conf['TsAdjNs'] = 0

        self._main_misc_halt = main_conf, misc_conf, halt_conf

    def complete(self):
        main_conf, misc_conf, halt_conf = self._main_misc_halt  # type: dict, dict, str

        edge_mg = self._edge_mg
        edge_mg.install()
        edge_mg.save_main_conf(main_conf)
        edge_mg.save_misc_conf(misc_conf)
        edge_mg.save_haltaction_conf(halt_conf)
        edge_mg.save_setupstore_conf(active=True)

    def rollback(self):
        pass

    def notify(self):
        """
        :rtype: None
        """
        dg.LifeCycle.sigint_dgmod(self.dgmod.type, self.dgmod.id)


class MgDeleter:
    def __init__(self, dgmod):
        """
        :type dgmod: dg.DgMod
        """
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod):
        """
        :type entry_uimod: ph.UiMod
        :rtype: MgDeleter | None
        """
        if entry_uimod.type == ph.UI_TYPE_MV_CAMERA:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_MEDIAGRABBER, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self):
        self._dgmod.uninstall()
