from itertools import ifilter

from edge.com import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator


# noinspection PyUnusedLocal
def rc_get_creators(entry_uimod, all_used_uimods, uimod_to_newconf_map, forced=False):
    """
    :type entry_uimod: ph.UiMod | ph.UiMod2
    :type all_used_uimods: set[ph.UiMod, ph.UiMod2]
    :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
    :type uimod_to_newconf_map: dict
    :type forced: bool
    :rtype: list of BaseCreator
    """
    if entry_uimod.type != ph.UI_TYPE_MV_CAMERA:
        return []

    try:
        cam_uimod, lpr_uimod, fusion_uimod = (
            next(ifilter(lambda x: x.type == ph.UI_TYPE_MV_CAMERA, all_used_uimods)),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_LPR, all_used_uimods)),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods)))

        activate = any([
            uimod_to_newconf_map.get(cam_uimod)
        ])

        if forced or activate:
            return [RcCreator(cam_uimod, lpr_uimod, fusion_uimod)]

    except StopIteration:
        pass

    return []


class RcCreator(BaseCreator):
    def __init__(self, cam_uimod, lpr_uimod, fusion_uimod):
        """
        :type cam_uimod: ph.UiMod
        :type lpr_uimod: ph.UiMod
        :type fusion_uimod: ph.UiMod
        """
        BaseCreator.__init__(self)
        self._cam_uimod = cam_uimod
        self._lpr_uimod = lpr_uimod
        self._fusion_uimod = fusion_uimod
        self._rc_dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_RAWCODEC, cam_uimod.pid, cam_uimod.id)
        self._main_ss = None

    @property
    def dgmod(self):
        """
        :rtype: dg.DgMod
        """
        return self._rc_dgmod

    def prepare(self, uimod_to_newconf_map):
        """
        :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
        :type uimod_to_newconf_map: dict
        """
        ui_pid = self._cam_uimod.pid
        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()
        spool_dir = ph.get_spool_dir(ui_pid)

        # create main.conf
        main_conf = {
            "FileApi": {
                "Store": fs.join_path(spool_dir, 'rc{0}-rc{0}-store'.format(self._cam_uimod.id)),
                "FileCountLow": 18,
                "FileCountHigh": 20,
                "AddSceneReact": {
                    "Dir": fs.join_path(spool_dir, 'mg{0}-rc{0}-scene'.format(self._cam_uimod.id))
                },
                "SceneEvent": [
                    fs.join_path(spool_dir, 'rc{}-lr{}-scene'.format(
                        self._cam_uimod.id, self._lpr_uimod.id
                    )),
                    fs.join_path(spool_dir, 'rc{}-tr{}-scene'.format(
                        self._cam_uimod.id, self._fusion_uimod.id
                    )),
                    fs.join_path(spool_dir, 'mg{0}-mcgst{0}-scene'.format(self._cam_uimod.id))
                ]
            },
            "codec": {
                "Scale": {
                    "MaxWidth": 0,
                    "MaxHeight": 0,
                    "Interpolation": 1
                },
                "Encode": {
                    "UsedThreadMax": 1
                }
            }
        }

        # create setupstore.conf
        if cam_uiconf['ro']['raw-format'] == 'MONO8T':
            ss_active = False
        else:
            # enable gc
            if cam_uiconf['rw']['videostream']['control']['profile'] == 'auto-70-150':
                main_conf['FileApi']['SceneEvent'].append(
                    fs.join_path(spool_dir, 'mg{0}-gc{0}-scene'.format(self._cam_uimod.id)))
            ss_active = True

        self._main_ss = main_conf, ss_active

    def complete(self):
        main_conf, ss_active = self._main_ss  # type:  dict, bool

        edge_rc = self._rc_dgmod
        edge_rc.install()
        edge_rc.save_main_conf(main_conf)
        edge_rc.save_setupstore_conf(ss_active)

    def rollback(self):
        pass

    def notify(self):
        """
        :rtype: None
        """
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class RcDeleter:
    def __init__(self, dgmod):
        """
        :type dgmod: dg.DgMod
        """
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod):
        """
        :type entry_uimod: ph.UiMod
        :rtype: RcDeleter | None
        """
        if entry_uimod.type == ph.UI_TYPE_MV_CAMERA:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_RAWCODEC, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self):
        self._dgmod.uninstall()

