from itertools import ifilter

from edge.com import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator
from edge.webserver.sysconf.uiconf.time import TimeUiconf


# noinspection PyUnusedLocal
def or_get_creators(uimod, all_used_uimods, uimod_to_newconf_map, forced=False):
    """
    :type uimod: ph.UiMod | ph.UiMod2
    :type all_used_uimods: set[ph.UiMod, ph.UiMod2]
    :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
    :type uimod_to_newconf_map: dict
    :type forced: bool
    :rtype: list of BaseCreator
    """
    if uimod.type not in [ph.UI_TYPE_HTTP_STREAMING, ph.UI_TYPE_RTSP_STREAMING]:
        return []

    try:
        ra_uimod, rstreaming_uimod, hstreaming_uimod, fusion_uimod, ws_uimod, term_uimod, gnss_uimod, cam_uimod, \
            roadway_uimod, location_uimod = (
            next(ifilter(lambda x: x.type == ph.UI_TYPE_RADAR, all_used_uimods), None),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_RTSP_STREAMING, all_used_uimods), None),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_HTTP_STREAMING, all_used_uimods), None),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods)),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_WEBSERVER, all_used_uimods), None),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_TERMINAL, all_used_uimods)),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_GNSS, all_used_uimods), None),
            next(ifilter(lambda x: x.type in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA], all_used_uimods), None),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_ROADWAY, all_used_uimods), None),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_LOCATION, all_used_uimods), None))

        spotspeed_uimods = filter(lambda x: x.type == ph.UI_TYPE_SPOT_SPEED, all_used_uimods)
        spotspeed_uimod = None
        if len(spotspeed_uimods) == 1:
            spotspeed_uimod = spotspeed_uimods[0]

        if uimod.type == ph.UI_TYPE_HTTP_STREAMING:
            streaming_uimod = hstreaming_uimod

        elif uimod.type == ph.UI_TYPE_RTSP_STREAMING:
            streaming_uimod = rstreaming_uimod

        else:
            raise NotImplementedError

        activate = any([
            uimod_to_newconf_map.get(streaming_uimod),
            uimod_to_newconf_map.get(gnss_uimod),
            uimod_to_newconf_map.get(cam_uimod),
            uimod_to_newconf_map.get(roadway_uimod),
            uimod_to_newconf_map.get(term_uimod),
            uimod_to_newconf_map.get(spotspeed_uimod),
            uimod_to_newconf_map.get(location_uimod)
        ])

        if forced or activate:
            creator = _Creator(
                streaming_uimod, fusion_uimod, gnss_uimod, roadway_uimod, term_uimod, location_uimod, cam_uimod,
                webserver_uimod=ws_uimod, radar_uimod=ra_uimod, spotspeed_uimod=spotspeed_uimod)

            return [creator]

    except StopIteration:
        pass

    return []


class _Creator(BaseCreator):
    def __init__(self, streaming_uimod, fusion_uimod, gnss_uimod, roadway_uimod, terminal_uimod, location_uimod, cam_uimod=None,
                 webserver_uimod=None, radar_uimod=None, spotspeed_uimod=None):
        """
        :type streaming_uimod: ph.UiMod
        :type fusion_uimod: ph.UiMod
        :type gnss_uimod: ph.UiMod
        :type roadway_uimod: ph.UiMod
        :type terminal_uimod: ph.UiMod
        :type location_uimod: ph.UiMod | None
        :type cam_uimod: ph.UiMod | None
        :type radar_uimod: ph.UiMod | None
        :type spotspeed_uimod: ph.UiMod | None
        :type webserver_uimod: ph.UiMod | None
        """
        BaseCreator.__init__(self)
        self._streaming_uimod = streaming_uimod
        self._fusion_uimod = fusion_uimod
        self._gnss_uimod = gnss_uimod
        self._roadway_uimod = roadway_uimod
        self._terminal_uimod = terminal_uimod
        self._webserver_uimod = webserver_uimod
        self._radar_uimod = radar_uimod
        self._spotspeed_uimod = spotspeed_uimod
        self._location_uimod = location_uimod
        self._cam_uimod = cam_uimod
        self._main_ss = None

        self._rn_dgmod = dg.DgMod.create(dg.EDGE_TYPE_RENDERER, self._streaming_uimod.pid, self._streaming_uimod.id)

    @property
    def dgmod(self):
        """
        :rtype: dg.DgMod
        """
        return self._rn_dgmod

    def prepare(self, uimod_to_newconf_map):
        """
        :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
        :type uimod_to_newconf_map: dict
        """
        spool_dir = ph.get_spool_dir(self._streaming_uimod.pid)

        if self._streaming_uimod.type == ph.UI_TYPE_RTSP_STREAMING:
            codec = "RGBA"
            scene_dir = fs.join_path(spool_dir, 'or{}-rs{}-scene'.format(
                self._streaming_uimod.id, self._streaming_uimod.id))

        elif self._streaming_uimod.type == ph.UI_TYPE_HTTP_STREAMING:
            codec = "JPEG"
            scene_dir = fs.join_path(spool_dir, 'or{}-ws{}-scene'.format(
                self._streaming_uimod.id, self._webserver_uimod.id))

        else:
            raise NotImplementedError

        source_dir = fs.join_path(spool_dir, 'tr{}-or{}-scene'.format(
            self._fusion_uimod.id, self._streaming_uimod.id
        ))

        roadway_uiconf = uimod_to_newconf_map.get(self._roadway_uimod) or self._roadway_uimod.load_conf()
        roadway_pub = roadway_uiconf['rw']
        roadway_args = roadway_pub['args']

        streaming_uiconf = uimod_to_newconf_map.get(self._streaming_uimod) or self._streaming_uimod.load_conf()
        streaming_pub = streaming_uiconf['rw']  # type: dict
        streaming_live = streaming_pub['live']  # type: dict
        streaming_vs = streaming_live['videostream']  # type: dict
        streaming_v = streaming_live['renderer']['vehicle']  # type: dict

        image_w = streaming_vs['frame-w']
        main_conf = {
            "Source": {
                "Url": source_dir
            },
            "services": {
                'camera': self._cam_uimod.conf_path if self._cam_uimod else None,
                'origin': ph.TermAboutUiconf.get_conf_path(self._streaming_uimod.pid),
                "datetime": TimeUiconf.CONF_PATH,
                "geolocation": self._gnss_uimod.conf_path,
                "monitoring": {
                    "path": "/var/spool/edge/pipeline-{}/zabbix/or{}".format(
                        self._streaming_uimod.pid, self._streaming_uimod.id),
                    "tout": 30
                }
            },
            "Renderer": {
                "profile": {
                    "width": image_w,
                    "annotation": {
                        "vehicle": {
                            "lr": {
                                "plate": streaming_v.get('video-plate', 'never'),
                                "dir": streaming_v.get('video-dir', 'never'),
                                "lane": streaming_v.get('video-lane', 'never'),
                                "trace": streaming_v.get('video-plate-trace', 'never'),
                                "center": streaming_v.get('video-plate-center', 'never')
                            }

                        }
                    },
                    "footer": streaming_live['footer']
                }
            },
            "image-store": {
                "format": codec,
                "endpoint": "/var/spool/edge/pipeline-{0}/or{1}-or{1}-store/".format(
                    self._streaming_uimod.pid, self._streaming_uimod.id),
                "max-count": 8
            },
            "scene-event": {
                "max-file-count": 8,
                "endpoints": [scene_dir]
            }
        }
        if False:
            main_conf['tl-args'] = [
                "/var/spool/edge/pipeline-%s/tl%s-or%s-track/" % (self._streaming_uimod.pid,
                                                                  self._streaming_uimod.id,
                                                                  self._streaming_uimod.id), 1
            ]

        if self._location_uimod:
            main_conf['services']['location'] = self._location_uimod.conf_path

        if self._radar_uimod:
            main_conf['Renderer']['profile']['annotation']['vehicle']['ra'] = {
                "target": streaming_v.get('radar-target', 'never'),
                "vehicle": streaming_v.get('radar-vehicle', 'never'),
                "speed": streaming_v.get('radar-speed', 'never'),
                "len": streaming_v.get('radar-length', 'never'),
                "dir": streaming_v.get('radar-dir', 'never')
            }

        if self._spotspeed_uimod is not None:
            spotspeed_limits_uiconf = uimod_to_newconf_map.get(self._spotspeed_uimod) or self._spotspeed_uimod.load_conf()
            spot_speed_rules = spotspeed_limits_uiconf['rw']['args']['rules']
            # do not render overspeed if there is > 1 spot-speed rules
            if len(spot_speed_rules) == 1 and spot_speed_rules[0]['enabled']:
                speed_limit = spot_speed_rules[0]['limits']['speed-limit']
                max_overspeed = spot_speed_rules[0]['limits']['max-overspeed']
                main_conf['Renderer']['YellowSpeedThreshold'] = speed_limit
                main_conf['Renderer']['RedSpeedThreshold'] = speed_limit + max_overspeed

        if len(roadway_args['lanes']) > 0 and streaming_v['video-lane'] != 'never':
            main_conf['road-mask'] = {
                'lanes': roadway_args['lanes'],
                'scene-wh': roadway_args['scene-wh']
            }

        ss_active = bool(streaming_uiconf['rw']['enabled'])
        self._main_ss = main_conf, ss_active

    def complete(self):
        rn_dgmod = self.dgmod
        main_conf, ss_active = self._main_ss  # type: dict, bool
        rn_dgmod.install()
        rn_dgmod.save_main_conf(main_conf)
        rn_dgmod.save_setupstore_conf(ss_active)

    def rollback(self):
        pass

    def notify(self):
        """
        :rtype: None
        """
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class OrDeleter:
    def __init__(self, dgmod):
        """
        :type dgmod: dg.DgMod
        """
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod):
        """
        :type entry_uimod: ph.UiMod
        :rtype: OrDeleter
        """
        if entry_uimod.type in [ph.UI_TYPE_HTTP_STREAMING, ph.UI_TYPE_RTSP_STREAMING]:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_RENDERER, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self):
        self._dgmod.uninstall()
