from itertools import ifilter

from edge.com import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator


# noinspection PyUnusedLocal
def rg_get_creators(entry_uimod, all_used_uimods, uimod_to_newconf_map, forced=False):
    """
    :type entry_uimod: ph.UiMod | ph.UiMod2
    :type all_used_uimods: set[ph.UiMod, ph.UiMod2]
    :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
    :type uimod_to_newconf_map: dict
    :type forced: bool
    :rtype: list of BaseCreator
    """
    if entry_uimod.type != ph.UI_TYPE_RTSP_CAMERA:
        return []

    try:
        cam_uimod, fusion_uimod, lpr_uimod, rl_uimods, railroad_crossing_uimods = (
            next(ifilter(lambda x: x.type == ph.UI_TYPE_RTSP_CAMERA, all_used_uimods)),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods)),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_LPR, all_used_uimods)),
            filter(lambda x: x.type == ph.UI_TYPE_RED_LIGHT, all_used_uimods),
            filter(lambda x: x.type == ph.UI_TYPE_RAILROAD_CROSSING, all_used_uimods))

        activate = any([
            uimod_to_newconf_map.get(cam_uimod),
            any(map(lambda x: uimod_to_newconf_map.get(x), rl_uimods)),
            any(map(lambda x: uimod_to_newconf_map.get(x), railroad_crossing_uimods))
        ])

        return [RgCreator(cam_uimod, lpr_uimod, fusion_uimod, rl_uimods, railroad_crossing_uimods)]

    except StopIteration:
        pass

    return []


class RgCreator(BaseCreator):
    def __init__(self, cam_uimod, lpr_uimod, fusion_uimod, rl_uimods, railroad_crossing_uimods):
        """
        :type cam_uimod: ph.UiMod
        :type lpr_uimod: ph.UiMod
        :type fusion_uimod: ph.UiMod
        :type rl_uimods: list[ph.UiMod]
        :type railroad_crossing_uimods: list[ph.UiMod]
        """
        BaseCreator.__init__(self)
        self._cam_uimod = cam_uimod
        self._lpr_uimod = lpr_uimod
        self._fusion_uimod = fusion_uimod
        self._rl_uimods = rl_uimods
        self._railroad_crossing_uimods = railroad_crossing_uimods
        self._rg_dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_RTSP_GRABBER, cam_uimod.pid, cam_uimod.id)
        self._main_halt = None

    @property
    def dgmod(self):
        """
        :rtype: dg.DgMod
        """
        return self._rg_dgmod

    def prepare(self, uimod_to_newconf_map):
        """
        :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
        :type uimod_to_newconf_map: dict
        """
        ui_pid = self._cam_uimod.pid
        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()

        use_udp = False
        transport = cam_uiconf['rw']['conf']['video'].get('transport', ["RTP/AVP/UDP", "RTP/AVP/TCP"])
        if "RTP/AVP/UDP" in transport:
            use_udp = True

        skip_rate = cam_uiconf['rw']['conf']['video'].get('skip-rate', 1) - 1
        spool_dir = ph.get_spool_dir(ui_pid)

        ui_format_to_edge_format = {
            'video/H264': 'h264',
            'video/JPEG': 'jpeg'
        }
        use_gpu = 'hw/auto' == cam_uiconf['rw']['conf']['video'].get('decoder', 'hw/auto')
        # create main.conf
        main_conf = {
            "InputStreamType": ui_format_to_edge_format.get(cam_uiconf['rw']['conf']['video']['format']),
            "UseUDP": use_udp,
            "SkipRate": skip_rate,
            "TargetHeight": 0,
            "Url": cam_uiconf['rw']['conf']['url'],
            "TargetWidth": 0,
            "UseGpu": use_gpu,
            "Store": {
                "Directory": fs.join_path(spool_dir, 'rg{0}-rg{0}-store'.format(self._cam_uimod.id)),
                "FileCountLow": 40,
                "FileCountHigh": 45
            },
            "FileApi": {
                "SceneEventMono8t": [],
                "SceneEventRaw": [],
                "HaltEvent": {
                    "SignalAbsentTout": 5000,
                    "DeviceAbsentTout": 600000
                },
            },
            "Monitoring": {
                "StatTout": 5,
                "StatPath": fs.join_path(spool_dir, 'zabbix', 'vs0')
            }
        }
        rgbs = cam_uiconf['ro'].get("rgbs", False)
        mcgst_scene = cam_uiconf['ro'].get("mcgst-scene", True)

        if rgbs:
            main_conf['StoreLbsv'] = [self._rg_dgmod.id, '0']

        color_mode_grayscale = cam_uiconf['rw']['conf']['video'].get('grayscale', False)
        if color_mode_grayscale:
            main_conf['FileApi']['SceneEventMono8t'].append(
                fs.join_path(spool_dir, 'rg{}-lr{}-scene'.format(
                    self._cam_uimod.id, self._lpr_uimod.id
                )))

            main_conf['FileApi']['SceneEventMono8t'].append(
                fs.join_path(spool_dir, 'rg{}-tr{}-scene'.format(
                    self._cam_uimod.id, self._fusion_uimod.id
                )))

            if mcgst_scene and not rgbs:
                main_conf['FileApi']['SceneEventMono8t'].append(
                    fs.join_path(spool_dir, 'mg{0}-mcgst{0}-scene'.format(self._cam_uimod.id)))

        else:
            main_conf['FileApi']['SceneEventRaw'].append(
                fs.join_path(spool_dir, 'rg{}-lr{}-scene'.format(
                    self._cam_uimod.id, self._lpr_uimod.id
                )))

            main_conf['FileApi']['SceneEventRaw'].append(
                fs.join_path(spool_dir, 'rg{}-tr{}-scene'.format(
                    self._cam_uimod.id, self._fusion_uimod.id
                )))

            if mcgst_scene and not rgbs:
                main_conf['FileApi']['SceneEventRaw'].append(
                    fs.join_path(spool_dir, 'mg{0}-mcgst{0}-scene'.format(self._cam_uimod.id)))

        # add mg to rl path
        if len(self._rl_uimods) == 1:
            rl_uimod = self._rl_uimods[0]
            red_light_jnode = uimod_to_newconf_map.get(rl_uimod) or rl_uimod.load_conf()
            vendor = red_light_jnode['ro'].get('vendor', 'zonechecker')
            if vendor == 'zonechecker':
                main_conf['FileApi']['SceneEventMono8t'].append(
                    fs.join_path(spool_dir, 'mg{}-zc{}-scene'.format(
                        self._cam_uimod.id, rl_uimod.id
                    )))
        elif len(self._railroad_crossing_uimods) == 1:
            rc_uimod = self._railroad_crossing_uimods[0]
            railroad_crossing_jnode = uimod_to_newconf_map.get(rc_uimod) or rc_uimod.load_conf()
            rc_uiconf_pri = railroad_crossing_jnode['ro']
            vendor = rc_uiconf_pri.get('vendor', 'zonechecker')
            if vendor == 'zonechecker':
                main_conf['FileApi']['SceneEventRaw'].append(
                    fs.join_path(spool_dir, 'mg{}-zc{}-scene'.format(
                        self._cam_uimod.id, rc_uimod.id
                    )))

        # create haltaction.conf
        def_halt_list = [
            "0:python2 /opt/edge/bin/psuctl run led-camera color G blink 0HZ",
            "1:python2 /opt/edge/bin/psuctl run led-camera color R blink 1HZ",
            "2:python2 /opt/edge/bin/psuctl run led-camera color R blink 0HZ"
        ]
        halt_list = cam_uiconf['ro'].get('haltaction', def_halt_list)
        if halt_list:
            halt_conf = '\n'.join(halt_list)

        else:
            halt_conf = ''

        self._main_halt = main_conf, halt_conf

    def complete(self):
        main_conf, halt_conf = self._main_halt  # type: dict, str

        edge_rg = self._rg_dgmod
        edge_rg.install()
        edge_rg.save_main_conf(main_conf)
        edge_rg.save_haltaction_conf(halt_conf)
        edge_rg.save_setupstore_conf(active=True)

    def rollback(self):
        pass

    def notify(self):
        """
        :rtype: None
        """
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class RgDeleter:
    def __init__(self, dgmod):
        """
        :type dgmod: dg.DgMod
        """
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod):
        """
        :type entry_uimod: ph.UiMod
        :rtype: RgDeleter | None
        """
        if entry_uimod.type == ph.UI_TYPE_RTSP_CAMERA:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_RTSP_GRABBER, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self):
        self._dgmod.uninstall()
