from itertools import ifilter

from edge import eventlog
from edge.com import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator


# noinspection PyUnusedLocal
def rs_get_creators(entry_uimod, all_used_uimods, uimod_to_newconf_map, forced=False):
    """
    :type entry_uimod: ph.UiMod | ph.UiMod2
    :type all_used_uimods: set[ph.UiMod, ph.UiMod2]
    :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
    :type uimod_to_newconf_map: dict
    :type forced: bool
    :rtype: list of BaseCreator
    """
    if entry_uimod.type != ph.UI_TYPE_RTSP_STREAMING:
        return []

    try:
        cam_uimod, rstreaming_uimod = (
            next(ifilter(lambda x: x.type == ph.UI_TYPE_MV_CAMERA, all_used_uimods), None),
            next(ifilter(lambda x: x.type == ph.UI_TYPE_RTSP_STREAMING, all_used_uimods)))

        activate = any([
            uimod_to_newconf_map.get(cam_uimod),
            uimod_to_newconf_map.get(rstreaming_uimod)
        ])

        if forced or activate:
            return [RsCreator(rstreaming_uimod, cam_uimod)]

    except StopIteration:
        pass

    return []


class RsCreator(BaseCreator):
    def __init__(self, rtsps_uimod, cam_uimod):
        """
        :type rtsps_uimod: ph.UiMod
        :type cam_uimod: ph.UiMod
        """
        BaseCreator.__init__(self)
        self._rtsps_uimod = rtsps_uimod
        self._cam_uimod = cam_uimod
        self._rs_dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_RTSP_SERVER, rtsps_uimod.pid, rtsps_uimod.id)
        self._info_main_ss = None

    @property
    def dgmod(self):
        """
        :rtype: dg.DgMod
        """
        return self._rs_dgmod

    def prepare(self, uimod_to_newconf_map):
        """
        :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
        :type uimod_to_newconf_map: dict
        """
        ui_pid = self._rtsps_uimod.pid

        rtsps_uiconf = uimod_to_newconf_map.get(self._rtsps_uimod) or self._rtsps_uimod.load_conf()
        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()

        # get main.conf#/Resolution
        ui_videostream_jnode = rtsps_uiconf['rw']['live']['videostream']

        # get main.conf#/Stream/UdpPorts
        ui_streaming_jnode = rtsps_uiconf['rw']['streaming']
        if ui_streaming_jnode['transport'] == 'rtp-over-ucast-udp-or-tcp':
            if ui_streaming_jnode['min-udp-port'] < ui_streaming_jnode['max-udp-port']:
                edge_udp_ports = ui_streaming_jnode['min-udp-port'], ui_streaming_jnode['max-udp-port']
            else:
                edge_udp_ports = ui_streaming_jnode['max-udp-port'], ui_streaming_jnode['min-udp-port']

        elif ui_streaming_jnode['transport'] == 'rtp-over-tcp':
            edge_udp_ports = None, None

        else:
            raise NotImplementedError

        # get main.conf#/Stream/Link and main.conf#/Quality
        ui_tcp_port = ui_streaming_jnode['tcp-port']
        if ui_videostream_jnode['format'] == "h264-cqp":
            link = 'rtsp://@:%s/live-h264' % ui_tcp_port
            quality = ui_videostream_jnode['format-h264-cqp-compression']

        elif ui_videostream_jnode['format'] == "h264-cbr":
            link = 'rtsp://@:%s/live-h264' % ui_tcp_port
            quality = ui_videostream_jnode['format-h264-cbr-bitrate-max']

        elif ui_videostream_jnode['format'] == "mjpeg":
            link = 'rtsp://@:%s/live-mjpeg' % ui_tcp_port
            quality = ui_videostream_jnode['format-mjpeg-quality']

        else:
            raise NotImplementedError

        spool_dir = ph.get_spool_dir(ui_pid)

        main_conf = {
            "Stream": {
                "Link": link,
                "UdpPorts": edge_udp_ports
            },
            "UseGPU": not ui_videostream_jnode['sw-codec'],
            "Quality": quality,
            "FileApi": {
                "AddSceneImgReact": {
                    "Url": fs.join_path(spool_dir, 'or{0}-rs{1}-scene'.format(
                        self._rtsps_uimod.id, self._rtsps_uimod.id))
                }
            },
            "Fps": cam_uiconf['rw']['videostream']['fps'] // ui_videostream_jnode['skip-rate'],
            "Resolution": {
                "MaxHeight": 65535,
                "MaxWidth": ui_videostream_jnode['frame-w']
            }
        }
        # create setupstore.conf
        if rtsps_uiconf['rw']['enabled']:
            ss_active = True
        else:
            ss_active = False

        self._info_main_ss = main_conf, ss_active

    def complete(self):
        main_conf, ss_active = self._info_main_ss  # type: dict, bool

        edge_rs = self._rs_dgmod
        edge_rs.install()
        edge_rs.save_main_conf(main_conf)
        edge_rs.save_setupstore_conf(ss_active)

    def rollback(self):
        pass

    def notify(self):
        """
        :rtype: None
        """
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class RsDeleter:
    def __init__(self, dgmod):
        """
        :type dgmod: dg.DgMod
        """
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod):
        """
        :type entry_uimod: ph.UiMod
        :rtype: RsDeleter | None
        """
        if entry_uimod.type == ph.UI_TYPE_RTSP_STREAMING:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_RTSP_SERVER, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self):
        self._dgmod.uninstall()
