# coding=utf-8
import os
import errno

from edge import eventlog
from edge.webserver import backup, file_lock
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator


def _locked_generate(backup_dir, creators, used_uimods, used_dgmods, uimod_to_newconf_map):
    """
    :type backup_dir: str
    :type creators: list[BaseCreator]
    :type used_uimods: set[ph.UiMod, ph.UiMod2]
    :type used_dgmods: set[dg.DgMod]
    :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
    :type uimod_to_newconf_map: dict
    """
    try:
        os.makedirs(backup_dir)
    except os.error as x:
        if x.errno != errno.EEXIST:
            eventlog.log_e('fail: makedirs {} ({})'.format(backup_dir, errno.errorcode[x.errno]))
        raise NotImplementedError

    backup_error = False

    dgmods_backup = []
    # eventlog.log_i('dg modes to backup {}'.format(map(str, used_dgmods)))
    for dgmod in used_dgmods:
        if dgmod.type == dg.EDGE_TYPE_SETUPSTORE:
            continue

        dgmod_backup = backup.DgmodBackup.create(backup_dir, dgmod)
        if dgmod_backup is None:
            eventlog.log_w('fail: backup dgmod {}'.format(str(dgmod)))
            backup_error = True
            continue

        dgmods_backup.append(dgmod_backup)

    uimods_backup = []
    for uimod in used_uimods:
        if uimod.type in ph.FAKE_UI_TYPES:
            continue

        uimod_backup = backup.UimodBackup.create(backup_dir, uimod)
        if uimod_backup is None:
            eventlog.log_w('fail: backup uimod')
            backup_error = True
            continue

        uimods_backup.append(uimod_backup)

    for uimod, newconf in uimod_to_newconf_map.iteritems():
        if uimod.type in ph.FAKE_UI_TYPES:
            continue

        uimod_backup = backup.NewFileBackup.create(backup_dir, uimod, newconf)
        if uimod_backup is None:
            eventlog.log_w('fail: NewFileBackup')
            backup_error = True

        uimods_backup.append(uimod_backup)

    if backup_error:
        raise RuntimeError

    # noinspection PyBroadException
    try:
        restore_backup = False
        try:
            map(lambda creator: creator.prepare(uimod_to_newconf_map), creators)
            map(lambda creator: creator.complete(), creators)
        except Exception:
            map(lambda creator: creator.rollback(), creators)
            raise

        for uimod, newconf in uimod_to_newconf_map.iteritems():
            uimod.save_conf(newconf)

        # noinspection PyBroadException
        try:
            for dgmod_backup in dgmods_backup:
                dgmod_backup.remove()

            for uimod_backup in uimods_backup:
                uimod_backup.remove()

            if not os.listdir(backup_dir):
                os.rmdir(backup_dir)

        except Exception:
            eventlog.log_x()

    except Exception:
        restore_backup = True
        eventlog.log_x()

    if restore_backup:
        # TODO: what if except inside?
        map(lambda bak: bak.restore(), uimods_backup)
        map(lambda bak: bak.restore(), dgmods_backup)
        raise RuntimeError


def generate(backup_dir, entry_uirefs, uimod_to_newconf_map=None, forced=False):
    """
    :type backup_dir: str
    :type entry_uirefs: set[str] | list[str]
    :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
    :type uimod_to_newconf_map: dict | None
    :type forced: bool
    :rtype: set[dg.DgMod]
    """
    uimod_to_newconf_map = uimod_to_newconf_map or dict()

    from edge.webserver.sysconf.dgmod.eventstore import es_get_creators
    from edge.webserver.sysconf.dgmod.grabcontrol import gc_get_creators
    from edge.webserver.sysconf.dgmod.lpr import lr_get_creators
    from edge.webserver.sysconf.dgmod.rawcodec import rc_get_creators
    from edge.webserver.sysconf.dgmod.mediagrabber import mg_get_creators
    from edge.webserver.sysconf.dgmod.radar import ra_get_creators
    from edge.webserver.sysconf.dgmod.renderer import or_get_creators
    from edge.webserver.sysconf.dgmod.rtspgrabber import rg_get_creators
    from edge.webserver.sysconf.dgmod.rtspserver import rs_get_creators
    from edge.webserver.sysconf.dgmod.speedcontrol import sc_get_creators
    from edge.webserver.sysconf.dgmod.tracker import tr_get_creators
    from edge.webserver.sysconf.dgmod.webserver import ws_get_creators
    from edge.webserver.sysconf.dgmod.mediacodecgst import mc_get_creators
    from edge.webserver.sysconf.dgmod.blobstore import bs_get_creators
    from edge.webserver.sysconf.dgmod.trafficlight import tl_get_creators
    from edge.webserver.sysconf.dgmod.lbsapp import lbs_get_creators

    fn_list = [
        or_get_creators, rg_get_creators, lr_get_creators,
        tr_get_creators, mc_get_creators, bs_get_creators,
        sc_get_creators, es_get_creators, ws_get_creators,
        mg_get_creators, ra_get_creators, rs_get_creators,
        gc_get_creators, rc_get_creators,
        tl_get_creators, lbs_get_creators]

    root_info, root_links = ph.RootInfoUiconf.get_info_links()
    # eventlog.log_i('entry_uimod = {}'.format(str(entry_uimod)))

    all_used_uimods = set()  # type: set
    all_used_creators = []  # type: list

    for uiref in entry_uirefs:
        linked_uirefs = ph.RootInfoUiconf.get_linked_uirefs(uiref, root_links)
        linked_uimods = ph.RootInfoUiconf.get_uimods(linked_uirefs, root_info)

        # import ujson
        # eventlog.log_i("uiref {} :used_uimods: {} root_info : {}".format(
        #     uiref, map(str, used_uimods), ujson.dumps(root_info)))
        # eventlog.log_i("uiref {} :linked_uimods: {}".format(uiref, map(str, linked_uimods)))

        uimod = ph.RootInfoUiconf.get_uimods({uiref}, root_info).pop()
        # eventlog.log_i('uimod {}'.format(str(uimod)))
        for factory_fn in fn_list:
            # eventlog.log_i(factory_fn)
            all_used_creators.extend(factory_fn(uimod, linked_uimods, uimod_to_newconf_map, forced))

        all_used_uimods.update(linked_uimods)

    all_used_dgmods = set(map(lambda x: x.dgmod, all_used_creators))
    answer = all_used_dgmods
    if forced:
        all_used_dgmods = {dg.CTRL_ES_DGMOD, dg.CTRL_SS_DGMOD, dg.CTRL_WS_DGMOD}

    #  eventlog.log_i('================================len {}'.format(len(all_used_ui_modules2)))
    lock = file_lock.UimodsLocker(all_used_uimods)
    try:
        _locked_generate(backup_dir, all_used_creators, all_used_uimods, all_used_dgmods, uimod_to_newconf_map)
    finally:
        lock.release()

    map(lambda x: x.notify(), all_used_creators)

    return answer
