import os

from datetime import datetime

from edge import eventlog
from edge.com import fs
from edge.exceptions import *
from edge.webserver.sysconf.uiconf.uiconf2 import UiConf2
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf
from edge.webserver.sysconf.uiconf.rtsp_streaming import RtspStreamingUiconf

SEMVER_PATH = '/opt/edge/semver'
CTRL_SPOOL_DIR = '/var/spool/edge/'
CTRL_ARCH_DIR = '/var/opt/edge/'
CTRL_BACKUP_DIR = '/etc/edge/setupstore/backup/'
CTRL_DGCONF_DIR = '/etc/edge/'
CTRL_UICONF_DIR = '/etc/edge/setupstore/admin/actions/'
CTRL_UICONF_PATH = '/etc/edge/setupstore/about.json'
CTRL_UISCHEMA_DIR = '/etc/edge/setupstore/schemas/'
CTRL_DEF_UICONF_DIR = '/opt/edge/share/setupstore/ui/'
CTRL_DEF_DGCONF_DIR = '/opt/edge/share/etc/edge/'

TERM_UICONF_DIR = fs.join_path(CTRL_UICONF_DIR, 'pipeline')
TERM_UISCHEMA_DIR = fs.join_path(CTRL_UISCHEMA_DIR, 'pipeline')
TERM_DEF_UICONF_DIR = fs.join_path(CTRL_DEF_UICONF_DIR, 'pipeline')


UI_ROLE_ROOT = 'root'
UI_ROLE_SYSTEM = 'system'
UI_ROLE_CONTROLLER = 'controller'
UI_ROLE_TERMINAL = 'terminal'
UI_ROLE_SERVICE = 'service'
UI_ROLE_DETECTOR = 'detector'
UI_ROLE_ARCHIVE = 'archive'
UI_ROLE_UNIT = 'unit'

UI_TYPE_TIME = 'time'
UI_TYPE_USERS = 'users'
UI_TYPE_SIGN = 'sign'
UI_TYPE_NETWORK = 'network'
UI_TYPE_ZABBIX_AGENT = 'zabbix-agent'
UI_TYPE_FACTORY = 'factory'
UI_TYPE_CONTROLLER = 'controller'
UI_TYPE_TERMINAL = 'terminal'
UI_TYPE_GNSS = 'gnss'
UI_TYPE_LOCATION = 'location'
UI_TYPE_RADAR = 'radar'
UI_TYPE_MV_CAMERA = 'mv-camera'
UI_TYPE_RTSP_CAMERA = 'rtsp-camera'
UI_TYPE_ROADWAY = 'lane-enforcement'
UI_TYPE_LENS = 'lens'
UI_TYPE_LPR = 'lpr'
UI_TYPE_MEDIASTORE = 'mediastore'
UI_TYPE_FUSION = 'fusion'
UI_TYPE_RTSP_STREAMING = 'rtsp-streaming'
UI_TYPE_HTTP_STREAMING = 'http-streaming'
UI_TYPE_TRAFFIC_RULES = 'traffic-rules-engine'
UI_TYPE_WEBSERVER = 'webserver'
UI_TYPE_EVENTSTORE = 'eventstore'
UI_TYPE_TRAFFIC_LIGHT = 'traffic-light'
UI_TYPE_RTSP_SERVER = 'rtsp-server'
UI_TYPE_TCP_RECIPIENT = 'tcp-recipient'
UI_TYPE_LEDDAR = 'leddar'
UI_TYPE_WIEGAND_PCB = 'wiegand-pcb'
UI_TYPE_WIEGAND_PCB_RELAY_PORT = 'wiegand-pcb~relay-port'
UI_TYPE_WIEGAND_PCB_INPUT_PORT = 'wiegand-pcb~input-port'
UI_TYPE_IPCOM_WAGA = 'ipcom-waga'

UI_TYPE_IIT_CA = 'iit-ca'
UI_TYPE_IIT_CA_END_USER = 'iit-ca~end-user'

UI_TYPE_TRANSIT = 'transit'
UI_TYPE_PARKING = 'parking'
UI_TYPE_RED_LIGHT = 'red-light'
UI_TYPE_RAILROAD_CROSSING = 'railroad-crossing'
UI_TYPE_SPOT_SPEED = 'spot-speed'
UI_TYPE_SECTION_SPEED = 'section-speed'
UI_TYPE_WRONG_WAY = 'wrong-way'
UI_TYPE_BUS_LANE = 'bus-lane'
UI_TYPE_WANTED = 'wanted'
UI_TYPE_SPOT_SPEED_WAGA = 'spot-speed-waga'

UIMOD2_TYPES = [
    UI_TYPE_WIEGAND_PCB,
    UI_TYPE_WIEGAND_PCB_RELAY_PORT,
    UI_TYPE_WIEGAND_PCB_INPUT_PORT,
    UI_TYPE_IIT_CA,
    UI_TYPE_IIT_CA_END_USER
]

FAKE_UI_TYPES = [
    UI_TYPE_WEBSERVER,
    UI_TYPE_MEDIASTORE,
    UI_TYPE_FUSION,
    UI_TYPE_RADAR,
    UI_TYPE_TRAFFIC_RULES,
    UI_TYPE_TRAFFIC_LIGHT,
    UI_TYPE_RTSP_SERVER
]

MMR_CLASSES = [
    'UNKNOWN',
    'CAR',
    'MINIVAN',
    'VAN',
    'CROSSOVER',
    'OFFROAD',
    'BUS',
    'LT',
    'HT'
]


class UiMod:
    def __init__(self, uimod_pid, uimod_type, uimod_id):
        """
        :type uimod_pid: str
        :type uimod_type: str
        :type uimod_id: str
        """
        self.pid = uimod_pid
        self.type = uimod_type
        self.id = uimod_id

        self._conf = None

    def __hash__(self):
        return hash(self.type + self.id)

    def __str__(self):
        return self.ref

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                getattr(other, 'pid', None) == self.pid and
                getattr(other, 'type', None) == self.type and
                getattr(other, 'id', None) == self.id)

    @classmethod
    def create(cls, uimod_ref):
        """
        :type uimod_ref: str
        :rtype: UiMod
        """
        try:
            uimod_type, uimod_pid_id = uimod_ref.split('.', 1)
            uimod_pid, uimod_id = uimod_pid_id.split('_', 1)
        except Exception:
            eventlog.log_d('fail: NotFoundError ({})'.format(uimod_ref))
            raise NotFoundError

        return cls(uimod_pid, uimod_type, uimod_id)

    @classmethod
    def create_term(cls, term_pid):
        """
        :type term_pid: str
        :rtype: UiMod
        """
        return cls(term_pid, UI_TYPE_TERMINAL, str(int(term_pid) * 10))

    @classmethod
    def create_controller(cls, controller_id='0'):
        """
        :type controller_id: str
        :rtype: UiMod
        """
        return cls(controller_id, UI_TYPE_CONTROLLER, str(int(controller_id) * 10))

    def assert_type(self, uimod_type):
        """
        :type uimod_type: str
        :rtype: None
        """
        if self.type != uimod_type:
            eventlog.log_d('fail: NotFoundError ({})'.format(self.ref))
            raise NotFoundError

    @property
    def ref(self):
        # return '{}.{}'.format(self.type, str(10 * int(self.pid) + int(self.id)))
        return '{}.{}_{}'.format(self.type, self.pid, self.id)

    @property
    def conf_dir(self):
        """
        :rtype: str
        """
        return fs.join_path(TERM_UICONF_DIR, self.pid, self.type)

    @property
    def conf_path(self):
        """
        :rtype: str
        """
        old_pid, old_id = divmod(int(self.id), 10)
        return fs.join_path(self.conf_dir, '{}.json'.format(old_id))
        # return fs.join_path(self.conf_dir, '{}.json'.format(self.id))

    def load_conf(self, **kwargs):
        """
        :rtype: dict
        :raise FsError:
        """
        if self._conf is None:
            self._conf = fs.load_json3(self.conf_path, **kwargs)

        return self._conf

    def save_conf(self, conf_jnode):
        """
        :type conf_jnode: dict
        :rtype: None
        """
        fs.write_json_file(conf_jnode, self.conf_path, pretty=True)

    @property
    def schema_path(self):
        """
        :rtype: str
        """
        old_pid, old_id = divmod(int(self.id), 10)
        return fs.join_path(TERM_UISCHEMA_DIR, self.pid, self.type, '{}.json'.format(old_id))
        # return fs.join_path(BASE_UI_PIPELINE_SCHEMA_DIRPATH, self.pid, self.type, '{}.json'.format(self.id))

    @property
    def schema_limits(self):
        """
        :rtype: dict| None
        """
        schema_jnode = fs.load_json3(self.schema_path)
        return schema_jnode.get('xLimits')

    @property
    def lock_path(self):
        """
        :rtype: str
        """
        return fs.join_path('/var/run/edge', '{}.lock').format(self.ref)

    def create_backup_dirpath(self):
        """
        :rtype: str
        """
        import time
        backup_action = '{:d}-{}'.format(int(round(time.time() * 1e6)), self.ref)
        backup_dir = fs.join_path(CTRL_BACKUP_DIR, backup_action)
        return backup_dir

    def _get_defconf_path(self, factory_conf=None):
        """
        :type factory_conf: dict | None
        :rtype: str
        """
        import fnmatch

        if self.type == UI_TYPE_MV_CAMERA:
            if factory_conf is None:
                eventlog.log_e_param('factory_conf')
                raise NotImplementedError

            cam_vendor = factory_conf['vendor'].lower()  # TODO workaround for migrate script
            cam_model = str.replace(str(factory_conf['model']), ' ', '-')

            model_files = os.listdir(fs.join_path(TERM_DEF_UICONF_DIR, UI_TYPE_MV_CAMERA))
            model_files2 = fnmatch.filter(model_files, '{}_{}*jnode.json'.format(cam_vendor, cam_model))
            for model_file in model_files2:
                # TODO: what to do if there are jnodes for one camera with different focal-lengths?
                return fs.join_path(TERM_DEF_UICONF_DIR, self.type, model_file)

            eventlog.log_e('fail: unsupported (vendor:{}, model:{})'.format(cam_vendor, cam_model))
            raise NotImplementedError

        #
        type_to_name_map = {}
        if self.type != UI_TYPE_TERMINAL:
            root_info = RootInfoUiconf.get_info()
            _, term_info = InfoUiconf.find_unit_parent_r(root_info, unit_ref=self.ref, parent_type=UI_TYPE_TERMINAL)
            if term_info is None:
                eventlog.log_e('fail: unsupported')
                raise NotImplementedError

            elif InfoUiconf.find_unit_r(term_info, unit_type=UI_TYPE_RADAR):
                type_to_name_map = {
                    UI_TYPE_HTTP_STREAMING: 'speed-jnode.json',
                    UI_TYPE_RTSP_STREAMING: 'speed-jnode.json',
                    UI_TYPE_LOCATION: 'ummr0a-type29-jnode.json',
                    UI_TYPE_TRANSIT: 'def-speed-jnode.json'
                }

            elif InfoUiconf.find_unit_r(term_info, unit_type=UI_TYPE_MV_CAMERA):
                type_to_name_map = {
                    UI_TYPE_HTTP_STREAMING: 'lpr-jnode.json',
                    UI_TYPE_RTSP_STREAMING: 'lpr-jnode.json'
                }

            elif InfoUiconf.find_unit_r(term_info, unit_type=UI_TYPE_IPCOM_WAGA):
                type_to_name_map = {
                    UI_TYPE_HTTP_STREAMING: 'lpr-jnode.json',
                    UI_TYPE_RTSP_CAMERA: 'rtsp-jnode.json',
                    UI_TYPE_TRANSIT: 'def-speed-jnode.json'
                }

            elif InfoUiconf.find_unit_r(term_info, unit_type=UI_TYPE_RTSP_CAMERA):
                type_to_name_map = {
                    UI_TYPE_HTTP_STREAMING: 'lpr-jnode.json',
                    UI_TYPE_RTSP_CAMERA: 'rtsp-jnode.json'
                }

            else:
                eventlog.log_e('unsupported')
                raise NotImplementedError

        conf_name = type_to_name_map.get(self.type, 'def-jnode.json')
        return fs.join_path(TERM_DEF_UICONF_DIR, self.type, conf_name)

    def _install_conf(self, factory_conf=None):
        """
        :type factory_conf: dict | None
        :rtype: None
        """
        eventlog.log_i('_install_conf {}'.format(str(self)))
        conf_path = self.conf_path
        if fs.path_exists(conf_path):
            eventlog.log_w('fail: already installed')
            return

        defconf_path = self._get_defconf_path(factory_conf)

        fs.silent_mkdir_r(self.conf_dir)
        fs.copy_file(defconf_path, conf_path)

        if self.type == UI_TYPE_MV_CAMERA:
            uiconf = self.load_conf()
            uiconf['rw']['serial'] = str(factory_conf['serial'])
            self.save_conf(uiconf)

        elif self.type == UI_TYPE_RTSP_CAMERA:
            if factory_conf.get('with-intrinsic', False):
                uiconf = self.load_conf()
                uiconf['rw']['conf']['intrinsic'] = {
                    'active': False,
                    'frustum': {
                        'sensor-wh': [
                            11.3, 7.1
                        ],
                        'focal-dist': 50
                    },
                    'image-wh': [
                        1920, 1200
                    ]
                }
                self.save_conf(uiconf)

        elif self.type == UI_TYPE_TERMINAL:
            uiconf = self.load_conf()
            uiconf['ro']['uuid'] = factory_conf['term-uuid']
            uiconf['ro']['ctime'] = datetime.now().isoformat()
            uiconf['rw']['origin-name'] = factory_conf['term-pid']
            uiconf['rw']['origin-place'] = factory_conf['term-pid']
            self.save_conf(uiconf)

        elif self.type == UI_TYPE_RTSP_STREAMING:
            uiconf = self.load_conf()
            uiconf['rw']['streaming']['tcp-port'] += int(factory_conf['term-pid'])
            uiconf['rw']['streaming']['url'] = RtspStreamingUiconf(uiconf['rw']).rtsp_url
            self.save_conf(uiconf)

        eventlog.log_i('done: cp {} {}'.format(defconf_path, conf_path))

    def install_schema(self):
        """
        :rtype: None
        """
        conf_path = self.conf_path
        if fs.path_exists(conf_path):
            fs.ln_schema(self.schema_path, conf_path)

    def install(self, factory_conf=None):
        """
        :type factory_conf: dict | None
        :rtype: None
        """
        if self.type not in FAKE_UI_TYPES:
            self._install_conf(factory_conf)
            self.install_schema()

    def uninstall(self):
        """
        :rtype: None
        """
        if self.type not in FAKE_UI_TYPES:
            fs.silent_rm_r(self.conf_path)
            fs.silent_rm_r(self.schema_path)


class UiMod2:
    def __init__(self, ref):
        """
        :type ref: str
        """
        assert ref

        self._ref = ref
        self._conf_jnode = None
        self._container_path = None

    def __eq__(self, other):
        return isinstance(other, self.__class__) and getattr(other, 'ref', None) == self.ref

    def __hash__(self):
        return hash(self.ref)

    def __str__(self):
        return self.ref

    @property
    def ref(self):
        """
        :rtype: str
        """
        return self._ref

    @property
    def type(self):
        """
        :rtype: str
        """
        segments = self.ref.split('~')
        subtypes = map(lambda x: x.split('.')[0], segments)
        return '~'.join(subtypes)

    @property
    def lock_path(self):
        """
        :rtype: str
        """
        return self.uiconf.lock_path

    def create_backup_dirpath(self):
        """
        :rtype: str
        """
        import time
        backup_action = '{:d}-{}'.format(int(round(time.time() * 1e6)), self.ref)
        backup_dir = fs.join_path(CTRL_BACKUP_DIR, backup_action)
        return backup_dir

    def assert_ref(self):
        """
        :rtype: None
        :raises NotFoundError:
        """
        self._load_conf()

    def assert_type(self, other_type):
        """
        :type other_type: str
        :rtype: None

        :raises NotFoundError:
        """
        if self.type != other_type:
            eventlog.log_d('fail: NotFoundError ({})'.format(self.ref))
            raise NotFoundError

    @property
    def root_ref(self):
        """
        :rtype: str
        :raises NotFoundError:
        """
        return self.ref.split('~', 1)[0]

    @property
    def parent_ref(self):
        """
        :rtype: str
        :raises NotFoundError:
        """
        parent_ref = '~'.join(self.ref.split('~')[:-1])
        if not parent_ref:
            raise NotFoundError

        return parent_ref

    @property
    def parent(self):
        """
        :rtype: UiMod2
        :raises NotFoundError:
        """
        return UiMod2(self.parent_ref)

    def child_ref(self, child_type, child_chan):
        """
        :type child_type: str
        :type child_chan: str | int
        :rtype: str
        """
        if not child_type.startswith(self.type):
            raise NotFoundError

        child_type_segments = child_type.split('~')
        if not len(self.type.split('~')) == len(child_type_segments) - 1:
            raise NotFoundError

        return '{}~{}.{}'.format(self.ref, child_type_segments[-1], child_chan)

    def child(self, child_type, child_chan):
        """
        :type child_type: str
        :type child_chan: str | int
        :rtype: UiMod2
        """
        return UiMod2(self.child_ref(child_type, child_chan))

    def _load_conf(self):
        """
        :rtype: dict
        :raise NotFoundError:
        """
        if None in [self._container_path, self._conf_jnode]:
            container_names_desc = []
            segments = self.ref.split('~')
            for index in range(0, len(segments)):
                container_name = '~'.join(segments[0:index + 1]) + '.json'
                container_names_desc.append(container_name)

            for container_name in reversed(container_names_desc):
                container_path = fs.join_path(CTRL_UICONF_DIR, container_name)
                container_jnode = fs.load_json3(container_path, default=None) or dict()
                conf_jnode = container_jnode.get(self.ref)
                if conf_jnode:
                    self._container_path = container_path
                    self._conf_jnode = conf_jnode
                    return conf_jnode

            raise NotFoundError

        return self._conf_jnode

    @property
    def conf_path(self):
        """
        :rtype: str
        """
        import warnings
        warnings.warn('message', PendingDeprecationWarning, stacklevel=2)

        self.assert_ref()
        return self._container_path

    def save_conf(self, conf_jnode):
        """
        :type conf_jnode: dict
        :rtype: None
        """
        self.assert_ref()
        container_jnode = fs.load_json3(self._container_path, default=None) or dict()
        container_jnode[self.ref] = conf_jnode
        fs.write_json_file(container_jnode, self._container_path, pretty=True)

        self._conf_jnode = conf_jnode

    @property
    def uiconf(self):
        """
        :rtype: UiConf2
        :raise NotFoundError:
        """
        return UiConf2(self._load_conf())

    def save_uiconf(self, conf):
        """
        :type conf: UiConf2
        :rtype: None
        """
        with open(SEMVER_PATH, 'r') as f:
            ver = f.readlines()[0].strip()

        conf.set_ver(ver)
        self.save_conf(conf.raw)

    def install(self, conf, container_ref=None):
        """
        :type conf: UiConf2
        :type container_ref: str | None
        :rtype: None
        """
        try:
            self.assert_ref()
            eventlog.log_e('fail: InternalError (already installed)')
            raise InternalError

        except NotFoundError:
            pass

        if not self.ref.startswith(container_ref):
            raise NotImplementedError

        container_name = '{}.json'.format(container_ref or self.ref)
        container_path = fs.join_path(CTRL_UICONF_DIR, container_name)
        container_jnode = fs.load_json3(container_path, default=None) or dict()
        container_jnode[self.ref] = conf.raw
        fs.write_json_file(container_jnode, container_path, pretty=True)

        self._container_path = container_path
        self._conf_jnode = conf.raw

    def uninstall(self):
        """
        :rtype: None
        """
        self.assert_ref()
        container_jnode = fs.load_json3(self._container_path, default=None) or dict()
        del container_jnode[self.ref]
        if len(container_jnode) > 0:
            fs.write_json_file(container_jnode, self._container_path, pretty=True)
        else:
            fs.silent_rm_r(self._container_path)


def get_spool_dir(ui_pid):
    """
    :type ui_pid: str
    :rtype: str
    """
    return fs.join_path(CTRL_SPOOL_DIR, 'pipeline-{}'.format(ui_pid))


# noinspection PyClassHasNoInit
class TermAboutUiconf:
    @staticmethod
    def get_conf_path(ui_pid):
        """
        :type ui_pid: str
        :rtype: str
        """
        return fs.join_path(TERM_UICONF_DIR, ui_pid, 'about.json')

    @classmethod
    def save_conf(cls, term_pid, term_uuid, term_ctime=None, term_place=None, term_name=None):
        """
        :type term_pid: str
        :type term_uuid: str
        :type term_ctime: str | None
        :type term_place: str | None
        :type term_name: str | None
        :rtype: None
        """
        conf_path = cls.get_conf_path(term_pid)
        conf = fs.load_json3(conf_path, default={
            'origin': dict()
        })

        conf_origin = conf['origin']  # type: dict
        conf_origin['uuid'] = term_uuid
        conf_origin['place'] = term_place or term_pid
        conf_origin['name'] = term_name or term_pid

        if term_ctime:
            conf_origin['time'] = term_ctime

        elif 'time' not in conf_origin:
            conf_origin['time'] = datetime.now().isoformat()

        fs.write_json_file(conf, conf_path)


# noinspection PyClassHasNoInit
class InfoUiconf:
    @classmethod
    def get_info_links(cls, info_path):
        """
        :type info_path: str
        :rtype: dict, list of (str, str)
        """
        uiconf = fs.load_json3(info_path, default=None)
        if uiconf is None:
            raise NotImplementedError

        uiconf_pri = uiconf['pri']  # type: dict
        uiconf_pub = uiconf['pub']  # type: dict

        root_links = uiconf_pri['links']
        root_info = uiconf_pub['info']  # type: dict

        return root_info, root_links

    @classmethod
    def map_units_r(cls, root_info, map_fn=None, unit_types=None):
        """
        :type root_info: dict
        :type map_fn: (dict) -> Any
        :type unit_types: list[str] | None
        :rtype: list
        """
        def finds_units_r(root, unit_list):
            """
            :type root: dict
            :type unit_list: list[dict]
            :rtype: list
            """
            subunits = root.get('units', [])
            for subunit in subunits:
                finds_units_r(subunit, unit_list)

            if unit_types is None:
                unit_list.append(root)

            elif root['type'] in unit_types:
                unit_list.append(root)

            return unit_list

        units_info = finds_units_r(root_info, [])
        # eventlog.log_i('{}'.format(units_info))
        return map(map_fn, units_info) if map_fn else units_info

    @classmethod
    def find_unit_refs_r(cls, root_info, unit_types=None):
        """
        :type root_info: dict
        :type unit_types: list[str] | None
        :rtype: list[str]
        """
        return cls.map_units_r(root_info, lambda x: x['ref'], unit_types)

    @classmethod
    def find_unit_r(cls, root_info, unit_ref=None, unit_type=None):
        """
        :type root_info: dict
        :type unit_ref: str | None
        :type unit_type: str | None
        :rtype: dict | None
        """
        # eventlog.log_i('{} {}'.format(unit_ref, unit_type))
        assert any([unit_ref, unit_type])

        if root_info['ref'] == unit_ref or root_info['type'] == unit_type:
            return root_info

        # eventlog.log_i('12')
        subunits = root_info.get('units', [])
        # eventlog.log_i('subunits {}'.format(subunits))
        for subunit in subunits:
            # eventlog.log_i('20 {}'.format(subunit))
            unit = cls.find_unit_r(subunit, unit_ref, unit_type)
            # eventlog.log_i('21')
            if unit:
                return unit

    @classmethod
    def find_unit_parent_r(cls, root_info, unit_ref=None, unit_type=None, parent_type=None):
        """
        :type root_info: dict
        :type unit_ref: str | None
        :type unit_type: str | None
        :type parent_type: str | None

        :rtype: (dict, dict) | (None, None)
        """
        assert any([unit_ref, unit_type])

        def find_unit_parents_r(root, parents):
            """
            :type root: dict
            :type parents: list[dict]
            :rtype: dict | None, list[dict]
            """
            if root['ref'] == unit_ref or root['type'] == unit_type:
                return root, parents

            subunits = root.get('units', [])
            for subunit in subunits:
                unit, _ = find_unit_parents_r(subunit, parents)
                if unit:
                    parents.append(root)
                    return unit, parents

            return None, parents

        unit_info, parents_info = find_unit_parents_r(root_info, [])
        for parent in parents_info:
            if parent_type is None:
                return unit_info, parent

            elif parent_type == parent['type']:
                return unit_info, parent

        return None, None


# noinspection PyClassHasNoInit
class RootInfoUiconf(InfoUiconf):
    _DEFCONF_PATH = '/opt/edge/share/setupstore/ui/root-info.json'
    _CONF_PATH = '/etc/edge/setupstore/root-info.json'

    @classmethod
    def get_ver(cls):
        """
        :rtype: str | None
        """
        uiconf = fs.load_json3(cls._CONF_PATH, default=None)
        if uiconf is None:
            raise NotImplementedError

        return uiconf.get('ver')

    @classmethod
    def set_info_links(cls, root_info, root_links, ver=None):
        """
        :type root_info: dict
        :type root_links: dict
        :type ver: str | None
        :rtype: None
        """
        new_ver = ver or cls.get_ver()
        uiconf = {
            'ver': new_ver,
            'pri': {
                'links': root_links
            },
            'pub': {
                'info': root_info
            }
        }

        fs.write_json_file(uiconf, cls._CONF_PATH, True)

    @classmethod
    def get_info_links(cls, info_path=None):
        """
        :type info_path: str
        :rtype: dict, list of (str, str)
        """
        return InfoUiconf.get_info_links(info_path or cls._CONF_PATH)

    @classmethod
    def get_info(cls):
        """
        :rtype: dict
        """
        return cls.get_info_links()[0]

    @classmethod
    def get_links(cls):
        """
        :rtype: list of (str, str)
        """
        return cls.get_info_links()[1]

    @classmethod
    def set_unit_name(cls, unit_ref, unit_name):
        """
        :type unit_ref: str
        :type unit_name: str
        :rtype: None
        """
        root_info, root_links = cls.get_info_links()
        unit_info = cls.find_unit_r(root_info, unit_ref)
        if unit_info is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        unit_info['name'] = unit_name
        cls.set_info_links(root_info, root_links)

    @classmethod
    def get_linked_uirefs(cls, used_uiref, links=None):
        """
        :type used_uiref: str
        :type links: list[(str, str)] | None
        :rtype: set[str]
        """
        if links is None:
            links = cls.get_links()

        used_uirefs = set()
        for link in links:
            # eventlog.log_i('used_uiref {} {}'.format(used_uiref, link))
            if used_uiref in link:
                used_uirefs.update(link)

        return used_uirefs

    @classmethod
    def get_uimods(cls, uirefs, root_info=None):
        """
        :type uirefs: list[str] | set[str]
        :type root_info: dict | None
        :rtype: set[UiMod | UiMod2]
        """
        if root_info is None:
            root_info = cls.get_info()

        uimods = set()
        for uiref in uirefs:
            uimod_info = cls.find_unit_r(root_info, unit_ref=uiref)
            if uimod_info is None:
                eventlog.log_w('fail: find_unit_r (uiref:{})'.format(uiref))
                continue

            unit_type = uimod_info.get('type')
            if unit_type == UI_TYPE_WEBSERVER:
                uimod = UiMod('0', UI_TYPE_WEBSERVER, '0')
            elif unit_type in UIMOD2_TYPES:
                uimod = UiMod2(uimod_info['ref'])
            else:
                uimod = UiMod.create(uimod_info['ref'])

            uimods.add(uimod)

        return uimods

    @classmethod
    def new_uimod_pid_type_id(cls, root_info, uimod_type, uimod_pid=None):
        """
        :type uimod_type: str
        :type root_info: dict | None
        :type uimod_pid: str | None
        :rtype: (str | None, str, str)
        """
        used_refs = cls.find_unit_refs_r(root_info, unit_types=[uimod_type])

        if uimod_type == UI_TYPE_TERMINAL:
            if uimod_pid is None:
                # 'terminal.1_10' => 1
                import sys
                used_pids = map(lambda x: int(UiMod.create(x).pid), used_refs)
                for candidate in xrange(sys.maxint):
                    if candidate not in used_pids:
                        uimod_pid = candidate
                        break

            uimod_id = 10 * int(uimod_pid)
            eventlog.log_d('done: pid:{}, type:{}, id:{}'.format(uimod_pid, uimod_type, uimod_id))
            return str(uimod_pid), uimod_type, str(uimod_id)

        elif root_info['type'] in [UI_TYPE_CONTROLLER, UI_TYPE_WIEGAND_PCB, UI_TYPE_IIT_CA]:
            # 'eventstore.14' => 14
            used_ids = map(lambda x: int(x.split('.')[-1]), used_refs)
            for candidate in xrange(10):
                if candidate not in used_ids:
                    uimod_id = candidate
                    eventlog.log_d('done: pid:{}, type:{}, id:{}'.format(None, uimod_type, uimod_id))
                    return None, uimod_type, str(uimod_id)

        elif root_info['type'] in [UI_TYPE_TERMINAL, UI_TYPE_TRAFFIC_RULES, UI_TYPE_MV_CAMERA, UI_TYPE_RTSP_CAMERA, UI_TYPE_LPR]:
            # 'wrong-way.1_14' => 14 => 4
            used_uimod_rel_ids = cls.map_units_r(
                root_info, lambda x: int(UiMod.create(x['ref']).id) % 10,
                unit_types=[uimod_type])

            uimod_pid = uimod_pid or UiMod.create(root_info['ref']).pid
            for candidate in xrange(10):
                if candidate not in used_uimod_rel_ids:
                    uimod_id = 10 * int(uimod_pid) + candidate
                    eventlog.log_d('done: pid:{}, type:{}, id:{}'.format(uimod_pid, uimod_type, uimod_id))
                    return str(uimod_pid), uimod_type, str(uimod_id)

        raise NotImplementedError

    @classmethod
    def validate_term(cls, term_pid):
        """
        :type term_pid: str
        :rtype: bool
        """
        root_info = cls.get_info()
        term_uimod = UiMod.create_term(term_pid)
        term_info = cls.find_unit_r(root_info, unit_ref=term_uimod.ref)
        return bool(term_info)

    @classmethod
    def get_term_info(cls, term_uuid):
        """
        :type term_uuid: str
        :rtype: dict | None
        """
        root_info = cls.get_info()
        terms_info = cls.map_units_r(root_info, unit_types=[UI_TYPE_TERMINAL])

        for term_info in terms_info:
            term_uimod = UiMod.create(term_info['ref'])
            term_uiconf = term_uimod.load_conf()
            if term_uuid == term_uiconf['ro']['uuid']:
                return term_info

    @classmethod
    def get_term_pid(cls, term_uuid):
        """
        :type term_uuid: str
        :rtype: str | None
        """
        term_info = cls.get_term_info(term_uuid)
        if term_info:
            term_uimod = UiMod.create(term_info['ref'])
            return term_uimod.pid

    @classmethod
    def get_term_event_types(cls, term_info):
        """
        :type term_info: dict
        :rtype: list[str]
        """
        det_types = [
            UI_TYPE_TRANSIT,
            UI_TYPE_PARKING,
            UI_TYPE_WRONG_WAY,
            UI_TYPE_RED_LIGHT,
            UI_TYPE_WANTED,
            UI_TYPE_SECTION_SPEED,
            UI_TYPE_BUS_LANE
        ]

        if cls.find_unit_r(term_info, unit_type=UI_TYPE_RADAR):
            det_types.append(UI_TYPE_SPOT_SPEED)

        return det_types

    @classmethod
    def get_raterm_pid(cls):
        """
        :rtype: str | None
        """
        root_info = cls.get_info()
        radar_info = cls.find_unit_r(root_info, unit_type=UI_TYPE_RADAR)
        if radar_info:
            return UiMod.create(radar_info['ref']).pid

    @classmethod
    def add_term(cls, ctrl_uiref, factory_info, factory_links, internal=False):
        """
        :type ctrl_uiref: str
        :type factory_info: dict
        :type factory_links: list[[str, str]]
        :type internal: dict
        :rtype: dict, dict
        """
        root_info, root_links = cls.get_info_links()

        uimod_pid, _, _ = cls.new_uimod_pid_type_id(root_info, UI_TYPE_TERMINAL)

        def reduce_fn(acc, old_uiref):
            """
            :type acc: dict
            :param old_uiref: str
            :rtype: dict
            """
            uimod_type = UiMod.create(old_uiref).type
            pid_type_id = cls.new_uimod_pid_type_id(factory_info, uimod_type, uimod_pid)
            acc[old_uiref] = UiMod(*pid_type_id).ref
            return acc

        old_factory_refs = cls.find_unit_refs_r(factory_info)
        old_to_new_factory_refs = reduce(reduce_fn, old_factory_refs, dict())

        # links
        for factory_link in factory_links:
            factory_link[0] = old_to_new_factory_refs.get(factory_link[0], factory_link[0])
            factory_link[1] = old_to_new_factory_refs.get(factory_link[1], factory_link[1])

        root_links.extend(factory_links)

        # conf
        def replace_unit_ref_fn(unit_info):
            """
            :type unit_info: dict
            :rtype: None
            """
            unit_info['ref'] = old_to_new_factory_refs[unit_info['ref']]

        term_info = factory_info
        term_info['internal'] = internal
        cls.map_units_r(term_info, replace_unit_ref_fn)

        ctrl_info = cls.find_unit_r(root_info, unit_ref=ctrl_uiref)
        ctrl_info['units'].append(term_info)

        def find_link(links_list, ref1, ref2):
            """
            :type links_list: list([str, str])
            :type ref1: str
            :type ref2: str
            :rtype: bool
            """
            for link in links_list:
                if ref1 in link and ref2 in link:
                    return True

            return False

        tr_uirefs = cls.find_unit_refs_r(root_info, unit_types=[UI_TYPE_TRAFFIC_RULES])
        iit_ca_uirefs = cls.find_unit_refs_r(root_info, unit_types=[UI_TYPE_IIT_CA])
        iit_ca_eu_uirefs = cls.find_unit_refs_r(root_info, unit_types=[UI_TYPE_IIT_CA_END_USER])

        for tr_uiref in tr_uirefs:
            for iit_ca_uiref in iit_ca_uirefs:
                if not find_link(root_links, tr_uiref, iit_ca_uiref):
                    root_links.append([tr_uiref, iit_ca_uiref])

            for iit_ca_eu_uiref in iit_ca_eu_uirefs:
                if not find_link(root_links, tr_uiref, iit_ca_eu_uiref):
                    root_links.append([tr_uiref, iit_ca_eu_uiref])

        cls.set_info_links(root_info, root_links)
        return root_info, term_info

    @classmethod
    def del_unit_r(cls, unit_ref):
        """
        :type unit_ref: str
        :rtype: dict
        """
        root_info, root_links = cls.get_info_links()

        unit_info, parent_info = cls.find_unit_parent_r(root_info, unit_ref=unit_ref)
        if parent_info:
            parent_info['units'] = filter(lambda x: x['ref'] != unit_ref, parent_info['units'])
        else:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        unit_refs = cls.find_unit_refs_r(unit_info)
        for unit_ref in unit_refs:
            root_links = filter(lambda link: unit_ref not in link, root_links)

        cls.set_info_links(root_info, root_links)
        return root_info

    @classmethod
    def add_detector(cls, det_uimod_pid, det_uimod_type, det_name):
        """
        :type det_uimod_pid: str
        :type det_uimod_type: str
        :type det_name: str
        :rtype: (str, dict) | None
        """
        root_info, root_links = cls.get_info_links()

        term_uimod = UiMod.create_term(det_uimod_pid)
        term_info = cls.find_unit_r(root_info, unit_ref=term_uimod.ref)
        if term_info is None:
            eventlog.log_d('fail: NotFoundError ({})'.format(UI_TYPE_TERMINAL))
            raise NotFoundError

        rules_info = cls.find_unit_r(term_info, unit_type=UI_TYPE_TRAFFIC_RULES)
        if rules_info is None:
            eventlog.log_d('fail: NotFoundError ({})'.format(UI_TYPE_TRAFFIC_RULES))
            raise NotFoundError

        _, _, det_uimod_id = cls.new_uimod_pid_type_id(rules_info, det_uimod_type)
        det_uimod = UiMod(det_uimod_pid, det_uimod_type, det_uimod_id)
        rules_info['units'].append({
            'ref': det_uimod.ref,
            'type': det_uimod.type,
            'name': det_name,
            'role': UI_ROLE_DETECTOR,
            'internal': False
        })

        if det_uimod_type == UI_TYPE_SPOT_SPEED:
            http_streaming_info = cls.find_unit_r(term_info, unit_type=UI_TYPE_HTTP_STREAMING)
            if http_streaming_info is None:
                eventlog.log_d('fail: NotFoundError ({})'.format(UI_TYPE_HTTP_STREAMING))
            else:
                root_links.append([det_uimod.ref, http_streaming_info['ref']])

            rtsp_streaming_info = cls.find_unit_r(term_info, unit_type=UI_TYPE_RTSP_STREAMING)
            if rtsp_streaming_info is None:
                eventlog.log_d('fail: NotFoundError ({})'.format(UI_TYPE_RTSP_STREAMING))
            else:
                root_links.append([det_uimod.ref, rtsp_streaming_info['ref']])

        elif det_uimod_type in [UI_TYPE_RED_LIGHT, UI_TYPE_RAILROAD_CROSSING]:
            traffic_light_info = cls.find_unit_r(term_info, unit_type=UI_TYPE_TRAFFIC_LIGHT)
            if traffic_light_info is None:
                eventlog.log_d('fail: NotFoundError ({})'.format(UI_TYPE_TRAFFIC_LIGHT))
            else:
                root_links.append([det_uimod.ref, traffic_light_info['ref']])

            rtsp_cam_info = cls.find_unit_r(term_info, unit_type=UI_TYPE_RTSP_CAMERA)
            if rtsp_cam_info is None:
                eventlog.log_d('fail: NotFoundError ({})'.format(UI_TYPE_RTSP_CAMERA))
            else:
                root_links.append([det_uimod.ref, rtsp_cam_info['ref']])

            mv_cam_info = cls.find_unit_r(term_info, unit_type=UI_TYPE_MV_CAMERA)
            if mv_cam_info is None:
                eventlog.log_d('fail: NotFoundError ({})'.format(UI_TYPE_MV_CAMERA))
            else:
                root_links.append([det_uimod.ref, mv_cam_info['ref']])

        root_links.append([det_uimod.ref, rules_info['ref']])
        cls.set_info_links(root_info, root_links)
        return det_uimod_id, root_info

    @classmethod
    def add_service(cls, service_uimod_pid, service_uimod_type, service_name, parent_type, types_to_link):
        """
        :type service_uimod_pid: str
        :type service_uimod_type: str
        :type service_name: str
        :type parent_type: str
        :type types_to_link: list(str)
        :rtype: (str, dict) | None
        """
        root_info, root_links = cls.get_info_links()

        term_uimod = UiMod.create_term(service_uimod_pid)
        term_info = cls.find_unit_r(root_info, unit_ref=term_uimod.ref)
        if term_info is None:
            eventlog.log_d('fail: NotFoundError ({})'.format(UI_TYPE_TERMINAL))
            raise NotFoundError

        parent_info = cls.find_unit_r(term_info, unit_type=parent_type)
        if parent_info is None:
            eventlog.log_d('fail: NotFoundError ({})'.format(parent_type))
            raise NotFoundError

        _, _, service_uimod_id = cls.new_uimod_pid_type_id(parent_info, service_uimod_type)
        service_uimod = UiMod(service_uimod_pid, service_uimod_type, service_uimod_id)

        if parent_info.get('units') is None:
            parent_info['units'] = []

        parent_info['units'].append({
            'ref': service_uimod.ref,
            'type': service_uimod.type,
            'name': service_name,
            'role': UI_ROLE_SERVICE,
            'internal': False
        })

        for type_to_link in types_to_link:
            unit_to_link_info = cls.find_unit_r(term_info, unit_type=type_to_link)
            root_links.append([service_uimod.ref, unit_to_link_info['ref']])

        cls.set_info_links(root_info, root_links)
        return service_uimod_id, root_info

    @classmethod
    def add_controller_service(cls, service_uimod_type, service_name, types_to_link, controller_id='0'):
        """
        :type service_uimod_type: str
        :type service_name: str
        :type controller_id: str
        :type types_to_link: list(str)
        :rtype: (str, dict) | None
        """
        root_info, root_links = cls.get_info_links()

        controller_uimod = UiMod.create_controller(controller_id)
        controller_info = cls.find_unit_r(root_info, unit_ref=controller_uimod.ref)
        if controller_info is None:
            eventlog.log_d('fail: NotFoundError ({})'.format(UI_TYPE_TERMINAL))
            raise NotFoundError

        _, _, service_uimod_id = cls.new_uimod_pid_type_id(controller_info, service_uimod_type)
        service_uimod = UiMod('0', service_uimod_type, service_uimod_id)

        if controller_info.get('units') is None:
            controller_info['units'] = []

        controller_info['units'].append({
            'ref': service_uimod.ref,
            'type': service_uimod.type,
            'name': service_name,
            'role': UI_ROLE_SERVICE,
            'internal': False
        })

        for type_to_link in types_to_link:
            unit_to_link_info = cls.find_unit_r(controller_info, unit_type=type_to_link)
            root_links.append([service_uimod.ref, unit_to_link_info['ref']])

        cls.set_info_links(root_info, root_links)
        return service_uimod_id, root_info

    @classmethod
    def get_ext_root_info(cls, terms_info=None):
        """
        :type terms_info: dict | None
        :rtype: dict
        """

        if terms_info is None:
            terms_info = cls.get_info()

        return terms_info
        #
        # terms_info.sort(key=lambda x: x['id'])
        #
        # for term_info in terms_info:
        #     ui_pid = term_info['id']
        #     modules_info = term_info['modules']
        #
        #     try:
        #         DgStats.test_cam_status(ui_pid)
        #         cam_error = False
        #     except CamServiceError:
        #         cam_error = True
        #
        #     # set pipeline status
        #     cam_dep_types = [
        #         UI_TYPE_MV_CAMERA,
        #         UI_TYPE_RTSP_CAMERA,
        #         UI_TYPE_LPR,
        #         UI_TYPE_ROADWAY,
        #         UI_TYPE_HTTP_STREAMING
        #     ]
        #
        #     for module_info in modules_info:
        #         if module_info['type'] in cam_dep_types:
        #             module_info['error'] = cam_error
        #         else:
        #             module_info['error'] = False
        #
        #     term_uimod = UiMod(ui_pid, UI_TYPE_TERMINAL, str(10 * int(ui_pid)))
        #     term_info['limits'] = {
        #         'detectors': get_term_detectors_types(term_uimod.ref)
        #     }
        #
        # return terms_info

        # ctrl_info = InfoUiconf.find_unit_r(root_info, unit_type=UI_TYPE_CONTROLLER)
        # ctrl_units = ctrl_info['units']
        #
        # for unit_info in ctrl_units:
        #     if unit_info['type'] != 'terminal':
        #         continue
        #
        #     unit_info['factory'] = [
        #         {
        #             'role': 'detector',
        #             'type': 'parking'
        #         },
        #         {
        #             'role': 'detector',
        #             'type': 'wrong-way'
        #         },
        #         {
        #             'role': 'detector',
        #             'type': 'red-light'
        #         },
        #         {
        #             'role': 'detector',
        #             'type': 'wanted'
        #         },
        #         {
        #             'role': 'detector',
        #             'type': 'section-speed'
        #         }
        #     ]
        #
        #     term_units_info = unit_info['units']
        #     ra_units_info = filter(lambda subunit_jnode: subunit_jnode['type'] == 'radar', term_units_info)
        #     if len(ra_units_info) > 0:
        #         unit_info['factory'].append({
        #             'role': 'detector',
        #             'type': 'spot-speed'
        #         })


# noinspection PyClassHasNoInit
class CamDgStats:
    def __init__(self, jnode_path):
        """
        :type jnode_path: str
        """
        self._jnode = fs.load_json3(jnode_path, default=None)

    @classmethod
    def create(cls, term_pid):
        """
        :type term_pid: str
        :rtype: CamDgStats
        """
        ui_id = '0'
        jnode_path = fs.join_path('/var/spool/edge/pipeline-{}/zabbix/vs{}'.format(term_pid, ui_id))
        return cls(jnode_path)

    def _get_status(self):
        """
        :rtype: int | None
        """
        stats_jnode = self._jnode
        if stats_jnode:
            stream_stats = stats_jnode['stream']  # type: dict
            return stream_stats['error']

    def test_availability(self):
        """
        :rtype: None
        """
        cam_status = self._get_status()
        if cam_status is None:
            eventlog.log_w('fail: CamServiceError (cam-status:<none>)')
            raise CamServiceError

        elif cam_status != 0:
            eventlog.log_w('fail: CamServiceError (cam-status:<{}>)'.format(cam_status))
            raise CamServiceError

    def get_stream_wh(self):
        """
        :rtype: (int, int) | None
        """
        ui_id = '0'
        # if ui_vs_jnode is None:
        #     ui_vs_info = UiMod(ui_pid, UI_TYPE_VIDEOSOURCE, ui_id)
        #     ui_vs_path = ui_vs_info.conf_path
        #     ui_vs_jnode = fs.load_json2(ui_vs_path)

        # ro_jnode = ui_vs_jnode.get('ro', dict())
        # stats_dir = ro_jnode.get('stats-dirpath')
        # TODO: fix this

        stats_jnode = self._jnode
        if stats_jnode:
            stream_stats = stats_jnode['stream']  # type: dict
            if stream_stats['error']:
                eventlog.log_d('stat-file error: {}'.format(stream_stats['error']))

            return stream_stats.get('wh')

    def get_stats(self, term_pid):
        """
        :type term_pid: str
        :rtype: dict
        """
        term_uimod = UiMod.create_term(term_pid)
        term_uuid = TerminalUiconf(term_uimod.conf_path).get_uuid()
        stream_stats = {
            'stream': 'api/v2/origins/media/{}/live/stream'.format(term_uuid)
        }

        cam_stream_wh = self.get_stream_wh()
        if cam_stream_wh:
            stream_stats['resolution'] = cam_stream_wh

        return {
            'media': [stream_stats]
        }


class CamUiStats:
    def __init__(self, intrinsic_jnode):
        """
        :type intrinsic_jnode: dict | None
        """
        self._intrinsic_jnode = intrinsic_jnode

    @classmethod
    def create(cls, term_pid):
        """
        :type term_pid: str
        :rtype: CamUiStats
        """
        mv_cam_uimod = UiMod(term_pid, UI_TYPE_MV_CAMERA, '0')
        rtsp_cam_uimod = UiMod(term_pid, UI_TYPE_RTSP_CAMERA, '0')

        mv_jnode = mv_cam_uimod.load_conf(default=None) or dict()

        mv_intr = mv_jnode.get('ro', {}).get('intrinsic')
        if mv_intr:
            return cls(mv_intr)
        else:
            rtsp_jnode = rtsp_cam_uimod.load_conf(default=None) or dict()
            rtsp_intr = rtsp_jnode.get('rw', {}).get('conf', {}).get('intrinsic')
            if rtsp_intr:
                return cls(rtsp_intr)

        return cls(None)

    def get_whf(self):
        """
        :rtype: (float, float, float) | None
        """
        if self._intrinsic_jnode:
            cam_uiconf_pri_frustrum = self._intrinsic_jnode['frustum']
            sensor_w, sensor_h = cam_uiconf_pri_frustrum['sensor-wh']

            return sensor_w, sensor_h, cam_uiconf_pri_frustrum['focal-dist']


def get_terminal_stats(term_pid):
    """
    :type term_pid: str
    :rtype: dict
    """
    root_info = RootInfoUiconf.get_info()
    term_uimod = UiMod.create_term(term_pid)
    term_uuid = TerminalUiconf(term_uimod.conf_path).get_uuid()
    video_wh = CamDgStats.create(term_pid).get_stream_wh()

    stats_jnode = {
        'media': []
    }

    term_info = RootInfoUiconf.find_unit_r(root_info, unit_ref=term_uimod.ref)
    if term_info is None:
        eventlog.log_d('fail: NotFoundError ({})'.format(UI_TYPE_TERMINAL))
        raise NotFoundError

    http_streaming_info = RootInfoUiconf.find_unit_r(term_info, unit_type=UI_TYPE_HTTP_STREAMING)
    if http_streaming_info is not None:
        http_server_uimod = UiMod.create(http_streaming_info['ref'])
        http_server_uiconf = http_server_uimod.load_conf()
        http_server_uiconf_pub = http_server_uiconf['rw']
        http_enabled = http_server_uiconf_pub.get('enabled', False)
        if http_enabled:
            http_stats = {
                'stream': 'api/v2/origins/media/{}/live/stream'.format(term_uuid)
            }
            if video_wh:
                http_stats['resolution'] = video_wh

            stats_jnode['media'].append(http_stats)

    rtsp_streaming_info = RootInfoUiconf.find_unit_r(term_info, unit_type=UI_TYPE_RTSP_STREAMING)
    if rtsp_streaming_info is not None:
        rtsp_server_uimod = UiMod.create(rtsp_streaming_info['ref'])
        rtsp_server_uiconf = rtsp_server_uimod.load_conf()
        rtsp_server_uiconf_pub = rtsp_server_uiconf['rw']
        rtsp_enabled = rtsp_server_uiconf_pub.get('enabled', False)
        if rtsp_enabled:
            rtsp_stats = {
                'stream': rtsp_server_uiconf_pub.get('streaming', {}).get('url', 'try again later') # TODO
            }
            stats_jnode['media'].append(rtsp_stats)

    return stats_jnode


def get_detector_limits_stats(term_pid, vehicle=True, scene=True, lanes=False, classes=False):
    """
    :type term_pid: str
    :type vehicle: bool
    :type scene: bool
    :type lanes: bool
    :type classes: bool
    :rtype: dict
    """
    root_info = RootInfoUiconf.get_info()

    def store_factory(es_info):
        """
        :type es_info: dict
        :rtype: dict
        """
        return {
            'id': es_info['ref'].split('.')[-1],  # TODO: for current ui
            'ref': es_info['ref'],
            'name': es_info.get('name')
        }

    term_uimod = UiMod.create_term(term_pid)
    term_uuid = TerminalUiconf(term_uimod.conf_path).get_uuid()
    video_wh = CamDgStats.create(term_pid).get_stream_wh()
    video_w, video_h = (640, 480) if video_wh is None else video_wh

    stats_jnode = {
        'media': [{
            'resolution': [video_w, video_h],
            'stream': 'api/v2/origins/media/{}/live/stream'.format(term_uuid)
        }]
    }

    limits_jnode = {
        'images': dict(),
        'stores': InfoUiconf.map_units_r(root_info, store_factory, unit_types=[UI_TYPE_EVENTSTORE])
    }

    if vehicle:
        limits_jnode['images']['vehicle'] = {'max-width': 640}

    if scene:
        limits_jnode['images']['scene'] = {'max-width': video_w}

    if lanes:
        roadway_uimod = UiMod(term_pid, UI_TYPE_ROADWAY, '0')  # TODO
        roadway_uiconf = roadway_uimod.load_conf()
        roadway_lanes = roadway_uiconf['rw']['args']['lanes']
        limits_jnode['lanes'] = map(lambda x: x['name'], roadway_lanes)

    if classes:
        limits_jnode['classes'] = MMR_CLASSES

    return {
        'limits': limits_jnode,
        'stats': stats_jnode
    }


if __name__ == '__main__':
    pass
    # InfoUiconf._CONF_PATH = '/home/user/Documents/BitBucket/edgeserver/_share/edge-python-tools/edge-python-tools/share/misc/root.json'
    # RootInfoUiconf._CONF_PATH = '/home/user/Documents/BitBucket/edgeserver/_share/edge-python-tools/edge-python-tools/share/misc/root.json'
    #
    # root_info, links = RootInfoUiconf.get_info_links()
    # term_pid = '0'
    # term_uimod = UiMod.create_term(term_pid)
    #
    # term_info, parent_info = InfoUiconf.find_unit_parent(root_info, UI_TYPE_CONTROLLER, term_uimod.ref)
    #
    # def rm_link(unit_info):
    #     ref = unit_info['ref']
    #     print ref
    #     a = filter(lambda link: ref not in link, links)
    #     return a
    #
    # # reply = InfoUiconf.map_units(term_info, rm_link)
    # units_info = InfoUiconf.map_units(term_info, lambda x: x)
    # for unit_info in units_info:
    #     ref = unit_info['ref']
    #     links = filter(lambda link: ref not in link, links)
    #
    # import ujson
    # print ujson.dumps(links)
