from collections import defaultdict


from edge import eventlog
from edge.com import db as edge_db, net
from edge.service import db_value, tz
from edge.service.jnode.types import JsonTzServiceArgs
from edge.webserver.sysconf import ph
from edge.webserver.systest.test import Test, TestError, TestLimitsError
from edge.webserver.systest.misc import *
from edge.webserver.sysconf.api_controller import ControllerApi
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf


def _get_section_jnode(db):
    """
    :param db: should be under the db transaction context (operator 'with')
    :type db: edge.service.db_value.DbValueApi
    :rtype: dict
    """
    section_jnode = db.get_value(DB_SECTION)
    if section_jnode is None:
        section_jnode = {
            DB_KEY_TESTPACK: {
                'rw': {
                    'sessions': [],
                    'results': {}
                }
            }
        }

    return section_jnode


def _get_testpack_jnode(db):
    """
    :param db: should be under the db transaction context (operator 'with')
    :type db: edge.service.db_value.DbValueApi
    :rtype: dict
    """
    testpack_jnode = db.get_value(DB_SECTION, DB_KEY_TESTPACK)
    if testpack_jnode is None:
        testpack_jnode = {
            'rw': {
                'sessions': [],
                'results': {}
            }
        }

    return testpack_jnode


def _get_session_jnode(testpack_jnode, session_ref):
    """
    :type testpack_jnode: dict
    :type session_ref: str
    :rtype: dict

    :raise KeyError:
    """
    sessions_jnode = filter(
        lambda jnode: jnode['session'] == session_ref,
        testpack_jnode['rw']['sessions'])

    if len(sessions_jnode) == 1:
        return sessions_jnode[0]

    elif not sessions_jnode:
        raise KeyError

    raise NotImplementedError


def _get_session_limits(section_jnode, session_ref):
    """
    :type section_jnode: dict
    :rtype: dict
    """
    def get_limits_map():
        """
        :return: test-to-origin-map
        :rtype: dict
        """
        test_to_origin = defaultdict(list)
        root_info = ph.RootInfoUiconf.get_info()

        term_infos = ph.InfoUiconf.map_units_r(root_info, unit_types=[ph.UI_TYPE_TERMINAL])
        for term_info in term_infos:
            term_ref = term_info['ref']
            term_uimod = ph.UiMod.create(term_ref)
            term_uuid = TerminalUiconf(term_uimod.conf_path).get_uuid()

            ra_info = ph.InfoUiconf.find_unit_r(term_info, unit_type=ph.UI_TYPE_RADAR)

            test_to_origin['datetime'] += [term_uuid]
            if ra_info is not None:
                test_to_origin['position'] += [term_uuid]
                test_to_origin['spot-speed'] += [term_uuid]

        if net.query_gps_count() > 0:
            test_to_origin['geolocation'] += ['device']

        return test_to_origin

    def get_tests_map():
        """
        :return: test-to-origin-map
        :rtype: dict
        """
        session_jnode = _get_session_jnode(section_jnode[DB_KEY_TESTPACK], session_ref)
        tests_jnode = session_jnode['tests']

        test_to_origin = defaultdict(list)
        for test_jnode in tests_jnode:
            test_type, test_run = test_jnode['test']

            db_key = mk_test_db_key(session_jnode['session'], (test_type, test_run))
            ro_jnode = section_jnode[db_key].get('ro') or dict()
            origin = ro_jnode.get('origin', 'device')

            test_to_origin[test_type] += [origin]

        return test_to_origin

    limits_map = get_limits_map()
    tests_map = get_tests_map()

    def get_origin_to_test_map():
        """
        :return: origin-to-test-map
        :rtype: dict
        """
        origin_to_test = defaultdict(list)
        for test_type, src_origins in limits_map.iteritems():
            dst_origins = tests_map.get(test_type, [])
            for origin in src_origins:
                if origin not in dst_origins:
                    origin_to_test[origin] += [test_type]

        return origin_to_test

    rest_map = get_origin_to_test_map()

    def mk_limits_jnode():
        limits_jnode = {
            'origins': list(),
            'device': {
                'tests': list()
            }
        }

        origins_jnode = limits_jnode['origins']
        device_jnode = limits_jnode['device']
        for origin, tests in rest_map.iteritems():
            if origin != 'device':
                origins_jnode.append({
                    'tests': map(lambda test_type: dict(test=test_type), tests),
                    'origin': {
                        'uuid': origin,
                    }
                })

            else:
                device_jnode['tests'] = map(lambda test_type: dict(test=test_type), tests)

        return limits_jnode

    return mk_limits_jnode()


def _mk_test(services, session_ref, test_ref):
    """
    :type services: Services
    :type session_ref: str
    :type test_ref: str, str
    :rtype: Test
    """
    test_type = test_ref[0]
    if test_type == 'geolocation':
        from .geolocation_test import GeolocationTest
        return GeolocationTest(services, session_ref, test_ref)

    elif test_type == 'datetime':
        from .datetime_test import DatetimeTest
        return DatetimeTest(services, session_ref, test_ref)

    elif test_type == 'spot-speed':
        from .spot_speed_test import SpotSpeedTest
        return SpotSpeedTest(services, session_ref, test_ref)

    elif test_type == 'position':
        from .position_test import PositionTest
        return PositionTest(services, session_ref, test_ref)

    raise NotImplementedError


class TestSession:
    def __init__(self, services, session_ref):
        """
        :type services: Services
        :type session_ref: str
        """
        self.services = services
        self._ref = session_ref

    @property
    def ref(self):
        """
        :rtype: str
        """
        return self._ref

    @staticmethod
    def _add_test(db, session_ref, session_jnode, test_jnode):
        """
        :param db: should be under the db transaction context (operator 'with')
        :type db: edge.service.db_value.DbValueApi
        :type session_ref: str
        :type session_jnode: dict
        :type test_jnode: dict
        :rtype: str, str
        """
        tests_jnode = session_jnode['tests']

        test_type = test_jnode['test']
        test_run = mk_ref(map(lambda x: x['test'][1], tests_jnode))
        test_ref = test_type, test_run

        db.set_value(DB_SECTION, mk_test_db_key(session_ref, test_ref), value_jnode={
            'ro': test_jnode.get('args'),
            'rw': {
                'log': []
            }
        })

        tests_jnode.append({'test': test_ref})
        return test_ref

    def add_test(self, args):  # TODO: validate args againts limits
        """
        :type args: dict

        :return: (test-type, test-run)
        :rtype: str, str
        """
        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)

            session_jnode = _get_session_jnode(testpack_jnode, self.ref)
            test_ref = self._add_test(db_service, self.ref, session_jnode, args)

            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)
            return test_ref

    def add_tests(self):
        with self.services.db as db_service:
            section_jnode = _get_section_jnode(db_service)
            testpack_jnode = section_jnode[DB_KEY_TESTPACK]

            session_jnode = _get_session_jnode(testpack_jnode, self.ref)
            limits_jnode = _get_session_limits(section_jnode, self.ref)

            # make device tests
            for test_jnode in limits_jnode['device']['tests']:
                self._add_test(db_service, self.ref, session_jnode, {
                    'test': test_jnode['test']
                })

            # make origins tests
            for origin_jnode in limits_jnode['origins']:
                origin = origin_jnode['origin']['uuid']
                for test_jnode in origin_jnode['tests']:
                    self._add_test(db_service, self.ref, session_jnode, {
                        'test': test_jnode['test'],
                        'args': {
                            'origin': origin
                        }
                    })

            # sort tests
            def _cmp_test(left, right):
                """
                :type left: dict
                :type right: dict
                :rtype: int
                """
                l_type = left['test'][0]
                r_type = right['test'][0]
                _type_to_order_map = {
                    'datetime': 0,
                    'geolocation': 1,
                    'spot-speed': 2,
                    'position': 3
                }
                return _type_to_order_map[l_type] - _type_to_order_map[r_type]

            session_jnode['tests'].sort(cmp=_cmp_test)

            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)

    def get_test(self, ref):
        """
        :param ref: (test-type, test-run)
        :type ref: (str, str)
        :rtype: Test | None
        """
        with self.services.db as db_service:
            test_jnode = db_service.get_value(DB_SECTION, mk_test_db_key(self.ref, ref))
            if not test_jnode:
                return None

            return _mk_test(self.services, self.ref, ref)

    def del_test(self, ref):
        """
        :param ref: (test-type, test-run)
        :type ref: (str, str)
        :rtype bool
        """
        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)
            sessions_jnode = _get_session_jnode(testpack_jnode, self.ref)

            last_tests = sessions_jnode['tests']
            next_tests = filter(
                lambda test_jnode: test_jnode['test'] != ref,
                last_tests)

            if len(next_tests) == len(last_tests):
                return False

            sessions_jnode['tests'] = next_tests
            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)
            db_service.drop_value(DB_SECTION, key=mk_test_db_key(self.ref, ref))
            return True

    def set_summary(self, jnode):
        """
        :type jnode: dict
        """
        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)
            sessions_jnode = _get_session_jnode(testpack_jnode, self.ref)

            sessions_jnode['summary'] = jnode.get('summary')
            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)


class TestPack:
    def __init__(self, services):
        """
        :type services: Services
        """
        self.services = services

    def add_results(self, args):
        """
        :type args: dict | None
        """
        raw_crt_jnode = args.get('certificate') if args else None

        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)
            crt_jnode = dict()
            if raw_crt_jnode:
                crt_jnode['ref-number'] = raw_crt_jnode['ref-number']
                crt_jnode['max-time'] = raw_crt_jnode['max-time']

                min_time = raw_crt_jnode.get('min-time')
                if min_time:
                    crt_jnode['min-time'] = min_time

                authority = raw_crt_jnode.get('authority')
                if authority:
                    crt_jnode['authority'] = authority

            if crt_jnode:
                testpack_jnode['rw']['results']['certificate'] = crt_jnode
            else:
                try:
                    del testpack_jnode['rw']['results']['certificate']
                except KeyError:
                    pass

            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)

        if crt_jnode:
            ControllerApi.install_systest_crt(crt_jnode)
        else:
            ControllerApi.uninstall_systest_crt()

    def add_report(self, tmp_blob_filepath):
        """
        :type tmp_blob_filepath: str
        """

        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)
            old_report_ref = testpack_jnode['rw']['results'].get('report')
            if old_report_ref:
                ControllerApi.uninstall_systest_report(old_report_ref)

            report_ref = '{}.pdf'.format(mk_ref())
            testpack_jnode['rw']['results']['report'] = report_ref
            ControllerApi.install_systest_report(report_ref, tmp_blob_filepath)
            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)

    def delete_report(self, report_ref):
        """
        :type report_ref: str
        :raise TestError:
        """
        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)

            if testpack_jnode['rw']['results'].get('report') != report_ref:
                raise TestError

            del testpack_jnode['rw']['results']['report']
            ControllerApi.uninstall_systest_report(report_ref)
            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)

    def add_session(self):
        """
        :rtype: TestSession
        :raise TestLimitsError:
        :raise TestError:
        """
        # reconfig radar
        try:
            ra_origin_pid = ph.RootInfoUiconf.get_raterm_pid()
            if ra_origin_pid:
                min_rcs = get_rcs(ra_origin_pid, systest_active=True)
                if min_rcs is not None:
                    net.query_ra_alignment(ra_origin_pid, min_rcs=min_rcs, alignment=True)
        except net.NetError:
            eventlog.log_e('fail: setup radar')
            raise TestError

        ControllerApi.start_systest()

        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)
            sessions_jnode = testpack_jnode['rw']['sessions']
            if len(sessions_jnode) > 0:
                raise TestLimitsError

            session_ref = mk_ref(known_refs=list())
            session_jnode = {
                'session': session_ref,
                'tests': []
            }

            testpack_jnode['rw']['sessions'] = [session_jnode]
            db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)
            return TestSession(self.services, session_ref)

    def get_session(self, ref):
        """
        :type ref: str
        :rtype: TestSession | None
        """
        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)

        try:
            _get_session_jnode(testpack_jnode, ref)
            return TestSession(self.services, ref)
        except KeyError:
            pass

    def del_session(self, ref):
        """
        :type ref: str
        :rtype: bool

        :raise TestError:
        """
        try:
            ra_origin_pid = ph.RootInfoUiconf.get_raterm_pid()
            if ra_origin_pid:
                min_rcs = get_rcs(ra_origin_pid, systest_active=False)
                if min_rcs is not None:
                    net.query_ra_alignment(ra_origin_pid, min_rcs=min_rcs, alignment=False)
        except net.NetError:
            eventlog.log_e('fail: setup radar')
            raise TestError

        ControllerApi.finish_systest()

        with self.services.db as db_service:
            testpack_jnode = _get_testpack_jnode(db_service)
            sessions_jnode = testpack_jnode['rw']['sessions']
            session_jnode = None
            for jnode in sessions_jnode:
                if jnode['session'] == ref:
                    session_jnode = jnode

            if session_jnode:
                sessions_jnode.remove(session_jnode)
                db_service.set_value(DB_SECTION, DB_KEY_TESTPACK, testpack_jnode)
                db_service.drop_value(DB_SECTION, mk_test_db_glob(ref))
                return True

        return False

    def pack(self):
        """
        :rtype: dict
        """
        with self.services.db as db_service:
            section_jnode = _get_section_jnode(db_service)
            testpack_rw = section_jnode[DB_KEY_TESTPACK]['rw']

        for session_jnode in testpack_rw['sessions']:
            session_jnode['limits'] = _get_session_limits(
                section_jnode, session_jnode['session'])

        return testpack_rw


def init(db_args=None):
    """
    :type db_args: edge_db.DbArgs | None
    :rtype: TestPack
    """
    if db_args is None:
        db_args = edge_db.DbArgs.deserialize2()

    db_service = db_value.init_service(db_args)

    tz_args = JsonTzServiceArgs.create()
    tz_service = tz.init_service(tz_args)

    services = Services(tz_service, db_service)
    return TestPack(services)


# noinspection PyUnusedLocal
def fini(testpack):
    """
    :type testpack: TestPack
    """
    db_value.fini_service(testpack.services.db)
    tz.fini_service(testpack.services.tz)
