import edge.webserver.sysconf.dgstats.renderer

__all__ = [
    'TestStep', 'Test',
    'Summary', 'StatSummary'
]

import os
from edge import eventlog
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.api_terminal import TerminalApi
from .misc import *
from .exception import *

_TEST_TYPE_GEOLOCATION = 'geolocation'
_TEST_TYPE_DATE_TIME = 'datetime'
_TEST_TYPE_SPOT_SPEED = 'spot-speed'
_TEST_TYPE_POSITION = 'position'


class Summary:
    def __init__(self, comment=''):
        """
        :type comment: str
        """
        self.comment = comment

    @classmethod
    def repack(cls, jnode):
        """
        :type jnode: dict
        :rtype: dict

        :raise ValueError:
        """
        summary = cls.unpack(jnode)
        return summary.pack()

    @staticmethod
    def unpack(jnode):
        """
        :type jnode: dict | None
        :rtype: Summary

        :raise ValueError:
        """
        if not jnode:
            return Summary()

        comment = jnode.get('comment', '')
        if type(comment) not in [unicode, str]:
            eventlog.log_i('fail: param #/comment has invalid type')
            raise ValueError

        elif not 0 <= len(comment) <= 512:
            eventlog.log_i('fail: param #/comment has invalid len')
            raise ValueError

        return Summary(comment)

    def pack(self):
        """
        :rtype: dict
        """
        answer = dict()
        if len(self.comment):
            answer = {
                'comment': self.comment
            }

        return answer


class StatSummary(Summary):
    def __init__(self, status, comment=''):
        """
        :type status: str
        :type comment: str
        """
        Summary.__init__(self, comment)
        self.status = status

    @staticmethod
    def unpack(jnode):
        """
        :type jnode: dict
        :rtype: StatSummary

        :raise ValueError:
        """
        summary = Summary.unpack(jnode)

        # unpack #/status
        status = jnode.get('status')
        if status not in ['passed', 'failed']:
            eventlog.log_i('fail: param #/status has invalid value')
            raise ValueError

        return StatSummary(status, summary.comment)

    def pack(self):
        """
        :rtype: dict
        """
        answer = Summary.pack(self)
        answer['status'] = self.status
        return answer


def _mk_step(services, session_ref, test_ref, step_ref):
    """
    :type services: Services
    :type session_ref: str
    :type test_ref: (str, str)
    :type step_ref: str
    :rtype: TestStep
    """
    test_type = test_ref[0]
    if test_type == _TEST_TYPE_GEOLOCATION:
        from .geolocation_test import GeolocationTestStep
        return GeolocationTestStep(services, session_ref, test_ref, step_ref)

    elif test_type == _TEST_TYPE_DATE_TIME:
        from .datetime_test import DatetimeTestStep
        return DatetimeTestStep(services, session_ref, test_ref, step_ref)

    elif test_type == _TEST_TYPE_SPOT_SPEED:
        from .spot_speed_test import SpotSpeedTestStep
        return SpotSpeedTestStep(services, session_ref, test_ref, step_ref)

    elif test_type == _TEST_TYPE_POSITION:
        from .position_test import PositionTestStep
        return PositionTestStep(services, session_ref, test_ref, step_ref)

    else:
        return _mk_step(services, session_ref, test_ref, step_ref)


class TestStep:
    def __init__(self, services, session_ref, test_ref, step_ref):
        """
        :type services: Services
        :type session_ref: str
        :type test_ref: (str, str)
        :type step_ref: str
        """
        self._services = services

        self._session_ref = session_ref
        self._test_ref = test_ref
        self._step_ref = step_ref

    def _edit_log_item_jnode(self, log_item_jnode, body_jnode):
        """
        :type log_item_jnode: dict
        :type body_jnode: dict
        :rtype: dict
        """
        log_item_jnode['data'] = body_jnode['data']
        log_item_jnode['summary'] = body_jnode.get('summary')
        return log_item_jnode

    def edit(self, jnode):
        """
        :type jnode: dict
        """
        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))
            if not test_jnode:
                eventlog.log_e('fail: invalid test-ref')
                raise NotImplementedError

            for log_item_jnode in test_jnode['rw']['log']:
                if log_item_jnode['item'] == self._step_ref:
                    self._edit_log_item_jnode(log_item_jnode, jnode)
                    break

            db_service.set_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref),
                test_jnode)

            return test_jnode['rw']


class Test:
    _MAX_LOG_ITEMS = 10

    def __init__(self, services, session_ref, test_ref):
        """
        :type services: Services
        :type session_ref: str
        :type test_ref: (str, str)
        """
        self._services = services

        self._session_ref = session_ref
        self._test_ref = test_ref

    @property
    def ref(self):
        """
        :return: (test-type, test-run)
        :rtype: (str, str)
        """
        return self._test_ref

    @staticmethod
    def _get_live_image(ui_pid, bs_chan='0'):
        """
        :type ui_pid: str
        :type bs_chan: str
        :rtype: long, str
        """
        # get last overlay imagepath
        or_id = str(int(ui_pid) * 10)
        or_store = "/var/spool/edge/pipeline-{0}/or{1}-or{1}-store/".format(ui_pid, or_id)
        file_names = os.listdir(or_store)

        def unpack_ts(file_name):
            """
            :type file_name: str
            :rtype: int
            """
            return int(file_name.split('-')[-1])

        blob_name = max(file_names, key=unpack_ts)
        blob_path = os.path.join(or_store, blob_name)
        blob_ts = unpack_ts(blob_name)

        # save image to bs.65535
        from edge.com import bs
        bs_session = bs.Session.open()
        try:
            blob_id = bs_session.save_blob2(bs_chan, blob_ts / 1000, blob_path)
            blob_ref = '{}_{}.jpeg'.format(bs_chan, blob_id)
            return blob_ts, blob_ref

        finally:
            bs_session.close()

    def _mk_step_jnode(self, test_jnode, step_jnode, args_jnode):
        """
        :type test_jnode: dict
        :type step_jnode: dict
        :type args_jnode: dict | None
        """
        return step_jnode

    def set_steps_report(self, report):
        """
        :type report: bool
        :rtype: bool
        """
        with self._services.db as db_service:
            test_jnode = db_service.get_value(DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))
            if not test_jnode:
                return False

            for step_jnode in test_jnode['rw']['log']:
                step_jnode['report'] = report

            db_service.set_value(DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref), test_jnode)

        return True

    def add_step(self, jnode):
        """
        :type jnode: dict
        :rtype: None

        :raise TestError
        """
        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))
            if not test_jnode:
                raise TestError

            if len(test_jnode['rw']['log']) >= self._MAX_LOG_ITEMS:
                eventlog.log_i('fail: http409. item limit exceeded')
                raise TestLimitsError

            data_jnode = jnode.get('data')
            args_jnode = jnode.get('args')
            step_ref = mk_ref(map(lambda x: x['item'], test_jnode['rw']['log']))
            if data_jnode:
                blob_ts = data_jnode.get('ts')
                if blob_ts is not None:
                    time_isostr = self._services.tz.get_datetime_isostr(blob_ts / 1e6)
                else:
                    time_isostr = self._services.tz.get_datetime_now_isostr()

                step_jnode = {
                    'item': step_ref,
                    'time': time_isostr,
                    'data': data_jnode
                }

            else:
                step_jnode = self._mk_step_jnode(test_jnode, {
                    'item': step_ref,
                    'time': self._services.tz.get_datetime_now_isostr()
                }, args_jnode)

            if step_jnode:
                test_jnode['rw']['log'].append(step_jnode)
                db_service.set_value(
                    DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref),
                    test_jnode)

    def get_step(self, ref):
        """
        :type ref: str
        :rtype: TestStep | None
        """
        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))
            if not test_jnode:
                return None

            log_items = filter(
                lambda log_item_jnode: log_item_jnode['item'] == ref,
                test_jnode['rw']['log'])

            if len(log_items) != 1:
                return None

            return _mk_step(self._services, self._session_ref, self.ref, ref)

    def del_step(self, step_ref):
        """
        :type step_ref: str
        :rtype: bool
        """
        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))
            if not test_jnode:
                return False

            last_items = test_jnode['rw']['log']
            next_items = filter(
                lambda log_item_jnode: log_item_jnode['item'] != step_ref,
                last_items)

            if len(last_items) == len(next_items):
                return False

            test_jnode['rw']['log'] = next_items
            db_service.set_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref),
                test_jnode)

            return True

    def del_steps(self, step_refs):
        """
        :type step_refs: list[str]
        :rtype: bool
        """
        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))
            if not test_jnode:
                return False

            if step_refs:
                last_items = test_jnode['rw']['log']
                next_items = filter(
                    lambda log_item_jnode: log_item_jnode['item'] not in step_refs,
                    last_items)

            else:
                next_items = []

            test_jnode['rw']['log'] = next_items
            db_service.set_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref),
                test_jnode)

            return True

    def set_summary(self, jnode):
        """
        :type jnode: dict | None
        """
        summary_jnode = None
        if jnode:
            summary_jnode = StatSummary.repack(jnode['summary'])

        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))

            test_jnode['rw']['summary'] = summary_jnode
            db_service.set_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref),
                test_jnode)

            return test_jnode['rw']

    def pack(self):
        """
        :rtype: dict
        """
        from edge.webserver.sysconf.dgstats.renderer import RnDgStats

        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))

            test_jnode['rw']['limits'] = {
                'log': {
                    'max-items': self._MAX_LOG_ITEMS
                }
            }
            test_ro_jnode = test_jnode.get('ro') or dict()
            term_uuid = test_ro_jnode.get('origin')
            if term_uuid:
                test_jnode['rw']['limits']['origin'] = {
                    'uuid': term_uuid
                }

                term_info = ph.RootInfoUiconf.get_term_info(term_uuid)
                if term_info:
                    term_uimod = ph.UiMod.create(term_info['ref'])
                    result = RnDgStats.create(term_uimod.pid).get_stream_wh_dh()
                    if result:
                        w, h, dh = result
                        test_jnode['rw']['limits']['origin']['live'] = {
                            'wh': [w, h]
                        }

                    ra_type = TerminalApi.get_ra_type(term_uimod.pid)
                    if self._test_ref[0] in [_TEST_TYPE_SPOT_SPEED, _TEST_TYPE_POSITION]:
                        if ra_type:
                            test_jnode['rw']['limits']['simulator'] = [
                                SIMULATOR_TYPE_EKTSDG,
                                SIMULATOR_TYPE_IS24D,
                                SIMULATOR_TYPE_SAPSAN3_1,
                                SIMULATOR_TYPE_SAPSAN3_2
                            ]
                        else:
                            test_jnode['rw']['limits']['simulator'] = []

            return test_jnode['rw']
